@@ -3,7 +3,7 @@
  *
  *	written 11-feb-86 by Daniel Lawrence
  *
- * $Header: /usr/build/vile/vile/RCS/bind.c,v 1.336 2009/05/26 21:36:17 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/bind.c,v 1.375 2016/07/16 13:54:52 tom Exp $
  *
  */
 
@@ -20,15 +20,10 @@
 				   over those shorter.  e.g. display "quit"
 				   instead of "q" if possible */
 
-extern const int nametblsize;
+extern const int nametbl_size;
 
-#if OPT_CASELESS
 #define Strcmp(s,d)      cs_strcmp(case_insensitive, s, d)
 #define StrNcmp(s,d,len) cs_strncmp(case_insensitive, s, d, len)
-#else
-#define Strcmp(s,d)      strcmp(s, d)
-#define StrNcmp(s,d,len) strncmp(s, d, len)
-#endif
 
 #if OPT_REBIND
 #define	isSpecialCmd(k) \
@@ -42,6 +37,7 @@ extern const int nametblsize;
 static int key_to_bind(const CMDFUNC * kcmd);
 static int update_binding_list(BUFFER *bp);
 static void makebindlist(LIST_ARGS);
+static void makebindkeyslist(LIST_ARGS);
 #endif /* OPT_REBIND */
 
 #if OPT_NAMEBST
@@ -119,6 +115,8 @@ static struct {
     },
 };
 
+static UINT which_current;
+
 #if OPT_REBIND
 static BINDINGS *bindings_to_describe = &dft_bindings;
 #endif
@@ -151,22 +149,26 @@ old_namebst(BI_NODE * a)
     if (!(a->value.n_flags & NBST_READONLY)) {
 	CMDFUNC *cmd = TYPECAST(CMDFUNC, a->value.n_cmd);
 	if (cmd != 0) {
+	    if ((cmd->c_flags & CMD_TYPE) != CMD_FUNC) {
 #if OPT_ONLINEHELP
-	    if (cmd->c_help)
-		free(TYPECAST(char, cmd->c_help));
+		if (cmd->c_help)
+		    free(TYPECAST(char, cmd->c_help));
 #endif
 #if OPT_MACRO_ARGS
-	    if (cmd->c_args) {
-		int n;
-		for (n = 0; cmd->c_args[n].pi_text != 0; ++n) {
-		    free(cmd->c_args[n].pi_text);
+		if (cmd->c_args) {
+		    int n;
+		    for (n = 0; cmd->c_args[n].pi_text != 0; ++n) {
+			free(cmd->c_args[n].pi_text);
+		    }
+		    free(cmd->c_args);
 		}
-		free(cmd->c_args);
-	    }
 #endif
-	    free(cmd);
+		free(cmd);
+		free(TYPECAST(char, BI_KEY(a)));
+	    }
+	} else {
+	    free(TYPECAST(char, BI_KEY(a)));
 	}
-	free(TYPECAST(char, BI_KEY(a)));
     }
     free(a);
     endofDisplay();
@@ -196,17 +198,62 @@ static void
 dpy_namebst(BI_NODE * a GCC_UNUSED, int level GCC_UNUSED)
 {
 #if OPT_TRACE
-    TRACE(("%s%p -> %s (%d)\n",
-	   trace_indent(level, '.'), (void *) a, BI_KEY(a), a->balance));
+    char *indent = strmalloc(trace_indent(level, '.'));
+    TRACE(("[%d]%s%p -> %s%s (%d)\n",
+	   level, indent,
+	   (void *) a,
+	   (a->value.n_flags & NBST_READONLY)
+	   ? "*"
+	   : "",
+	   BI_KEY(a), a->balance));
+    free(indent);
 #endif
 }
 
+static void
+xcg_namebst(BI_NODE * a, BI_NODE * b)
+{
+    BI_DATA temp = a->value;
+    a->value = b->value;
+    b->value = temp;
+}
+
 #define BI_DATA0 {{0}, 0, {0,0,0}}
 #define BI_TREE0 0, 0, BI_DATA0
+
 static BI_TREE namebst =
-{new_namebst, old_namebst, dpy_namebst, BI_TREE0};
+{new_namebst, old_namebst, dpy_namebst, xcg_namebst, BI_TREE0};
+
+static BI_TREE glbsbst =
+{new_namebst, old_namebst, dpy_namebst, xcg_namebst, BI_TREE0};
+
 static BI_TREE redefns =
-{new_namebst, old_namebst, dpy_namebst, BI_TREE0};
+{new_namebst, old_namebst, dpy_namebst, xcg_namebst, BI_TREE0};
+
+/*----------------------------------------------------------------------------*/
+/*
+ * Name-completion for global commands uses a different table (a subset) from
+ * the normal command-table.  Provide a way to select that table.
+ */
+
+static BI_TREE *
+bst_pointer(UINT which)
+{
+    BI_TREE *result;
+
+    if (which == GLOBOK) {
+	result = &glbsbst;
+    } else {
+	result = &namebst;
+    }
+    return result;
+}
+
+static BI_TREE *
+bst_current(void)
+{
+    return bst_pointer(which_current);
+}
 #endif /* OPT_NAMEBST */
 
 /*----------------------------------------------------------------------------*/
@@ -235,7 +282,7 @@ vl_help(int f GCC_UNUSED, int n GCC_UNUSED)
 	returnCode(FALSE);
 
     if (bp->b_active == FALSE) {	/* never been used */
-	hname = cfg_locate(helpfile, FL_ANYWHERE | FL_READABLE);
+	hname = cfg_locate(helpfile, LOCATE_HELP);
 	if (!hname) {
 	    mlforce("[Sorry, can't find the help information]");
 	    (void) zotbuf(bp);
@@ -251,8 +298,7 @@ vl_help(int f GCC_UNUSED, int n GCC_UNUSED)
 	set_bname(bp, HELP_BufName);
 	set_rdonly(bp, hname, MDVIEW);
 
-	make_local_b_val(bp, MDIGNCASE);	/* easy to search, */
-	set_b_val(bp, MDIGNCASE, TRUE);
+	set_local_b_val(bp, MDIGNCASE, TRUE);	/* easy to search, */
 	b_set_scratch(bp);
 	b_set_recentlychanged(bp);
 	if (!alreadypopped)
@@ -406,9 +452,8 @@ set_termchrs(int f GCC_UNUSED, int n GCC_UNUSED)
     /* get the table-entry */
     tb_scopy(&name, "");
     if ((s = kbd_reply("Terminal setting: ", &name, chr_eol,
-		       ' ', 0, chr_complete)) == TRUE) {
-
-	j = chr_lookup(tb_values(name));
+		       ' ', 0, chr_complete)) == TRUE
+	&& (j = chr_lookup(tb_values(name))) >= 0) {
 	switch (TermChrs[j].how_to) {
 	case 's':
 	default:
@@ -505,7 +550,7 @@ unbindchar(BINDINGS * bs, int c)
     KBIND *skbp;		/* saved pointer into the command table */
 
     /* if it's a simple character, it will be in the normal[] array */
-    if (!isSpecial(c)) {
+    if (is8Bits(c)) {
 	if (bs->kb_normal[CharOf(c)]) {
 	    bs->kb_normal[CharOf(c)] = 0;
 	    return TRUE;
@@ -666,7 +711,7 @@ install_bind(int c, const CMDFUNC * kcmd, BINDINGS * bs)
     reset_prefix(-1, kcod2fnc(bs, c), bs);
     reset_prefix(c, kcmd, bs);
 
-    if (!isSpecial(c)) {
+    if (is8Bits(c)) {
 	bs->kb_normal[CharOf(c)] = TYPECAST(CMDFUNC, kcmd);
     } else if ((kbp = kcode2kbind(bs, c)) != 0) {	/* change it in place */
 	kbp->k_cmd = kcmd;
@@ -698,7 +743,7 @@ bind_any_key(BINDINGS * bs)
 
     /* prompt the user to type in a key to bind */
     /* and get the function name to bind it to */
-    fnp = kbd_engl("Bind function whose full name is: ", cmd);
+    fnp = kbd_engl("Bind function whose full name is: ", cmd, 0);
 
     if (fnp == NULL || (kcmd = engl2fnc(fnp)) == NULL) {
 	return no_such_function(fnp);
@@ -745,6 +790,7 @@ bind_s_key(int f GCC_UNUSED, int n GCC_UNUSED)
 /* remember whether we last did "apropos" or "describe-bindings" */
 static char *last_lookup_string;
 static CMDFLAGS last_whichcmds;
+static int last_whichmode;
 static int append_to_binding_list;
 
 /* ARGSUSED */
@@ -756,7 +802,7 @@ update_binding_list(BUFFER *bp GCC_UNUSED)
 }
 
 static int
-describe_any_bindings(char *lookup, int whichcmd)
+describe_any_bindings(char *lookup, CMDFLAGS whichcmd)
 {
     last_lookup_string = lookup;
     last_whichcmds = whichcmd;
@@ -768,7 +814,60 @@ describe_any_bindings(char *lookup, int whichcmd)
 int
 desbind(int f GCC_UNUSED, int n GCC_UNUSED)
 {
-    return describe_any_bindings((char *) 0, 0);
+    return describe_any_bindings((char *) 0, (CMDFLAGS) 0);
+}
+
+/* ARGSUSED */
+static int
+update_bindkeys_list(BUFFER *bp GCC_UNUSED)
+{
+    return liststuff(bindings_to_describe->bufname, append_to_binding_list,
+		     makebindkeyslist, (int) last_whichcmds, (void *) last_lookup_string);
+}
+
+static int
+describe_all_bindings(char *lookup, CMDFLAGS whichcmd, int mode)
+{
+    last_lookup_string = lookup;
+    last_whichcmds = whichcmd;
+    last_whichmode = mode;
+
+    return update_bindkeys_list((BUFFER *) 0);
+}
+
+/* ARGSUSED */
+int
+desall_bind(int f GCC_UNUSED, int n GCC_UNUSED)
+{
+    return describe_all_bindings((char *) 0, (CMDFLAGS) 0, (f ? n : -1));
+}
+
+static int
+describe_all_keys_bindings(BINDINGS * bs, int mode)
+{
+    int code;
+    bindings_to_describe = bs;
+    code = describe_all_bindings((char *) 0, (CMDFLAGS) 0, mode);
+    bindings_to_describe = &dft_bindings;
+    return code;
+}
+
+int
+desall_i_bind(int f GCC_UNUSED, int n GCC_UNUSED)
+{
+    return describe_all_keys_bindings(&ins_bindings, (f ? n : -1));
+}
+
+int
+desall_c_bind(int f GCC_UNUSED, int n GCC_UNUSED)
+{
+    return describe_all_keys_bindings(&cmd_bindings, (f ? n : -1));
+}
+
+int
+desall_s_bind(int f GCC_UNUSED, int n GCC_UNUSED)
+{
+    return describe_all_keys_bindings(&sel_bindings, (f ? n : -1));
 }
 
 static int
@@ -776,7 +875,7 @@ describe_alternate_bindings(BINDINGS * bs)
 {
     int code;
     bindings_to_describe = bs;
-    code = describe_any_bindings((char *) 0, 0);
+    code = describe_any_bindings((char *) 0, (CMDFLAGS) 0);
     bindings_to_describe = &dft_bindings;
     return code;
 }
@@ -813,6 +912,13 @@ desopers(int f GCC_UNUSED, int n GCC_UNUSED)
     return describe_any_bindings((char *) 0, OPER);
 }
 
+/* ARGSUSED */
+int
+desglobals(int f GCC_UNUSED, int n GCC_UNUSED)
+{
+    return describe_any_bindings((char *) 0, GLOBOK);
+}
+
 /* lookup function by substring */
 /* ARGSUSED */
 int
@@ -821,11 +927,11 @@ dessubstr(int f GCC_UNUSED, int n GCC_UNUSED)
     int s;
     static char substring[NSTRING];
 
-    s = mlreply("Apropos string: ", substring, sizeof(substring));
+    s = mlreply("Apropos string: ", substring, (UINT) sizeof(substring));
     if (s != TRUE)
 	return (s);
 
-    return describe_any_bindings(substring, 0);
+    return describe_any_bindings(substring, (CMDFLAGS) 0);
 }
 
 static char described_cmd[NLINE + 1];	/* string to match cmd names to */
@@ -842,12 +948,12 @@ desfunc(int f GCC_UNUSED, int n GCC_UNUSED)
     described_cmd[0] = '^';
 
     fnp = kbd_engl("Describe function whose full name is: ",
-		   described_cmd + 1);
+		   described_cmd + 1, 0);
     if (fnp == NULL || engl2fnc(fnp) == NULL) {
 	s = no_such_function(fnp);
     } else {
 	append_to_binding_list = TRUE;
-	s = describe_any_bindings(described_cmd, 0);
+	s = describe_any_bindings(described_cmd, (CMDFLAGS) 0);
 	append_to_binding_list = FALSE;
     }
     return s;
@@ -979,9 +1085,9 @@ prompt_describe_key(BINDINGS * bs)
 
     /* describe it */
     described_cmd[0] = '^';
-    (void) strcpy(described_cmd + 1, temp.n_name);
+    (void) vl_strncpy(described_cmd + 1, temp.n_name, sizeof(described_cmd));
     append_to_binding_list = TRUE;
-    s = describe_any_bindings(described_cmd, 0);
+    s = describe_any_bindings(described_cmd, (CMDFLAGS) 0);
     append_to_binding_list = FALSE;
 
     mlwrite("Key sequence '%s' is bound to function \"%s\"",
@@ -1033,7 +1139,7 @@ static unsigned
 converted_len(char *buffer)
 {
     unsigned len = 0, c;
-    while ((c = *buffer++) != EOS) {
+    while ((c = CharOf(*buffer++)) != EOS) {
 	if (c == '\t')
 	    len |= 7;
 	len++;
@@ -1151,15 +1257,15 @@ show_onlinehelp(const CMDFUNC * cmd)
 	return FALSE;
 #if OPT_PROCEDURES
     if ((cmd->c_flags & CMD_TYPE) == CMD_PROC)
-	flags &= !(UNDO | REDO);
+	flags &= ~(UNDO | REDO);
 #endif
 #if OPT_PERL
     if ((cmd->c_flags & CMD_TYPE) == CMD_PERL)
-	flags &= !(UNDO | REDO);
+	flags &= ~(UNDO | REDO);
 #endif
     if (flags & (RANGE | UNDO | REDO | GLOBOK)) {
 	const char *gaps = "";
-	char *next = outseq;
+	char *next;
 
 	next = lsprintf(outseq, "  ( ");
 	if (cmd->c_flags & (UNDO | REDO)) {
@@ -1226,7 +1332,7 @@ btree_walk(BI_NODE * node, int (*func) (BI_NODE *, const void *),
 static int
 clearflag_func(BI_NODE * n, const void *d GCC_UNUSED)
 {
-    n->value.n_flags &= (UCHAR) (~NBST_DONE);
+    clr_typed_flags(n->value.n_flags, UCHAR, NBST_DONE);
     return 0;
 }
 
@@ -1300,9 +1406,10 @@ makebind_func(BI_NODE * node, const void *d)
 static void
 makebindlist(int whichmask, void *mstring)
 {
+    BI_TREE *my_bst = bst_current();
     struct bindlist_data data;
 
-    data.mask = whichmask;
+    data.mask = (UINT) whichmask;
     data.min = SHORT_CMD_LEN;
     data.apropos = (char *) mstring;
     data.bs = bindings_to_describe;
@@ -1311,19 +1418,72 @@ makebindlist(int whichmask, void *mstring)
     mlwrite("[Building binding list]");
 
     /* clear the NBST_DONE flag */
-    btree_walk(&namebst.head, clearflag_func, 0);
+    btree_walk(&(my_bst->head), clearflag_func, 0);
 
     /* create binding list */
-    if (btree_walk(&namebst.head, makebind_func, &data))
+    if (btree_walk(&(my_bst->head), makebind_func, &data))
 	return;
 
     /* catch entries with no synonym > SHORT_CMD_LEN */
     data.min = 0;
-    if (btree_walk(&namebst.head, makebind_func, &data))
+    if (btree_walk(&(my_bst->head), makebind_func, &data))
 	return;
 
     mlerase();			/* clear the message line */
 }
+
+/* build a binding list (limited or full) */
+/* ARGSUSED */
+static void
+makebindkeyslist(int whichmask, void *mstring)
+{
+    char outseq[NSTRING];	/* output buffer for command sequence */
+    NTAB temp;			/* name table pointer */
+    int ch;
+    int state;
+
+    (void) mstring;
+
+    /* let us know this is in progress */
+    mlwrite("[Building binding list]");
+    for (state = 0; state < 3; ++state) {
+	int modify = 0;
+	int found = 0;
+	const char *what = "?";
+	switch (state) {
+	case 0:
+	    what = "Normal keys";
+	    break;
+	case 1:
+	    what = "^A-keys";
+	    modify = CTLA;
+	    break;
+	case 2:
+	    what = "^X-keys";
+	    modify = CTLX;
+	    break;
+	}
+	bprintf("%s-- %s", state ? "\n\n" : "", what);
+	for (ch = 0; ch < N_chars; ++ch) {
+	    (void) kcod2prc(ch + modify, outseq);
+	    if (fnc2ntab(&temp, kcod2fnc(bindings_to_describe, ch + modify))) {
+		if (whichmask && !(temp.n_cmd->c_flags & (CMDFLAGS) whichmask))
+		    continue;
+		bputc('\n');
+		bprintf("%d\t%s\t", ch, outseq);
+		quoted(0, temp.n_name);
+		++found;
+	    } else if (last_whichmode > 1) {
+		bputc('\n');
+		bprintf("%d\t%s", ch, outseq);
+	    }
+	}
+	if (found) {
+	    bprintf("\n-- %s total: %d", what, found);
+	}
+    }
+    mlerase();			/* clear the message line */
+}
 #else /* OPT_NAMEBST */
 /* fully describe a function into the current buffer, given a pointer to
  * its name table entry */
@@ -1380,7 +1540,7 @@ makebindlist(int whichmask, void *mstring)
     char *listed;
 
     beginDisplay();
-    listed = typecallocn(char, (size_t) nametblsize);
+    listed = typecallocn(char, (size_t) nametbl_size);
     endofDisplay();
 
     if (listed == 0) {
@@ -1491,12 +1651,16 @@ static int
 check_file_access(char *fname, UINT mode)
 {
     int result = ffaccess(fname, mode);
+
+    TRACE((T_CALLED "check_file_access(%s, %#x) = %d\n", fname, mode, result));
 #if SYS_UNIX
 #define CHK_OWNER (FL_HOME | FL_CDIR)
     if (result) {
 	int doit = FALSE;
-	int check;
 	int owner;
+#ifdef GVAL_CHK_ACCESS
+	int check;
+#endif
 
 	/*
 	 * Suppress ownership check for absolute pathnames.
@@ -1511,18 +1675,25 @@ check_file_access(char *fname, UINT mode)
 
 	if (mode & FL_EXECABLE) {
 	    doit = TRUE;
-	} else if ((check = global_g_val(GVAL_CHK_ACCESS)) != 0) {
+	}
+#ifdef GVAL_CHK_ACCESS
+	else if ((check = global_g_val(GVAL_CHK_ACCESS)) != 0) {
 	    /*
 	     * The values for check-access mode are single bits.  But we treat
 	     * those as an ordered list.  Hence, setting "home" implies we
 	     * also check "current", since that is the first item in the list.
 	     */
-	    if ((mode
-		 & (check * 2 - 1)
-		 & ~(FL_EXECABLE | FL_WRITEABLE | FL_READABLE)) != 0)
+	    if ((mode & FL_ALWAYS) && !((UINT) check & FL_ALWAYS)) {
+		doit = FALSE;
+	    } else if ((mode
+			& ((UINT) check * 2 - 1)
+			& ~(FL_EXECABLE | FL_WRITEABLE | FL_READABLE)) != 0) {
 		doit = TRUE;
+	    }
 	}
-	if (doit) {
+#endif
+
+	if (doit && !(mode & FL_INSECURE)) {
 	    char *dname = (char *) malloc(NFILEN + strlen(fname) + 10);
 	    char *leaf;
 
@@ -1540,7 +1711,7 @@ check_file_access(char *fname, UINT mode)
     }
 #endif
 
-    return result;
+    returnCode(result);
 }
 
 static char *
@@ -1561,9 +1732,10 @@ locate_file_in_list(char *list, char *fname, UINT mode)
     const char *cp;
     char *sp;
     char dir_name[NFILEN];
+    int first = TRUE;
 
     if ((cp = list) != 0) {
-	while ((cp = parse_pathlist(cp, dir_name)) != 0) {
+	while ((cp = parse_pathlist(cp, dir_name, &first)) != 0) {
 	    if ((sp = locate_fname(dir_name, fname, mode)) != 0)
 		return sp;
 	}
@@ -1586,12 +1758,12 @@ PATH_value(void)
     if (!tb_length(myfiles)) {
 	char mypath[NFILEN];
 
-	(void) strcpy(mypath, NONNULL(prog_arg));
-	if ((tmp = vms_pathleaf(mypath)) == mypath)
-	    (void) strcpy(mypath,
-			  current_directory(FALSE));
-	else
+	(void) vl_strncpy(mypath, NONNULL(prog_arg), sizeof(mypath));
+	if ((tmp = vms_pathleaf(mypath)) == mypath) {
+	    (void) vl_strncpy(mypath, current_directory(FALSE), sizeof(mypath));
+	} else {
 	    *tmp = EOS;
+	}
 
 	if (!tb_init(&myfiles, EOS)
 	    || !tb_sappend(&myfiles, mypath)
@@ -1622,10 +1794,14 @@ char *
 cfg_locate(char *fname, UINT which)
 {
     char *sp;
-    UINT mode = (which & (FL_EXECABLE | FL_WRITEABLE | FL_READABLE));
+    UINT mode = (which & (FL_ALWAYS |
+			  FL_EXECABLE |
+			  FL_WRITEABLE |
+			  FL_READABLE |
+			  FL_INSECURE));
 
 #define FL_BIT(name) ((which & FL_##name) ? " " #name : "")
-    TRACE((T_CALLED "cfg_locate('%s',%s%s%s%s%s%s%s%s%s)\n", NonNull(fname),
+    TRACE((T_CALLED "cfg_locate('%s',%s%s%s%s%s%s%s%s%s%s%s)\n", NonNull(fname),
 	   FL_BIT(EXECABLE),
 	   FL_BIT(WRITEABLE),
 	   FL_BIT(READABLE),
@@ -1634,7 +1810,9 @@ cfg_locate(char *fname, UINT which)
 	   FL_BIT(EXECDIR),
 	   FL_BIT(STARTPATH),
 	   FL_BIT(PATH),
-	   FL_BIT(LIBDIR)));
+	   FL_BIT(LIBDIR),
+	   FL_BIT(ALWAYS),
+	   FL_BIT(INSECURE)));
 
     /* take care of special cases */
     if (!fname || !fname[0] || isSpace(fname[0]))
@@ -1748,9 +1926,10 @@ list_which_file_in_list(char *list, char *fname, UINT mode)
 {
     const char *cp;
     char dir_name[NFILEN];
+    int first = TRUE;
 
     if ((cp = list) != 0) {
-	while ((cp = parse_pathlist(cp, dir_name)) != 0) {
+	while ((cp = parse_pathlist(cp, dir_name, &first)) != 0) {
 	    list_which_fname(dir_name, fname, mode);
 	}
     }
@@ -1759,8 +1938,9 @@ list_which_file_in_list(char *list, char *fname, UINT mode)
 static void
 list_which(LIST_ARGS)
 {
+    UINT uflag = (UINT) flag;
     char *fname = (char *) ptr;
-    UINT mode = (flag & (FL_EXECABLE | FL_WRITEABLE | FL_READABLE));
+    UINT mode = (uflag & (FL_EXECABLE | FL_WRITEABLE | FL_READABLE));
 
     /* take care of special cases */
     if (!fname || !fname[0] || isSpace(fname[0]))
@@ -1773,27 +1953,27 @@ list_which(LIST_ARGS)
 	    fname);
 
     /* look in the current directory */
-    if (flag & FL_CDIR) {
+    if (uflag & FL_CDIR) {
 	bprintf("\n$cwd");
 	list_one_fname(fname, FL_CDIR | mode);
     }
 
-    if (flag & FL_HOME) {	/* look in the home directory */
+    if (uflag & FL_HOME) {	/* look in the home directory */
 	bprintf("\n$HOME");
 	list_which_fname(home_dir(), fname, FL_HOME | mode);
     }
 
-    if (flag & FL_EXECDIR) {	/* look in vile's bin directory */
+    if (uflag & FL_EXECDIR) {	/* look in vile's bin directory */
 	bprintf("\n$exec-path");
 	list_which_fname(exec_pathname, fname, FL_EXECDIR | mode);
     }
 
-    if (flag & FL_STARTPATH) {	/* look along "VILE_STARTUP_PATH" */
+    if (uflag & FL_STARTPATH) {	/* look along "VILE_STARTUP_PATH" */
 	bprintf("\n$startup-path");
 	list_which_file_in_list(startup_path, fname, FL_STARTPATH | mode);
     }
 
-    if (flag & FL_PATH) {	/* look along "PATH" */
+    if (uflag & FL_PATH) {	/* look along "PATH" */
 #if OPT_PATHLOOKUP
 	bprintf("\n$PATH");
 	list_which_file_in_list(PATH_value(), fname, FL_PATH | mode);
@@ -1801,7 +1981,7 @@ list_which(LIST_ARGS)
 
     }
 
-    if (flag & FL_LIBDIR) {	/* look along "VILE_LIBDIR_PATH" */
+    if (uflag & FL_LIBDIR) {	/* look along "VILE_LIBDIR_PATH" */
 	bprintf("\n$libdir-path");
 	list_which_file_in_list(libdir_path, fname, FL_LIBDIR | mode);
     }
@@ -1815,7 +1995,7 @@ show_which_file(char *fname, UINT mode, int f, int n)
 	mlwrite("%s", result);
     }
     if (f)
-	liststuff(WHICH_BufName, n > 2, list_which, mode, fname);
+	liststuff(WHICH_BufName, n > 2, list_which, (int) mode, fname);
     return (result != 0);
 }
 
@@ -1851,7 +2031,7 @@ which_exec(int f, int n)
 char *
 kcod2pstr(int c, char *seq, int limit)
 {
-    seq[0] = (char) kcod2escape_seq(c, &seq[1], limit - 1);
+    seq[0] = (char) kcod2escape_seq(c, &seq[1], (size_t) (limit - 1));
     return seq;
 }
 
@@ -1874,7 +2054,7 @@ static const struct {
 #endif
 
 #define ADD_KCODE(src) \
-	if ((ptr - base) + (len = strlen(src)) + 3 < limit) { \
+	if (((size_t) (ptr - base) + (len = strlen(src)) + need + 2) < limit) { \
 	    strcpy(ptr, src); \
 	    ptr += len; \
 	}
@@ -1886,15 +2066,34 @@ int
 kcod2escape_seq(int c, char *ptr, size_t limit)
 {
     char *base = ptr;
+    char temp[20];
+    size_t need;
 
-    if (base != 0 && limit > 5) {
-	if (c & CTLA)
+#if OPT_MULTIBYTE
+    int ch = (c & ((1 << MaxCBits) - 1));
+    const char *s;
+
+    if ((ch >= 256)
+	&& (cmd_encoding >= enc_UTF8
+	    || (cmd_encoding <= enc_AUTO && okCTYPE2(vl_wide_enc)))
+	&& (s = vl_mb_to_utf8(ch)) != 0) {
+	vl_strncpy(temp, s, sizeof(temp));
+    } else
+#endif
+    {
+	temp[0] = (char) c;
+	temp[1] = EOS;
+    }
+    need = strlen(temp);
+
+    if (base != 0 && limit > (4 + need)) {
+	if ((UINT) c & CTLA)
 	    *ptr++ = (char) cntl_a;
-	else if (c & CTLX)
+	else if ((UINT) c & CTLX)
 	    *ptr++ = (char) cntl_x;
 
 #if OPT_KEY_MODIFY
-	if (c & mod_KEY) {
+	if ((UINT) c & mod_KEY) {
 	    size_t len;
 	    unsigned n;
 	    for (n = 0; n < TABLESIZE(key_modifiers); ++n) {
@@ -1916,10 +2115,10 @@ kcod2escape_seq(int c, char *ptr, size_t limit)
 #endif
 	}
 #endif
-	if (c & SPEC)
+	if ((UINT) c & SPEC)
 	    *ptr++ = (char) poundc;
-	*ptr++ = (char) c;
-	*ptr = EOS;
+	strcpy(ptr, temp);
+	ptr += need;
     }
     return (int) (ptr - base);
 }
@@ -1975,11 +2174,11 @@ kcod2prc(int c, char *seq)
     char temp[NSTRING];
     int length;
 
-    length = kcod2pstr(c, temp, sizeof(temp))[0];
+    length = kcod2pstr(c, temp, (int) sizeof(temp))[0];
 #if OPT_KEY_MODIFY
-    if ((c & mod_KEY) != 0 && (length != 0)) {
+    if (((UINT) c & mod_KEY) != 0 && (length != 0)) {
 	(void) strcpy(seq, temp + 1);
-	if (length < (int) (1 + strlen(temp + 1) + (CharOf(c) == 0))) {
+	if (length < (int) (1 + strlen(temp + 1) + (size_t) (CharOf(c) == 0))) {
 	    (void) bytes2prc(seq + length - 1, temp + length, 1);
 	}
     } else
@@ -2022,7 +2221,7 @@ kcod2fnc(const BINDINGS * bs, int c)
     const CMDFUNC *result = 0;
 
     if (bs != 0) {
-	if (isSpecial(c)) {
+	if (!is8Bits(c)) {
 	    KBIND *kp = kcode2kbind(bs, c);
 	    result = (kp != 0) ? kp->k_cmd : 0;
 	} else {
@@ -2060,7 +2259,7 @@ fnc2pstr(const CMDFUNC * f)
     if (c == -1)
 	return NULL;
 
-    return kcod2pstr(c, seq, sizeof(seq));
+    return kcod2pstr(c, seq, (int) sizeof(seq));
 }
 #endif
 
@@ -2097,7 +2296,7 @@ fnc2ntab(NTAB * result, const CMDFUNC * cfp)
 #if OPT_NAMEBST
 	result->n_name = 0;
 	result->n_cmd = cfp;
-	btree_walk(&namebst.head, match_cmdfunc, result);
+	btree_walk(&(bst_current()->head), match_cmdfunc, result);
 	found = (result->n_name != 0);
 #else
 	switch (cfp->c_flags & CMD_TYPE) {
@@ -2161,29 +2360,14 @@ fnc2engl(const CMDFUNC * cfp)
 	translate english name to function pointer
 		 return any match or NULL if none
  */
-#define BINARY_SEARCH_IS_BROKEN 0
-
 const CMDFUNC *
 engl2fnc(const char *fname)
 {
 #if OPT_NAMEBST
-    BI_NODE *n = btree_pmatch(BI_RIGHT(&namebst.head), TRUE, fname);
+    BI_NODE *n = btree_pmatch(BI_RIGHT(&(bst_current()->head)), TRUE, fname);
 
     if (n != NULL)
 	return n->value.n_cmd;
-#else
-#if BINARY_SEARCH_IS_BROKEN	/* then use the old linear look-up */
-    NTAB *nptr;			/* pointer to entry in name binding table */
-    size_t len = strlen(fname);
-
-    if (len != 0) {		/* scan through the table, returning any match */
-	nptr = nametbl;
-	while (nptr->n_cmd != NULL) {
-	    if (strncmp(fname, nptr->n_name, len) == 0)
-		return nptr->n_cmd;
-	    ++nptr;
-	}
-    }
 #else
     /* this runs 10 times faster for 'nametbl[]' */
     int lo, hi, cur;
@@ -2194,7 +2378,7 @@ engl2fnc(const char *fname)
 
 	/* scan through the table, returning any match */
 	lo = 0;
-	hi = nametblsize - 2;	/* don't want last entry -- it's NULL */
+	hi = nametbl_size - 2;	/* don't want last entry -- it's NULL */
 
 	while (lo <= hi) {
 	    cur = (lo + hi) >> 1;
@@ -2212,7 +2396,6 @@ engl2fnc(const char *fname)
 	    }
 	}
     }
-#endif /* binary vs linear */
 #endif /* OPT_NAMEBST */
     return NULL;
 }
@@ -2260,7 +2443,7 @@ decode_prefix(const char *kk, UINT * prefix)
 		    bits = CTLX;
 		if (isCntrl(poundc) && ch == (UCHAR) toalpha(poundc))
 		    bits = SPEC;
-	    } else if (!strncmp(kk, "FN", 2)) {
+	    } else if (!strncmp(kk, "FN", (size_t) 2)) {
 		bits = SPEC;
 	    }
 	    if (bits != 0)
@@ -2299,13 +2482,13 @@ prc2kcod(const char *kk)
 		break;
 	    if (s == kk || s[1] == '\n' || s[1] == '\0')
 		break;
-	    mklower(vl_strncpy(temp, kk, len + 1));
+	    mklower(vl_strncpy(temp, kk, (size_t) (len + 1)));
 	    if (isLower(temp[0]))
 		temp[0] = (char) toUpper(temp[0]);
 	    for (n = 0; n < TABLESIZE(key_modifiers); ++n) {
-		if (!strncmp(key_modifiers[n].name, temp, len)) {
+		if (!strncmp(key_modifiers[n].name, temp, (size_t) len)) {
 		    found = TRUE;
-		    pref |= (key_modifiers[n].code | mod_KEY);
+		    pref |= ((UINT) key_modifiers[n].code | mod_KEY);
 		    break;
 		}
 	    }
@@ -2331,7 +2514,7 @@ prc2kcod(const char *kk)
     }
 #endif
 
-    if (strlen(kk) > 2 && !strncmp(kk, "M-", 2)) {
+    if (strlen(kk) > 2 && !strncmp(kk, "M-", (size_t) 2)) {
 	pref |= HIGHBIT;
 	kk += 2;
     }
@@ -2339,7 +2522,7 @@ prc2kcod(const char *kk)
     if (*kk == '^' && kk[1] != EOS) {	/* control character */
 	kcod = (UCHAR) kk[1];
 	if (isLower(kcod))
-	    kcod = toUpper(kcod);
+	    kcod = (UINT) toUpper(kcod);
 	kcod = tocntrl(kcod);
 	kk += 2;
     } else {			/* any single char, control or not */
@@ -2378,17 +2561,18 @@ prc2engl(const char *kk)
  * we will splice calls.
  */
 char *
-kbd_engl(const char *prompt, char *buffer)
+kbd_engl(const char *prompt, char *buffer, UINT which)
 {
-    if (kbd_engl_stat(prompt, buffer, 0) == TRUE)
-	return buffer;
-    return NULL;
+    char *result = NULL;
+    if (kbd_engl_stat(prompt, buffer, which, 0) == TRUE)
+	result = buffer;
+    return result;
 }
 
 /* sound the alarm! */
 #if OPT_TRACE
 void
-trace_alarm(char *file, int line)
+trace_alarm(const char *file, int line)
 #else
 void
 kbd_alarm(void)
@@ -2407,22 +2591,47 @@ kbd_alarm(void)
     warnings++;
 }
 
+void
+kbd_start(KBD_DATA * data)
+{
+    beginDisplay();
+
+#if OPT_MULTIBYTE
+    make_local_b_val(bminip, VAL_FILE_ENCODING);
+    if (cmd_encoding == enc_AUTO) {
+	set_b_val(bminip, VAL_FILE_ENCODING, b_val(curbp, VAL_FILE_ENCODING));
+    } else {
+	set_b_val(bminip, VAL_FILE_ENCODING, cmd_encoding);
+    }
+#endif
+
+    data->save_bp = curbp;
+    data->save_wp = curwp;
+    data->save_mk = MK;
+
+    curbp = bminip;
+    curwp = wminip;
+    MK = DOT;
+}
+
+void
+kbd_finish(KBD_DATA * data)
+{
+    curbp = data->save_bp;
+    curwp = data->save_wp;
+    MK = data->save_mk;
+
+    endofDisplay();
+}
+
 /* put a character to the keyboard-prompt */
 int
 kbd_putc(int c)
 {
-    BUFFER *savebp;
-    WINDOW *savewp;
-    MARK savemk;
+    KBD_DATA save;
     int status;
 
-    beginDisplay();
-    savebp = curbp;
-    savewp = curwp;
-    curbp = bminip;
-    curwp = wminip;
-    savemk = MK;
-    MK = DOT;
+    kbd_start(&save);
     if ((kbd_expand <= 0) && isreturn(c)) {
 	status = kbd_erase_to_end(0);
     } else {
@@ -2435,10 +2644,7 @@ kbd_putc(int c)
 	TRACE(("mini:%2d:%s\n", llength(DOT.l), lp_visible(DOT.l)));
 #endif
     }
-    curbp = savebp;
-    curwp = savewp;
-    MK = savemk;
-    endofDisplay();
+    kbd_finish(&save);
 
     return status;
 }
@@ -2455,65 +2661,41 @@ kbd_puts(const char *s)
 void
 kbd_erase(void)
 {
-    BUFFER *savebp;
-    WINDOW *savewp;
-    MARK savemk;
+    if (vl_echo && !clhide) {
+	KBD_DATA save;
 
-    if (!vl_echo || clhide)
-	return;
-
-    beginDisplay();
-    savebp = curbp;
-    savewp = curwp;
-    curbp = bminip;
-    curwp = wminip;
-    savemk = MK;
-    MK = DOT;
-    if (DOT.o > 0) {
-	backchar_to_bol(TRUE, 1);
-	ldel_chars(1, FALSE);
-    }
+	kbd_start(&save);
+	if (DOT.o > 0) {
+	    backchar_to_bol(TRUE, 1);
+	    ldel_chars((B_COUNT) 1, FALSE);
+	}
 #ifdef VILE_DEBUG
-    TRACE(("MINI:%2d:%s\n", llength(DOT.l), lp_visible(DOT.l)));
+	TRACE(("MINI:%2d:%s\n", llength(DOT.l), lp_visible(DOT.l)));
 #endif
-    curbp = savebp;
-    curwp = savewp;
-    MK = savemk;
-    endofDisplay();
+	kbd_finish(&save);
+    }
 }
 
 int
 kbd_erase_to_end(int column)
 {
-    BUFFER *savebp;
-    WINDOW *savewp;
-    MARK savemk;
-
     if (vl_echo && !clhide) {
-	beginDisplay();
-	savebp = curbp;
-	savewp = curwp;
-	curbp = bminip;
-	curwp = wminip;
-	savemk = MK;
-	MK = DOT;
+	KBD_DATA save;
+
+	kbd_start(&save);
 	if (llength(DOT.l) > 0) {
 	    DOT.o = column;
 	    if (llength(DOT.l) > DOT.o)
-		ldel_bytes(llength(DOT.l) - DOT.o, FALSE);
+		ldel_bytes((B_COUNT) (llength(DOT.l) - DOT.o), FALSE);
 	    TRACE(("NULL:%2d:%s\n", llength(DOT.l), lp_visible(DOT.l)));
 	}
-	curbp = savebp;
-	curwp = savewp;
-	MK = savemk;
-	endofDisplay();
+	kbd_finish(&save);
     }
 
     return TRUE;
 }
 
-#if OPT_CASELESS
-static int
+int
 cs_strcmp(
 	     int case_insensitive,
 	     const char *s1,
@@ -2524,7 +2706,7 @@ cs_strcmp(
     return strcmp(s1, s2);
 }
 
-static int
+int
 cs_strncmp(
 	      int case_insensitive,
 	      const char *s1,
@@ -2535,7 +2717,6 @@ cs_strncmp(
 	return strnicmp(s1, s2, n);
     return strncmp(s1, s2, n);
 }
-#endif /* OPT_CASELESS */
 
 /* definitions for name-completion */
 #define	NEXT_DATA(p)	((p)+size_entry)
@@ -2551,7 +2732,7 @@ cs_strncmp(
 /*ARGSUSED*/
 static const char *
 skip_partial(
-		int case_insensitive GCC_UNUSED,
+		int case_insensitive,
 		char *buf,
 		size_t len,
 		const char *table,
@@ -2656,7 +2837,7 @@ makecmpllist(int case_insensitive, void *cinfop)
 	    maxlen = l;
     }
 
-    slashcol = (int) (pathleaf(buf) - buf);
+    slashcol = (size_t) (pathleaf(buf) - buf);
     if (slashcol != 0) {
 	char b[NLINE];
 	(void) strncpy(b, buf, slashcol);
@@ -2671,7 +2852,7 @@ makecmpllist(int case_insensitive, void *cinfop)
     if (cmplcols == 0)
 	cmplcols = 1;
 
-    nentries = (int) ((last - first) / size_entry);
+    nentries = (int) ((size_t) (last - first) / size_entry);
     cmplrows = nentries / cmplcols;
     cmpllen = term.cols / cmplcols;
     if (cmplrows * cmplcols < nentries)
@@ -2681,7 +2862,9 @@ makecmpllist(int case_insensitive, void *cinfop)
 	for (j = 0; j < cmplcols; j++) {
 	    int idx = cmplrows * j + i;
 	    if (idx < nentries) {
-		const char *s = THIS_NAME(first + (idx * size_entry)) + slashcol;
+		const char *s = (THIS_NAME(first
+					   + ((size_t) idx * size_entry))
+				 + slashcol);
 		if (j == cmplcols - 1)
 		    bprintf("%s\n", s);
 		else
@@ -2782,7 +2965,7 @@ popdown_completions(const char *old_bname, WINDOW *old_wp)
  */
 static size_t
 fill_partial(
-		int case_insensitive GCC_UNUSED,
+		int case_insensitive,
 		char *buf,
 		size_t pos,
 		const char *first,
@@ -2799,6 +2982,7 @@ fill_partial(
 	   TRACE_NULL(THIS_NAME(first)),
 	   TRACE_NULL(THIS_NAME(last))));
 
+    assert(buf != 0);
 #if 0				/* case insensitive reply correction doesn't work reliably yet */
     if (!clexec && case_insensitive) {
 	int spos = pos;
@@ -2827,7 +3011,7 @@ fill_partial(
 
 	/* scan through the candidates */
 	for (p = first; p != last; p = NEXT_DATA(p)) {
-	    if (StrNcmp(&THIS_NAME(p)[n], &buf[n], 1) != 0) {
+	    if (StrNcmp(&THIS_NAME(p)[n], &buf[n], (size_t) 1) != 0) {
 		buf[n] = EOS;
 		if (n == pos
 #if OPT_POPUPCHOICE
@@ -3112,7 +3296,7 @@ eol_command(EOL_ARGS)
  * This procedure is invoked from 'kbd_string()' to setup the command-name
  * completion and query displays.
  */
-static int
+int
 cmd_complete(DONE_ARGS)
 {
     int status;
@@ -3149,18 +3333,18 @@ cmd_complete(DONE_ARGS)
 }
 
 int
-kbd_engl_stat(const char *prompt, char *buffer, int stated)
+kbd_engl_stat(const char *prompt, char *buffer, UINT which, int stated)
 {
     KBD_OPTIONS kbd_flags = KBD_EXPCMD | KBD_NULLOK | ((NAMEC != ' ') ? 0 : KBD_MAYBEC);
     int code;
     static TBUFF *temp;
     size_t len = NLINE;
 
+    which_current = which;
+
     tb_scopy(&temp, "");
-#if COMPLETE_FILES
     init_filec(FILECOMPLETION_BufName);
-#endif
-    kbd_flags |= stated;
+    kbd_flags |= (KBD_OPTIONS) stated;
     code = kbd_reply(
 			prompt,	/* no-prompt => splice */
 			&temp,	/* in/out buffer */
@@ -3171,27 +3355,36 @@ kbd_engl_stat(const char *prompt, char *buffer, int stated)
     if (len > tb_length(temp))
 	len = tb_length(temp);
     strncpy0(buffer, tb_values(temp), len);
+
+    which_current = 0;
     return code;
 }
 
 #if OPT_NAMEBST
 int
-insert_namebst(const char *name, const CMDFUNC * cmd, int ro)
+insert_namebst(const char *name, const CMDFUNC * cmd, int ro, UINT which)
 {
     int result;
 
+    TRACE2((T_CALLED "insert_namebst(%s,%s)\n", name, ro ? "ro" : "rw"));
     if (name != 0) {
+	BI_TREE *my_bst = bst_pointer(which);
 	BI_DATA temp, *p;
 
-	if ((p = btree_search(&namebst, name)) != 0) {
+	if ((p = btree_search(my_bst, name)) != 0) {
 	    if ((p->n_flags & NBST_READONLY)) {
-		if (btree_insert(&redefns, p) == 0
-		    || !btree_delete(&namebst, name)) {
-		    return FALSE;
+		if (btree_insert(&redefns, p) == 0) {
+		    returnCode(FALSE);
+		} else {
+		    if (!btree_delete(my_bst, name)) {
+			returnCode(FALSE);
+		    }
 		}
 		mlwrite("[Redefining builtin '%s']", name);
-	    } else if (!delete_namebst(name, TRUE)) {
-		return FALSE;
+	    } else {
+		if (!delete_namebst(name, TRUE, TRUE, which)) {
+		    returnCode(FALSE);
+		}
 	    }
 	}
 
@@ -3199,11 +3392,11 @@ insert_namebst(const char *name, const CMDFUNC * cmd, int ro)
 	temp.n_cmd = cmd;
 	temp.n_flags = (UCHAR) (ro ? NBST_READONLY : 0);
 
-	result = (btree_insert(&namebst, &temp) != 0);
+	result = (btree_insert(my_bst, &temp) != 0);
     } else {
 	result = FALSE;
     }
-    return result;
+    return2Code(result);
 }
 
 static void
@@ -3246,44 +3439,48 @@ remove_cmdfunc_ref(BINDINGS * bs, const CMDFUNC * cmd)
  * Lookup a name in the binary-search tree, remove it if found
  */
 int
-delete_namebst(const char *name, int release)
+delete_namebst(const char *name, int release, int redefining, UINT which)
 {
-    BI_DATA *p = btree_search(&namebst, name);
-    int code;
+    BI_TREE *my_bst = bst_pointer(which);
+    BI_DATA *p = btree_search(my_bst, name);
+    int code = TRUE;
 
+    TRACE((T_CALLED "delete_namebst(%s,%d) %p\n", name, release, (void *) p));
     /* not a named procedure */
-    if (!p)
-	return TRUE;
+    if ((p = btree_search(my_bst, name)) != 0) {
 
-    /* we may have to free some stuff */
-    if (p && release) {
-	remove_cmdfunc_ref(&dft_bindings, p->n_cmd);
-	remove_cmdfunc_ref(&ins_bindings, p->n_cmd);
-	remove_cmdfunc_ref(&cmd_bindings, p->n_cmd);
+	/* we may have to free some stuff */
+	if (p && release) {
+	    remove_cmdfunc_ref(&dft_bindings, p->n_cmd);
+	    remove_cmdfunc_ref(&ins_bindings, p->n_cmd);
+	    remove_cmdfunc_ref(&cmd_bindings, p->n_cmd);
 
-	/* free stuff */
+	    /* free stuff */
 #if OPT_PERL
-	if (p->n_cmd->c_flags & CMD_PERL)
-	    perl_free_sub(CMD_U_PERL(p->n_cmd));
+	    if (p->n_cmd->c_flags & CMD_PERL)
+		perl_free_sub(CMD_U_PERL(p->n_cmd));
 #endif
 
-	if (!(p->n_flags & NBST_READONLY)) {
-	    beginDisplay();
-	    free(TYPECAST(char, p->n_cmd->c_help));
-	    free(TYPECAST(char, p->n_cmd));
-	    endofDisplay();
+	    if (!(p->n_flags & NBST_READONLY)) {
+		beginDisplay();
+		free(TYPECAST(char, p->n_cmd->c_help));
+		free(TYPECAST(char, p->n_cmd));
+		endofDisplay();
+	    }
+	    p->n_cmd = 0;	/* ...so old_namebst won't free this too */
 	}
-	p->n_cmd = 0;		/* ...so old_namebst won't free this too */
-    }
 
-    if ((code = btree_delete(&namebst, name)) == TRUE) {
-	if ((p = btree_search(&redefns, name)) != 0) {
-	    mlwrite("[Restoring builtin '%s']", name);
-	    code = (btree_insert(&namebst, p) != 0);
-	    (void) btree_delete(&redefns, p->bi_key);
+	if ((code = btree_delete(my_bst, name)) == TRUE) {
+	    if (!redefining) {
+		if ((p = btree_search(&redefns, name)) != 0) {
+		    mlwrite("[Restoring builtin '%s']", name);
+		    code = (btree_insert(my_bst, p) != 0);
+		    (void) btree_delete(&redefns, p->bi_key);
+		}
+	    }
 	}
     }
-    return code;
+    returnCode(code);
 }
 
 /*
@@ -3292,33 +3489,35 @@ delete_namebst(const char *name, int release)
  * name-completions.
  */
 int
-rename_namebst(const char *oldname, const char *newname)
+rename_namebst(const char *oldname, const char *newname, UINT which)
 {
     BI_DATA *prior;
     char name[NBUFN];
 
     /* not a named procedure */
-    if ((prior = btree_search(&namebst, oldname)) == 0)
+    if ((prior = btree_search(bst_pointer(which), oldname)) == 0)
 	return TRUE;
 
     /* remove the entry if the new name is not a procedure (bracketed) */
     if (!is_scratchname(newname))
-	return delete_namebst(oldname, TRUE);
+	return delete_namebst(oldname, TRUE, FALSE, which);
 
     /* add the new name */
     strip_brackets(name, newname);
-    if ((insert_namebst(name, prior->n_cmd,
-			prior->n_flags & NBST_READONLY)) != TRUE)
+    if ((insert_namebst(name,
+			prior->n_cmd,
+			prior->n_flags & NBST_READONLY,
+			which)) != TRUE)
 	return FALSE;
 
     /* delete the old (but don't free the data) */
-    return delete_namebst(oldname, FALSE);
+    return delete_namebst(oldname, FALSE, FALSE, which);
 }
 
 int
-search_namebst(const char *name)
+search_namebst(const char *name, UINT which)
 {
-    return (btree_search(&namebst, name) != 0);
+    return (btree_search(bst_pointer(which), name) != 0);
 }
 
 /*
@@ -3326,11 +3525,12 @@ search_namebst(const char *name)
  * do this in a binary-search manner to get a balanced tree.
  */
 void
-build_namebst(const NTAB * nptr, int lo, int hi)
+build_namebst(const NTAB * nptr, int lo, int hi, UINT which)
 {
-    for (; lo < hi; lo++)
-	if (!insert_namebst(nptr[lo].n_name, nptr[lo].n_cmd, TRUE))
+    for (; lo < hi; lo++) {
+	if (!insert_namebst(nptr[lo].n_name, nptr[lo].n_cmd, TRUE, which))
 	    tidy_exit(BADEXIT);
+    }
 }
 
 /*
@@ -3338,20 +3538,21 @@ build_namebst(const NTAB * nptr, int lo, int hi)
  * contents.
  */
 static int
-kbd_complete_bst(
-		    unsigned flags GCC_UNUSED,
-		    int c,	/* TESTC, NAMEC or isreturn() */
-		    char *buf,
-		    size_t *pos)
+kbd_complete_bst(unsigned flags,
+		 int c,		/* TESTC, NAMEC or isreturn() */
+		 char *buf,
+		 size_t *pos)
 {
-    unsigned cpos = (unsigned) *pos;
+    size_t cpos = *pos;
     int status = FALSE;
     const char **nptr;
 
+    (void) flags;
+
     kbd_init();			/* nothing to erase */
     buf[cpos] = EOS;		/* terminate it for us */
 
-    if ((nptr = btree_parray(&namebst, buf, cpos)) != 0) {
+    if ((nptr = btree_parray(bst_current(), buf, cpos)) != 0) {
 	status = kbd_complete(0, c, buf, pos, (const char *) nptr,
 			      sizeof(*nptr));
 	beginDisplay();
@@ -3443,6 +3644,7 @@ free_ext_bindings(BINDINGS * bs)
 void
 bind_leaks(void)
 {
+    btree_printf(&namebst);
 #if OPT_REBIND
     free_all_bindings(&dft_bindings);
     free_ext_bindings(&ins_bindings);
@@ -3452,6 +3654,7 @@ bind_leaks(void)
 #if OPT_NAMEBST
     btree_freeup(&redefns);
     btree_freeup(&namebst);
+    btree_freeup(&glbsbst);
 #endif
 }
 #endif /* NO_LEAKS */
