@@ -5,7 +5,7 @@
  * keys. Like everyone else, they set hints
  * for the display system.
  *
- * $Header: /usr/build/vile/vile/RCS/buffer.c,v 1.340 2009/04/02 23:58:18 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/buffer.c,v 1.366 2016/07/16 15:11:33 tom Exp $
  *
  */
 
@@ -254,35 +254,38 @@ zotwp(BUFFER *bp)
     WINDOW dummy;
     int s = FALSE;
 
-    TRACE((T_CALLED "zotwp(%s)\n", bp->b_bname));
+    TRACE((T_CALLED "zotwp(%s)\n", bp ? bp->b_bname : "null"));
 
-    /*
-     * Locate buffer to switch to after deleting windows.  It can't be
-     * the same as the buffer which is being deleted and if it's an
-     * invisible or scratch buffer, it must have been already visible.
-     * From the set of allowable buffers, choose the most recently
-     * used.
-     */
-    for_each_buffer(nbp) {
-	if (nbp != bp && (!b_is_temporary(nbp) || nbp->b_nwnd != 0)) {
-	    if (obp == 0)
-		obp = nbp;
-	    else if (obp->b_last_used < nbp->b_last_used)
-		obp = nbp;
+    if (bp != 0) {
+
+	/*
+	 * Locate buffer to switch to after deleting windows.  It can't be
+	 * the same as the buffer which is being deleted and if it's an
+	 * invisible or scratch buffer, it must have been already visible.
+	 * From the set of allowable buffers, choose the most recently
+	 * used.
+	 */
+	for_each_buffer(nbp) {
+	    if (nbp != bp && (!b_is_temporary(nbp) || nbp->b_nwnd != 0)) {
+		if (obp == 0)
+		    obp = nbp;
+		else if (obp->b_last_used < nbp->b_last_used)
+		    obp = nbp;
+	    }
 	}
-    }
 
-    /* Delete window instances... */
-    for_each_window(wp) {
-	if (wp->w_bufp == bp && valid_window(wheadp->w_wndp)) {
-	    dummy.w_wndp = wp->w_wndp;
-	    s = delwp(wp);
-	    wp = &dummy;
+	/* Delete window instances... */
+	for_each_window(wp) {
+	    if (wp->w_bufp == bp && valid_window(wheadp->w_wndp)) {
+		dummy.w_wndp = wp->w_wndp;
+		s = delwp(wp);
+		wp = &dummy;
+	    }
 	}
-    }
-    if (valid_buffer(obp))
-	s = swbuffer(obp);
+	if (valid_buffer(obp))
+	    s = swbuffer(obp);
 
+    }
     returnCode(s);
 }
 
@@ -416,6 +419,14 @@ run_buffer_hook(void)
 
 #define HIST_SHOW 9
 
+/*
+ * Display the names of the buffers along with their index numbers on the
+ * minibuffer line.  This is normally done in response to "_", i.e.,
+ * "historical-buffer".
+ *
+ * Returns the index of the "#" buffer in the list, so that we can use this in
+ * histbuff0() to select "#" when "_" is repeated (toggled).
+ */
 static int
 hist_show(int lo_limit, int hi_limit, int cycle)
 {
@@ -426,8 +437,9 @@ hist_show(int lo_limit, int hi_limit, int cycle)
 
     TRACE((T_CALLED "hist_show(%d..%d) %d\n", lo_limit, hi_limit, cycle));
 
-    if (!global_g_val(GMDABUFF))
+    if (!global_g_val(GMDABUFF)) {
 	abp = find_alt();
+    }
 
     (void) strcpy(line, "");
     for (i = lo_limit; i <= hi_limit; ++i) {
@@ -437,6 +449,10 @@ hist_show(int lo_limit, int hi_limit, int cycle)
 	if (bp != 0 && bp != curbp) {
 	    if (first < 0)
 		first = j;
+	    if (!global_g_val(GMDABUFF)) {
+		if (abp && abp == bp)
+		    first = j;
+	    }
 	    (void) lsprintf(line + strlen(line), "  %d%s%s %s",
 			    j,
 			    b_is_changed(bp) ? "*" : "",
@@ -468,6 +484,12 @@ bp2any_wp(BUFFER *bp)
 	if (wp->w_bufp == bp)
 	    break;
     }
+    /*
+     * Check for special case where we're running scripts before the screen is
+     * initialized, and want to ensure we do not put buffers into wminip.
+     */
+    if (wp == 0 && !is_vtinit() && wnullp != 0 && wnullp->w_bufp == bp)
+	wp = wnullp;
     return wp;
 }
 
@@ -506,7 +528,7 @@ histbuff0(int f, int n, int this_window)
 		if (++cycle >= hi_limit)
 		    cycle = 0;
 		c = -1;
-	    } else if (isBackTab(c)) {
+	    } else if (isBackTab((UINT) c)) {
 		if (--cycle < 0)
 		    cycle = hi_limit - 1;
 		c = -1;
@@ -529,7 +551,7 @@ histbuff0(int f, int n, int this_window)
 	    mlwarn("[No such buffer.]");
 	    return FALSE;
 	}
-	/* first assume its a buffer name, then a file name */
+	/* first assume it is a buffer name, then a file name */
 	if ((bp = find_b_name(bufn)) == NULL)
 	    return getfile(bufn, TRUE);
 
@@ -772,7 +794,7 @@ init_bname_cmpl(void)
     char **list = 0;
     BUFFER *bp;
 
-    if ((count = countBuffers()) != 0) {
+    if ((count = (size_t) countBuffers()) != 0) {
 	beginDisplay();
 	if ((list = typeallocn(char *, count + 1)) != 0) {
 	    for_each_buffer(bp) {
@@ -794,7 +816,7 @@ init_bname_cmpl(void)
     return list;
 }
 
-static int
+int
 bname_complete(DONE_ARGS)
 {
     size_t cpos = *pos;
@@ -946,6 +968,12 @@ popdown_buffer(int f GCC_UNUSED, int n GCC_UNUSED)
 }
 #endif /* !SMALLER */
 
+void
+set_last_bp(BUFFER *bp)
+{
+    last_bp = bp;
+}
+
 /* switch back to the first buffer (i.e., ":rewind") */
 /* ARGSUSED */
 int
@@ -953,7 +981,7 @@ firstbuffer(int f GCC_UNUSED, int n GCC_UNUSED)
 {
     int s = histbuff(TRUE, 0);
     if (!global_g_val(GMDABUFF))
-	last_bp = s ? curbp : 0;
+	set_last_bp(s ? curbp : 0);
     return s;
 }
 
@@ -964,6 +992,7 @@ nextbuffer(int f GCC_UNUSED, int n GCC_UNUSED)
 {
     BUFFER *bp;
     BUFFER *stopatbp;
+    int result = FALSE;
 
     TRACE((T_CALLED "nextbuffer()\n"));
 
@@ -974,28 +1003,34 @@ nextbuffer(int f GCC_UNUSED, int n GCC_UNUSED)
 	    bp = bheadp;
 	    while (bp != 0 && bp->b_bufp != stopatbp)
 		bp = bp->b_bufp;
-	    /* if that one's invisible, back up and try again */
-	    if (b_is_invisible(bp))
-		stopatbp = bp;
-	    else
-		returnCode(swbuffer(bp));
+	    if (bp == 0)
+		break;
+	    stopatbp = bp;
+	    /* if that one's invisible, keep trying */
+	    if (!b_is_invisible(bp))
+		break;
 	}
+	/* we're back to the top -- they were all invisible */
+	result = swbuffer(stopatbp);
     } else {			/* go forward thru args-list */
-	if (curbp == 0)
-	    (void) find_nth_created(1);
-	if (last_bp == 0)
-	    last_bp = find_b_hist(0);
-	if (last_bp != 0) {
+	if (last_bp == 0) {
+	    set_last_bp(curbp);
+	}
+	if ((bp = last_bp) != 0) {
 	    for (bp = last_bp->b_bufp; bp; bp = bp->b_bufp) {
-		if (b_is_argument(bp))
-		    returnCode(swbuffer(last_bp = bp));
+		if (b_is_argument(bp)) {
+		    break;
+		}
 	    }
 	}
-	mlforce("[No more files to edit]");
-	returnCode(FALSE);
+	if (bp != 0) {
+	    result = swbuffer(bp);
+	} else {
+	    mlforce("[No more files to edit]");
+	    result = FALSE;
+	}
     }
-    /* we're back to the top -- they were all invisible */
-    returnCode(swbuffer(stopatbp));
+    returnCode(result);
 }
 
 #if !SMALLER
@@ -1006,6 +1041,7 @@ prevbuffer(int f GCC_UNUSED, int n GCC_UNUSED)
 {
     BUFFER *bp;			/* eligible buffer to switch to */
     BUFFER *stopatbp;		/* eligible buffer to switch to */
+    int result = FALSE;
 
     TRACE((T_CALLED "prevbuffer()\n"));
 
@@ -1016,18 +1052,26 @@ prevbuffer(int f GCC_UNUSED, int n GCC_UNUSED)
 	while (valid_buffer(bp) && (bp = bp->b_bufp) != 0) {
 	    /* get the next buffer in the list */
 	    /* if that one's invisible, skip it and try again */
-	    if (!b_is_invisible(bp))
-		returnCode(swbuffer(bp));
+	    if (!b_is_invisible(bp)) {
+		stopatbp = bp;
+		break;
+	    }
 	}
+	/* we're back to the top -- they were all invisible */
+	result = swbuffer(stopatbp);
     } else {			/* go backward thru args-list */
-	if (curbp == 0)
-	    (void) find_nth_created(1);
-	else if (find_nth_created(curbp->b_created - 1) == 0)
+	if (curbp == 0) {
+	    bp = find_nth_created(1);
+	} else {
+	    bp = find_nth_created(curbp->b_created - 1);
+	}
+	if (bp != 0) {
+	    result = swbuffer(bp);
+	} else {
 	    mlforce("[No more files to edit]");
-	returnCode(FALSE);
+	}
     }
-    /* we're back to the top -- they were all invisible */
-    returnCode(swbuffer(stopatbp));
+    returnCode(result);
 }
 #endif /* !SMALLER */
 
@@ -1140,17 +1184,17 @@ found_modeline(LINE *lp, int *first, int *last)
 	limit = modeline_limit;
 
     for (n = 0; n < TABLESIZE(mls_patterns); ++n) {
-	regexp *prog = mls_regcomp(n);
+	regexp *prog = mls_regcomp((int) n);
 	if (lregexec(prog, lp, 0, limit)) {
 	    int j = mls_patterns[n].mark;
-	    *first = prog->startp[j] - prog->startp[0];
-	    *last = prog->endp[j] - prog->startp[0];
+	    *first = (int) (prog->startp[j] - prog->startp[0]);
+	    *last = (int) (prog->endp[j] - prog->startp[0]);
 
 	    /*
 	     * Real vi may have modes that we do not want to set.
 	     */
 	    if (prog->endp[1] - prog->startp[1] == 2
-		&& !strncmp(prog->startp[1], "vi", 2))
+		&& !strncmp(prog->startp[1], "vi", (size_t) 2))
 		rc = 2;
 	    else
 		rc = 1;
@@ -1174,7 +1218,7 @@ do_one_modeline(LINE *lp, int vi, int first, int last)
 	while (last > first && isSpace(lgetc(lp, last - 1)))
 	    --last;
 
-	tb_bappend(&data, lvalue(lp) + first, last - first);
+	tb_bappend(&data, lvalue(lp) + first, (size_t) (last - first));
 	tb_append(&data, EOS);
 
 	in_modeline = vi;
@@ -1189,13 +1233,25 @@ do_one_modeline(LINE *lp, int vi, int first, int last)
 void
 do_modelines(BUFFER *bp)
 {
+    WINDOW *wp;
+
     if (b_val(bp, MDMODELINE) &&
-	b_val(bp, VAL_MODELINES) > 0) {
+	b_val(bp, VAL_MODELINES) > 0 &&
+	(wp = bp2any_wp(bp)) != 0) {
 	LINE *lp;
 	int once = b_val(bp, VAL_MODELINES);
 	int limit = 2 * once;
 	int count;
 	int rc, first = 0, last = 0;
+	BUFFER *save_bp = curbp;
+	WINDOW *save_wp = curwp;
+
+	/*
+	 * If we're sourcing a file, then curbp is not necessarily the same.
+	 * Fix that.
+	 */
+	curwp = wp;
+	curbp = bp;
 
 	bsizes(bp);
 
@@ -1221,6 +1277,8 @@ do_modelines(BUFFER *bp)
 		do_one_modeline(lp, rc, first, last);
 	    }
 	}
+	curbp = save_bp;
+	curwp = save_wp;
 	returnVoid();
     }
 }
@@ -1270,132 +1328,146 @@ swbuffer_lfl(BUFFER *bp, int lockfl, int this_window)
 
     if (!bp) {
 	mlforce("BUG:  swbuffer passed null bp");
-	returnCode(FALSE);
-    }
-
-    TRACE(("swbuffer(%s) nwnd=%d\n", bp->b_bname, bp->b_nwnd));
-    if (curbp == bp
-	&& valid_window(curwp)
-	&& DOT.l != 0
-	&& curwp->w_bufp == bp) {	/* no switching to be done */
-
-	if (!b_is_reading(bp) &&
-	    !bp->b_active)	/* on second thought, yes there is */
-	    status = suckitin(bp, TRUE, lockfl);
+	status = FALSE;
     } else {
-#if !WINMARK
-	/* Whatever else we do, make sure MK isn't bogus when we leave */
-	MK = nullmark;
-#endif
+	TRACE(("swbuffer(%s) nwnd=%d\n", bp->b_bname, bp->b_nwnd));
+	if (curbp == bp
+	    && valid_window(curwp)
+	    && DOT.l != 0
+	    && curwp->w_bufp == bp) {	/* no switching to be done */
 
-	if (valid_buffer(curbp)) {
-	    /* if we'll have to take over this window, and it's the last */
-	    if ((this_window || bp->b_nwnd == 0)
-		&& curbp->b_nwnd != 0 && --(curbp->b_nwnd) == 0) {
-#if !SMALLER
-		if (!this_window)
-#endif
-		    undispbuff(curbp, curwp);
-	    } else if (DOT.l != 0 && (this_window || bp->b_nwnd == 0)) {
-		/* Window is still getting taken over so make a
-		   copy of the traits for a possible future
-		   set-window */
-		copy_traits(&(curbp->b_wtraits), &(curwp->w_traits));
+	    if (!b_is_reading(bp) &&
+		!bp->b_active) {	/* on second thought, yes there is */
+		status = suckitin(bp, TRUE, lockfl);
 	    }
-	}
+	} else {
+#if !WINMARK
+	    /* Whatever else we do, make sure MK isn't bogus when we leave */
+	    MK = nullmark;
+#endif
 
-	/* don't let make_current() call the hook -- there's
-	   more to be done down below */
-	DisableHook(&bufhook);
-	make_current(bp);	/* sets curbp */
-	EnableHook(&bufhook);
+	    if (valid_buffer(curbp)) {
+		/* if we'll have to take over this window, and it's the last */
+		if ((this_window || bp->b_nwnd == 0)
+		    && curbp->b_nwnd != 0 && --(curbp->b_nwnd) == 0) {
+#if !SMALLER
+		    if (!this_window)
+#endif
+			undispbuff(curbp, curwp);
+		} else if (DOT.l != 0 && (this_window || bp->b_nwnd == 0)) {
+		    /* Window is still getting taken over so make a
+		       copy of the traits for a possible future
+		       set-window */
+		    copy_traits(&(curbp->b_wtraits), &(curwp->w_traits));
+		}
+	    }
 
-	bp = curbp;		/* if running the bufhook caused an error,
+	    /* don't let make_current() call the hook -- there's
+	       more to be done down below */
+	    DisableHook(&bufhook);
+	    make_current(bp);	/* sets curbp */
+	    EnableHook(&bufhook);
+
+	    bp = curbp;		/* if running the bufhook caused an error,
 				   we may be in a different buffer than we
 				   thought we were going to */
-
-#if !SMALLER
-	if (this_window) {
-	    /* Put buffer in this window even if it's already on screen */
-	    LINE *lp;
-	    int trait_matches;
-
-	    if (curwp == 0 || curwp->w_bufp == 0)
-		returnCode(FALSE);
-	    else if (curwp->w_bufp == bp)
-		returnCode(TRUE);
-
-	    if (curwp->w_bufp->b_nwnd == 0) {
-		undispbuff(curwp->w_bufp, curwp);
-	    } else {
-		copy_traits(&(curwp->w_bufp->b_wtraits),
-			    &(curwp->w_traits));
+	    if (bp == 0) {
+		mlforce("BUG:  make_current set null bp");
+		status = FALSE;
 	    }
+#if !SMALLER
+	    else if (this_window) {
+		/*
+		 * Put buffer in this window even if it's already on the screen
+		 */
+		LINE *lp;
+		int trait_matches;
 
-	    /* Initialize the window using the saved buffer traits if possible. 
-	     * If they don't pass a sanity check, simply initialize the newest
-	     * window to be a clone of any other window on the buffer, or
-	     * initialize it as we do in popupbuff().
-	     */
-	    trait_matches = 0;
+		if (curwp == 0 || curwp->w_bufp == 0)
+		    returnCode(FALSE);
+		else if (curwp->w_bufp == bp)
+		    returnCode(TRUE);
+
+		if (curwp->w_bufp->b_nwnd == 0) {
+		    undispbuff(curwp->w_bufp, curwp);
+		} else {
+		    copy_traits(&(curwp->w_bufp->b_wtraits),
+				&(curwp->w_traits));
+		}
+
+		/*
+		 * Initialize the window using the saved buffer traits if
+		 * possible.  If they don't pass a sanity check, simply
+		 * initialize the newest window to be a clone of any other
+		 * window on the buffer, or initialize it as we do in
+		 * popupbuff().
+		 */
+		trait_matches = 0;
 #if WINMARK
 #define TRAIT_MATCHES_NEEDED 5
 #else
 #define TRAIT_MATCHES_NEEDED 4
 #endif
-	    for_each_line(lp, bp) {
-		if (lp == bp->b_dot.l && llength(lp) > bp->b_dot.o)
-		    trait_matches++;
+		for_each_line(lp, bp) {
+		    if (lp == bp->b_dot.l && llength(lp) > bp->b_dot.o)
+			trait_matches++;
 #if WINMARK
-		if (lp == bp->b_mark.l && llength(lp) > bp->b_mark.o)
-		    trait_matches++;
+		    if (lp == bp->b_mark.l && llength(lp) > bp->b_mark.o)
+			trait_matches++;
 #endif
-		if (lp == bp->b_lastdot.l && llength(lp) > bp->b_lastdot.o)
-		    trait_matches++;
-		if (lp == bp->b_tentative_lastdot.l && llength(lp) > bp->b_tentative_lastdot.o)
-		    trait_matches++;
-		if (lp == bp->b_wline.l && llength(lp) > bp->b_wline.o)
-		    trait_matches++;
+		    if (lp == bp->b_lastdot.l && llength(lp) > bp->b_lastdot.o)
+			trait_matches++;
+		    if (lp == bp->b_tentative_lastdot.l && llength(lp) > bp->b_tentative_lastdot.o)
+			trait_matches++;
+		    if (lp == bp->b_wline.l && llength(lp) > bp->b_wline.o)
+			trait_matches++;
+		    if (trait_matches == TRAIT_MATCHES_NEEDED)
+			break;
+		}
 		if (trait_matches == TRAIT_MATCHES_NEEDED)
-		    break;
-	    }
-	    if (trait_matches == TRAIT_MATCHES_NEEDED)
-		copy_traits(&(curwp->w_traits), &(bp->b_wtraits));
-	    else if ((wp = bp2any_wp(bp)) == 0)
-		init_window(curwp, bp);
-	    else
-		clone_window(curwp, wp);
+		    copy_traits(&(curwp->w_traits), &(bp->b_wtraits));
+		else if ((wp = bp2any_wp(bp)) == 0)
+		    init_window(curwp, bp);
+		else
+		    clone_window(curwp, wp);
 
-	    curwp->w_bufp = bp;
-	    status = suckitin(bp, (bp->b_nwnd++ == 0), lockfl);
-	} else
+		curwp->w_bufp = bp;
+		status = suckitin(bp, (bp->b_nwnd++ == 0), lockfl);
+	    } else
 #endif /* !SMALLER */
-	    /* get it already on the screen if possible */
-	if (bp->b_nwnd != 0) {	/* then it's on the screen somewhere */
-	    if ((wp = bp2any_wp(bp)) == 0)
-		mlforce("BUG: swbuffer: wp still NULL");
-	    curwp = wp;
-	    upmode();
+		/* get it already on the screen if possible */
+	    if (bp->b_nwnd != 0) {
+		/* then it's on the screen somewhere */
+		if ((wp = bp2any_wp(bp)) == 0)
+		    mlforce("BUG: swbuffer: wp still NULL");
+		curwp = wp;
+		upmode();
 #ifdef MDCHK_MODTIME
-	    (void) check_file_changed(bp, bp->b_fname);
+		(void) check_file_changed(bp, bp->b_fname);
 #endif
 #if OPT_UPBUFF
-	    if (bp != find_BufferList())
-		updatelistbuffers();
+		if (bp != find_BufferList())
+		    updatelistbuffers();
 #endif
-	    run_buffer_hook();
-	    status = (find_bp(bp) != 0);
-	} else if (curwp == 0) {
-	    status = FALSE;	/* we haven't started displaying yet */
-	} else {
-	    /* oh well, suck it into this window */
-	    curwp->w_bufp = bp;
-	    status = suckitin(bp, (bp->b_nwnd++ == 0), lockfl);
+		run_buffer_hook();
+		status = (find_bp(bp) != 0);
+	    } else if (curwp == 0) {
+		status = FALSE;	/* we haven't started displaying yet */
+	    } else {
+		if (!is_vtinit()) {
+		    curwp = wnullp;
+		}
+		/* oh well, suck it into this window */
+		curwp->w_bufp = bp;
+		status = suckitin(bp, (bp->b_nwnd++ == 0), lockfl);
+	    }
 	}
-    }
 #if OPT_TITLE
-    set_editor_title();
+	set_editor_title();
 #endif
+    }
+    if (status == TRUE)
+	set_last_bp(bp);
     returnCode(status);
 }
 
@@ -1416,6 +1488,12 @@ buf_win_sanity(void)
 }
 #endif
 
+int
+eql_bname(BUFFER *bp, const char *name)
+{
+    return (cs_strcmp(global_g_val(GMDFILENAME_IC), bp->b_bname, name) == 0);
+}
+
 void
 undispbuff(BUFFER *bp, WINDOW *wp)
 {
@@ -1483,11 +1561,7 @@ has_C_suffix(BUFFER *bp)
 {
     int s;
     int save = ignorecase;
-#if OPT_CASELESS
-    ignorecase = TRUE;
-#else
-    ignorecase = FALSE;
-#endif
+    ignorecase = global_g_val(GMDFILENAME_IC);
     s = nregexec(global_g_val_rexp(GVAL_CSUFFIXES)->reg,
 		 bp->b_fname, (char *) 0, 0, -1);
     ignorecase = save;
@@ -1501,7 +1575,8 @@ kill_that_buffer(BUFFER *bp)
     int s;
     char bufn[NFILEN];
 
-    (void) strcpy(bufn, bp->b_bname);	/* ...for info-message */
+    /* ...for info-message */
+    (void) vl_strncpy(bufn, bp->b_bname, sizeof(bufn));
     if (buffer_is_solo(bp)) {
 	return FALSE;
     }
@@ -1536,7 +1611,7 @@ make_buffer_list(char *bufn)
     BUFFER *bp;
     int count = 0;
     int limit = countBuffers();
-    char **result = typecallocn(char *, limit + 1);
+    char **result = typecallocn(char *, (size_t) limit + 1);
 
     if (result != 0) {
 #if OPT_FORBUFFERS_CHOICES
@@ -1635,7 +1710,7 @@ for_buffers(int f, int n)
 		    /*
 		     * FIXME: does namedcmd() modify what execstr points to?
 		     */
-		    execstr = strcpy(this_command, command);
+		    execstr = vl_strncpy(this_command, command, sizeof(this_command));
 		    clexec = TRUE;
 
 		    if (namedcmd(f, n))
@@ -1751,7 +1826,7 @@ killbuffer(int f, int n)
 	DOT = save_DOT;
 	DOT.o = getgoal(DOT.l);
 	curwp->w_line = save_TOP;
-	curwp->w_flag &= (USHORT) ~ WFMOVE;
+	clr_typed_flags(curwp->w_flag, USHORT, WFMOVE);
     }
 #endif
     return s;
@@ -1819,7 +1894,6 @@ int
 zotbuf(BUFFER *bp)
 {
     int status;
-    int didswitch = FALSE;
 
     TRACE((T_CALLED "zotbuf(bp=%p)\n", (void *) bp));
 
@@ -1855,14 +1929,10 @@ zotbuf(BUFFER *bp)
     }
 #endif
     /* Blow text away.      */
-    if ((status = bclear(bp)) != TRUE) {
-	/* the user must have answered no */
-	if (didswitch)
-	    (void) swbuffer(bp);
-    } else {
+    if ((status = bclear(bp)) == TRUE) {
 #if OPT_NAMEBST
 	if (tb_values(bp->b_procname) != 0) {
-	    delete_namebst(tb_values(bp->b_procname), TRUE);
+	    delete_namebst(tb_values(bp->b_procname), TRUE, FALSE, 0);
 	    tb_free(&(bp->b_procname));
 	}
 #endif
@@ -1872,6 +1942,14 @@ zotbuf(BUFFER *bp)
     returnCode(status);
 }
 
+BUFFER *
+zotbuf2(BUFFER *bp)
+{
+    if (zotbuf(bp))
+	bp = 0;
+    return bp;
+}
+
 /* Rename a buffer given a buffer pointer and new buffer name. */
 int
 renamebuffer(BUFFER *rbp, char *bufname)
@@ -1882,7 +1960,7 @@ renamebuffer(BUFFER *rbp, char *bufname)
 
     TRACE((T_CALLED "renamebuffer(%s, %s)\n", rbp->b_bname, bufname));
 
-    if (*mktrimmed(strcpy(bufn, bufname)) == EOS)
+    if (*mktrimmed(vl_strncpy(bufn, bufname, sizeof(bufn))) == EOS)
 	returnCode(ABORT);
 
     bp = find_b_name(bufn);
@@ -1897,8 +1975,8 @@ renamebuffer(BUFFER *rbp, char *bufname)
     if (is_scratchname(rbp->b_bname)) {
 	char procname[NBUFN];
 	(void) strip_brackets(procname, rbp->b_bname);
-	if (search_namebst(procname)
-	    && rename_namebst(procname, bufn) != TRUE)
+	if (search_namebst(procname, 0)
+	    && rename_namebst(procname, bufn, 0) != TRUE)
 	    returnCode(ABORT);
     }
 #endif
@@ -1921,7 +1999,7 @@ rename_specific_buffer(BUFFER *bp)
 
     /* prompt for and get the new buffer name */
     do {
-	if (mlreply(prompt, bufn, sizeof(bufn)) != TRUE)
+	if (mlreply(prompt, bufn, (UINT) sizeof(bufn)) != TRUE)
 	    return (FALSE);
 	prompt = "That name's been used.  New name: ";
 	status = renamebuffer(bp, bufn);
@@ -2357,37 +2435,46 @@ add_line_at(BUFFER *bp, LINE *prevp, const char *text, int len)
     LINE *nextp;
     LINE *lp;
     int ntext;
+    int result = FALSE;
 
-    nextp = lforw(prevp);
-    ntext = (len < 0) ? (int) strlen(NONNULL(text)) : len;
-    newlp = lalloc(ntext, bp);
-    if (newlp == 0)
-	return (FALSE);
+    if (len < 0) {
+	if (text != 0) {
+	    result = add_line_at(bp, prevp, text, (int) strlen(text));
+	} else {
+	    result = add_line_at(bp, prevp, "", 0);
+	}
+    } else {
+	nextp = lforw(prevp);
+	ntext = len;
+	newlp = lalloc(ntext, bp);
+	if (newlp != 0) {
 
-    lp = newlp;
-    if (ntext > 0)
-	(void) memcpy(lvalue(lp), text, (size_t) ntext);
+	    lp = newlp;
+	    if (ntext > 0)
+		(void) memcpy(lvalue(lp), text, (size_t) ntext);
 
-    /* try to maintain byte/line counts? */
-    if (b_is_counted(bp)) {
-	if (nextp == buf_head(bp)) {
-	    make_local_b_val(bp, MDNEWLINE);
-	    set_b_val(bp, MDNEWLINE, TRUE);
-	    bp->b_bytecount += (ntext + 1);
-	    bp->b_linecount += 1;
+	    /* try to maintain byte/line counts? */
+	    if (b_is_counted(bp)) {
+		if (nextp == buf_head(bp)) {
+		    set_local_b_val(bp, MDNEWLINE, TRUE);
+		    bp->b_bytecount += (B_COUNT) (ntext + 1);
+		    bp->b_linecount += 1;
 #if !SMALLER			/* tradeoff between codesize & data */
-	    lp->l_number = bp->b_linecount;
+		    lp->l_number = bp->b_linecount;
 #endif
-	} else
-	    b_clr_counted(bp);
+		} else
+		    b_clr_counted(bp);
+	    }
+
+	    set_lforw(prevp, newlp);	/* link into the buffer */
+	    set_lback(newlp, prevp);
+	    set_lback(nextp, newlp);
+	    set_lforw(newlp, nextp);
+
+	    result = TRUE;
+	}
     }
-
-    set_lforw(prevp, newlp);	/* link into the buffer */
-    set_lback(newlp, prevp);
-    set_lback(nextp, newlp);
-    set_lforw(newlp, nextp);
-
-    return (TRUE);
+    return result;
 }
 
 /*
@@ -2421,9 +2508,9 @@ next_buffer_line(const char *bname)
 	return NULL;
 
     /* how much left on the current line? */
-    blen = llength(bp->b_dot.l);
+    blen = (B_COUNT) llength(bp->b_dot.l);
     if (blen > (B_COUNT) bp->b_dot.o)
-	blen -= bp->b_dot.o;
+	blen -= (B_COUNT) bp->b_dot.o;
     else
 	blen = 0;
 
@@ -2501,7 +2588,7 @@ set_bname(BUFFER *bp, const char *name)
     int j, k;
     char *d;
 
-    (void) strncpy0(bp->b_bname, name, NBUFN);
+    (void) strncpy0(bp->b_bname, name, (size_t) NBUFN);
 
     d = bp->b_bname;
     for (j = 0, k = -1; d[j]; j++) {
@@ -2595,9 +2682,8 @@ bfind(const char *bname, UINT bflag)
 #if OPT_ENCRYPT
 	    if (!b_is_temporary(bp)
 		&& cryptkey != 0 && *cryptkey != EOS) {
-		(void) strcpy(bp->b_cryptkey, cryptkey);
-		make_local_b_val(bp, MDCRYPT);
-		set_b_val(bp, MDCRYPT, TRUE);
+		(void) vl_strncpy(bp->b_cryptkey, cryptkey, sizeof(bp->b_cryptkey));
+		set_local_b_val(bp, MDCRYPT, TRUE);
 	    } else
 		bp->b_cryptkey[0] = EOS;
 #endif
@@ -2682,7 +2768,12 @@ bclear(BUFFER *bp)
     if (!b_is_temporary(bp)	/* Not invisible or scratch */
 	&&b_is_changed(bp)) {	/* Something changed    */
 	char ques[30 + NBUFN];
-	(void) strcat(strcpy(ques, "Discard changes to "), bp->b_bname);
+
+	(void) vl_strncat(vl_strncpy(ques,
+				     "Discard changes to ",
+				     sizeof(ques)),
+			  bp->b_bname,
+			  sizeof(ques));
 	if (mlyesno(ques) != TRUE)
 	    return FALSE;
     }
@@ -2700,8 +2791,7 @@ bclear(BUFFER *bp)
 #endif
 
     while ((lp = lforw(buf_head(bp))) != buf_head(bp)) {
-	lremove(bp, lp);
-	lfree(lp, bp);
+	lremove2(bp, lp);
     }
 
     if (bp->b_ulinep != 0) {
@@ -2750,14 +2840,14 @@ bsizes(BUFFER *bp)
     if (valid_buffer(curbp)
 	&& !b_is_counted(bp)) {
 	LINE *lp;		/* current line */
-	B_COUNT ending = len_record_sep(bp);
+	B_COUNT ending = (B_COUNT) len_record_sep(bp);
 	B_COUNT numchars = 0;	/* # of chars in file */
 	L_NUM numlines = 0;	/* # of lines in file */
 
 	/* count chars and lines */
 	for_each_line(lp, bp) {
 	    ++numlines;
-	    numchars += llength(lp) + ending;
+	    numchars += ((B_COUNT) llength(lp) + ending);
 #if !SMALLER			/* tradeoff between codesize & data */
 	    lp->l_number = numlines;
 #endif
@@ -2788,7 +2878,7 @@ bsizes(BUFFER *bp)
  * Mark a buffer 'changed'
  */
 void
-chg_buff(BUFFER *bp, USHORT flag)
+chg_buff(BUFFER *bp, unsigned flag)
 {
     WINDOW *wp;
 
@@ -2810,7 +2900,7 @@ chg_buff(BUFFER *bp, USHORT flag)
 #endif
     for_each_visible_window(wp) {
 	if (wp->w_bufp == bp) {
-	    wp->w_flag |= flag;
+	    wp->w_flag |= (USHORT) flag;
 #ifdef WMDLINEWRAP
 	    /* The change may affect the line-height displayed
 	     * on the screen.  Assume the worst-case.
@@ -2826,7 +2916,7 @@ chg_buff(BUFFER *bp, USHORT flag)
  * Mark a buffer 'unchanged'
  */
 void
-unchg_buff(BUFFER *bp, USHORT flag)
+unchg_buff(BUFFER *bp, unsigned flag)
 {
     WINDOW *wp;
 
@@ -2841,7 +2931,7 @@ unchg_buff(BUFFER *bp, USHORT flag)
 
 	for_each_visible_window(wp) {
 	    if (wp->w_bufp == bp)
-		wp->w_flag |= flag;
+		wp->w_flag |= (USHORT) flag;
 	}
 #if OPT_UPBUFF
 	if (update_on_chg(bp))
@@ -2944,15 +3034,15 @@ writeall(int f, int n, int promptuser, int leaving, int autowriting, int all)
 		char tmp[NFILEN];
 
 		mlforce("\n");
-		outlen = (term.cols - 1) -
-		    (sizeof(SAVE_FAILED_FMT) - 3);
+		outlen = ((term.cols - 1) -
+			  (int) (sizeof(SAVE_FAILED_FMT) - 3));
 		mlforce(SAVE_FAILED_FMT,
 			path_trunc(is_internalname(bp->b_fname)
 				   ? bp->b_bname
 				   : bp->b_fname,
 				   outlen,
 				   tmp,
-				   sizeof(tmp)));
+				   (int) sizeof(tmp)));
 		/* dirtymsgline still TRUE */
 		break;
 #undef SAVE_FAILED_FMT
