@@ -1,5 +1,5 @@
 /*
- * $Id: charsets.c,v 1.66 2009/02/27 00:32:38 tom Exp $
+ * $Id: charsets.c,v 1.79 2015/09/07 00:58:42 tom Exp $
  *
  * see
  http://msdn.microsoft.com/library/default.asp?url=/library/en-us/intl/unicode_42jv.asp
@@ -29,7 +29,7 @@ static const UCHAR mark_UTF32BE[] = { 0x00, 0x00, 0xfe, 0xff };
 typedef struct {
     BOM_CODES code;
     const UCHAR *mark;
-    unsigned size;
+    size_t size;
 } BOM_TABLE;
 
 #define DATA(name) { bom_##name, mark_##name, sizeof(mark_##name) }
@@ -46,8 +46,8 @@ static const BOM_TABLE bom_table[] = {
 
 /******************************************************************************/
 
-static void
-allow_decoder(BUFFER *bp, unsigned need)
+static int
+allow_decoder(BUFFER *bp, size_t need)
 {
     if (need > bp->decode_utf_len) {
 	bp->decode_utf_len = (need + 1) * 2;
@@ -55,10 +55,11 @@ allow_decoder(BUFFER *bp, unsigned need)
 					  bp->decode_utf_buf,
 					  bp->decode_utf_len);
     }
+    return (bp->decode_utf_buf != 0);
 }
 
-static void
-allow_encoder(BUFFER *bp, unsigned need)
+static int
+allow_encoder(BUFFER *bp, size_t need)
 {
     if (need > bp->encode_utf_len) {
 	bp->encode_utf_len = (need + 1) * 2;
@@ -66,6 +67,7 @@ allow_encoder(BUFFER *bp, unsigned need)
 					  bp->encode_utf_buf,
 					  bp->encode_utf_len);
     }
+    return (bp->encode_utf_buf != 0);
 }
 
 int
@@ -147,58 +149,90 @@ vl_conv_to_utf8(UCHAR * target, UINT source, B_COUNT limit)
 }
 
 int
-vl_conv_to_utf32(UINT * target, const char *source, B_COUNT limit)
+vl_check_utf8(const char *source, B_COUNT limit)
 {
-#define CH(n) (UCHAR)((*target) >> ((n) * 8))
     int rc = 0;
     int j;
-    UINT mask = 0;
 
     /*
      * Find the number of bytes we will need from the source.
      */
     if ((*source & 0x80) == 0) {
 	rc = 1;
-	mask = (UINT) * source;
     } else if ((*source & 0xe0) == 0xc0) {
 	rc = 2;
-	mask = (UINT) (*source & 0x1f);
     } else if ((*source & 0xf0) == 0xe0) {
 	rc = 3;
-	mask = (UINT) (*source & 0x0f);
     } else if ((*source & 0xf8) == 0xf0) {
 	rc = 4;
-	mask = (UINT) (*source & 0x07);
     } else if ((*source & 0xfc) == 0xf8) {
 	rc = 5;
-	mask = (UINT) (*source & 0x03);
     } else if ((*source & 0xfe) == 0xfc) {
 	rc = 6;
-	mask = (UINT) (*source & 0x01);
-    }
-
-    if ((B_COUNT) rc > limit) {	/* whatever it is, we cannot decode it */
-	TRACE2(("limit failed %d/%ld in vl_conv_to_utf32\n", rc, limit));
-	rc = 0;
     }
 
     /*
      * sanity-check.
      */
     if (rc > 1) {
-	for (j = 1; j < rc; j++) {
+	int have = rc;
+
+	if ((int) limit < have)
+	    have = (int) limit;
+
+	for (j = 1; j < have; j++) {
 	    if ((source[j] & 0xc0) != 0x80)
 		break;
 	}
-	if (j != rc) {
-	    TRACE2(("check failed %d/%d in vl_conv_to_utf32\n", j, rc));
+	if (j != have) {
+	    TRACE2(("check failed %d/%d in vl_check_utf8\n", j, rc));
 	    rc = 0;
 	}
     }
+    return rc;
+}
+
+int
+vl_conv_to_utf32(UINT * target, const char *source, B_COUNT limit)
+{
+#define CH(n) (UCHAR)((*target) >> ((n) * 8))
+    int rc = vl_check_utf8(source, limit);
+
+    if ((B_COUNT) rc > limit) {	/* whatever it is, we cannot decode it */
+	TRACE2(("limit failed %d/%ld in vl_conv_to_utf32\n", rc, limit));
+	rc = 0;
+    }
 
     if (target != 0) {
+	UINT mask = 0;
+	int j;
 	int shift = 0;
 	*target = 0;
+
+	switch (rc) {
+	case 1:
+	    mask = (UINT) * source;
+	    break;
+	case 2:
+	    mask = (UINT) (*source & 0x1f);
+	    break;
+	case 3:
+	    mask = (UINT) (*source & 0x0f);
+	    break;
+	case 4:
+	    mask = (UINT) (*source & 0x07);
+	    break;
+	case 5:
+	    mask = (UINT) (*source & 0x03);
+	    break;
+	case 6:
+	    mask = (UINT) (*source & 0x01);
+	    break;
+	default:
+	    mask = 0;
+	    break;
+	}
+
 	for (j = 1; j < rc; j++) {
 	    *target |= (UINT) (source[rc - j] & 0x3f) << shift;
 	    shift += 6;
@@ -215,14 +249,14 @@ vl_conv_to_utf32(UINT * target, const char *source, B_COUNT limit)
 }
 
 static const BOM_TABLE *
-find_mark_info(int code)
+find_mark_info(BOM_CODES code)
 {
     const BOM_TABLE *result = 0;
     unsigned n;
 
     for (n = 0; n < TABLESIZE(bom_table); ++n) {
 	const BOM_TABLE *mp = bom_table + n;
-	if (mp->code == (BOM_CODES) code) {
+	if (mp->code == code) {
 	    result = mp;
 	    break;
 	}
@@ -352,25 +386,28 @@ dump_as_utfXX(BUFFER *bp, const char *buf, int nbuf, const char *ending)
     const BOM_TABLE *mp = find_mark_info2(bp);
 
     if (mp != 0 && mp->size > 1) {
-	unsigned j = 0;
-	unsigned k = 0;
-	unsigned need = nbuf + strlen(ending);
-	unsigned lend = strlen(ending);
+	size_t j = 0;
+	size_t k = 0;
+	size_t need = (size_t) nbuf + strlen(ending);
+	size_t lend = strlen(ending);
+
+	if (!allow_encoder(bp, need * mp->size))
+	    goto finish;
+	if (!allow_decoder(bp, need))
+	    goto finish;
 
-	allow_encoder(bp, need * mp->size);
-	allow_decoder(bp, need);
 	while (j < (unsigned) nbuf) {
 	    int skip = vl_conv_to_utf32(bp->decode_utf_buf + k++,
 					buf + j,
-					nbuf - j);
+					(B_COUNT) ((UINT) nbuf - j));
 	    if (skip == 0)
 		goto finish;
-	    j += skip;
+	    j += (UINT) skip;
 	}
 	while (*ending != 0) {
 	    int skip = vl_conv_to_utf32(bp->decode_utf_buf + k++,
 					ending++,
-					lend--);
+					(B_COUNT) (lend--));
 	    if (skip == 0)
 		goto finish;
 	}
@@ -409,7 +446,7 @@ dump_as_utfXX(BUFFER *bp, const char *buf, int nbuf, const char *ending)
 		break;
 	    }
 	}
-	rc = j;
+	rc = (int) j;
     }
   finish:
     return rc;
@@ -421,8 +458,7 @@ set_byteorder_mark(BUFFER *bp, int value)
 {
     if (value != ENUM_UNKNOWN
 	&& value != global_b_val(VAL_BYTEORDER_MARK)) {
-	make_local_b_val(bp, VAL_BYTEORDER_MARK);
-	set_b_val(bp, VAL_BYTEORDER_MARK, value);
+	set_local_b_val(bp, VAL_BYTEORDER_MARK, value);
 
 	TRACE(("set_byteorder_mark for '%s' to %s\n",
 	       bp->b_bname,
@@ -435,8 +471,7 @@ set_encoding(BUFFER *bp, int value)
 {
     if (value != ENUM_UNKNOWN
 	&& value != global_b_val(VAL_FILE_ENCODING)) {
-	make_local_b_val(bp, VAL_FILE_ENCODING);
-	set_b_val(bp, VAL_FILE_ENCODING, value);
+	set_local_b_val(bp, VAL_FILE_ENCODING, value);
 
 	TRACE(("set_encoding for '%s' to %s\n",
 	       bp->b_bname,
@@ -453,13 +488,13 @@ load_as_utf8(BUFFER *bp, LINE *lp)
 
     if (mp != 0 && mp->size > 1) {
 	int pass;
-	unsigned j, k;
-	unsigned need = llength(lp);
-	unsigned used;
+	size_t j, k;
+	size_t need = (size_t) llength(lp);
+	size_t used;
 
 	TRACE2(("load_as_utf8:%d:%s\n", need, lp_visible(lp)));
-	allow_decoder(bp, need);
-	if (bp->decode_utf_buf != 0) {
+	if (allow_decoder(bp, need)) {
+	    rc = TRUE;
 	    if (need) {
 		for (j = k = 0; j < need; ++k) {
 		    UCHAR ch = CH(j);
@@ -480,26 +515,26 @@ load_as_utf8(BUFFER *bp, LINE *lp)
 			break;
 		    case bom_UTF16LE:
 			bp->decode_utf_buf[k] = (CH(j)
-						 + (CH(j + 1) << 8));
+						 + (UINT) (CH(j + 1) << 8));
 			break;
 		    case bom_UTF16BE:
 			bp->decode_utf_buf[k] = (CH(j + 1)
-						 + (CH(j) << 8));
+						 + (UINT) (CH(j) << 8));
 			break;
 		    case bom_UTF32LE:
 			bp->decode_utf_buf[k] = (CH(j + 0)
-						 + (CH(j + 1) << 8)
-						 + (CH(j + 2) << 16)
-						 + (CH(j + 3) << 24));
+						 + (UINT) (CH(j + 1) << 8)
+						 + (UINT) (CH(j + 2) << 16)
+						 + (UINT) (CH(j + 3) << 24));
 			break;
 		    case bom_UTF32BE:
 			bp->decode_utf_buf[k] = (CH(j + 3)
-						 + (CH(j + 2) << 8)
-						 + (CH(j + 1) << 16)
-						 + (CH(j + 0) << 24));
+						 + (UINT) (CH(j + 2) << 8)
+						 + (UINT) (CH(j + 1) << 16)
+						 + (UINT) (CH(j + 0) << 24));
 			break;
 		    }
-		    j += mp->size;
+		    j += (UINT) mp->size;
 		}
 		used = k;
 
@@ -508,10 +543,10 @@ load_as_utf8(BUFFER *bp, LINE *lp)
 		    for (j = k = 0; j < used; ++j) {
 			int nn = vl_conv_to_utf8(buffer,
 						 bp->decode_utf_buf[j],
-						 used + 1 - j);
+						 (B_COUNT) (used + 1 - j));
 			if (buffer != 0)
 			    buffer += nn;
-			k += nn;
+			k += (UINT) nn;
 		    }
 		    if (pass == 1) {
 			TRACE2(("need %d, have %d\n", k, lp->l_size));
@@ -523,21 +558,20 @@ load_as_utf8(BUFFER *bp, LINE *lp)
 			     * of the buffer, do not want to allow undo.  Just
 			     * go ahead and reallocate the line's text buffer.
 			     */
-			    if ((ntext = castalloc(char, k)) == NULL) {
+			    if ((ntext = castalloc(char, (k + 1))) == NULL) {
 				rc = FALSE;
 				break;
 			    }
 			    ltextfree(lp, bp);
 			    lvalue(lp) = ntext;
 			    lp->l_size = k;
-			    llength(lp) = k;
+			    llength(lp) = (int) k;
 			} else {
-			    llength(lp) = k;
+			    llength(lp) = (int) k;
 			}
 		    }
 		}
 	    }
-	    rc = TRUE;
 	} else {
 	    bp->decode_utf_len = 0;
 	}
@@ -557,7 +591,7 @@ remove_crlf_nulls(BUFFER *bp, UCHAR * buffer, B_COUNT * length)
     const BOM_TABLE *mp = find_mark_info2(bp);
     UCHAR mark_cr[4];
     UCHAR mark_lf[4];
-    unsigned marklen = 0;
+    size_t marklen = 0;
 
     if (mp != 0) {
 	memset(mark_cr, 0, sizeof(mark_cr));
@@ -604,13 +638,13 @@ remove_crlf_nulls(BUFFER *bp, UCHAR * buffer, B_COUNT * length)
 		else if (!memcmp(mark_lf, buffer + src, marklen))
 		    skip = '\n';
 		if (skip) {
-		    buffer[dst++] = skip;
+		    buffer[dst++] = (UCHAR) skip;
 		    skip = 0;
 		} else {
 		    memcpy(buffer + dst, buffer + src, marklen);
-		    dst += marklen;
+		    dst += (B_COUNT) marklen;
 		}
-		src += marklen;
+		src += (B_COUNT) marklen;
 	    }
 	    *length = dst;
 	}
@@ -628,19 +662,21 @@ riddled_buffer(const BOM_TABLE * mp, UCHAR * buffer, B_COUNT length)
 {
     int result = 0;
     B_COUNT total = 0;
-    int offset = 0;
-    unsigned j, k;
+    size_t offset = 0;
+    size_t j, k;
 
     if (mp->size && !(mp->size % 2)) {
-	TRACE(("checking if %s / %d-byte\n",
+	TRACE(("checking if %s / %u-byte\n",
 	       byteorder2s(mp->code),
-	       mp->size));
+	       (UINT) mp->size));
 
 	/* Check the line-length.  If it is not a multiple of the pattern
 	 * size, just give up.
 	 */
 	if ((length + offset) % mp->size) {
-	    TRACE(("length %ld vs pattern %d - give up\n", length, mp->size));
+	    TRACE(("length %ld vs pattern %u - give up\n",
+		   length,
+		   (UINT) mp->size));
 	} else {
 	    /*
 	     * Now walk through the line and measure the pattern against it.
@@ -656,11 +692,13 @@ riddled_buffer(const BOM_TABLE * mp, UCHAR * buffer, B_COUNT length)
 		    }
 		}
 		if (found) {
-		    total += mp->size;
+		    total += (B_COUNT) mp->size;
 		}
 	    }
 	}
-	result = (int) (length ? (((100.0 * total) / length)) : 0);
+	result = (int) (length
+			? (((100.0 * (double) total) / (double) length))
+			: 0);
 
 	TRACE(("...%ld/%ld ->%d%%\n", total, length, result));
     }
@@ -668,7 +706,7 @@ riddled_buffer(const BOM_TABLE * mp, UCHAR * buffer, B_COUNT length)
 }
 
 static void
-set_encoding_from_bom(BUFFER *bp, int bom_value)
+set_encoding_from_bom(BUFFER *bp, BOM_CODES bom_value)
 {
     const BOM_TABLE *mp;
     int result;
@@ -713,8 +751,7 @@ set_bom_from_encoding(BUFFER *bp, int enc_value)
 	default:
 	    if (result != ENUM_UNKNOWN
 		&& result != global_b_val(VAL_BYTEORDER_MARK)) {
-		make_local_b_val(bp, VAL_BYTEORDER_MARK);
-		set_b_val(bp, VAL_BYTEORDER_MARK, bom_NONE);
+		set_local_b_val(bp, VAL_BYTEORDER_MARK, bom_NONE);
 	    }
 	    break;
 	}
@@ -778,7 +815,8 @@ decode_bom(BUFFER *bp, UCHAR * buffer, B_COUNT * length)
     }
 
     if (b_val(bp, VAL_BYTEORDER_MARK) > bom_NONE
-	&& (mp = find_mark_info(b_val(bp, VAL_BYTEORDER_MARK))) != 0
+	&& (mp = find_mark_info((BOM_CODES) b_val(bp,
+						  VAL_BYTEORDER_MARK))) != 0
 	&& line_has_mark(mp, buffer, *length)) {
 	for (n = 0; n < *length - mp->size; ++n) {
 	    buffer[n] = buffer[n + mp->size];
@@ -786,9 +824,9 @@ decode_bom(BUFFER *bp, UCHAR * buffer, B_COUNT * length)
 	while (n < *length) {
 	    buffer[n++] = 0;
 	}
-	*length -= mp->size;
+	*length -= (B_COUNT) mp->size;
 
-	set_encoding_from_bom(bp, b_val(bp, VAL_BYTEORDER_MARK));
+	set_encoding_from_bom(bp, (BOM_CODES) b_val(bp, VAL_BYTEORDER_MARK));
 	code = TRUE;
     }
     returnCode(code);
@@ -843,7 +881,7 @@ deduce_charset(BUFFER *bp, UCHAR * buffer, B_COUNT * length, int always)
 	    int check = riddled_buffer(&bom_table[n], buffer, *length);
 	    if (check > match) {
 		match = check;
-		found = n;
+		found = (int) n;
 	    }
 	}
 	if (found > 0 && match >= b_val(bp, VAL_PERCENT_UTF8)) {
@@ -885,13 +923,13 @@ deduce_charset(BUFFER *bp, UCHAR * buffer, B_COUNT * length, int always)
 int
 check_utf8(UCHAR * buffer, B_COUNT length)
 {
-    unsigned n;
+    B_COUNT n;
     int check = TRUE;
     int skip = 0;
     int found;
     UINT target;
 
-    for (n = found = 0; n < length - 1; n += skip) {
+    for (n = 0, found = 0; n < length - 1; n += (B_COUNT) skip) {
 	skip = vl_conv_to_utf32(&target,
 				(char *) (buffer + n),
 				length - n);
@@ -929,8 +967,8 @@ write_bom(BUFFER *bp)
     const BOM_TABLE *mp;
     int status = FIOSUC;
 
-    if ((mp = find_mark_info(b_val(bp, VAL_BYTEORDER_MARK))) != 0) {
-	status = ffputline((const char *) mp->mark, mp->size, NULL);
+    if ((mp = find_mark_info((BOM_CODES) b_val(bp, VAL_BYTEORDER_MARK))) != 0) {
+	status = ffputline((const char *) mp->mark, (int) mp->size, NULL);
     }
     return status;
 }
@@ -961,7 +999,7 @@ chgd_byteorder(BUFFER *bp,
 	       int testing)
 {
     if (!testing && !glob_vals) {
-	set_encoding_from_bom(bp, args->local->vp->i);
+	set_encoding_from_bom(bp, (BOM_CODES) args->local->vp->i);
     }
     return TRUE;
 }
