@@ -12,11 +12,12 @@
 */
 
 /*
- * $Header: /usr/build/vile/vile/RCS/estruct.h,v 1.689 2009/05/17 23:39:11 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/estruct.h,v 1.751 2016/12/17 23:26:12 tom Exp $
  */
 
 #ifndef _estruct_h
 #define _estruct_h 1
+/* *INDENT-OFF* */
 
 #ifdef HAVE_CONFIG_H
 #include <config.h>
@@ -180,7 +181,7 @@
  * probably correct for MSC (Microsoft C) and ZTC (Zortech), but I'm not
  * sure of those.  (It implies a lot of ANSI and POSIX behavior.)
  */
-#if CC_TURBO || CC_WATCOM || CC_MSC || CC_DJGPP || SYS_WINNT || CC_CSETPP || CC_MSVC || CC_LCC_WIN32
+#if CC_TURBO || CC_WATCOM || CC_MSC || CC_DJGPP || SYS_WINNT || CC_CSETPP || CC_MSVC || CC_LCC_WIN32 || SYS_OS2_EMX
 # define CC_NEWDOSCC 1
 #endif
 
@@ -216,6 +217,8 @@
 #if SYS_WINNT
 # define HAVE_ENVIRON		1
 # define HAVE_PUTENV		1
+# define HAVE_STRICMP		1
+# define HAVE_STRINCMP		1
 #endif
 
 /*
@@ -275,9 +278,26 @@
 # endif
 #endif
 
+#if SYS_MSDOS || SYS_OS2 || SYS_VMS || SYS_WINNT
+#ifdef OPT_PERL
+typedef unsigned short	mode_t;
+/* perl's win32.h typedef's mode_t */
+#else
+#define mode_t int
+#endif
+#endif
+
+#if SYS_OS2 && !SYS_OS2_EMX
+#define inline /* nothing */
+#endif
+
+#if SYS_OS2 || SYS_OS2_EMX || SYS_WINNT || SYS_MSDOS
+#define MISSING_EXTERN_CRYPT 1
+#endif
+
 #endif /* HAVE_CONFIG_H */
 
-#if defined(WIN32) || CC_TURBO
+#if !(SYS_CYGWIN) && (defined(WIN32) || CC_TURBO || defined(WIN32_LEAN_AND_MEAN))
 #include "w32vile.h"
 #endif
 
@@ -488,10 +508,6 @@
 #define NO_WIDGETS 0
 #endif
 
-#ifndef OL_WIDGETS
-#define OL_WIDGETS 0
-#endif
-
 /*
  * Constants for ifdef'ing out chunks of code
  */
@@ -508,7 +524,7 @@
 #define	OPT_COLOR (DISP_ANSI || IBM_VIDEO || DISP_TERMCAP || DISP_CURSES || DISP_X11)
 #define	OPT_16_COLOR (IBM_VIDEO || DISP_TERMCAP || DISP_CURSES || DISP_X11)
 
-#define OPT_DUMBTERM (DISP_TERMCAP || DISP_CURSES || DISP_VMSVT)
+#define OPT_DUMBTERM (DISP_TERMCAP || DISP_CURSES || DISP_VMSVT || DISP_X11)
 
 /* Feature turnon/turnoff */
 #define OPT_CACHE_VCOL  !SMALLER /* cache mk_to_vcol() starting point          */
@@ -607,8 +623,7 @@
 
 /* systems with MSDOS-like filename syntax */
 #define OPT_MSDOS_PATH  (SYS_MSDOS || SYS_OS2 || SYS_WINNT || SYS_OS2_EMX)
-#define OPT_CASELESS	(SYS_MSDOS || SYS_OS2 || SYS_WINNT || SYS_OS2_EMX || SYS_CYGWIN || SYS_VMS)
-#define OPT_UNC_PATH	(SYS_WINNT)
+#define OPT_UNC_PATH	(SYS_WINNT || SYS_CYGWIN)
 
 /* individual features that are (normally) controlled by SMALLER */
 #define OPT_AUTOCOLOR	(!SMALLER && OPT_COLOR)	/* autocolor support */
@@ -627,18 +642,20 @@
 #define OPT_HOOKS	!SMALLER		/* read/write hooks, etc. */
 #define OPT_ISO_8859    !SMALLER		/* ISO 8859 characters */
 #define OPT_ISRCH       !SMALLER		/* Incremental searches */
+#define OPT_KEEP_POS    !SMALLER		/* keep-position mode */
 #define OPT_LINEWRAP    !SMALLER		/* line-wrap mode */
 #define OPT_MAJORMODE   !SMALLER		/* majormode support */
 #define OPT_MACRO_ARGS	(!SMALLER && OPT_EVAL)	/* macro argument parsing */
 #define OPT_MLFORMAT    !SMALLER		/* modeline-format */
 #define OPT_MODELINE    !SMALLER		/* vi-style modeline-support */
-#define OPT_MULTIBYTE   (!SMALLER && OPT_LOCALE) /* experimental multibyte support */
+#define OPT_MULTIBYTE   (!SMALLER && OPT_LOCALE) /* multibyte support */
 #define OPT_NAMEBST     !SMALLER		/* name's stored in a bst */
 #define OPT_ONLINEHELP  !SMALLER		/* short per-command help */
 #define OPT_POPUPCHOICE !SMALLER		/* popup-choices mode */
 #define OPT_POPUP_MSGS  !SMALLER		/* popup-msgs mode */
 #define OPT_POSFORMAT   !SMALLER		/* position-format */
 #define OPT_REBIND      !SMALLER		/* permit rebinding of keys at run-time	*/
+#define OPT_REGS_CMPL   !SMALLER		/* name-completion for registers */
 #define OPT_SHOW_WHICH	!SMALLER		/* which-source, etc. */
 #define OPT_TAGS_CMPL   (!SMALLER && OPT_TAGS)	/* name-completion for tags */
 #define OPT_TERMCHRS    !SMALLER		/* set/show-terminal */
@@ -689,7 +706,9 @@
 
 	/* menus */
 #define	OPT_MENUS	(!SMALLER && DISP_X11 && (MOTIF_WIDGETS||ATHENA_WIDGETS))
-#define OPT_MENUS_COLORED 0	/* (MOTIF_WIDGETS && OPT_MENUS) */
+#ifndef OPT_MENUS_COLORED
+#define OPT_MENUS_COLORED 0
+#endif
 
 	/* icons */
 #define OPT_X11_ICON	DISP_X11 /* use compiled-in X icon */
@@ -705,41 +724,37 @@
 #define OWN_SELECTION()	/*EMPTY*/
 #endif
 
-/*
- * Unix vi-style file encryption is available only on some platforms
- */
-#ifdef HAVE_CRYPT
-#define OPT_ENCRYPT     !SMALLER
-#else
-#define OPT_ENCRYPT     0
-#endif
-
 /*
  * Symbols that turn on tables related to OPT_ENUM_MODES in nefsms.h
  */
 #define OPT_COLOR_SCHEMES          (OPT_ENUM_MODES && !SMALLER && OPT_COLOR)
 
 #define OPT_ACCESS_CHOICES         !SMALLER
-#define OPT_BACKUP_CHOICES	   (OPT_ENUM_MODES && OPT_FILEBACK)
-#define OPT_BOOL_CHOICES	   !SMALLER
+#define OPT_BACKUP_CHOICES         (OPT_ENUM_MODES && OPT_FILEBACK)
+#define OPT_BOOL_CHOICES           !SMALLER
 #define OPT_BYTEORDER_MARK_CHOICES OPT_MULTIBYTE
-#define OPT_CHARCLASS_CHOICES	   OPT_SHOW_CTYPE
-#define OPT_COLOR_CHOICES	   (OPT_ENUM_MODES && OPT_COLOR)
+#define OPT_CHARCLASS_CHOICES      OPT_SHOW_CTYPE
+#define OPT_CMD_ENCODING_CHOICES   OPT_MULTIBYTE
+#define OPT_COLOR_CHOICES          (OPT_ENUM_MODES && OPT_COLOR)
 #define OPT_CURTOKENS_CHOICES      OPT_CURTOKENS
 #define OPT_DIRECTIVE_CHOICES      !SMALLER
+#define OPT_ENCRYPT                !SMALLER
 #define OPT_FILE_ENCODING_CHOICES  OPT_MULTIBYTE
 #define OPT_FORBUFFERS_CHOICES     !SMALLER
-#define OPT_HILITE_CHOICES	   (OPT_ENUM_MODES && OPT_HILITEMATCH)
+#define OPT_HILITE_CHOICES         (OPT_ENUM_MODES && OPT_HILITEMATCH)
+#define OPT_KBD_ENCODING_CHOICES   OPT_MULTIBYTE
+#define OPT_KEEP_POS_CHOICES       !SMALLER
 #define OPT_LOOKUP_CHOICES         !SMALLER
 #define OPT_MMQUALIFIERS_CHOICES   OPT_MAJORMODE
 #define OPT_PARAMTYPES_CHOICES     OPT_MACRO_ARGS
 #define OPT_PATH_CHOICES           !SMALLER
-#define OPT_POPUP_CHOICES	   (OPT_ENUM_MODES && OPT_POPUPCHOICE)
+#define OPT_POPUP_CHOICES          (OPT_ENUM_MODES && OPT_POPUPCHOICE)
 #define OPT_READERPOLICY_CHOICES   !SMALLER
 #define OPT_RECORDATTRS_CHOICES    (OPT_ENUM_MODES && SYS_VMS)
 #define OPT_RECORDFORMAT_CHOICES   (OPT_ENUM_MODES && SYS_VMS)
 #define OPT_RECORDSEP_CHOICES      !SMALLER
 #define OPT_SHOWFORMAT_CHOICES     !SMALLER
+#define OPT_TITLE_ENCODING_CHOICES OPT_MULTIBYTE
 #define OPT_VIDEOATTRS_CHOICES     (OPT_ENUM_MODES && OPT_COLOR_SCHEMES)
 #define OPT_VTFLASHSEQ_CHOICES     (OPT_ENUM_MODES && VTFLASH_HOST && OPT_FLASH)
 #define OPT_XCOLORS_CHOICES        (OPT_ENUM_MODES && OPT_EXTRA_COLOR)
@@ -751,7 +766,7 @@
 #define	GLOB_SINGLE	'?'
 #define	GLOB_ELLIPSIS	"..."	/* implemented on VMS-only */
 #define	GLOB_RANGE	"[]"
-#define GLOB_ENVIRON	"$"	/* unimplemented */
+#define	GLOB_ENVIRON	"$"	/* unimplemented */
 #define	GLOB_NEGATE	"^!"
 
 /*
@@ -820,8 +835,11 @@ extern	char *	sys_errlist[];
 #define	set_errno(code)	errno = code
 
 	/* bit-mask definitions */
-#define	lBIT(n)	((ULONG)(1L<<(n)))
-#define	iBIT(n) ((UINT)(1 <<(n)))
+#define	lBIT(n)	((ULONG)(1UL<<(n)))
+#define	iBIT(n) ((UINT)(1U <<(n)))
+
+#define clr_typed_flags(dst,type,flags)  (dst) = (type) ((dst) & ~(flags))
+#define clr_flags(dst,flags)             (dst) = ((dst) & ~(flags))
 
 #if !(defined(HAVE_STRCHR) && defined(HAVE_STRRCHR))
 #define USE_INDEX 1
@@ -836,6 +854,8 @@ extern char *rindex (const char *s, int c);
 #endif
 #endif /* USE_INDEX */
 
+#define vl_index (strchr)
+
 #if defined(STDC_HEADERS) || defined(HAVE_STRING_H)
 # include <string.h>
   /* An ANSI string.h and pre-ANSI memory.h might conflict.  */
@@ -961,10 +981,14 @@ extern int MainProgram(int argc, char *argv[]);
 #endif
 
 	/* semaphore may be needed to prevent interrupt of display-code */
-#if defined(SIGWINCH) || OPT_WORKING
-# if OPT_TRACE > 2
+#if OPT_WORKING && OPT_TRACE
 extern void beginDisplay(void);
 extern void endofDisplay(void);
+#endif
+
+#if defined(SIGWINCH) || OPT_WORKING
+# if OPT_WORKING && (OPT_TRACE > 2)
+	/* no macros */
 # else
 # define beginDisplay() ++im_displaying
 # define endofDisplay() if (im_displaying > 0) --im_displaying; else assert(im_displaying > 0)
@@ -1003,14 +1027,12 @@ extern void endofDisplay(void);
  * we use those definitions to avoid trouble when using OS/2 include
  * files.
  */
-# include <os2def.h>
-#else
-# if !defined(WIN32)
+#  include <os2def.h>
+#elif SYS_CYGWIN || !(defined(WIN32) || defined(WIN32_LEAN_AND_MEAN))
 #  define UCHAR  unsigned char
 #  define UINT   unsigned int
 #  define USHORT unsigned short
 #  define ULONG  unsigned long
-# endif
 #endif
 
 /*	internal constants	*/
@@ -1071,6 +1093,7 @@ extern void endofDisplay(void);
 #define CLIP_KCHR  ';'
 #define SEL_KCHR   '.'
 #define KEYST_KCHR '<'
+#define UNAME_KCHR '@'
 
 #define	NBLOCK	16			/* line block chunk size	*/
 #define MINWLNS	3			/* min # lines, window/screen	*/
@@ -1082,6 +1105,12 @@ extern void endofDisplay(void);
 
 #define MinCBits   8			/* bits in N_chars		*/
 
+#if OPT_MULTIBYTE && (DISP_TERMCAP || DISP_CURSES || DISP_BORLAND)
+#define OPT_VL_OPEN_MBTERM 1		/* uses vl_open_mbterm		*/
+#else
+#define OPT_VL_OPEN_MBTERM 0
+#endif
+
 #if OPT_MULTIBYTE
 #define MaxCBits   16			/* allow UTF-16 internal	*/
 #else
@@ -1112,7 +1141,8 @@ extern void endofDisplay(void);
 #define mod_NOMOD  (~(mod_KEY|mod_SHIFT|mod_CTRL|mod_ALT))
 #endif
 
-#define kcod2key(c)	(int)((c) & (UINT)(iBIT(MaxCBits)-1)) /* strip off the above prefixes */
+#define kcod2key(c)	(int)((UINT)(c) & (UINT)(iBIT(MaxCBits)-1)) /* strip off the above prefixes */
+#define	is8Bits(c)	(((UINT)(c) & (UINT)~0xff) == 0)
 #define	isSpecial(c)	(((UINT)(c) & (UINT)~(iBIT(MaxCBits)-1)) != 0)
 
 #define	char2int(c)	((int)kcod2key(c))	/* mask off sign-extension, etc. */
@@ -1136,14 +1166,16 @@ extern void endofDisplay(void);
 #define isLegalExp(s,x) ((s = (x)) != 0 && !isErrorVal(s))
 
 /* protect against losing namespaces */
+#undef	VL_ERROR
 #undef	FALSE
 #undef	TRUE
 #undef	ABORT
 #undef	SORTOFTRUE
 
-#define FALSE	0			/* False, no, bad, etc.		*/
-#define TRUE	1			/* True, yes, good, etc.	*/
-#define ABORT	2			/* Death, ESC, abort, etc.	*/
+#define VL_ERROR	-1		/* Error			*/
+#define FALSE		0		/* False, no, bad, etc.		*/
+#define TRUE		1		/* True, yes, good, etc.	*/
+#define ABORT		2		/* Death, ESC, abort, etc.	*/
 #define	SORTOFTRUE	3		/* really!	*/
 
 /* keystroke replay states */
@@ -1210,6 +1242,8 @@ typedef enum {
 	, XCOLOR_ENUM
 	, XCOLOR_HYPERTEXT
 	, XCOLOR_ISEARCH
+	, XCOLOR_LINEBREAK
+	, XCOLOR_LINENUMBER
 	, XCOLOR_MODELINE
 	, XCOLOR_NUMBER
 	, XCOLOR_REGEX
@@ -1297,6 +1331,12 @@ typedef enum {
 			          ? enc_8BIT \
 			          : b_val(bp, VAL_FILE_ENCODING)))
 
+typedef enum {
+    	KPOS_VILE = 0
+	, KPOS_NVI
+	, KPOS_VI
+} KEEP_POS_CHOICES;
+
 typedef enum {
 	MMQ_ANY = 0
 	, MMQ_ALL
@@ -1332,6 +1372,7 @@ typedef enum {
 	, PT_FILE
 	, PT_INT
 	, PT_MODE
+	, PT_REG
 	, PT_STR
 	, PT_VAR
 #if OPT_MAJORMODE
@@ -1385,6 +1426,8 @@ typedef enum {
 #define FL_STARTPATH iBIT(6)	/* look in environment $VILE_STARTUP_PATH */
 #define FL_PATH      iBIT(7)	/* look in environment $PATH */
 #define FL_LIBDIR    iBIT(8)	/* look in environment $VILE_LIBDIR_PATH */
+#define FL_ALWAYS    iBIT(9)	/* file is not a script, but data */
+#define FL_INSECURE  iBIT(10)	/* do not care if insecure */
 
 #define FL_ANYWHERE  (FL_CDIR|FL_HOME|FL_EXECDIR|FL_STARTPATH|FL_PATH|FL_LIBDIR)
 
@@ -1397,6 +1440,8 @@ typedef enum {
 #endif
 
 #define LOCATE_EXEC   (FL_PATH | FL_LIBDIR | FL_EXECABLE)
+#define LOCATE_HELP   (FL_READABLE | FL_ALWAYS | FL_ANYWHERE)
+#define LOCATE_TAGS   (FL_READABLE | FL_ALWAYS | FL_CDIR)
 
 /* definitions for name-completion */
 #define	NAMEC		name_cmpl /* char for forcing name-completion */
@@ -1722,8 +1767,16 @@ typedef struct	LINE {
 #define LTRIMMED lBIT(2)	/* line doesn't have newline to display */
 
 /* macros to ease the use of lines */
+#define dot_next_bol()  do { \
+				DOT.l = lforw(DOT.l); \
+				DOT.o = b_left_margin(curbp); \
+			} while (0)
+#define dot_prev_bol()  do { \
+				DOT.l = lback(DOT.l); \
+				DOT.o = b_left_margin(curbp); \
+			} while (0)
 #define	for_each_line(lp,bp) for (lp = lforw(buf_head(bp)); \
-					lp != buf_head(bp); \
+					(lp != 0) && (lp != buf_head(bp)); \
 					lp = lforw(lp))
 
 #define l_nxtundo		l.l_stklnk
@@ -1749,7 +1802,7 @@ typedef struct	LINE {
 #define lputc(lp, n, c)		(lvalue(lp)[(n)]=(char)(c))
 #define lvalue(lp)		((lp)->l_text)
 #define llength(lp)		((lp)->l_used)
-#define line_length(lp)		((B_COUNT)(llength(lp)+len_rs)) /* counting recordsep */
+#define line_length(lp)		((B_COUNT)llength(lp)+len_rs) /* counting recordsep */
 
 #define liscopied(lp)		((lp)->l.l_undo_cookie == current_undo_cookie)
 #define lsetcopied(lp)		((lp)->l.l_undo_cookie = current_undo_cookie)
@@ -1757,12 +1810,12 @@ typedef struct	LINE {
 
 #define lismarked(lp)		((lp)->l.l_flag & LGMARK)
 #define lsetmarked(lp)		((lp)->l.l_flag |= LGMARK)
-#define lsetnotmarked(lp)	((lp)->l.l_flag &= ~LGMARK)
+#define lsetnotmarked(lp)	(clr_typed_flags((lp)->l.l_flag, USHORT, LGMARK))
 #define lflipmark(lp)		((lp)->l.l_flag ^= LGMARK)
 
 #define listrimmed(lp)		((lp)->l.l_flag & LTRIMMED)
 #define lsettrimmed(lp)		((lp)->l.l_flag |= LTRIMMED)
-#define lsetnottrimmed(lp)	((lp)->l.l_flag &= ~LTRIMMED)
+#define lsetnottrimmed(lp)	((lp)->l.l_flag &= (USHORT) ~LTRIMMED)
 #define lsetclear(lp)		((lp)->l.l_flag = (lp)->l.l_undo_cookie = 0)
 
 #define lisreal(lp)		(llength(lp) >= 0)
@@ -1874,9 +1927,9 @@ typedef UCHAR VIDEO_ATTR;
  */
 #define GROW(ptr, type, oldsize, newsize) \
 { \
-	unsigned tmpold = oldsize; \
+	UINT tmpold = (UINT) oldsize; \
 	type *tmpp; \
-	tmpp = typeallocn(type, newsize); \
+	tmpp = typeallocn(type, (UINT) newsize); \
 	if (tmpp == NULL) \
 		return FALSE; \
  \
@@ -1887,7 +1940,7 @@ typedef UCHAR VIDEO_ATTR;
 		tmpold = 0; \
 	} \
 	ptr = tmpp; \
-	(void) memset ((char *)(ptr+tmpold), 0, (newsize - tmpold) * sizeof(type)); \
+	(void) memset ((char *)(ptr+tmpold), 0, ((UINT) newsize - tmpold) * sizeof(type)); \
 }
 
 /*
@@ -2083,8 +2136,7 @@ typedef struct {
 	for (vals = get_submode_vals(bp, n = m); vals != 0; vals = get_submode_valx(bp, m, &n))
 #else
 #define is_c_mode(bp) (b_val(bp,MDCMOD))
-#define fix_cmode(bp,value)	make_local_b_val(bp, MDCMOD), \
-				set_b_val(bp, MDCMOD, value)
+#define fix_cmode(bp,value)	set_local_b_val(bp, MDCMOD, value)
 #define for_each_modegroup(bp,n,m,vals) vals = bp->b_values.bv;
 #endif
 
@@ -2201,9 +2253,9 @@ typedef struct	BUFFER {
 #if OPT_MULTIBYTE
 	BOM_CODES implied_BOM;		/* fix decode/encode if BOM missing */
 	UINT	*decode_utf_buf;	/* workspace for decode_charset() */
-	UINT	decode_utf_len;
+	size_t	decode_utf_len;
 	char	*encode_utf_buf;	/* workspace for encode_charset() */
-	UINT	encode_utf_len;
+	size_t	encode_utf_len;
 #endif
 #if OPT_PERL || OPT_TCL || OPT_PLUGIN
 	void *	b_api_private;		/* pointer to private perl, tcl, etc.
@@ -2270,6 +2322,10 @@ typedef struct	BUFFER {
 #define is_local_b_val(bp,which)  \
 	is_local_val(bp->b_values.bv,which)
 
+#define set_local_b_val(bp,mode,value) \
+		make_local_b_val(bp, mode), \
+		set_b_val(bp, mode, value)
+
 #define is_empty_buf(bp) (lforw(buf_head(bp)) == buf_head(bp))
 
 #define b_dot     b_wtraits.w_dt
@@ -2280,16 +2336,25 @@ typedef struct	BUFFER {
 #define b_tentative_lastdot b_wtraits.w_tld
 #define b_wline   b_wtraits.w_ln
 
-#if OPT_CASELESS
-#define eql_bname(bp,name) !stricmp(bp->b_bname, name)
+#ifndef HAVE_STRICMP
+#undef stricmp
+#define stricmp vl_stricmp
+#endif
+
+#ifndef HAVE_STRNICMP
+#undef strnicmp
+#define strnicmp vl_strnicmp
+#endif
+
+#if (SYS_MSDOS || SYS_OS2 || SYS_WINNT || SYS_OS2_EMX || SYS_CYGWIN || SYS_VMS)
+#define DFT_FILE_IC TRUE
 #else
-#define eql_bname(bp,name) !strcmp(bp->b_bname, name)
+#define DFT_FILE_IC FALSE
 #endif
 
 #define COPY_B_VAL(dst,src,val) \
 	if (is_local_b_val(src, val)) { \
-	    make_local_b_val(dst, val); \
-	    set_b_val(dst, val, b_val(src, val)); \
+	    set_local_b_val(dst, val, b_val(src, val)); \
 	}
 
 /* values for b_flag */
@@ -2428,6 +2493,8 @@ extern MARK *api_mark_iterator(BUFFER *bp, int *iter);
     } one_time
 #endif /* OPT_VIDEO_ATTRS */
 
+typedef int WIN_ID;
+
 /*
  * There is a window structure allocated for every active display window. The
  * windows are kept in a big list, in top to bottom screen order, with the
@@ -2456,7 +2523,7 @@ typedef struct	WINDOW {
 	int	w_ruler_col;
 #endif
 #if OPT_PERL || OPT_TCL
-	int	w_id;			/* Unique window id */
+	WIN_ID	w_id;			/* Unique window id */
 #endif
 }	WINDOW;
 
@@ -2496,6 +2563,15 @@ typedef struct	WINDOW {
 #define MK Mark
 #endif
 
+/*
+ * Data to save/restore when switching temporarily to the minibuffer for I/O.
+ */
+typedef struct {
+	BUFFER *save_bp;
+	WINDOW *save_wp;
+	MARK save_mk;
+} KBD_DATA;
+
 	/* we use left-margin for protecting the prefix-area of [Registers]
 	 * from cut/paste selection.
 	 */
@@ -2530,7 +2606,8 @@ typedef struct	{
 	int	rows;			/* current row count		*/
 	int	maxcols;		/* max column count		*/
 	int	cols;			/* current column count		*/
-	ENC_CHOICES encoding;		/* tell what the display can do	*/
+	void	(*set_enc)(ENC_CHOICES); /* tell what the display can do */
+	ENC_CHOICES (*get_enc)(void);	/* tell what the display can do	*/
 	void	(*open) (void);		/* Open terminal at the start.	*/
 	void	(*close) (void);	/* Close terminal at end.	*/
 	void	(*kopen) (void);	/* keyboard open		*/
@@ -2550,7 +2627,7 @@ typedef struct	{
 	int	(*setdescrip) (const char *f); /* reset display description */
 	void	(*setfore) (int f);	/* set foreground color		*/
 	void	(*setback) (int b);	/* set background color		*/
-	void	(*setpal) (const char *p); /* set color palette	*/
+	int	(*setpal) (const char *p); /* set color palette	*/
 	void	(*setccol) (int c);	/* set cursor color		*/
 	void	(*scroll) (int from, int to, int n); /* scroll region	*/
 	void	(*pflush) (void);	/* really flush			*/
@@ -2569,7 +2646,7 @@ typedef struct	{
 	void	(*mevent) (void);	/* mouse event			*/
 }	TERM;
 
-#define term_is_utfXX()         ((int) term.encoding >= enc_UTF8)
+#define term_is_utfXX()         (term.get_enc() >= enc_UTF8)
 
 #if DISP_CURSES && defined(HAVE_ADDNWSTR)
 #define WIDE_CURSES 1
@@ -2577,7 +2654,7 @@ typedef struct	{
 #define WIDE_CURSES 0
 #endif
 
-#if OPT_MULTIBYTE && (DISP_TERMCAP || WIDE_CURSES || DISP_X11 || (DISP_NTWIN && defined(UNICODE)))
+#if OPT_MULTIBYTE && (DISP_TERMCAP || WIDE_CURSES || DISP_X11 || ((DISP_NTCONS || DISP_NTWIN) && defined(UNICODE)))
 typedef USHORT VIDEO_TEXT;
 typedef USHORT VIDEO_CHAR;
 #else
@@ -2585,8 +2662,6 @@ typedef UCHAR VIDEO_TEXT;
 typedef char  VIDEO_CHAR;
 #endif
 
-#define VIDEO_MIN 4
-
 typedef struct  VIDEO {
 	UINT	v_flag;			/* Flags */
 #if	OPT_COLOR
@@ -2598,9 +2673,7 @@ typedef struct  VIDEO {
 #if	OPT_VIDEO_ATTRS
 	VIDEO_ATTR *v_attrs;		/* screen data attributes */
 #endif
-	/* allocate 4 bytes here, and malloc 4 bytes less than we need,
-		to keep malloc from rounding up. */
-	VIDEO_TEXT v_text[VIDEO_MIN];	/* Screen data. */
+	VIDEO_TEXT v_text[1];		/* Screen data. */
 }	VIDEO;
 
 #define VideoText(vp) (vp)->v_text
@@ -2796,6 +2869,7 @@ typedef struct {
 #define ZERO    argBIT(12)	/* allow 0 to be given as a line number */
 #define OPTREG  argBIT(13)	/* allow optional register-name */
 #define USEREG  argBIT(14)	/* expect register-name */
+#define FROM_TO argBIT(15)	/* % is all lines */
 #define FILES   (XFILE | EXTRA)	/* multiple extra files allowed */
 #define WORD1   (EXTRA | NOSPC)	/* one extra word allowed */
 #define FILE1   (FILES | NOSPC)	/* 1 file allowed, defaults to current file */
@@ -2804,8 +2878,8 @@ typedef struct {
 #define RANGE   (FROM  | TO)	/* range of linespecs allowed */
 
 /* these flags determine the type of cu.* */
-#define typBIT(n) cmdBIT(n+28)	/* ...to simplify adding bits */
-/* bits 27-29 */
+#define typBIT(n) cmdBIT(n+29)	/* ...to simplify adding bits */
+/* bits 29-31 */
 #define CMD_FUNC 0L		/* this is the default (CmdFunc) */
 #define CMD_PROC typBIT(0)	/* named procedure (BUFFER *) */
 #define CMD_OPER typBIT(1)	/* user-defined operator */
@@ -2913,19 +2987,27 @@ extern void exit (int code);
 extern void _exit (int code);
 #endif	/* HAVE_STDLIB_H */
 
+#if !(defined(intptr_t))
+/* get intptr_t */
+#if defined(HAVE_INTTYPES_H)
+#include <inttypes.h>
+#elif defined(HAVE_STDINT_H)
+#include <stdint.h>
+#elif !(defined(WIN32) || defined(WIN32_LEAN_AND_MEAN))
+typedef long intptr_t;
+#endif
+#endif
+
 /* array/table size */
 #define	TABLESIZE(v)	(sizeof(v)/sizeof(v[0]))
 
 #define PERCENT(num,den) ((den) ? (int)((100.0 * (num))/(den)) : 100)
 
 /* Quiet compiler warnings on places where we're being blamed incorrectly,
- * e.g., for casting away const, or for alignment problems.  It's generally
- * legal in c89 to cast a pointer to long w/o loss of precision.
- *
- * FIXME: c99 may require a wider type.
+ * e.g., for casting away const, or for alignment problems.
  */
 #ifdef __GNUC__
-#define TYPECAST(type,ptr) (type*)((long)(ptr))
+#define TYPECAST(type,ptr) (type*)((intptr_t)(ptr))
 #else
 #define TYPECAST(type,ptr) (type*)(ptr)
 #endif
@@ -2994,7 +3076,10 @@ extern void ExitProgram(int code);
 /*
  * If we have va_copy(), use it.
  */
-#if defined(va_copy) || defined(HAVE_VA_COPY)
+#if defined(HAVE___VA_COPY)
+#define begin_va_copy(dst,src)	__va_copy(dst, src)
+#define end_va_copy(dst)	va_end(dst)
+#elif defined(va_copy) || defined(HAVE_VA_COPY)
 #define begin_va_copy(dst,src)	va_copy(dst, src)
 #define end_va_copy(dst)	va_end(dst)
 #else
@@ -3078,13 +3163,13 @@ extern void ExitProgram(int code);
 #    define show_alloc() void __mp_summary()
 #  endif
 #  if CAN_TRACE && OPT_TRACE
-#    include "trace.h"
+#    include <trace.h>
 #  endif
-#else
-#  if CAN_TRACE && (NO_LEAKS || DOALLOC || OPT_TRACE)
-#    include "trace.h"
+#elif CAN_TRACE
+#  if (NO_LEAKS || DOALLOC || OPT_TRACE)
+#    include <trace.h>
 #  elif !defined(show_alloc) && NO_LEAKS
-#    include "trace.h"
+#    include <trace.h>
 #  endif
 #endif	/* USE_DBMALLOC */
 
@@ -3114,24 +3199,33 @@ extern void show_elapsed(void);
 /* Normally defined in "trace.h" */
 #ifndef TRACE
 #define TRACE(p) /* nothing */
-#define returnCode(c)   return(c)
-#define returnPtr(c)    return(c)
-#define returnString(c) return(c)
-#define returnVoid()    return
+#define returnCString(c) return(c)
+#define returnCode(c)    return(c)
+#define returnPtr(c)     return(c)
+#define returnString(c)  return(c)
+#define returnVoid()     return
 #endif
 
 #if OPT_TRACE > 1
-#define TRACE2(params)   TRACE(params)
-#define return2Code(c)   returnCode(c)
-#define return2Ptr(c)    returnPtr(c)
-#define return2String(c) returnString(c)
-#define return2Void()    returnVoid()
+#define TRACE2(params)    TRACE(params)
+#define return2CString(c) returnCString(c)
+#define return2Code(c)    returnCode(c)
+#define return2Ptr(c)     returnPtr(c)
+#define return2String(c)  returnString(c)
+#define return2Void()     returnVoid()
 #else
 #define TRACE2(params) /*nothing*/
-#define return2Code(c)   return(c)
-#define return2Ptr(c)    return(c)
-#define return2String(c) return(c)
-#define return2Void()    return
+#define return2CString(c) return(c)
+#define return2Code(c)    return(c)
+#define return2Ptr(c)     return(c)
+#define return2String(c)  return(c)
+#define return2Void()     return
+#endif
+
+#if OPT_TRACE > 2
+#define TRACE3(params)    TRACE(params)
+#else
+#define TRACE3(params) /*nothing*/
 #endif
 
 #if OPT_EVAL || OPT_DEBUGMACROS
@@ -3155,6 +3249,12 @@ extern void show_elapsed(void);
 #endif
 #endif /* gcc workarounds */
 
+#if defined(__GNUC__) && defined(_FORTIFY_SOURCE)
+#define IGNORE_RC(func) ignore_unused = (int) func
+#else
+#define IGNORE_RC(func) (void) func
+#endif /* gcc workarounds */
+
 /*
  * Local prototypes (must follow NO_LEAKS definition)
  */
@@ -3176,4 +3276,6 @@ extern void show_elapsed(void);
 #include "fakevms.h"
 #endif
 
+/* *INDENT-ON* */
+
 #endif /* _estruct_h */
