@@ -4,7 +4,7 @@
  *	original by Daniel Lawrence, but
  *	much modified since then.  assign no blame to him.  -pgf
  *
- * $Header: /usr/build/vile/vile/RCS/exec.c,v 1.331 2009/05/24 11:00:39 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/exec.c,v 1.361 2016/07/22 00:20:03 tom Exp $
  *
  */
 
@@ -181,6 +181,7 @@ parse_signed_number(const char *s, int *number)
 static const char *
 parse_linespec(const char *s, LINE **markptr)
 {
+    const char *base = s;
     int num;
     LINE *lp;			/* where the linespec takes us */
     int status;
@@ -233,17 +234,19 @@ parse_linespec(const char *s, LINE **markptr)
 	    last_srch_direc = (*save == '/') ? FORWARD : REVERSE;
 	    s = skip_pattern(save, &done);
 	    tb_init(&searchpat, EOS);
-	    tb_bappend(&searchpat, save + 1, s - save - 1 - done);
+	    tb_bappend(&searchpat, save + 1, (size_t) (s - save - 1 - done));
 	    tb_append(&searchpat, EOS);
 
-	    if ((gregexp = regcomp(tb_values(searchpat),
+	    if (tb_length(searchpat) > 1 &&
+		(gregexp = regcomp(tb_values(searchpat),
 				   tb_length(searchpat) - 1,
 				   b_val(curbp, MDMAGIC))) != 0) {
 
 		scanboundry(TRUE, DOT, last_srch_direc);
 		if (scanner(gregexp,
 			    last_srch_direc,
-			    TRUE,
+			    (DOT.o == 0),
+			    FALSE,
 			    (int *) 0)) {
 		    lp = DOT.l;
 		    found = TRUE;
@@ -260,9 +263,7 @@ parse_linespec(const char *s, LINE **markptr)
 
 	/* if linespec was faulty, quit now */
 	if (!lp) {
-	    *markptr = lp;
-	    swapmark();
-	    return s;
+	    break;
 	}
 
 	/* maybe add an offset */
@@ -274,7 +275,17 @@ parse_linespec(const char *s, LINE **markptr)
     } while (*s == ';' || isSign(*s));
 
     *markptr = lp;
-    swapmark();
+
+    /*
+     * If we're processing a script (for example), which has no line-specifiers,
+     * we'll get a null pointer here.  But there's no reason to flag a bug in
+     * swapmark().
+     */
+    if (lp == 0 && (s == base)) {
+	DOT = MK;
+    } else {
+	swapmark();
+    }
     return s;
 }
 
@@ -283,8 +294,11 @@ parse_linespec(const char *s, LINE **markptr)
  * Steve Kirkendall
  */
 static int
-rangespec(const char *specp, LINE **fromlinep, LINE **tolinep,
-	  CMDFLAGS * flagp)
+rangespec(const char *specp,
+	  LINE **fromlinep,
+	  LINE **tolinep,
+	  CMDFLAGS * flagp,
+	  int testonly)
 {
     const char *scan;		/* used to scan thru specp */
     LINE *fromline;		/* first linespec */
@@ -293,12 +307,12 @@ rangespec(const char *specp, LINE **fromlinep, LINE **tolinep,
     *flagp = 0;
 
     if (specp == 0)
-	return FALSE;
+	return -1;
 
     /* ignore command lines that start with a double-quote */
     if (*specp == '"') {
 	*fromlinep = *tolinep = DOT.l;
-	return TRUE;
+	return 1;
     }
 
     /* permit extra colons at the start of the line */
@@ -317,7 +331,7 @@ rangespec(const char *specp, LINE **fromlinep, LINE **tolinep,
 	fromline = lforw(buf_head(curbp));
 	toline = lback(buf_head(curbp));
 	scan++;
-	*flagp |= (FROM | TO);
+	*flagp |= FROM_TO | (FROM | TO);
     } else {
 	scan = parse_linespec(scan, &toline);
 	*flagp |= FROM;
@@ -330,8 +344,8 @@ rangespec(const char *specp, LINE **fromlinep, LINE **tolinep,
 	    scan = parse_linespec(scan, &toline);
 	    *flagp |= TO;
 	    if (toline == 0) {
-		/* faulty line spec */
-		return FALSE;
+		TRACE(("...faulty line spec\n"));
+		return -1;
 	    }
 	}
     }
@@ -345,15 +359,15 @@ rangespec(const char *specp, LINE **fromlinep, LINE **tolinep,
     /* skip whitespace */
     scan = skip_cblanks(scan);
 
-    if (*scan) {
-	/* dbgwrite("crud at end %s (%s)",specp, scan); */
-	return FALSE;
+    if (*scan && !testonly) {
+	TRACE(("...crud at end %s (%s)\n", specp, scan));
+	return -1;
     }
 
     *fromlinep = fromline;
     *tolinep = toline;
 
-    return TRUE;
+    return (int) (scan - specp);
 }
 
 /* special test for 'a style mark references */
@@ -400,14 +414,13 @@ making_pattern(EOL_ARGS)
 {
     int done = FALSE;
     int in_regexp = FALSE;
-    const char *src = buffer;
 
     (void) eolchar;
 
     TRACE(("making_pattern(buffer=%.*s, cpos=%d, c=%#x, eolchar=%#x)\n",
 	   (int) cpos, buffer, (int) cpos, c, eolchar));
     if (cpos != 0) {
-	src = skip_linespecs(buffer, cpos, &done);
+	(void) skip_linespecs(buffer, (int) cpos, &done);
 	in_regexp = !done || isPattern(c);
     } else if (isPattern(c)) {
 	in_regexp = TRUE;
@@ -481,6 +494,37 @@ more_named_cmd(void)
     return result;
 }
 
+/*
+ * Scrolling through the history buffer may bring up a selection containing
+ * a line/range specifier and arguments past the command-verb.  Check for,
+ * and adjust for these cases:
+ * a) no change is needed
+ * b) strip arguments
+ * c) back up to put the command-verb into the fifo, returning TRUE.
+ * If no back-up is needed, this function returns FALSE.
+ */
+static int
+resplice_command(TBUFF **lspec, const char *cspec)
+{
+    int result = FALSE;
+    int end_lspec = 0;
+    int end_input = (int) strlen(cspec);
+    LINE *fromline = 0;
+    LINE *toline = 0;
+    CMDFLAGS lflag = 0;
+
+    if (tb_length(*lspec) == 1 && *tb_values(*lspec) == EOS) {
+	end_lspec = rangespec(cspec, &fromline, &toline, &lflag, TRUE);
+	if (end_lspec > 0) {
+	    while (end_input-- > 0) {
+		unkeystroke(cspec[end_input]);
+	    }
+	    result = TRUE;
+	}
+    }
+    return result;
+}
+
 /*
  * namedcmd:	execute a named command even if it is not bound
  */
@@ -491,9 +535,21 @@ more_named_cmd(void)
 static				/* next procedure is local */
 #endif
 
+#ifndef DEBUG_KPOS
+#define DEBUG_KPOS 0
+#endif
+
+#if DEBUG_KPOS
+#define TRACE_KPOS(s) TRACE(s)
+#else
+#define TRACE_KPOS(s)		/* nothing */
+#endif
+
 int
 execute_named_command(int f, int n)
 {
+    static char empty[] = "";
+
     int status;
     const CMDFUNC *cfp;		/* function to execute */
     char *fnp;			/* ptr to the name of the cmd to exec */
@@ -513,6 +569,8 @@ execute_named_command(int f, int n)
     tb_scopy(&lspec, "");
     last_DOT = DOT;
 
+    TRACE_KPOS(("kpos before %d.%d\n", line_no(curbp, DOT.l), DOT.o));
+
     /* prompt the user to type a named command */
 #ifdef SMALLER
     mlprompt(": ");
@@ -546,7 +604,7 @@ execute_named_command(int f, int n)
 	    repeat_cmd = EOS;
 	} else {		/* looking for command-name */
 	    fnp = NULL;
-	    status = kbd_engl_stat((char *) 0, cspec, KBD_STATED);
+	    status = kbd_engl_stat((char *) 0, cspec, 0, KBD_STATED);
 	    if (status == TRUE) {
 		fnp = cspec;
 #if !SMALLER
@@ -574,11 +632,17 @@ execute_named_command(int f, int n)
 		    } else {	/* e.g., ":e#" */
 			c = end_string();
 			if (isPunct(c)
-			    && strchr(global_g_val_ptr(GVAL_EXPAND_CHARS),
-				      c) != 0) {
+			    && vl_index(global_g_val_ptr(GVAL_EXPAND_CHARS),
+					c) != 0) {
 			    unkeystroke(c);
 			}
 		    }
+		    if (resplice_command(&lspec, cspec)) {
+			cmode = 0;
+			hst_reset();
+			hst_init(EOS);
+			continue;
+		    }
 		    break;
 		}
 	    } else if (status == SORTOFTRUE) {
@@ -625,7 +689,7 @@ execute_named_command(int f, int n)
     }
 
     /* parse the accumulated lspec */
-    if (rangespec(tb_values(lspec), &fromline, &toline, &lflag) != TRUE) {
+    if (rangespec(tb_values(lspec), &fromline, &toline, &lflag, FALSE) < 0) {
 	mlforce("[Improper line range '%s']", tb_values(lspec));
 	return FALSE;
     }
@@ -642,7 +706,7 @@ execute_named_command(int f, int n)
 	    return FALSE;
 	} else {		/* range, no function */
 	    cfp = &f_gomark;
-	    fnp = "";
+	    fnp = empty;
 	}
     } else if ((cfp = engl2fnc(fnp)) == NULL) {		/* bad function */
 	return no_such_function(fnp);
@@ -711,6 +775,8 @@ execute_named_command(int f, int n)
 	mlforce("[Can't use a range with \"%s\" command.]", fnp);
 	return FALSE;
     }
+    TRACE_KPOS(("kpos from %d\n", line_no(curbp, fromline)));
+    TRACE_KPOS(("kpos to   %d\n", line_no(curbp, toline)));
 
     /* some commands have special default ranges */
     if (lflag & DFLALL) {
@@ -774,7 +840,7 @@ execute_named_command(int f, int n)
 	int last = maybe_num ? 2 : 1;
 
 	while (!end_of_cmd() && (that < last)) {
-	    status = mlreply_reg_count(that, &num, &that);
+	    status = mlreply_reg_count(fnc2engl(cfp), that, &num, &that);
 	    if (status == ABORT)
 		return status;
 	    else if (status != TRUE)
@@ -791,7 +857,7 @@ execute_named_command(int f, int n)
 		swapmark();
 	    } else {
 		ukb = (short) num;
-		kregflag = (short) ((that == 1) ? KAPPEND : 0);
+		kregflag = (USHORT) ((that == 1) ? KAPPEND : 0);
 		that = 1;
 		/* patch: need to handle recursion */
 	    }
@@ -799,6 +865,8 @@ execute_named_command(int f, int n)
     }
 
     save_DOT = DOT;
+    TRACE_KPOS(("kpos save %d.%d\n", line_no(curbp, DOT.l), DOT.o));
+
     if (!(flags & NOMOVE)
 	&& ((toline != 0) || (fromline != 0))) {
 	/* assume it's an absolute motion */
@@ -858,6 +926,51 @@ execute_named_command(int f, int n)
 	restore_dot(save_DOT);
     else if (status == ABORT)
 	restore_dot(last_DOT);
+#ifdef GVAL_KEEP_POS
+    else {
+	TRACE_KPOS(("kpos restoring\n"));
+	switch (global_g_val(GVAL_KEEP_POS)) {
+	case KPOS_VILE:
+	    break;
+	case KPOS_NVI:
+	    if ((lflag & FROM_TO) &&
+		(cfp == &f_operrshift || cfp == &f_operlshift)) {
+		flags |= NOMOVE;
+		TRACE_KPOS(("kpos restore bot\n"));
+	    } else if (toline != 0 &&
+		       fromline != 0) {
+		int dotline = line_no(curbp, save_DOT.l);
+		if (line_no(curbp, toline) < dotline) {
+		    save_DOT.l = toline;
+		    save_DOT.o = 0;
+		    flags |= NOMOVE;
+		    TRACE_KPOS(("kpos restore to %d\n", line_no(curbp, toline)));
+		} else if (line_no(curbp, fromline) > dotline) {
+		    save_DOT.l = toline;
+		    save_DOT.o = 0;
+		    flags |= NOMOVE;
+		    TRACE_KPOS(("kpos restore from %d\n", line_no(curbp, fromline)));
+		} else {
+		    flags |= NOMOVE;
+		    TRACE_KPOS(("kpos restore ignored to/from\n"));
+		}
+	    } else {
+		TRACE_KPOS(("kpos restore failed\n"));
+	    }
+	    break;
+	case KPOS_VI:
+	    save_DOT.l = toline;
+	    save_DOT.o = 0;
+	    flags |= NOMOVE;
+	    break;
+	}
+	if (flags & NOMOVE) {
+	    restore_dot(save_DOT);
+	}
+    }
+#endif
+
+    TRACE_KPOS(("kpos after %d.%d\n", line_no(curbp, DOT.l), DOT.o));
 
     return status;
 }
@@ -926,7 +1039,7 @@ docmd(char *cline, int execflag, int f, int n)
      */
     if (toktyp(token) != TOK_LITSTR || isDigit(token[0])) {
 	f = TRUE;
-	n = strtol(tokval(token), 0, 0);
+	n = (int) strtol(tokval(token), 0, 0);
 
 	/* and now get the command to execute */
 	if ((token = mac_unquotedarg(&tok)) == 0) {
@@ -986,7 +1099,7 @@ call_cmdfunc(const CMDFUNC * p, int f, int n)
 
 #if OPT_PERL
     case CMD_PERL:		/* perl subroutine */
-	status = perl_call_sub(CMD_U_PERL(p), p->c_flags & OPER, f, n);
+	status = perl_call_sub(CMD_U_PERL(p), (int) (p->c_flags & OPER), f, n);
 	break;
 #endif
 #endif /* OPT_NAMEBST */
@@ -1031,8 +1144,9 @@ execute(const CMDFUNC * execfunc, int f, int n)
 	/* but if we're in insertmode, it's okay, since we must
 	   be executing a function key, like an arrow key,
 	   that the user will want to have replayed later */
-	if ((curwp == NULL || !insertmode) && (flags & REDO) == 0)
+	if ((!insertmode) && (flags & REDO) == 0)
 	    dotcmdstop();
+
 	if (flags & UNDO) {
 	    /* undoable command can't be permitted when read-only */
 	    if (!(flags & VIEWOK)) {
@@ -1103,6 +1217,68 @@ get_token(char *src, TBUFF **tok, int (*endfunc) (EOL_ARGS), int eolchar, int *a
     return get_token2(src, tok, endfunc, eolchar, actual);
 }
 
+/*
+ * This is used for the "map" command, which doesn't honor anything except ^V.
+ */
+char *
+get_token1(char *src, TBUFF **tok, int (*endfunc) (EOL_ARGS), int eolchar, int *actual)
+{
+    int c, chr;
+
+    if (actual != 0)
+	*actual = EOS;
+    if (src == 0)
+	return src;
+
+    /* first scan past any whitespace in the source string */
+    src = skip_space_tab(src);
+
+    /* scan through the source string, which may be quoted */
+    while ((c = *src) != EOS) {
+	if (c == quotec) {
+	    if (*++src == EOS)
+		break;
+	}
+#if OPT_MODELINE
+	/*
+	 * Multiple settings on the line may be separated by colons.
+	 */
+	else if (c == ':' && in_modeline) {
+	    if (actual != 0)
+		*actual = ' ';
+	    src++;
+	    break;
+	} else
+#endif
+	if (c == eolchar) {
+	    if (actual != 0)
+		*actual = *src;
+	    if (!isBlank(c))
+		src++;
+	    break;
+	} else if (endfunc(tb_values(*tok), tb_length(*tok), c, eolchar)) {
+	    if (actual != 0)
+		*actual = *src;
+	    break;
+	} else if (isBlank(c)) {
+	    if (actual != 0)
+		*actual = *src;
+	    break;
+	}
+
+	chr = *src++;		/* record the character */
+	tb_append(tok, chr);
+    }
+
+    /* scan past any whitespace remaining in the source string */
+    src = skip_space_tab(src);
+    token_ended_line = isreturn(*src) || *src == EOS;
+
+    tb_append(tok, EOS);
+
+    return src;
+}
+
 /*
  * This is split-out from get_token() so we can handle the special case of
  * re-gluing a "!" to a quoted string from kbd_reply().
@@ -1315,7 +1491,7 @@ macroize(TBUFF **p, TBUFF *src, int skip)
 
 		TRACE(("macroizing %s\n", tb_visible(src)));
 		(void) tb_append(p, DQUOTE);
-		for (n = skip; n < len; n++) {
+		for (n = (size_t) skip; n < len; n++) {
 		    c = txt[n];
 		    if (multi && count++)
 			(void) tb_sappend(p, "\" \"");
@@ -1487,6 +1663,8 @@ setup_macro_buffer(TBUFF *name, int bufnum, UINT flags)
     char bname[NBUFN];		/* name of buffer to use */
     BUFFER *bp;
 
+    (void) flags;
+
     /* construct the buffer name for the macro */
     if (bufnum < 0)
 	(void) add_brackets(bname, tb_values(name));
@@ -1536,8 +1714,10 @@ setup_macro_buffer(TBUFF *name, int bufnum, UINT flags)
 		case TOK_LITSTR:
 		    if (cf->c_args == 0) {
 			cf->c_args = typeallocn(PARAM_INFO, limit + 1);
-			if (cf->c_args == 0)
+			if (cf->c_args == 0) {
+			    free(cf);
 			    return no_memory("Allocating parameter info");
+			}
 		    }
 		    if (decode_parameter_info(temp, &(cf->c_args[count]))) {
 			cf->c_args[++count].pi_type = PT_UNKNOWN;
@@ -1547,6 +1727,7 @@ setup_macro_buffer(TBUFF *name, int bufnum, UINT flags)
 		    /* FALLTHRU */
 #endif
 		default:
+		    free(cf);
 		    mlforce("[Unexpected token '%s']", tb_values(temp));
 		    return FALSE;
 		}
@@ -1562,7 +1743,7 @@ setup_macro_buffer(TBUFF *name, int bufnum, UINT flags)
 	tb_free(&helpstring);
 #endif
 
-	if (insert_namebst(tb_values(name), cf, FALSE) != TRUE)
+	if (insert_namebst(tb_values(name), cf, FALSE, 0) != TRUE)
 	    return FALSE;
 	tb_copy(&(bp->b_procname), name);
     }
@@ -1607,7 +1788,7 @@ store_macro(int f, int n)
 	mlforce("[Macro number out of range]");
 	status = FALSE;
     } else {
-	status = setup_macro_buffer((TBUFF *) 0, n, CMD_PROC);
+	status = setup_macro_buffer((TBUFF *) 0, n, (UINT) CMD_PROC);
     }
     return status;
 }
@@ -1635,7 +1816,7 @@ store_proc(int f, int n)
 	if ((status = kbd_reply("Procedure name: ", &name,
 				eol_command, ' ',
 				KBD_NORMAL, no_completion)) == TRUE) {
-	    status = setup_macro_buffer(name, -1, CMD_PROC);
+	    status = setup_macro_buffer(name, -1, (UINT) CMD_PROC);
 	}
     }
     return status;
@@ -1659,7 +1840,7 @@ store_op(int f, int n)
 	if ((status = kbd_reply("Operator name: ", &name,
 				eol_command, ' ',
 				KBD_NORMAL, no_completion)) == TRUE) {
-	    status = setup_macro_buffer(name, -1, CMD_OPER);
+	    status = setup_macro_buffer(name, -1, (UINT) CMD_OPER);
 	}
     }
     return status;
@@ -1678,7 +1859,7 @@ execproc(int f, int n)
     char bufn[NBUFN];		/* name of buffer to execute */
 
     if ((status = mlreply("Execute procedure: ",
-			  name, sizeof(name))) != TRUE) {
+			  name, (int) sizeof(name))) != TRUE) {
 	return status;
     }
 
@@ -2181,7 +2362,7 @@ setup_dobuf(BUFFER *bp, WHLOOP ** result)
 	if (i <= 0)
 	    continue;
 
-	dirnum = dname_to_dirnum(&cmdp, i);
+	dirnum = dname_to_dirnum(&cmdp, (size_t) i);
 	if (dirnum == D_WHILE ||
 	    dirnum == D_BREAK ||
 	    dirnum == D_ENDWHILE) {
@@ -2263,7 +2444,7 @@ setup_dobuf(BUFFER *bp, WHLOOP ** result)
 
 #if OPT_TRACE && !SMALLER
 static const char *
-TraceIndent(int level, char *cmdp, int length)
+TraceIndent(int level, char *cmdp, size_t length)
 {
     static const char indent[] = ".  .  .  .  .  .  .  .  ";
     switch (dname_to_dirnum(&cmdp, length)) {
@@ -2276,7 +2457,7 @@ TraceIndent(int level, char *cmdp, int length)
     default:
 	break;
     }
-    level = strlen(indent) - (3 * level);
+    level = (int) strlen(indent) - (3 * level);
     if (level < 0)
 	level = 0;
     return &indent[level];
@@ -2320,7 +2501,7 @@ compute_indent(char *cmd, size_t length, int *indent, int *indstate)
 	int next = 0;
 
 	cmd = skip_blanks(cmd);
-	dirnum = dname_to_dirnum(&cmd, (length - (cmd - base)));
+	dirnum = dname_to_dirnum(&cmd, (length - (size_t) (cmd - base)));
 	switch (dirnum) {
 	case D_IF:		/* FALLTHRU */
 	case D_WHILE:		/* FALLTHRU */
@@ -2343,8 +2524,10 @@ compute_indent(char *cmd, size_t length, int *indent, int *indstate)
 	    break;
 	}
 	*indstate = next;
-	if ((*indent += here) < 0)
-	    *indent = 0;
+	if (indent != 0) {
+	    if ((*indent += here) < 0)
+		*indent = 0;
+	}
     }
 }
 
@@ -2366,7 +2549,7 @@ handle_endm(void)
 	macrobuffer->b_dot.o = 0;
 	macrobuffer = 0;
 #if !SMALLER
-	compute_indent((char *) 0, 0, (int *) 0, (int *) 0);
+	compute_indent((char *) 0, (size_t) 0, (int *) 0, (int *) 0);
 #endif
     }
 }
@@ -2392,6 +2575,8 @@ perform_dobuf(BUFFER *bp, WHLOOP * whlist)
 
     static BUFFER *dobuferrbp = 0;
 
+    (void) whlist;
+
     TRACE((T_CALLED "perform_dobuf(bp=%p, whlist=%p) buffer '%s'\n",
 	   (void *) bp, (void *) whlist, bp->b_bname));
 
@@ -2406,10 +2591,11 @@ perform_dobuf(BUFFER *bp, WHLOOP * whlist)
 	if (llength(lp) <= 0)
 	    linlen = 0;
 	else
-	    linlen = llength(lp);
+	    linlen = (size_t) llength(lp);
 
 	if (glue) {
-	    if ((linebuf = castrealloc(char, linebuf, glue + linlen + 1)) == 0) {
+	    linebuf = castrealloc(char, linebuf, (size_t) glue + linlen + 1);
+	    if (linebuf == 0) {
 		status = no_memory("during macro execution");
 		break;
 	    }
@@ -2420,7 +2606,8 @@ perform_dobuf(BUFFER *bp, WHLOOP * whlist)
 #if !SMALLER
 	    compute_indent(lvalue(lp), linlen, &indent, &indstate);
 #endif
-	    if ((linebuf = cmdp = castalloc(char, indent + linlen + 1)) == 0) {
+	    linebuf = cmdp = castalloc(char, (size_t) indent + linlen + 1);
+	    if (linebuf == 0) {
 		status = no_memory("during macro execution");
 		break;
 	    }
@@ -2461,7 +2648,7 @@ perform_dobuf(BUFFER *bp, WHLOOP * whlist)
 	if (lforw(lp) != buf_head(bp)
 	    && linlen != 0
 	    && isEscaped(cmdp + linlen)) {
-	    glue = (int) (linlen + (cmdp - linebuf) - 1);
+	    glue = (int) (linlen + (size_t) (cmdp - linebuf) - 1);
 	    continue;
 	}
 	cmdp = skip_space_tab(linebuf);
@@ -2658,94 +2845,93 @@ dobuf(BUFFER *bp, int limit, int real_cmd_count)
 
     static int dobufnesting;	/* flag to prevent runaway recursion */
 
-    TRACE((T_CALLED "dobuf(%s, %d)\n", bp->b_bname, limit));
-    beginDisplay();
-    if (++dobufnesting < 9) {
+    if (bp != 0) {
+	TRACE((T_CALLED "dobuf(%s, %d)\n", bp->b_bname, limit));
+	beginDisplay();
+	if (++dobufnesting < 9) {
 
-	save_vl_msgs = vl_msgs;
-	save_cmd_count = cmd_count;
+	    save_vl_msgs = vl_msgs;
+	    save_cmd_count = cmd_count;
 
-	/* macro arguments are readonly, so we do this once */
-	if ((status = save_arguments(bp)) != ABORT) {
-	    vl_msgs = FALSE;
+	    /* macro arguments are readonly, so we do this once */
+	    if ((status = save_arguments(bp)) != ABORT) {
+		vl_msgs = FALSE;
 
-	    for (counter = 1; counter <= limit; counter++) {
-		if (dobufnesting == 1)
-		    cmd_count = ((real_cmd_count >= 0)
-				 ? real_cmd_count
-				 : counter);
+		for (counter = 1; counter <= limit; counter++) {
+		    if (dobufnesting == 1)
+			cmd_count = ((real_cmd_count >= 0)
+				     ? real_cmd_count
+				     : counter);
 
 #if ! SMALLER
-		if (setup_dobuf(bp, &whlist) != TRUE) {
-		    status = FALSE;
-		} else
+		    if (setup_dobuf(bp, &whlist) != TRUE) {
+			status = FALSE;
+		    } else
 #else
-		whlist = NULL;
+		    whlist = NULL;
 #endif
-		{
-		    IFSTK save_ifstk;
-		    push_buffer(&save_ifstk);
-		    status = perform_dobuf(bp, whlist);
-		    pop_buffer(&save_ifstk);
+		    {
+			IFSTK save_ifstk;
+			push_buffer(&save_ifstk);
+			status = perform_dobuf(bp, whlist);
+			pop_buffer(&save_ifstk);
+		    }
+
+		    handle_endm();
+		    free_all_whiles(whlist);
+
+		    if (status != TRUE)
+			break;
 		}
 
-		handle_endm();
-		free_all_whiles(whlist);
-
-		if (status != TRUE)
-		    break;
-	    }
-
-	    /*
-	     * If the caller set $return, use that value.
-	     */
+		/*
+		 * If the caller set $return, use that value.
+		 */
 #if OPT_EVAL
-	    if (this_macro_result != 0)
-		tb_copy(&macro_result, this_macro_result);
+		if (this_macro_result != 0)
+		    tb_copy(&macro_result, this_macro_result);
 #endif
 
-	    restore_arguments(bp);
-	    vl_msgs = save_vl_msgs;
-	    cmd_count = save_cmd_count;
+		restore_arguments(bp);
+		vl_msgs = save_vl_msgs;
+		cmd_count = save_cmd_count;
+	    } else {
+	    }
 	} else {
+	    mlwarn("[Too many levels of files]");
+	    tb_error(&macro_result);
 	}
-    } else {
-	mlwarn("[Too many levels of files]");
-	tb_error(&macro_result);
-    }
-    dobufnesting--;
+	dobufnesting--;
 
-    /*
-     * Set $_ from our TBUFF (preferred), or a readable form of the status
-     * codes.  In the latter case, this is not the same as $status, since we
-     * try to show the ABORT and SORTOFTRUE cases as well.
-     */
+	/*
+	 * Set $_ from our TBUFF (preferred), or a readable form of the status
+	 * codes.  In the latter case, this is not the same as $status, since
+	 * we try to show the ABORT and SORTOFTRUE cases as well.
+	 */
 #if OPT_EVAL
-    tb_free(&last_macro_result);
-    if (macro_result != 0) {
-	last_macro_result = macro_result;
-    } else {
-	switch (status) {
-	case FALSE:
-	    tb_scopy(&last_macro_result, "FALSE");
-	    break;
-	case TRUE:
-	    tb_scopy(&last_macro_result, "TRUE");
-	    break;
-	case ABORT:
-	    tb_scopy(&last_macro_result, "ABORT");
-	    break;
-	case SORTOFTRUE:
-	    tb_scopy(&last_macro_result, "SORTOFTRUE");
-	    break;
-	default:
-	    tb_error(&last_macro_result);
-	    break;
+	tb_free(&last_macro_result);
+	if (macro_result != 0) {
+	    last_macro_result = macro_result;
+	} else {
+	    switch (status) {
+	    case FALSE:
+		/* FALLTHRU */
+	    case TRUE:
+		/* FALLTHRU */
+	    case ABORT:
+		/* FALLTHRU */
+	    case SORTOFTRUE:
+		tb_scopy(&last_macro_result, status2s(status));
+		break;
+	    default:
+		tb_error(&last_macro_result);
+		break;
+	    }
 	}
-    }
 #endif /* OPT_EVAL */
 
-    endofDisplay();
+	endofDisplay();
+    }
 
     returnCode(status);
 }
