@@ -9,7 +9,7 @@
  * Extensions for vile by Paul Fox
  * Rewrote to use regular expressions - T.Dickey
  *
- * $Header: /usr/build/vile/vile/RCS/fences.c,v 1.86 2008/01/22 00:12:01 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/fences.c,v 1.95 2015/03/13 08:37:40 tom Exp $
  *
  */
 
@@ -48,7 +48,7 @@
 #define BlkEnd   b_val_rexp(curbp, VAL_FENCE_END)->reg
 
 #define CurrentChar() \
-	(is_at_end_of_line(DOT) ? '\n' : char_at(DOT))
+	(is_at_end_of_line(DOT) ? '\n' : CharAtDot())
 #define InDirection(sdir) \
 	((sdir == REVERSE) ? backchar(FALSE, 1) : forwchar(FALSE, 1))
 
@@ -62,13 +62,22 @@
 #endif
 
 	/*
-	 * ops are inclusive of the endpoint; account for this when saving
-	 * pre_op_dot
+	 * Start (or recur) a complex fence.
+	 */
+#define start_fence_op2(sdir, oldpos, oldpre) \
+	TRACE(("...start_fence_op2 %s@%d\n", __FILE__, __LINE__)); \
+	oldpos = DOT; \
+	oldpre = pre_op_dot
+
+	/*
+	 * Start a simple fence.
+	 *
+	 * This adds a special case, which is handled differently for complex
+	 * fences: ops are inclusive of the endpoint; account for this when
+	 * saving pre_op_dot
 	 */
 #define start_fence_op(sdir, oldpos, oldpre) \
-	TRACE(("...start_fence_op\n")); \
-	oldpos = DOT; \
-	oldpre = pre_op_dot; \
+	start_fence_op2(sdir, oldpos, oldpre); \
 	if (doingopcmd && sdir == REVERSE) { \
 		forwchar(TRUE,1); \
 		pre_op_dot = DOT; \
@@ -76,7 +85,7 @@
 	}
 
 #define test_fence_op(st, oldpos, oldpre) \
-	TRACE(("...test_fence_op, status=%d\n", st)); \
+	TRACE(("...test_fence_op, status=%d %s@%d\n", st, __FILE__, __LINE__)); \
 	if (st != TRUE) { \
 		DOT = oldpos; \
 		pre_op_dot = oldpre; \
@@ -97,10 +106,12 @@ static long iterations;
 static int
 next_line(int sdir)
 {
-    if (sdir == REVERSE)
+    DOT.o = b_left_margin(curbp);
+    if (sdir == REVERSE) {
 	DOT.l = lback(DOT.l);
-    else
+    } else {
 	DOT.l = lforw(DOT.l);
+    }
 
     if (is_header_line(DOT, curbp)) {
 	return FALSE;
@@ -112,10 +123,10 @@ next_line(int sdir)
 }
 
 #if OPT_TRACE
-static char *
+static const char *
 typeof_complex(int code)
 {
-    char *s;
+    const char *s;
     switch (code) {
     case CPP_IF:
 	s = "CPP_IF";
@@ -223,6 +234,10 @@ complex_fence(int sdir, int key, int group, int level, int *newkey)
     int result = -1;
     struct VAL *vals;
 
+    (void) group;
+    (void) level;
+    (void) newkey;
+
     TRACE(("ComplexFence(%d:%d) %d: %s %s\n",
 	   level,
 	   group,
@@ -398,7 +413,7 @@ find_complex(int sdir, int *newkey)
     for_each_modegroup(curbp, group, 0, vals) {
 	ignorecase = any_mode(vals, MDIGNCASE);
 	if ((key = match_complex(TRACEARG(group) DOT.l, vals)) != CPP_UNKNOWN) {
-	    start_fence_op(sdir, oldpos, oldpre);
+	    start_fence_op2(sdir, oldpos, oldpre);
 	    sdir = ((key == CPP_ENDIF)
 		    ? REVERSE
 		    : FORWARD);
@@ -438,7 +453,7 @@ find_one_complex(int sdir, int level, int group, int *newkey)
      */
     TRACE(("find_one_complex %4d:%s\n", line_no(curbp, DOT.l), lp_visible(DOT.l)));
     if ((key = match_complex(TRACEARG(group) DOT.l, vals)) != CPP_UNKNOWN) {
-	start_fence_op(sdir, oldpos, oldpre);
+	start_fence_op2(sdir, oldpos, oldpre);
 	if (level == 0)
 	    sdir = ((key == CPP_ENDIF)
 		    ? REVERSE
@@ -471,7 +486,7 @@ is_user_fence(int ch, int *sdirp)
     char *chp, och;
     if (!ch)
 	return 0;
-    chp = strchr(fences, ch);
+    chp = vl_index(fences, ch);
     if (!chp)
 	return 0;
     if ((chp - fences) & 1) {
@@ -529,7 +544,7 @@ comment_fence(int sdir)
 
     scanboundry(FALSE, DOT, sdir);
     if (scanner((sdir == FORWARD) ? BlkEnd : BlkBegin,
-		sdir, FALSE, (int *) 0)) {
+		sdir, (DOT.o == 0), FALSE, (int *) 0)) {
 	if (!doingopcmd || doingsweep) {
 	    sweephack = TRUE;
 	    if (sdir == FORWARD && (BlkEnd->mlen > 1))
@@ -572,7 +587,7 @@ getfence(int ch,		/* fence type to match against */
 	} else if (sdir == FORWARD) {
 	    if (oldpos.o < llength(oldpos.l)) {
 		do {
-		    ch = char_at(oldpos);
+		    ch = char_at_mark(oldpos);
 		} while (!is_user_fence(ch, (int *) 0) &&
 			 ++oldpos.o < llength(oldpos.l));
 	    }
@@ -582,7 +597,7 @@ getfence(int ch,		/* fence type to match against */
 	} else {
 	    if (oldpos.o >= 0) {
 		do {
-		    ch = char_at(oldpos);
+		    ch = char_at_mark(oldpos);
 		} while (!is_user_fence(ch, (int *) 0) &&
 			 --oldpos.o >= 0);
 	    }
