@@ -5,7 +5,7 @@
  * reading and writing of the disk are
  * in "fileio.c".
  *
- * $Header: /usr/build/vile/vile/RCS/file.c,v 1.427 2009/04/03 23:28:23 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/file.c,v 1.459 2016/07/15 00:52:52 tom Exp $
  */
 
 #include "estruct.h"
@@ -186,7 +186,7 @@ check_file_changed(BUFFER *bp, char *fn)
 #if OPT_LCKFILES
 	/* release own lock before read the file again */
 	if (global_g_val(GMDUSEFILELOCK)) {
-	    if (!b_val(curbp, MDLOCKED) && !b_val(curbp, MDVIEW))
+	    if (!b_val(bp, MDLOCKED) && !b_val(bp, MDVIEW))
 		release_lock(fn);
 	}
 #endif
@@ -281,6 +281,7 @@ CleanAfterPipe(int Wrote)
     if (must_clean_pipe) {
 #if SYS_UNIX || SYS_MSDOS
 	term.unclean();		/* may clear the screen as a side-effect */
+	term.open();		/* (re)open */
 	term.kopen();
 	term.flush();
 	if (Wrote)
@@ -437,7 +438,7 @@ same_fname(const char *fname, BUFFER *bp, int lengthen)
 /* support for "unique-buffers" mode -- file uids. */
 
 int
-fileuid_get(const char *fname, FUID * fuid)
+fileuid_get(const char *fname GCC_UNUSED, FUID * fuid GCC_UNUSED)
 {
 #ifdef CAN_CHECK_INO
     struct stat sb;
@@ -456,7 +457,7 @@ fileuid_get(const char *fname, FUID * fuid)
 }
 
 void
-fileuid_set(BUFFER *bp, FUID * fuid)
+fileuid_set(BUFFER *bp GCC_UNUSED, FUID * fuid GCC_UNUSED)
 {
 #ifdef CAN_CHECK_INO
     bp->b_fileuid = *fuid;	/* struct copy */
@@ -464,7 +465,7 @@ fileuid_set(BUFFER *bp, FUID * fuid)
 }
 
 void
-fileuid_invalidate(BUFFER *bp)
+fileuid_invalidate(BUFFER *bp GCC_UNUSED)
 {
 #ifdef CAN_CHECK_INO
     bp->b_fileuid.ino = 0;
@@ -580,6 +581,7 @@ reset_to_unnamed(BUFFER *bp)
 int
 fileread(int f GCC_UNUSED, int n GCC_UNUSED)
 {
+    BUFFER *bp = curbp;
     int status;
     char fname[NFILEN];
     W_VALUES *save_wvals;
@@ -590,21 +592,25 @@ fileread(int f GCC_UNUSED, int n GCC_UNUSED)
 	if ((status = mlreply_file("Replace with file: ", (TBUFF **) 0,
 				   FILEC_REREAD, fname)) != TRUE)
 	    returnCode(status);
-    } else if (b_is_temporary(curbp)) {
+    } else if (b_is_temporary(bp)) {
 	returnCode(cannot_reread());
-    } else if (!(global_g_val(GMDWARNREREAD) || curbp->b_fname[0] == EOS)
-	       || ((status = mlyesno("Reread current buffer")) == TRUE)) {
-	(void) vl_strncpy(fname, curbp->b_fname, sizeof(fname));
+    } else if (!(global_g_val(GMDWARNREREAD) || bp->b_fname[0] == EOS)
+	       || (mlyesno("Reread current buffer") == TRUE)) {
+	(void) vl_strncpy(fname, bp->b_fname, sizeof(fname));
 	/* Check if we are rereading an unnamed-buffer if it is not
 	 * associated with a file.
 	 */
-	if (is_internalname(curbp->b_fname)) {
-	    if (eql_bname(curbp, STDIN_BufName)
-		|| eql_bname(curbp, UNNAMED_BufName)) {
-		status = bclear(curbp);
+	if (is_internalname(bp->b_fname)) {
+	    WINDOW *wp;
+	    if (eql_bname(bp, STDIN_BufName)
+		|| eql_bname(bp, UNNAMED_BufName)) {
+		status = bclear(bp);
 		if (status == TRUE)
 		    mlerase();
-		curwp->w_flag |= (WFMODE | WFHARD);
+		for_each_visible_window(wp) {
+		    if (wp->w_bufp == bp)
+			wp->w_flag |= (WFMODE | WFHARD);
+		}
 		returnCode(status);
 	    }
 	    returnCode(cannot_reread());
@@ -616,31 +622,31 @@ fileread(int f GCC_UNUSED, int n GCC_UNUSED)
 #if OPT_LCKFILES
     /* release own lock before read the replaced file */
     if (global_g_val(GMDUSEFILELOCK)) {
-	if (!b_val(curbp, MDLOCKED) && !b_val(curbp, MDVIEW))
-	    release_lock(curbp->b_fname);
+	if (!b_val(bp, MDLOCKED) && !b_val(bp, MDVIEW))
+	    release_lock(bp->b_fname);
     }
 #endif
 
     /* we want no errors or complaints, so mark it unchanged */
-    b_clr_changed(curbp);
+    b_clr_changed(bp);
 
-    save_wvals = save_window_modes(curbp);
-    status = readin(fname, TRUE, curbp, TRUE);
+    save_wvals = save_window_modes(bp);
+    status = readin(fname, TRUE, bp, TRUE);
     if (status == ABORT) {
-	reset_to_unnamed(curbp);
+	reset_to_unnamed(bp);
     } else {
-	set_buffer_name(curbp);
+	set_buffer_name(bp);
 #if OPT_FINDERR
 	/*
 	 * Doing ":e!" on [Output] will reset the buffer name to match the
 	 * command.  Even if it was not [Output], it is still useful to make
 	 * the error-buffer match the last-read shell/pipe output.
 	 */
-	if (isShellOrPipe(curbp->b_fname))
-	    set_febuff(curbp->b_bname);
+	if (isShellOrPipe(bp->b_fname))
+	    set_febuff(bp->b_bname);
 #endif
     }
-    restore_window_modes(curbp, save_wvals);
+    restore_window_modes(bp, save_wvals);
 
     returnCode(status);
 }
@@ -731,6 +737,8 @@ set_files_to_edit(const char *prompt, int appflag)
 	    if (status == ABORT)
 		reset_to_unnamed(firstbp);
 	}
+	if (status == TRUE)
+	    set_last_bp(firstbp);
     }
     returnCode(status);
 }
@@ -757,8 +765,7 @@ viewfile(int f GCC_UNUSED, int n GCC_UNUSED)
 	    if ((s = getfile(actual, FALSE)) != TRUE)
 		break;
 	    /* if we succeed, put it in view mode */
-	    make_local_b_val(curwp->w_bufp, MDVIEW);
-	    set_b_val(curwp->w_bufp, MDVIEW, TRUE);
+	    set_local_b_val(curwp->w_bufp, MDVIEW, TRUE);
 	    markWFMODE(curwp->w_bufp);
 	}
     }
@@ -822,12 +829,12 @@ writelinesmsg(char *fn, int nline, B_COUNT nchar)
 	chars_len = (int) strlen(strchars);
 	outlen = (term.cols - 1) -
 	    (
-		(sizeof(WRITE_FILE_FMT) - 13) +
-		strlen(action) +
+		(int) (sizeof(WRITE_FILE_FMT) - 13) +
+		(int) strlen(action) +
 		lines_len +
 		chars_len +
-		strlen(PLURAL(nline)) +
-		strlen(PLURAL(nchar))
+		(int) strlen(PLURAL(nline)) +
+		(int) strlen(PLURAL(nchar))
 	    );
 	mlforce(WRITE_FILE_FMT,
 		action,
@@ -835,7 +842,7 @@ writelinesmsg(char *fn, int nline, B_COUNT nchar)
 		PLURAL(nline),
 		strchars,
 		PLURAL(nchar),
-		path_trunc(fn, outlen, tmp, sizeof(tmp)));
+		path_trunc(fn, outlen, tmp, (int) sizeof(tmp)));
     } else {
 	mlforce("[%d lines]", nline);
     }
@@ -847,13 +854,11 @@ set_rdonly_modes(BUFFER *bp, int always)
 {
     /* set view mode for read-only files */
     if ((global_g_val(GMDRONLYVIEW))) {
-	make_local_b_val(bp, MDVIEW);
-	set_b_val(bp, MDVIEW, TRUE);
+	set_local_b_val(bp, MDVIEW, TRUE);
     }
     /* set read-only mode for read-only files */
     if (always || global_g_val(GMDRONLYRONLY)) {
-	make_local_b_val(bp, MDREADONLY);
-	set_b_val(bp, MDREADONLY, TRUE);
+	set_local_b_val(bp, MDREADONLY, TRUE);
     }
 }
 
@@ -861,7 +866,7 @@ set_rdonly_modes(BUFFER *bp, int always)
 static void
 ff_read_directory(BUFFER *bp, char *fname)
 {
-    int count = fill_directory_buffer(bp, fname, 0);
+    int count = fill_directory_buffer(bp, fname, (size_t) 0);
 
     mlwrite("[Read %d line%s]", count, PLURAL(count));
     b_set_flags(bp, BFDIRS);
@@ -902,6 +907,9 @@ ff_load_directory(char *fname)
 static int
 kifile(char *fname)
 {
+#if OPT_ENCRYPT
+    BUFFER *bp = curbp;
+#endif
     int s;
     int nline;
     size_t i;
@@ -922,7 +930,7 @@ kifile(char *fname)
 
 	nline = 0;
 #if OPT_ENCRYPT
-	if ((s = vl_resetkey(curbp, fname)) == TRUE)
+	if ((s = vl_resetkey(bp, fname)) == TRUE)
 #endif
 	{
 	    mlwrite("[Reading...]");
@@ -1048,7 +1056,7 @@ getfile2bp(const char *fname,	/* file name to find */
 	    if (!ok_to_ask || !global_g_val(GMDWARNRENAME))
 		continue;
 	    hst_glue(' ');
-	    s = mlreply("Will use buffer name: ", bname, sizeof(bname));
+	    s = mlreply("Will use buffer name: ", bname, (int) sizeof(bname));
 	    if (s == ABORT)
 		return 0;
 	    if (s == FALSE || bname[0] == EOS)
@@ -1224,17 +1232,17 @@ explicit_dosmode(BUFFER *bp, RECORD_SEP record_sep)
  * Recompute the buffer size, redisplay the [Settings] buffer.
  */
 static int
-modified_record_sep(RECORD_SEP record_sep)
+modified_record_sep(BUFFER *bp, RECORD_SEP record_sep)
 {
     TRACE((T_CALLED "modified_record_sep(%s)\n",
 	   choice_to_name(&fsm_recordsep_blist, record_sep)));
 
-    explicit_dosmode(curbp, record_sep);
-    guess_dosmode(curbp);
-    explicit_dosmode(curbp, record_sep);	/* ignore the guess - only want to strip CR's */
+    explicit_dosmode(bp, record_sep);
+    guess_dosmode(bp);
+    explicit_dosmode(bp, record_sep);	/* ignore the guess - only want to strip CR's */
 
-    b_clr_counted(curbp);
-    bsizes(curbp);
+    b_clr_counted(bp);
+    bsizes(bp);
 
     returnCode(TRUE);
 }
@@ -1246,7 +1254,7 @@ modified_record_sep(RECORD_SEP record_sep)
 int
 set_rs_crlf(int f GCC_UNUSED, int n GCC_UNUSED)
 {
-    return modified_record_sep(RS_CRLF);
+    return modified_record_sep(curbp, RS_CRLF);
 }
 
 /* as above, but forces unix-mode instead */
@@ -1254,7 +1262,7 @@ set_rs_crlf(int f GCC_UNUSED, int n GCC_UNUSED)
 int
 set_rs_lf(int f GCC_UNUSED, int n GCC_UNUSED)
 {
-    return modified_record_sep(RS_LF);
+    return modified_record_sep(curbp, RS_LF);
 }
 
 /* as above, but forces macintosh-mode instead */
@@ -1262,7 +1270,7 @@ set_rs_lf(int f GCC_UNUSED, int n GCC_UNUSED)
 int
 set_rs_cr(int f GCC_UNUSED, int n GCC_UNUSED)
 {
-    return modified_record_sep(RS_CR);
+    return modified_record_sep(curbp, RS_CR);
 }
 #endif /* OPT_DOSFILES */
 
@@ -1280,11 +1288,9 @@ grab_lck_file(BUFFER *bp, char *fname)
 	if (!set_lock(fname, locker, sizeof(locker))) {
 	    /* we didn't get it */
 	    if ((locked_by = strmalloc(locker)) != 0) {
-		make_local_b_val(bp, MDVIEW);
-		set_b_val(bp, MDVIEW, TRUE);
-		make_local_b_val(bp, MDLOCKED);
-		set_b_val(bp, MDLOCKED, TRUE);
-		make_local_b_val(bp, VAL_LOCKER);
+		set_local_b_val(bp, MDVIEW, TRUE);
+		set_local_b_val(bp, MDLOCKED, TRUE);
+		make_local_b_val(bp, VAL_LOCKER, locked_by);
 		set_b_val_ptr(bp, VAL_LOCKER, locked_by);
 		markWFMODE(bp);
 	    }
@@ -1357,7 +1363,7 @@ guess_recordseparator(BUFFER *bp, UCHAR * buffer, B_COUNT length, L_NUM * lines)
 	} else if (global_b_val(MDDOS)) {
 	    if ((bp->b_flag & BFEXEC) != 0) {
 		result = (count_dos && !count_unix) ? RS_CRLF : RS_LF;
-	    } else if (check_percent_crlf(bp, count_dos, count_unix)) {
+	    } else if (check_percent_crlf(bp, (int) count_dos, (int) count_unix)) {
 		result = RS_CRLF;
 	    }
 	} else if (global_b_val(VAL_RECORD_SEP) == RS_CR
@@ -1366,12 +1372,12 @@ guess_recordseparator(BUFFER *bp, UCHAR * buffer, B_COUNT length, L_NUM * lines)
 	}
 	if (buffer[length - 1] != '\n')
 	    ++count_lf;
-	*lines = (result == RS_CR) ? count_cr : count_lf;
+	*lines = (L_NUM) ((result == RS_CR) ? count_cr : count_lf);
     } else if (count_cr != 0) {
 	result = RS_CR;
 	if (buffer[length - 1] != '\r')
 	    ++count_cr;
-	*lines = count_cr;
+	*lines = (L_NUM) count_cr;
     } else {
 	*lines = 1;		/* we still need a line, to allocate data */
 	result = RS_SYS_DEFAULT;
@@ -1433,6 +1439,8 @@ quickreadf(BUFFER *bp, int *nlinep)
     UCHAR *buffer;
     int rc;
 
+    (void) lineno;
+
     TRACE((T_CALLED "quickreadf(buffer=%s, file=%s)\n", bp->b_bname, bp->b_fname));
 
     beginDisplay();
@@ -1468,7 +1476,7 @@ quickreadf(BUFFER *bp, int *nlinep)
 #if OPT_ENCRYPT
 	if (b_val(bp, MDCRYPT)
 	    && bp->b_cryptkey[0]) {	/* decrypt the file */
-	    vl_setup_encrypt(bp->b_cryptkey);
+	    vl_setup_encrypt(bp->b_cryptkey, vl_seed);
 	    vl_encrypt_blok((char *) buffer, (UINT) length);
 	}
 #endif
@@ -1491,7 +1499,7 @@ quickreadf(BUFFER *bp, int *nlinep)
 	}
 
 	/* allocate all of the line structs we'll need */
-	if ((bp->b_LINEs = typeallocn(LINE, nlines)) == NULL) {
+	if ((bp->b_LINEs = typeallocn(LINE, (unsigned) nlines)) == NULL) {
 	    free(buffer);
 	    ffrewind();
 	    rc = FIOMEM;
@@ -1513,10 +1521,10 @@ quickreadf(BUFFER *bp, int *nlinep)
 #if !SMALLER
 		lp->l_number = ++lineno;
 #endif
-		llength(lp) = next - offset - 1;
+		llength(lp) = (C_NUM) (next - offset - 1);
 		if (!b_val(bp, MDNEWLINE) && next == length)
 		    llength(lp) += 1;
-		lp->l_size = llength(lp) + 1;
+		lp->l_size = (size_t) (llength(lp) + 1);
 		lvalue(lp) = (char *) (buffer + offset);
 		set_lforw(lp, lp + 1);
 		if (lp != bp->b_LINEs)
@@ -1576,7 +1584,7 @@ readin(char *fname, int lockfl, BUFFER *bp, int mflg)
 #endif
 
     TRACE((T_CALLED "readin(fname=%s, lockfl=%d, bp=%p, mflg=%d)\n",
-	   NONNULL(fname), lockfl, (void *) bp, mflg));
+	   fname, lockfl, (void *) bp, mflg));
 
 #if SYS_WINNT && DISP_NTWIN
     ffshadow = file_is_closed;	/* an assumption */
@@ -1596,8 +1604,7 @@ readin(char *fname, int lockfl, BUFFER *bp, int mflg)
 #if OPT_ENCRYPT
     /* bclear() gets rid of local flags */
     if (local_crypt) {
-	make_local_b_val(bp, MDCRYPT);
-	set_b_val(bp, MDCRYPT, TRUE);
+	set_local_b_val(bp, MDCRYPT, TRUE);
     }
 #endif
 
@@ -1605,16 +1612,14 @@ readin(char *fname, int lockfl, BUFFER *bp, int mflg)
     ch_fname(bp, fname);
     fname = bp->b_fname;	/* this may have been b_fname! */
 #if OPT_DOSFILES
-    make_local_b_val(bp, MDDOS);
     /* assume that if our OS wants it, that the buffer will have CRLF
      * lines.  this may change when the file is read, based on actual
      * line counts, below.  otherwise, if there's an error, or the
      * file doesn't exist, we will keep this default.
      */
-    set_b_val(bp, MDDOS, CRLF_LINES);
+    set_local_b_val(bp, MDDOS, system_crlf);
 #endif
-    make_local_b_val(bp, MDNEWLINE);
-    set_b_val(bp, MDNEWLINE, TRUE);	/* assume we've got it */
+    set_local_b_val(bp, MDNEWLINE, TRUE);
 
     if ((s = ffropen(fname)) == FIOERR) {	/* Hard file error */
 	/* do nothing -- error has been reported,
@@ -1638,10 +1643,10 @@ readin(char *fname, int lockfl, BUFFER *bp, int mflg)
 	    int outlen;
 	    char tmp[NFILEN];
 
-	    outlen = (term.cols - 1) -
-		(sizeof(READING_FILE_FMT) - 3);
+	    outlen = ((term.cols - 1) -
+		      (int) (sizeof(READING_FILE_FMT) - 3));
 	    mlforce(READING_FILE_FMT,
-		    path_trunc(fname, outlen, tmp, sizeof(tmp)));
+		    path_trunc(fname, outlen, tmp, (int) sizeof(tmp)));
 #undef READING_FILE_FMT
 	}
 #if OPT_VMS_PATH
@@ -1881,7 +1886,7 @@ slowreadf(BUFFER *bp, int *nlinep)
 	TRACE(("...try looking for UTF-8\n"));
 	for_each_line(lp, bp) {
 	    if (llength(lp) > 0) {
-		check = check_utf8((UCHAR *) lvalue(lp), llength(lp));
+		check = check_utf8((UCHAR *) lvalue(lp), (B_COUNT) llength(lp));
 		if (check == FALSE) {
 		    found = FALSE;
 		} else if (check == TRUE) {
@@ -1905,6 +1910,57 @@ slowreadf(BUFFER *bp, int *nlinep)
     returnCode(s);
 }
 
+/*
+ * Append to the buffer-name characters up to the allowed length, which will
+ * make a well-formed name.  The 'blanks' option is true to allow embedded
+ * blanks, e.g., for a shell command.  A trailing null is added after the
+ * 'len' bytes have been appended.
+ */
+static char *
+add_to_name(char *target, const char *source, size_t len, int blanks)
+{
+    while (len-- != 0) {
+	UINT ch;
+#if OPT_MULTIBYTE
+	if (vl_encoding >= enc_UTF8) {
+	    int used;
+	    int check;
+	    int code;
+
+	    used = vl_conv_to_utf32(&ch, source, (B_COUNT) len + 1);
+	    if (used <= 0)
+		break;
+	    source += used;
+
+	    if ((code = vl_ucs_to_8bit(&check, (int) ch)) != 0) {
+		if (ch <= 127)
+		    ch = (UINT) check;
+		else
+		    code = 0;
+	    }
+	    if (!code) {
+		if (isPrint(ch)) {
+		    ch = 'X';
+		} else {
+		    ch = '?';
+		}
+	    }
+	} else {
+	    ch = CharOf(*source++);
+	}
+#else
+	ch = CharOf(*source++);
+#endif
+	if (!isPrint(ch))
+	    break;
+	if (isSpace(ch))
+	    ch = (UINT) (blanks ? ' ' : '-');
+	*target++ = (char) ch;
+    }
+    *target = '\0';		/* buffer names are always null-terminated */
+    return target;
+}
+
 /*
  * Take a (null-terminated) file name, and from it fabricate a buffer name.
  * This routine knows about the syntax of file names on the target system.  I
@@ -1916,7 +1972,6 @@ makename(char *bname, const char *fname)
 {
     char *fcp;
     char *bcp;
-    int j;
     char temp[NFILEN];
 
     TRACE(("makename(%s)\n", fname));
@@ -1958,44 +2013,19 @@ makename(char *bname, const char *fname)
 		++fcp;
 	    } while (isSpace(*fcp));
 
-	    (void) vl_strncpy(bcp, fcp, NBUFN - 2);
-
-	    for (j = 4; (j < NBUFN) && isPrint(*bcp); j++) {
-		bcp++;
-	    }
+	    bcp = add_to_name(bcp, fcp, (size_t) (NBUFN - 4), TRUE);
 	    (void) strcpy(bcp, SCRTCH_RIGHT);
 
 	} else {
 
-	    (void) vl_strncpy(bcp, pathleaf(fcp), NBUFN);
+	    (void) add_to_name(bcp, pathleaf(fcp), (size_t) (NBUFN - 1), FALSE);
 
-#if SYS_UNIX
-	    /* UNIX filenames can have any characters (other than EOS!).
-	     * Refuse (rightly) to deal with leading/trailing blanks, but allow
-	     * embedded blanks.  For this special case, ensure that the buffer
-	     * name has no blanks, otherwise it is difficult to reference from
-	     * commands.
-	     */
-	    for (j = 0; j < NBUFN; j++) {
-		if (*bcp == EOS)
-		    break;
-		if (isSpace(*bcp))
-		    *bcp = '-';
-		bcp++;
-	    }
-#endif
 	}
 
 #else /* !(SYS_UNIX||SYS_VMS||SYS_MSDOS) */
 
-	bcp = skip_string(fcp);
-	{
-	    char *cp2 = bname;
-	    vl_strncpy(bname, bcp, NBUFN);
-	    cp2 = strchr(bname, ':');
-	    if (cp2)
-		*cp2 = EOS;
-	}
+	(void) add_to_name(bcp, fcp, NBUFN - 1, FALSE);
+
 #endif
     }
     if (*bname == EOS)
@@ -2020,7 +2050,7 @@ unqname(char *name)
 	j = strlen(strcpy(name, NO_BNAME));
 
     /* check to see if this name is in use */
-    vl_strncpy(newname, name, NBUFN);
+    add_to_name(newname, name, (size_t) (NBUFN - 1), TRUE);
     adjust = is_scratchname(newname);
     strcpy(suffixbuf, "-");
     while (find_b_name(newname) != NULL) {
@@ -2031,8 +2061,11 @@ unqname(char *name)
 	/* that is, put suffix at end if it fits, or else
 	   overwrite some of the name to make it fit */
 	/* the suffix is in "base 36" */
-	suffixlen = format_int(suffixbuf + 1, ++i, 36) + 1;
-	k = NBUFN - 1 - suffixlen;
+	suffixlen = format_int(suffixbuf + 1, (UINT)++ i, 36) + 1;
+	if ((suffixlen + 2) >= NBUFN)
+	    break;
+
+	k = (size_t) (NBUFN - 1 - suffixlen);
 	if (j < k)
 	    k = j;
 	if (adjust) {
@@ -2041,7 +2074,7 @@ unqname(char *name)
 	} else
 	    strcpy(&newname[k], suffixbuf);
     }
-    strncpy0(name, newname, NBUFN);
+    strncpy0(name, newname, (size_t) NBUFN);
     TRACE(("unqname ->%s\n", name));
 }
 
@@ -2055,17 +2088,20 @@ filewrite(int f, int n)
     int s;
     char fname[NFILEN];
     int forced = (f && n == SPECIAL_BANG_ARG);
+    BUFFER *bp = curbp;
 
     if (more_named_cmd()) {
 	if ((s = mlreply_file("Write to file: ", (TBUFF **) 0,
-			      forced ? FILEC_WRITE2
-			      : FILEC_WRITE, fname)) != TRUE)
+			      (UINT) (forced
+				      ? FILEC_WRITE2
+				      : FILEC_WRITE),
+			      fname)) != TRUE)
 	    return s;
     } else
-	(void) vl_strncpy(fname, curbp->b_fname, sizeof(fname));
+	(void) vl_strncpy(fname, bp->b_fname, sizeof(fname));
 
-    if ((s = writeout(fname, curbp, forced, TRUE)) == TRUE)
-	unchg_buff(curbp, 0);
+    if ((s = writeout(fname, bp, forced, TRUE)) == TRUE)
+	unchg_buff(bp, 0);
     return s;
 }
 
@@ -2081,12 +2117,13 @@ filesave(int f, int n)
 {
     int s;
     int forced = (f && n == SPECIAL_BANG_ARG);	/* then it was :w! */
+    BUFFER *bp = curbp;
 
-    if (no_file_name(curbp->b_fname))
+    if (no_file_name(bp->b_fname))
 	return FALSE;
 
-    if ((s = writeout(curbp->b_fname, curbp, forced, TRUE)) == TRUE)
-	unchg_buff(curbp, 0);
+    if ((s = writeout(bp->b_fname, bp, forced, TRUE)) == TRUE)
+	unchg_buff(bp, 0);
     return s;
 }
 
@@ -2129,78 +2166,31 @@ write_encoded_text(BUFFER *bp, const char *buf, int nbuf, const char *ending)
 #define FFPUTLINE(bp,buf,len,end) ffputline(buf,len,end)
 #endif
 
-static int
-actually_write(REGION * rp, char *fn, int msgf, BUFFER *bp, int forced, int encoded)
+int
+write_region(BUFFER *bp, REGION * rp, int encoded, int *nline, B_COUNT * nchar)
 {
     int s;
     LINE *lp;
-    int nline;
-    B_COUNT nchar;
     const char *ending = get_record_sep(bp);
-    int len_rs = len_record_sep(bp);
+    B_COUNT len_rs = (B_COUNT) len_record_sep(bp);
     C_NUM offset = rp->r_orig.o;
 
-    /* this is adequate as long as we cannot write parts of lines */
-    int whole_file = ((rp->r_orig.l == lforw(buf_head(bp)))
-		      && (rp->r_end.l == buf_head(bp)));
-
-    TRACE((T_CALLED " actually_write %s\n",
-	   encoded ? "encoded" : "not encoded"));
-#if OPT_HOOKS
-    if (run_a_hook(&writehook)) {
-	/*
-	 * The write-hook may have modified the buffer.  Assume
-	 * the worst, and reconstruct the region.
-	 */
-	(void) bsizes(bp);
-	if (whole_file
-	    || line_no(bp, rp->r_orig.l) > bp->b_linecount
-	    || line_no(bp, rp->r_end.l) > bp->b_linecount) {
-	    setup_file_region(bp, rp);
-	} else {
-	    DOT = rp->r_orig;
-	    MK = rp->r_end;
-	    (void) getregion(rp);
-	}
-	offset = rp->r_orig.o;
-    }
-#endif
-
-#if OPT_ENCRYPT
-    if ((s = vl_resetkey(curbp, fn)) != TRUE)
-	returnCode(s);
-#endif
-
-    /* open writes error message, if needed */
-    if ((s = ffwopen(fn, forced)) != FIOSUC)
-	returnCode(FALSE);
-
-    /* disable "working..." while we are writing - not reading - a pipe, since
-     * it may be a quasi-interactive process that we don't want to modify its
-     * display.
-     */
-    if (isShellOrPipe(fn)) {
-	beginDisplay();
-    }
-
-    /* tell us we're writing */
-    if (msgf == TRUE)
-	mlwrite("[Writing...]");
-
-    CleanToPipe(TRUE);
+    (void) encoded;
 
     lp = rp->r_orig.l;
-    nline = 0;			/* Number of lines     */
-    nchar = 0;			/* Number of chars     */
+    *nline = 0;			/* Number of lines     */
+    *nchar = 0;			/* Number of chars     */
 
 #if OPT_MULTIBYTE
     if ((s = write_bom(bp)) != FIOSUC) {
 	goto out;
     }
+#else
+    s = FIOSUC;
 #endif
 
     /* first (maybe partial) line and succeeding whole lines */
-    while ((rp->r_size + offset) >= line_length(lp)) {
+    while ((rp->r_size + (B_COUNT) offset) >= (B_COUNT) line_length(lp)) {
 	C_NUM len = llength(lp) - offset;
 	char *text = lvalue(lp) + offset;
 
@@ -2232,21 +2222,113 @@ actually_write(REGION * rp, char *fn, int msgf, BUFFER *bp, int forced, int enco
 	if ((s = FFPUTLINE(bp, text, len, ending)) != FIOSUC)
 	    goto out;
 
-	++nline;
-	nchar += line_length(lp);
+	*nline += 1;
+	*nchar += line_length(lp);
 	offset = 0;
 	lp = lforw(lp);
     }
 
     /* last line (fragment) */
     if (rp->r_size > 0) {
-	if ((s = FFPUTLINE(bp, lvalue(lp), rp->r_size, NULL)) != FIOSUC)
+#if OPT_SELECTIONS
+	C_NUM len = llength(lp);
+	char *text = lvalue(lp);
+
+	if (encoded) {
+	    TBUFF *temp;
+	    if ((temp = encode_attributes(lp, bp, rp)) != 0) {
+		text = tb_values(temp);
+		len = (int) tb_length(temp);
+	    }
+	    if ((s = FFPUTLINE(bp, text, len, NULL)) != FIOSUC) {
+		tb_free(&temp);
+		goto out;
+	    }
+	    tb_free(&temp);
+	} else
+#endif
+	if ((s = FFPUTLINE(bp, lvalue(lp), (int) rp->r_size, NULL)) != FIOSUC)
 	    goto out;
-	nchar += rp->r_size;
-	++nline;		/* it _looks_ like a line */
+	*nchar += rp->r_size;
+	*nline += 1;		/* it _looks_ like a line */
     }
 
   out:
+    return s;
+}
+
+static int
+actually_write(REGION * rp, char *fn, int msgf, BUFFER *bp, int forced, int encoded)
+{
+    int s;
+    int nline;
+    B_COUNT nchar;
+
+    /* this is adequate as long as we cannot write parts of lines */
+    int whole_file = ((rp->r_orig.l == lforw(buf_head(bp)))
+		      && (rp->r_end.l == buf_head(bp)));
+
+    TRACE((T_CALLED "actually_write %s\n",
+	   encoded ? "encoded" : "not encoded"));
+#if OPT_HOOKS
+    if (run_a_hook(&writehook)) {
+	/*
+	 * The write-hook may have modified the buffer.  Assume
+	 * the worst, and reconstruct the region.
+	 */
+	(void) bsizes(bp);
+	if (whole_file
+	    || line_no(bp, rp->r_orig.l) > bp->b_linecount
+	    || line_no(bp, rp->r_end.l) > bp->b_linecount) {
+	    setup_file_region(bp, rp);
+	} else {
+	    DOT = rp->r_orig;
+	    MK = rp->r_end;
+	    (void) getregion(bp, rp);
+	}
+    }
+    /*
+     * The write-hook may have set the buffer to view-only, or returned
+     * an error.  Do not write the buffer in that case.
+     */
+    switch (s2status(tb_values(last_macro_result))) {
+    case TRUE:
+    case SORTOFTRUE:
+	break;
+    default:
+	mlwarn("[write-hook returned %s]", tb_values(last_macro_result));
+	returnCode(FALSE);
+    }
+    if (same_fname(fn, bp, FALSE) && b_val(bp, MDVIEW)) {
+	mlwarn("[Buffer view-mode has changed]");
+	returnCode(FALSE);
+    }
+#endif
+
+#if OPT_ENCRYPT
+    if ((s = vl_resetkey(bp, fn)) != TRUE)
+	returnCode(s);
+#endif
+
+    /* open writes error message, if needed */
+    if (ffwopen(fn, forced) != FIOSUC)
+	returnCode(FALSE);
+
+    /* disable "working..." while we are writing - not reading - a pipe, since
+     * it may be a quasi-interactive process that we don't want to modify its
+     * display.
+     */
+    if (isShellOrPipe(fn)) {
+	beginDisplay();
+    }
+
+    /* tell us we're writing */
+    if (msgf == TRUE)
+	mlwrite("[Writing...]");
+
+    CleanToPipe(TRUE);
+
+    s = write_region(bp, rp, encoded, &nline, &nchar);
     if (s == FIOSUC) {		/* No write error.      */
 #if OPT_VMS_PATH
 	if (same_fname(fn, bp, FALSE))
@@ -2303,7 +2385,7 @@ file_protection(char *fn)
     if (!isShellOrPipe(fn)) {
 	if (file_stat(fn, &sb) == 0) {
 	    if (isFileMode(sb.st_mode))
-		result = sb.st_mode & 0777;
+		result = (int) (sb.st_mode & 0777);
 	}
     }
     return result;
@@ -2367,12 +2449,13 @@ writereg(REGION * rp,
 #endif
 			if (!ffaccess(fn, FL_WRITEABLE)
 			    && (protection = file_protection(fn)) >= 0) {
-			    chmod(SL_TO_BSL(fn), protection | 0600);
+			    (void) chmod(SL_TO_BSL(fn), (mode_t) (protection
+								  | 0600));
 			}
 		    }
 		    status = actually_write(rp, fn, msgf, bp, forced, encoded);
 		    if (protection > 0)
-			chmod(SL_TO_BSL(fn), protection);
+			(void) chmod(SL_TO_BSL(fn), (mode_t) protection);
 #if SYS_WINNT
 		    if (write_acl_added)
 			(void) w32_remove_write_acl(fn, prev_access_mask);
@@ -2417,6 +2500,7 @@ writeout(const char *fn, BUFFER *bp, int forced, int msgf)
 static int
 prompt_and_write_region(int encoded)
 {
+    BUFFER *bp = curbp;
     REGION region;
     int status;
     char fname[NFILEN];
@@ -2426,15 +2510,15 @@ prompt_and_write_region(int encoded)
 	    mlwrite("Range not written");
 	    return FALSE;
 	}
-	(void) vl_strncpy(fname, curbp->b_fname, sizeof(fname));
+	(void) vl_strncpy(fname, bp->b_fname, sizeof(fname));
     } else {
 	if ((status = mlreply_file("Write region to file: ",
 				   (TBUFF **) 0, FILEC_WRITE | FILEC_PROMPT,
 				   fname)) != TRUE)
 	    return status;
     }
-    if ((status = getregion(&region)) == TRUE)
-	status = writereg(&region, fname, TRUE, curbp, FALSE, encoded);
+    if ((status = getregion(bp, &region)) == TRUE)
+	status = writereg(&region, fname, TRUE, bp, FALSE, encoded);
     return status;
 }
 
@@ -2462,12 +2546,15 @@ write_enc_region(void)
 int
 kwrite(char *fn, int msgf)
 {
+#if OPT_ENCRYPT
+    BUFFER *bp = curbp;
+#endif
     KILL *kp;			/* pointer into kill register */
     int nline;
     B_COUNT nchar;
     int s;
     int c;
-    int i;
+    unsigned i;
     char *sp;			/* pointer into string to insert */
 
     /* make sure there is something to put */
@@ -2477,7 +2564,7 @@ kwrite(char *fn, int msgf)
 	return FALSE;		/* not an error, just nothing */
     }
 #if OPT_ENCRYPT
-    if ((s = vl_resetkey(curbp, fn)) != TRUE)
+    if ((s = vl_resetkey(bp, fn)) != TRUE)
 	return s;
 #endif
     if ((s = ffwopen(fn, FALSE)) != FIOSUC) {	/* Open writes message. */
@@ -2528,6 +2615,7 @@ kwrite(char *fn, int msgf)
 int
 vl_filename(int f GCC_UNUSED, int n GCC_UNUSED)
 {
+    BUFFER *bp = curbp;
     int s;
     char fname[NFILEN];
 
@@ -2543,19 +2631,19 @@ vl_filename(int f GCC_UNUSED, int n GCC_UNUSED)
 
 #if OPT_LCKFILES
     if (global_g_val(GMDUSEFILELOCK)) {
-	if (!b_val(curbp, MDLOCKED) && !b_val(curbp, MDVIEW))
-	    release_lock(curbp->b_fname);
-	ch_fname(curbp, fname);
-	make_global_b_val(curbp, MDLOCKED);
-	make_global_b_val(curbp, VAL_LOCKER);
-	grab_lck_file(curbp, fname);
+	if (!b_val(bp, MDLOCKED) && !b_val(bp, MDVIEW))
+	    release_lock(bp->b_fname);
+	ch_fname(bp, fname);
+	make_global_b_val(bp, MDLOCKED);
+	make_global_b_val(bp, VAL_LOCKER);
+	grab_lck_file(bp, fname);
     } else
 #endif
-	ch_fname(curbp, fname);
+	ch_fname(bp, fname);
 
 #if SYS_WINNT && DISP_NTWIN
     /* remember new filename if later written out to disk */
-    b_clr_registered(curbp);
+    b_clr_registered(bp);
 #endif
 
     curwp->w_flag |= WFMODE;
@@ -2570,6 +2658,7 @@ vl_filename(int f GCC_UNUSED, int n GCC_UNUSED)
 int
 ifile(char *fname, int belowthisline, FILE *haveffp)
 {
+    BUFFER *bp = curbp;
     LINE *prevp;
     LINE *newlp;
     LINE *nextp;
@@ -2580,7 +2669,7 @@ ifile(char *fname, int belowthisline, FILE *haveffp)
     TRACE((T_CALLED "ifile(fname=%s, belowthisline=%d, haveffp=%p)\n",
 	   NONNULL(fname), belowthisline, (void *) haveffp));
 
-    b_clr_invisible(curbp);	/* we are not temporary */
+    b_clr_invisible(bp);	/* we are not temporary */
     if (haveffp == 0) {
 	if ((status = ffropen(fname)) == FIOERR)	/* Hard file open */
 	    goto out;
@@ -2589,13 +2678,13 @@ ifile(char *fname, int belowthisline, FILE *haveffp)
 #if COMPLETE_FILES || COMPLETE_DIRS
 	else if ((status = ff_load_directory(fname)) == FIOERR)
 	    goto out;
-	if (b_is_directory(curbp)) {
+	if (b_is_directory(bp)) {
 	    /* special case: contents are already added */
 	    goto out;
 	}
 #endif
 #if OPT_ENCRYPT
-	if ((status = vl_resetkey(curbp, fname)) != TRUE)
+	if ((status = vl_resetkey(bp, fname)) != TRUE)
 	    returnCode(status);
 #endif
 	mlwrite("[Inserting...]");
@@ -2605,14 +2694,14 @@ ifile(char *fname, int belowthisline, FILE *haveffp)
 	ffp = haveffp;
     }
     prevp = DOT.l;
-    DOT.o = 0;
+    DOT.o = b_left_margin(curbp);
     MK = DOT;
 
     nline = 0;
     nextp = 0;
     while ((status = ffgetline(&nbytes)) <= FIOSUC) {
 #if OPT_DOSFILES
-	if (b_val(curbp, MDDOS)
+	if (b_val(bp, MDDOS)
 	    && (nbytes != 0)
 	    && fflinebuf[nbytes - 1] == '\r')
 	    nbytes--;
@@ -2623,12 +2712,12 @@ ifile(char *fname, int belowthisline, FILE *haveffp)
 	}
 
 	beginDisplay();
-	if (add_line_at(curbp, prevp, fflinebuf, (int) nbytes) != TRUE) {
+	if (add_line_at(bp, prevp, fflinebuf, (int) nbytes) != TRUE) {
 	    status = FIOMEM;
 	    newlp = 0;
 	} else {
 	    newlp = lforw(prevp);
-	    if (OkUndo(curbp)) {
+	    if (OkUndo(bp)) {
 		if ((tag_for_undo(newlp)) == ABORT) {
 		    status = FIOMEM;
 		    /*
@@ -2636,7 +2725,7 @@ ifile(char *fname, int belowthisline, FILE *haveffp)
 		     * If we don't do this, undoworker() won't find a match
 		     * against the buffer, and will be deadlocked.
 		     */
-		    lremove(curbp, prevp);
+		    lremove(bp, prevp);
 		}
 	    }
 	}
@@ -2659,22 +2748,23 @@ ifile(char *fname, int belowthisline, FILE *haveffp)
     /* mark the window for changes.  could this be moved up to
      * where we actually insert a line? */
     if (nline)
-	chg_buff(curbp, WFHARD);
+	chg_buff(bp, WFHARD);
 
   out:
     /* copy window parameters back to the buffer structure */
-    copy_traits(&(curbp->b_wtraits), &(curwp->w_traits));
+    copy_traits(&(bp->b_wtraits), &(curwp->w_traits));
 
     if (status < FIOERR)
 	imply_alt(fname, FALSE, FALSE);
 
 #if COMPLETE_FILES || COMPLETE_DIRS
-    if (b_is_directory(curbp)) {
+    if (b_is_directory(bp)) {
 	DOT.l = lback(DOT.l);
     } else
 #endif
 	/* advance to the next line */
 	DOT.l = lforw(DOT.l);
+    DOT.o = b_left_margin(curbp);
 
     returnCode(FIO2Status(status));
 }
@@ -2714,7 +2804,7 @@ create_save_dir(char *dirnam)
 
     for (; n < TABLESIZE(tbl); n++) {
 	if (is_directory(tbl[n])) {
-	    int omask = vl_umask(0077);
+	    mode_t omask = vl_umask(0077);
 	    (void) pathcat(dirnam, tbl[n], "vileDXXXXXX");
 
 	    /* on failure, keep going */
@@ -2874,9 +2964,9 @@ imdying(int ACTUAL_SIG_ARGS)
 			      "echo errors:  check in current",
 			      "directory too.;");
 
-	    cp = lsprintf(cp, ") | %s %s", *mailcmdp, np);
+	    (void) lsprintf(cp, ") | %s %s", *mailcmdp, np);
 
-	    (void) system(cmd);
+	    IGNORE_RC(system(cmd));
 	}
     }
     term.cursorvis(TRUE);	/* ( this might work ;-) */
