@@ -2,7 +2,7 @@
  * The routines in this file read and write ASCII files from the disk. All of
  * the knowledge about files are here.
  *
- * $Header: /usr/build/vile/vile/RCS/fileio.c,v 1.192 2008/11/25 23:40:31 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/fileio.c,v 1.206 2014/09/30 21:12:07 tom Exp $
  *
  */
 
@@ -183,9 +183,9 @@ make_backup(char *fname)
 
     if (ffexists(fname)) {	/* if the file exists, attempt a backup */
 	char tname[NFILEN];
-	char *s = pathleaf(strcpy(tname, fname));
+	char *s = pathleaf(vl_strncpy(tname, fname, sizeof(tname)));
 	char *t = strrchr(s, '.');
-	char *gvalfileback = global_g_val_ptr(GVAL_BACKUPSTYLE);
+	char *gvalfileback = b_val_ptr(curbp, VAL_BACKUPSTYLE);
 
 	switch (choice_to_code(&fsm_backup_blist, gvalfileback, strlen(gvalfileback))) {
 	case bak_BAK:
@@ -274,18 +274,19 @@ file_stat(const char *fn, struct stat *sb)
 		    if (strlen(cache[n].fn) == 0)
 			break;
 		}
-		cache[n].rc = stat(SL_TO_BSL(strcpy(cache[n].fn, fn)),
+		cache[n].rc = stat(SL_TO_BSL(vl_strncpy(cache[n].fn, fn,
+							sizeof(cache[n].fn))),
 				   &(cache[n].sb));
 	    }
 	    rc = cache[n].rc;
 	    *sb = cache[n].sb;
 	} else if (found) {
-	    vl_strncpy(cache[n].fn, "", NFILEN);
+	    vl_strncpy(cache[n].fn, "", (size_t) NFILEN);
 	}
     } else {
 	fn = "";
 	for (n = 0; n < TABLESIZE(cache); ++n) {
-	    vl_strncpy(cache[n].fn, fn, NFILEN);
+	    vl_strncpy(cache[n].fn, fn, (size_t) NFILEN);
 	}
     }
     TRACE(("file_stat(%s) = %d%s\n", fn, rc,
@@ -373,6 +374,8 @@ ffwopen(char *fn, int forced)
     char *name;
     const char *mode = FOPEN_WRITE;
 
+    (void) forced;
+
     TRACE(("ffwopen(fn=%s, forced=%d)\n", fn, forced));
     ffstatus = file_is_closed;
     if (i_am_dead) {
@@ -409,7 +412,7 @@ ffwopen(char *fn, int forced)
 	    return (FIOERR);
 	}
 
-	if (*global_g_val_ptr(GVAL_BACKUPSTYLE) != 'o') {	/* "off" ? */
+	if (*b_val_ptr(curbp, VAL_BACKUPSTYLE) != 'o') {	/* "off" ? */
 	    if (!make_backup(fn)) {
 		if (!forced) {
 		    mlerror("making backup file");
@@ -459,8 +462,12 @@ ffaccess(char *fn, UINT mode)
      * the symbols for portability.
      */
 #ifndef X_OK
+#if defined(_WIN32)
+#define X_OK 0
+#else
 #define X_OK 1
 #endif
+#endif
 #ifndef W_OK
 #define W_OK 2
 #endif
@@ -545,14 +552,16 @@ ffsize(B_COUNT * have)
     long prev;
 
     prev = ftell(ffp);
+    *have = 0;
     if (fseek(ffp, 0, SEEK_END) >= 0) {
 	result = 0;
 	*have = ftell(ffp);
-	fseek(ffp, prev, SEEK_SET);
+	(void) fseek(ffp, prev, SEEK_SET);
     }
 #else
     struct stat statbuf;
 
+    *have = 0;
     if ((result = fstat(fileno(ffp), &statbuf)) == 0) {
 	*have = (B_COUNT) statbuf.st_size;
     }
@@ -641,9 +650,9 @@ ffread(char *buf, B_COUNT want, B_COUNT * have)
 	 * avoid overflow.
 	 */
 	while (want != 0) {
-	    long ask = (((want - *have) > LONG_MAX)
-			? LONG_MAX
-			: (want - *have));
+	    size_t ask = (((want - *have) > LONG_MAX)
+			  ? LONG_MAX
+			  : (want - *have));
 
 #if FFREAD_FREAD
 	    /* size_t may not fit in long, making a sign-extension */
@@ -653,7 +662,7 @@ ffread(char *buf, B_COUNT want, B_COUNT * have)
 #endif
 	    if (got <= 0)
 		break;
-	    *have += got;
+	    *have += (B_COUNT) got;
 	}
 	if (*have == 0)
 	    result = -1;
@@ -668,7 +677,7 @@ ffseek(B_COUNT n)
 #if SYS_VMS
     ffrewind();			/* see below */
 #endif
-    fseek(ffp, n, SEEK_SET);
+    (void) fseek(ffp, (long) n, SEEK_SET);
 }
 
 void
@@ -685,7 +694,7 @@ ffrewind(void)
     (void) fclose(ffp);
     ffp = fopen(temp, FOPEN_READ);
 #else
-    fseek(ffp, 0L, SEEK_SET);
+    (void) fseek(ffp, 0L, SEEK_SET);
 #endif
 }
 
@@ -699,7 +708,7 @@ ffclose(void)
 
     if (ffstatus == file_is_unbuffered) {
 	if (fflinelen) {
-	    write(ffd, fflinebuf, fflinelen);
+	    IGNORE_RC(write(ffd, fflinebuf, (unsigned) fflinelen));
 	    fflinelen = 0;
 	}
 	close(ffd);
@@ -779,12 +788,12 @@ ffputc(int c)
 
 #if OPT_ENCRYPT
     if (ffcrypting && (ffstatus != file_is_pipe))
-	d = vl_encrypt_char(d);
+	d = (char) vl_encrypt_char(d);
 #endif
     if (i_am_dead) {
 	fflinebuf[fflinelen++] = d;
 	if (fflinelen >= NSTRING) {
-	    write(ffd, fflinebuf, fflinelen);
+	    IGNORE_RC(write(ffd, fflinebuf, (unsigned) fflinelen));
 	    fflinelen = 0;
 	}
     } else {
@@ -823,7 +832,7 @@ alloc_linebuf(size_t needed)
 }
 
 #define ALLOC_LINEBUF(i) \
-	    if ((i) >= fflinelen && !alloc_linebuf(i)) \
+	    if ((i) >= fflinelen && !alloc_linebuf((size_t) i)) \
 		return (FIOMEM)
 
 /*
@@ -853,7 +862,7 @@ ffgetline(size_t *lenp)
 	    return (FIOEOF);
 	}
 	if (llength(ffcursor) > 0) {
-	    i = llength(ffcursor);
+	    i = (size_t) llength(ffcursor);
 	    ALLOC_LINEBUF(i);
 	    memcpy(fflinebuf, lvalue(ffcursor), i);
 	}
@@ -882,7 +891,7 @@ ffgetline(size_t *lenp)
 		 */
 		if (count_fline == 0) {
 		    UCHAR *buffer = (UCHAR *) fflinebuf;
-		    B_COUNT length = i;
+		    B_COUNT length = (B_COUNT) i;
 		    make_global_b_val(btempp, VAL_BYTEORDER_MARK);
 		    make_global_b_val(btempp, VAL_FILE_ENCODING);
 		    if (decode_bom(btempp, buffer, &length))
@@ -892,14 +901,14 @@ ffgetline(size_t *lenp)
 
 		if (b_val(btempp, VAL_FILE_ENCODING) > enc_UTF8) {
 		    UCHAR *buffer = (UCHAR *) fflinebuf;
-		    B_COUNT length = (i + 1);
+		    B_COUNT length = (B_COUNT) (i + 1);
 
 		    if (!aligned_charset(btempp, buffer, &length)) {
 			do {
 			    ALLOC_LINEBUF(i + 2);
 			    fflinebuf[i++] = (char) c;
 			    c = vl_getc(ffp);	/* expecting a null... */
-			    length = (i + 1);
+			    length = (B_COUNT) (i + 1);
 			    buffer = (UCHAR *) fflinebuf;
 			} while (!aligned_charset(btempp, buffer, &length));
 			fflinebuf[i] = (char) c;
@@ -988,8 +997,8 @@ ffhasdata(void)
 #endif
 #if defined(FIONREAD) && !SYS_WINNT
     {
-	long x;
-	if ((ioctl(fileno(ffp), FIONREAD, (void *) & x) >= 0) && x != 0)
+	long x = 0;
+	if ((ioctl(fileno(ffp), (long) FIONREAD, (void *) &x) >= 0) && x != 0)
 	    return TRUE;
     }
 #endif
