@@ -1,5 +1,5 @@
 /*
- * $Header: /usr/build/vile/vile/filters/RCS/pl-filt.c,v 1.98 2009/04/25 16:12:05 tom Exp $
+ * $Id: pl-filt.c,v 1.125 2016/12/17 17:30:10 tom Exp $
  *
  * Filter to add vile "attribution" sequences to perl scripts.  This is a
  * translation into C of an earlier version written for LEX/FLEX.
@@ -8,9 +8,9 @@
 #include <filters.h>
 
 #ifdef DEBUG
-DefineOptFilter("perl", "d");
+DefineOptFilter(perl, "d");
 #else
-DefineFilter("perl");
+DefineFilter(perl);
 #endif
 
 /*
@@ -18,7 +18,7 @@ DefineFilter("perl");
  * should consider '$' and '&', but reading the parser hints that might work
  * too).
  */
-#define QUOTE_DELIMS "#:/?|!:%',{}[]()@;=~\""
+#define QUOTE_DELIMS "$+&#:/?|!:%',{}[]()@;=~\""
 /* from Perl's S_scan_str() */
 #define LOOKUP_TERM "([{< )]}> )]}>"
 
@@ -48,7 +48,7 @@ typedef struct {
     int has_no_pattern;
 } AfterKey;
 
-static char *put_embedded(char *, int, char *);
+static char *put_embedded(char *, int, const char *);
 
 static char *Action_attr;
 static char *Comment_attr;
@@ -114,7 +114,7 @@ count_blanks(char *s)
 	}
 	++s;
     }
-    return s - base;
+    return (int) (s - base);
 }
 
 /******************************************************************************
@@ -186,15 +186,15 @@ is_STRINGS(char *s, int *err, int delim)
 	    }
 	    s++;
 	}
-	found = s - base;
+	found = (int) (s - base);
     }
     return found;
 }
 
 static int
-is_KEYWORD(char *s)
+is_KEYWORD(const char *s)
 {
-    char *base = s;
+    const char *base = s;
     int ch;
     int quote = 0;
 
@@ -223,7 +223,7 @@ is_KEYWORD(char *s)
     }
     if (quote == (s - base))
 	s = base;
-    return s - base;
+    return (int) (s - base);
 }
 
 /*
@@ -256,7 +256,7 @@ is_QIDENT(char *s)
 		break;
 	}
     }
-    return s - base;
+    return (int) (s - base);
 }
 
 /*
@@ -264,9 +264,9 @@ is_QIDENT(char *s)
  * If not double-quoted, look also for syntax such as $foo'bar
  */
 static int
-is_NORMALVARS(char *s, int dquoted)
+is_NORMALVARS(const char *s, int dquoted)
 {
-    char *base = s;
+    const char *base = s;
     int ch;
     int squoted = 0;
     int part1 = 0;
@@ -275,7 +275,7 @@ is_NORMALVARS(char *s, int dquoted)
     while (MORE(s)) {
 	ch = CharOf(*s);
 	if (s == base) {
-	    if (strchr(dquoted ? "$" : "&$%@", ch) == 0) {
+	    if (vl_index(dquoted ? "$" : "&$%@", ch) == 0) {
 		break;
 	    }
 	} else if (squoted && !dquoted) {
@@ -298,13 +298,13 @@ is_NORMALVARS(char *s, int dquoted)
 	}
 	s++;
     }
-    return (part1 && (dquoted || (squoted == part2))) ? (s - base) : 0;
+    return (part1 && (dquoted || (squoted == part2))) ? (int) (s - base) : 0;
 }
 
 static int
-is_OTHERVARS(char *s)
+is_OTHERVARS(const char *s)
 {
-    char *base = s;
+    const char *base = s;
     int ch;
     int part1 = 0;
     int part2 = 0;
@@ -337,7 +337,7 @@ is_OTHERVARS(char *s)
 	}
 	s++;
     }
-    return (part1 || part2) ? (s - base) : 0;
+    return (part1 || part2) ? (int) (s - base) : 0;
 }
 
 static int
@@ -400,9 +400,10 @@ is_NUMBER(char *s, int *err)
 	    } else {
 		break;
 	    }
-	} else if (value && (ch == 'e' || ch == 'E')) {
-	    if (state >= 2 || !value)
+	} else if (value && (radix != 16) && (ch == 'e' || ch == 'E')) {
+	    if (state >= 2 || !value) {
 		*err = 1;
+	    }
 	    state = 2;
 	} else {
 	    if (((state || (radix == 10)) && isdigit(ch))
@@ -428,7 +429,7 @@ is_NUMBER(char *s, int *err)
 	s++;
     }
 
-    return value ? (s - base) : 0;
+    return value ? (int) (s - base) : 0;
 }
 
 static char *
@@ -460,7 +461,7 @@ is_ELLIPSIS(char *s)
 }
 
 static int
-is_IDENT(char *s, int quoted)
+is_IDENT(const char *s, int quoted)
 {
     int found;
 
@@ -498,7 +499,7 @@ is_COMMENT(char *s)
 	}
 	s = t;
     }
-    return (s - base);
+    return (int) (s - base);
 }
 
 /* preprocessor stuff:
@@ -516,7 +517,7 @@ is_PREPROC(char *s)
     if (*s++ != '#')
 	return 0;
     s += is_BLANK(s);
-    if (strncmp(s, "line", 4))
+    if (strncmp(s, "line", (size_t) 4))
 	return 0;
     if ((skip = is_BLANK(s += 4)) == 0)
 	return 0;
@@ -530,18 +531,18 @@ is_PREPROC(char *s)
 	    return 0;
     }
     s += is_BLANK(s);
-    return (s - base);
+    return (int) (s - base);
 }
 
 /******************************************************************************
  ******************************************************************************/
 
 static int
-end_marker(char *s, char *marker, int only)
+end_marker(char *s, const char *marker, int only)
 {
-    int len = strlen(marker);
+    size_t len = strlen(marker);
 
-    return (ATLEAST(s, len)
+    return (ATLEAST(s, (int) len)
 	    && !strncmp(s, marker, len)
 	    && (!only || (s[len] == '\n')));
 }
@@ -552,11 +553,23 @@ end_marker(char *s, char *marker, int only)
  * "=" (the "=" is painted in a different color).  Otherwise return 0.
  */
 static int
-begin_POD(char *s)
+begin_POD(char *s, int emit)
 {
     int result = 0;
-    char *base = s;
-    int skip = 0;
+    char *base;
+    int skip = 2;
+    int maybe = 0;
+
+    while (s > the_file) {
+	if (s[0] != '\n' ||
+	    s[-1] != '\n' ||
+	    --skip <= 0) {
+	    break;
+	}
+	--s;
+    }
+    skip = 0;
+    base = s;
 
     while (MORE(s)) {
 	if (*s == '\n')
@@ -565,21 +578,36 @@ begin_POD(char *s)
 	    break;
 	++s;
     }
+    if ((base != the_file) && (skip == 1)) {
+	maybe = 1;
+	++skip;
+    }
     if (((base == the_file) || (skip >= 2))
 	&& ATLEAST(s, 2)
 	&& s[0] == '='
 	&& isalpha(CharOf(s[1]))) {
-	result = s + 1 - base;
-	while (base != s) {
-	    flt_putc(*base++);	/* skip the newlines */
+	result = (int) (s + 1 - base);
+	if (result && maybe)
+	    flt_error("expected a blank line");
+	if (emit) {
+	    while (base != s) {
+		flt_putc(*base++);	/* skip the newlines */
+	    }
 	}
     }
     return result;
 }
 
 static int
-end_POD(char *s)
+end_POD(char *s, int skip)
 {
+    if (skip) {
+	while (MORE(s)) {
+	    if (*s != '\n')
+		break;
+	    ++s;
+	}
+    }
     return end_marker(s, "=cut", 0);
 }
 
@@ -601,8 +629,8 @@ begin_HERE(char *s, int *quoted)
 	s += count_blanks(s);
 	base = s;
 	if ((ok = is_QIDENT(s)) != 0) {
-	    unsigned temp = 0;
-	    char *marker = do_alloc((char *) 0, ok + 1, &temp);
+	    size_t temp = 0;
+	    char *marker = do_alloc((char *) 0, (size_t) (ok + 1), &temp);
 	    char *d = marker;
 
 	    s += ok;
@@ -665,7 +693,7 @@ line_size(char *s)
 	    break;
 	s++;
     }
-    return s - base;
+    return (int) (s - base);
 }
 
 /*
@@ -683,6 +711,46 @@ begin_PATTERN(char *s)
     return 0;
 }
 
+/*
+ * Check for simple pattern, trying to distinguish '/' from its use for divide,
+ * and '$' from its use in variable names.
+ */
+static int
+is_PATTERN(char *s)
+{
+    char *base = s;
+    int delim = *s++;
+    int result = 0;
+    int escaped = 0;
+    char *content = s;
+
+    while (MORE(s)) {
+	int ch = *s++;
+	if (escaped) {
+	    escaped = 0;
+	} else if (ch == BACKSLASH) {
+	    escaped = 1;
+	} else if (ch == delim) {
+	    if (((s - 1) - content) != 0) {
+		/* content is nonempty */
+		char *next = 0;
+		(void) strtol(content, &next, 0);
+		if (next == 0 || next == content) {
+		    /* content was not a number */
+		    result = (int) (s - base);
+		}
+	    }
+	    break;
+	} else if (isspace(ch)) {
+	    break;
+	} else if (ch == '$') {
+	    if (*s != delim)
+		break;
+	}
+    }
+    return result;
+}
+
 /*
  * If we're pointing to a quote-like operator, return its length.
  * As a side-effect, set the number of delimiters (assuming '/') it needs.
@@ -692,12 +760,13 @@ is_QUOTE(char *s, int *delims)
 {
     char *base = s;
     size_t len;
+    int result = 0;
 
     *delims = 0;
     while (MORE(s) && isIdent(*s)) {
 	++s;
     }
-    if ((len = (s - base)) != 0) {
+    if ((len = (size_t) (s - base)) != 0) {
 	switch (len) {
 	case 1:
 	    if (*base == 'm' || *base == 'q') {
@@ -707,12 +776,12 @@ is_QUOTE(char *s, int *delims)
 	    }
 	    break;
 	case 2:
-	    if (!strncmp(base, "tr", 2)) {
+	    if (!strncmp(base, "tr", (size_t) 2)) {
 		*delims = 3;
-	    } else if (!strncmp(base, "qq", 2) ||
-		       !strncmp(base, "qx", 2) ||
-		       !strncmp(base, "qw", 2) ||
-		       !strncmp(base, "qr", 2)) {
+	    } else if (!strncmp(base, "qq", (size_t) 2) ||
+		       !strncmp(base, "qx", (size_t) 2) ||
+		       !strncmp(base, "qw", (size_t) 2) ||
+		       !strncmp(base, "qr", (size_t) 2)) {
 		*delims = 2;
 	    }
 	    break;
@@ -721,18 +790,20 @@ is_QUOTE(char *s, int *delims)
     if (*delims == 0)
 	s = base;
 
-    if (s != base) {
+    if ((result = (int) (s - base)) != 0) {
 	int test = char_after_blanks(s);
-	DPRINTF(("is_Quote(%.*s:%c)", (int) (s - base), base, test));
+	DPRINTF(("is_Quote(%.*s:%c)", result, base, test));
 	if (test == '#' && isspace(CharOf(*s)))
 	    test = 0;
-	if ((test == 0) || (strchr(QUOTE_DELIMS, test) == 0))
+	if ((test == 0) || (strchr(QUOTE_DELIMS "<>", test) == 0)) {
 	    s = base;
+	    result = 0;
+	}
 	DPRINTF(("is_QUOTE(%.*s)\n",
-		 (s != base) ? (int) (s - base) : 1,
-		 (s != base) ? base : ""));
+		 result ? result : 1,
+		 result ? base : ""));
     }
-    return (s - base);
+    return result;
 }
 
 static int
@@ -754,6 +825,7 @@ add_to_PATTERN(char *s)
 	int ignored = 0;
 	int escaped = 0;
 	int comment = 0;
+	int in_blocks = 0;
 	int in_curlys = 0;
 	int in_parens = 0;
 
@@ -791,8 +863,19 @@ add_to_PATTERN(char *s)
 		    escaped = 1;
 		} else {
 		    if (!escaped) {
+			ignored = 0;
 
 			switch (*s) {
+			case L_BLOCK:
+			    if (in_blocks == 0)
+				in_blocks = 1;
+			    break;
+			case R_BLOCK:
+			    if (in_blocks != 0) {
+				in_blocks = 0;
+				ignored = 1;
+			    }
+			    break;
 			case L_CURLY:
 			    ++in_curlys;
 			    break;
@@ -812,7 +895,8 @@ add_to_PATTERN(char *s)
 			if (*s == '#' && delim == R_CURLY && !in_curlys)
 			    comment = 1;
 
-			ignored = 0;
+			if (delim == L_BLOCK && in_blocks)
+			    ignored = 1;
 			if (delim == L_CURLY && in_curlys)
 			    ignored = 1;
 			if (delim == L_PAREN && in_parens)
@@ -827,7 +911,7 @@ add_to_PATTERN(char *s)
 			     */
 			    if (ATLEAST(s, 4)
 				&& need == 1
-				&& !strncmp(s + 1, "...", 3))
+				&& !strncmp(s + 1, "...", (size_t) 3))
 				need += 2;
 			    /*
 			     * check for /pattern/ {pattern}
@@ -849,7 +933,7 @@ add_to_PATTERN(char *s)
 	    s++;
 	}
 	DPRINTF(("*finally\n->%.*s\n", line_size(s), s));
-	return (s - base);
+	return (int) (s - base);
     }
     return 0;
 }
@@ -880,7 +964,7 @@ write_PATTERN(char *s, int len)
 	len -= skip;
     }
 
-    skip = skip_BLANKS(s) - s;
+    skip = (int) (skip_BLANKS(s) - s);
     if (skip) {
 	s += skip;
 	len -= skip;
@@ -955,7 +1039,7 @@ write_PATTERN(char *s, int len)
 	}
     }
     if (comment)
-	first += (skip_BLANKS(s) - s);
+	first += (int) (skip_BLANKS(s) - s);
     if (comment) {
 	flt_puts(s + first, (len - first), Comment_attr);
     } else {
@@ -1003,7 +1087,7 @@ is_Subscript(char *s, int len, char *delimp)
 {
     if (delimp != 0) {
 	int delim = *delimp;
-	int delim2 = delim;
+	int delim2;
 	char *next;
 
 	if (delim != 0 && (next = strchr(LOOKUP_TERM, delim)) != 0) {
@@ -1025,7 +1109,7 @@ put_newline(char *s)
 }
 
 static int
-var_embedded(char *s)
+var_embedded(const char *s)
 {
     if (*s == '$') {
 	if (s[1] == L_PAREN
@@ -1038,7 +1122,7 @@ var_embedded(char *s)
 }
 
 static char *
-put_embedded(char *s, int len, char *attr)
+put_embedded(char *s, int len, const char *attr)
 {
     int id;
     int j, k;
@@ -1112,37 +1196,39 @@ put_document(char *s)
 static void
 check_keyword(char *s, int ok, AfterKey * state)
 {
+    size_t len = (size_t) ok;
+
     state->may_have_pattern = 0;
     state->has_no_pattern = 0;
 
     switch (ok) {
     case 2:
-	state->may_have_pattern = (!strncmp(s, "if", ok)
-				   || !strncmp(s, "eq", ok)
-				   || !strncmp(s, "ge", ok)
-				   || !strncmp(s, "gt", ok)
-				   || !strncmp(s, "le", ok)
-				   || !strncmp(s, "lt", ok)
-				   || !strncmp(s, "ne", ok)
-				   || !strncmp(s, "or", ok));
+	state->may_have_pattern = (!strncmp(s, "if", len)
+				   || !strncmp(s, "eq", len)
+				   || !strncmp(s, "ge", len)
+				   || !strncmp(s, "gt", len)
+				   || !strncmp(s, "le", len)
+				   || !strncmp(s, "lt", len)
+				   || !strncmp(s, "ne", len)
+				   || !strncmp(s, "or", len));
 	break;
     case 3:
-	state->has_no_pattern = !strncmp(s, "sub", ok);
-	state->may_have_pattern = (!strncmp(s, "and", ok)
-				   || !strncmp(s, "cmp", ok)
-				   || !strncmp(s, "not", ok)
-				   || !strncmp(s, "xor", ok));
+	state->has_no_pattern = !strncmp(s, "sub", len);
+	state->may_have_pattern = (!strncmp(s, "and", len)
+				   || !strncmp(s, "cmp", len)
+				   || !strncmp(s, "not", len)
+				   || !strncmp(s, "xor", len));
 	break;
     case 4:
-	state->may_have_pattern = !strncmp(s, "grep", ok);
+	state->may_have_pattern = !strncmp(s, "grep", len);
 	break;
     case 5:
-	state->may_have_pattern = (!strncmp(s, "split", ok)
-				   || !strncmp(s, "until", ok)
-				   || !strncmp(s, "while", ok));
+	state->may_have_pattern = (!strncmp(s, "split", len)
+				   || !strncmp(s, "until", len)
+				   || !strncmp(s, "while", len));
 	break;
     case 6:
-	state->may_have_pattern = !strncmp(s, "unless", ok);
+	state->may_have_pattern = !strncmp(s, "unless", len);
 	break;
     }
 }
@@ -1157,7 +1243,7 @@ put_IDENT(char *s, int ok, AfterKey * if_wrd)
     char save = s[ok];
 
     s[ok] = '\0';
-    attr = keyword_attr(s);
+    attr = get_keyword_attr(s);
     s[ok] = save;
     flt_puts(s, ok, (attr != 0 && *attr != '\0') ? attr : Ident2_attr);
     check_keyword(s, ok, if_wrd);
@@ -1174,9 +1260,9 @@ put_NOKEYWORD(char *s, int ok, AfterKey * if_wrd)
     char save = s[ok];
 
     s[ok] = '\0';
-    attr = keyword_attr(s);
+    attr = get_keyword_attr(s);
     s[ok] = save;
-    flt_puts(s, ok, "");
+    flt_puts(s, ok, attr);
     check_keyword(s, ok, if_wrd);
     return s + ok;
 }
@@ -1197,7 +1283,7 @@ end_BACKTIC(char *s)
 	}
 	++s;
     }
-    return found ? (s - base) : 0;
+    return found ? (int) (s - base) : 0;
 }
 
 /*
@@ -1225,7 +1311,7 @@ is_GLOB(char *s)
 	    }
 	}
     }
-    return (wild && both) ? (s - base) : 0;
+    return (wild && both) ? (int) (s - base) : 0;
 }
 
 /*
@@ -1235,12 +1321,12 @@ is_GLOB(char *s)
 static int
 is_FORMAT(char *s, int len)
 {
-    if (len == 6 && !strncmp(s, "format", len)) {
+    if (len == 6 && !strncmp(s, "format", (size_t) len)) {
 	s += len;
 	s += is_BLANK(s);
 	s += is_NAME(s);
 	s += is_BLANK(s);
-	if (*s == '=')
+	if (*s == '=' && !ispunct(CharOf(s[1])))
 	    return 1;
     }
     return 0;
@@ -1258,13 +1344,13 @@ init_filter(int before GCC_UNUSED)
 #define opRightArrow() (had_op != 0 && old_op == had_op - 1 && *old_op == '-' && *had_op == '>')
 #define opBeforePattern() (had_op != 0 && strchr("{(|&=~!", *had_op) != 0)
 
-#define saveOp(p) { DPRINTF(("\nsaveOp @%d\n", __LINE__)); old_op = had_op; had_op = p; }
+#define saveOp(p) { DPRINTF(("\nsaveOp('%c') @%d\n", *p, __LINE__)); old_op = had_op; had_op = p; }
 #define clearOp() { DPRINTF(("\nclearOp @%d\n", __LINE__)); old_op = had_op = 0; }
 
 static void
 do_filter(FILE *input GCC_UNUSED)
 {
-    static unsigned used;
+    static size_t used;
     static char *line;
 
     AfterKey if_old;
@@ -1283,8 +1369,8 @@ do_filter(FILE *input GCC_UNUSED)
     int quoted = 0;
     int save;
     size_t request = 0;
-    unsigned actual = 0;
-    unsigned mark_len = 0;
+    size_t actual = 0;
+    size_t mark_len = 0;
 
     (void) input;
 
@@ -1322,7 +1408,6 @@ do_filter(FILE *input GCC_UNUSED)
 	the_last = the_file + the_size;
 
 	s = the_file;
-	if_old = nullKey;
 	if_wrd = nullKey;
 	while (MORE(s)) {
 	    if (*s == '\n' || s == the_file) {
@@ -1331,7 +1416,6 @@ do_filter(FILE *input GCC_UNUSED)
 		in_line++;
 	    }
 	    if_old = if_wrd;
-	    if_wrd = nullKey;
 	    DPRINTF(("(%s(%c) line:%d.%d(%d) if:%d.%d op:%c%c)%s\n",
 		     stateName(state), *s, in_line, in_stmt, parens,
 		     if_wrd.may_have_pattern,
@@ -1352,6 +1436,7 @@ do_filter(FILE *input GCC_UNUSED)
 		    state = eCODE;
 		    ++s;
 		}
+		if_wrd = nullKey;
 		break;
 
 	    case eCODE:
@@ -1359,36 +1444,47 @@ do_filter(FILE *input GCC_UNUSED)
 		    state = eIGNORED;
 		    flt_puts(s, 1, Preproc_attr);
 		    break;
-		} else if (!isspace(CharOf(*s)))
+		} else if (!isspace(CharOf(*s))) {
 		    ++in_stmt;
+		}
 
 		if (*s == BACKSLASH && ATLEAST(s, 2) && s[1] == BQUOTE) {
 		    flt_puts(s, 2, String_attr);
 		    s += 2;
+		    if_wrd = nullKey;
 		} else if (*s == BQUOTE) {
 		    flt_puts(s, 1, Action_attr);
 		    ++s;
 		    state = eBACKTIC;
+		    if_wrd = nullKey;
 		} else if ((ok = is_GLOB(s)) != 0) {
 		    s = put_embedded(s, ok, String_attr);
+		    if_wrd = nullKey;
 		} else if ((marker = begin_HERE(s, &quoted)) != 0) {
 		    state = eHERE;
 		    s = put_remainder(s, String_attr, quoted);
+		    if_wrd = nullKey;
 		} else if ((ok = begin_PATTERN(s)) != 0) {
 		    flt_puts(s, ok, "");
 		    s += ok;
 		    DPRINTF(("\nePATTERN:%d\n", __LINE__));
 		    state = ePATTERN;
-		} else if (in_line < 0
-			   && (ok = begin_POD(s))) {
+		    if_wrd = nullKey;
+		} else if (in_line <= 0
+			   && (ok = begin_POD(s, 0))) {
+		    char *base = s;
 		    state = ePOD;
-		    s = put_document(s + ok - 1);
+		    while (*s != '=')
+			flt_putc(*s++);
+		    s = put_document(s);
+		    if (end_POD(base, 1))
+			state = eCODE;
 		} else if (in_line == 0
 			   && (ok = is_PREPROC(s)) != 0) {
 		    flt_puts(s, ok, Preproc_attr);
 		    s += ok;
 		} else if ((ok = is_COMMENT(s)) != 0) {
-		    int skip = (skip_BLANKS(s) - s);
+		    int skip = (int) (skip_BLANKS(s) - s);
 		    ok -= skip;
 		    s += skip;
 		    flt_puts(s, ok, Comment_attr);
@@ -1407,6 +1503,7 @@ do_filter(FILE *input GCC_UNUSED)
 		} else if (*s == L_CURLY) {
 		    saveOp(s);
 		    flt_putc(*s++);
+		    if_wrd = nullKey;
 		} else if (*s == L_PAREN) {
 		    parens++;
 		    saveOp(s);
@@ -1415,17 +1512,21 @@ do_filter(FILE *input GCC_UNUSED)
 			DPRINTF(("\nePATTERN:%d\n", __LINE__));
 			state = ePATTERN;
 		    }
+		    if_wrd = nullKey;
 		} else if (*s == R_PAREN) {
 		    if (--parens < 0)
 			parens = 0;
 		    flt_putc(*s++);
 		    clearOp();
+		    if_wrd = nullKey;
 		} else if ((ok = is_ELLIPSIS(s)) != 0) {
 		    flt_puts(s, ok, "");
 		    s += ok;
+		    if_wrd = nullKey;
 		} else if ((ok = is_NUMBER(s, &err)) != 0) {
 		    clearOp();
 		    s = put_NUMBER(s, ok, &err);
+		    if_wrd = nullKey;
 		} else if ((ok = is_KEYWORD(s)) != 0) {
 		    if ((s != the_file)
 			&& (s[-1] == '*')) {	/* typeglob */
@@ -1446,7 +1547,7 @@ do_filter(FILE *input GCC_UNUSED)
 			    quoted = 0;
 			    state = eHERE;
 			    mark_len = 0;
-			    try_mark = do_alloc(0, 2, &mark_len);
+			    try_mark = do_alloc(0, (size_t) 2, &mark_len);
 			    if (try_mark != 0)
 				marker = strcpy(try_mark, ".");
 			}
@@ -1455,9 +1556,9 @@ do_filter(FILE *input GCC_UNUSED)
 			if (!strcmp(s, "__END__") || !strcmp(s, "__DATA__"))
 			    state = eIGNORED;
 			clearOp();
-			flt_puts(s, ok, keyword_attr(s));
+			flt_puts(s, ok, get_keyword_attr(s));
 			check_keyword(s, ok, &if_wrd);
-			s[ok] = save;
+			s[ok] = (char) save;
 			s += ok;
 		    }
 		} else if ((ok = is_Option(s)) != 0) {
@@ -1486,7 +1587,17 @@ do_filter(FILE *input GCC_UNUSED)
 			}
 			s += ok;
 		    }
+		    if_wrd = nullKey;
+		} else if (isPattern(*s) && (ok = is_PATTERN(s))) {
+		    s = write_PATTERN(s, ok);
+		    if_wrd = nullKey;
 		} else {
+		    if (!isspace(CharOf(*s))) {
+			if_wrd = nullKey;
+		    }
+		    if (*s == ';') {
+			in_stmt = 0;
+		    }
 		    if (parens) {
 			if (strchr("|&=~!->", *s) != 0) {
 			    saveOp(s);
@@ -1494,9 +1605,6 @@ do_filter(FILE *input GCC_UNUSED)
 			    clearOp();
 			}
 		    } else {
-			if (*s == ';') {
-			    in_stmt = 0;
-			}
 			if (strchr("|&=~!->", *s) != 0) {
 			    saveOp(s);
 			}
@@ -1512,6 +1620,7 @@ do_filter(FILE *input GCC_UNUSED)
 		    marker = 0;
 		}
 		s = put_remainder(s, String_attr, quoted);
+		if_wrd = nullKey;
 		break;
 
 	    case eIGNORED:
@@ -1535,12 +1644,16 @@ do_filter(FILE *input GCC_UNUSED)
 		    clearOp();
 		    state = eCODE;
 		}
+		if_wrd = nullKey;
 		break;
 
 	    case ePOD:
-		if (end_POD(s))
+		if (end_POD(s, 0))
 		    state = eCODE;
 		s = put_document(s);
+		if (begin_POD(s - 1, 0)) {
+		    state = ePOD;
+		}
 		break;
 	    }
 	}
