@@ -1,27 +1,35 @@
 /*
- * $Header: /usr/build/vile/vile/filters/RCS/rubyfilt.c,v 1.46 2009/04/25 16:12:14 tom Exp $
+ * $Header: /usr/build/vile/vile/filters/RCS/rubyfilt.c,v 1.83 2014/07/06 16:23:05 tom Exp $
  *
- * Filter to add vile "attribution" sequences to ruby scripts.  This is a
+ * Filter to add vile "attribution" sequences to ruby scripts.  This began as a
  * translation into C of an earlier version written for LEX/FLEX.
  *
  * Although the documentation says it has simpler syntax, Ruby borrows from
  * Perl the worst of its syntax, i.e., regular expressions which can be split
  * across lines, have embedded comments.
+ *
+ * TODO: %x is equivalent of back-quoted string
+ * TODO: %r is equivalent of regular expression
+ * TODO: %-quoting may accept a space as delimiter
+ * TODO: embed quotes using backslashes
+ * TODO: embed quotes by using double-quotes inside single
+ * TODO: embed quotes by using single-quotes inside double
+ * TODO: make var_embedded() display nested expressions
  */
 
 #include <filters.h>
 
 #ifdef DEBUG
-DefineOptFilter("ruby", "d");
+DefineOptFilter(ruby, "de");
 #else
-DefineFilter("ruby");
+DefineOptFilter(ruby, "e");
 #endif
 
 #define isIdent(c)   (isalnum(CharOf(c)) || c == '_')
 #define isIdent0(c)  (isalpha(CharOf(c)) || c == '_')
 
 #define MORE(s)	     ((s) < the_last)
-#define ATLEAST(s,n) (the_last - (s) > (n))
+#define ATLEAST(s,n) ((the_last - (s)) > (n))
 
 #ifdef DEBUG
 #define DPRINTF(params) if(FltOptions('d'))printf params
@@ -39,6 +47,7 @@ typedef enum {
     ,eDEF
     ,eHERE
     ,ePOD
+    ,eEND
 } States;
 
 typedef enum {
@@ -46,6 +55,7 @@ typedef enum {
     ,tBLANK
     ,tCHAR
     ,tCOMMENT
+    ,tERB
     ,tHERE
     ,tKEYWORD
     ,tNUMBER
@@ -55,6 +65,7 @@ typedef enum {
     ,tVARIABLE
 } TType;
 
+static char *Action_attr;
 static char *Comment_attr;
 static char *Error_attr;
 static char *Ident_attr;
@@ -64,6 +75,7 @@ static char *String_attr;
 static char *Number_attr;
 static char *Type_attr;
 
+static int line_size(char *);
 static int var_embedded(char *);
 
 /*
@@ -73,6 +85,19 @@ static char *the_file;
 static char *the_last;
 static size_t the_size;
 
+/*
+ * Stack of here-document tags
+ */
+#define HERE_TAGS struct _here_tags
+HERE_TAGS {
+    HERE_TAGS *next;
+    char *value;
+    int strip;
+    int quote;
+};
+
+static HERE_TAGS *here_tags;
+
 #ifdef DEBUG
 static char *
 stateName(States state)
@@ -98,6 +123,9 @@ stateName(States state)
     case ePOD:
 	result = "POD";
 	break;
+    case eEND:
+	result = "END";
+	break;
     }
     return result;
 }
@@ -120,6 +148,9 @@ tokenType(TType type)
     case tCOMMENT:
 	result = "COMMENT";
 	break;
+    case tERB:
+	result = "ERB";
+	break;
     case tHERE:
 	result = "HERE";
 	break;
@@ -215,7 +246,7 @@ is_STRINGS(char *s, int *err, int left_delim, int right_delim, int single)
 	    }
 	    s++;
 	}
-	found = s - base;
+	found = (int) (s - base);
 	DPRINTF(("...found(%d)\n", found));
     }
     return found;
@@ -229,6 +260,7 @@ is_QIDENT(char *s)
 {
     char *base = s;
     int ch;
+    int delim = 0;
     int leading = 1;
 
     while (MORE(s)) {
@@ -242,11 +274,19 @@ is_QIDENT(char *s)
 		   || ch == SQUOTE
 		   || ch == DQUOTE) {
 	    s++;
-	} else {
+	    if (delim) {
+		if (delim == ch)
+		    break;
+	    } else {
+		delim = ch;
+	    }
+	} else if (!delim) {
 	    break;
+	} else {
+	    ++s;
 	}
     }
-    return s - base;
+    return (int) (s - base);
 }
 
 /*
@@ -327,7 +367,7 @@ is_INSTANCE(char *s)
 	if (*s == '@')
 	    s++;
 	if ((found = is_KEYWORD(s)) != 0)
-	    found += (s - base);
+	    found += (int) (s - base);
     }
     return found;
 }
@@ -350,7 +390,7 @@ is_VARIABLE(char *s)
 }
 
 /*
- * pattern: \?(\\M-)?(\\C-)?.
+ * pattern: \?(\\M-a)?(\\C-a)?
  */
 static int
 is_CHAR(char *s, int *err)
@@ -358,16 +398,34 @@ is_CHAR(char *s, int *err)
     int found = 0;
 
     if (*s == '?' && ATLEAST(s, 5)) {
-	if ((*++s == BACKSLASH)
-	    && ((*++s == 'M') || (*s == 'C'))
-	    && (s[1] == '-')) {
-	    *err = 0;
-	    found = 5;
-	    if (*s == 'M' && ATLEAST(s, 5)
-		&& s[2] == BACKSLASH
-		&& s[3] == 'C'
-		&& s[4] == '-') {
-		found += 3;
+	if (*++s == BACKSLASH) {
+	    ++s;
+	    if (((*s == 'M') || (*s == 'C'))
+		&& (s[1] == '-')) {
+		*err = 0;
+		found = 5;
+		if (*s == 'M' && ATLEAST(s, 5)
+		    && s[2] == BACKSLASH
+		    && s[3] == 'C'
+		    && s[4] == '-'
+		    && isgraph(CharOf(s[5]))) {
+		    s += 5;
+		    found += 3;
+		} else if (*s == 'C') {
+		    s += 2;
+		}
+	    } else {
+		found = 3;
+	    }
+	} else {
+	    found = 2;
+	}
+	if (found) {
+	    /* s now points to the character, but it may be escaped */
+	    if (!isgraph(CharOf(*s))) {
+		found = 0;
+	    } else if (*s == BACKSLASH) {
+		++found;
 	    }
 	}
     }
@@ -447,14 +505,16 @@ is_NUMBER(char *s, int *err)
 	    } else {
 		break;
 	    }
-	} else if (dot && (ch == 'e' || ch == 'E')) {
-	    if (state != 1 || !value)
+	} else if (ch == 'e' || ch == 'E') {
+	    if ((state > 1 && radix < 16) || !value) {
 		*err = 1;
+	    }
 	    state = 2;
 	} else {
 	    if (((state || (radix == 10)) && isdigit(ch))
 		|| ((radix == 16) && isxdigit(ch))
-		|| ((radix == 8) && (ch >= '0' && ch < '8'))) {
+		|| ((radix == 8) && (ch >= '0' && ch < '8'))
+		|| ((radix == 2) && (ch >= '0' && ch < '2'))) {
 		value = 1;
 	    } else {
 		if (value) {
@@ -474,7 +534,7 @@ is_NUMBER(char *s, int *err)
 	s++;
     }
 
-    return value ? (s - base) : 0;
+    return value ? (int) (s - base) : 0;
 }
 
 /*
@@ -498,7 +558,7 @@ is_COMMENT(char *s)
 	}
 	s = t;
     }
-    return (s - base);
+    return (int) (s - base);
 }
 
 /******************************************************************************
@@ -510,14 +570,14 @@ is_COMMENT(char *s)
  * after the keyword.
  */
 static int
-end_marker(char *s, char *marker, int only)
+end_marker(char *s, const char *marker, int only)
 {
-    int len = strlen(marker);
+    int len = (int) strlen(marker);
 
     return (ATLEAST(s, len)
-	    && !strncmp(s, marker, len)
+	    && !strncmp(s, marker, (size_t) len)
 	    && isspace(CharOf(s[len]))
-	    && (!only || (s[len] == '\n')));
+	    && (!only || isreturn(s[len])));
 }
 
 /*
@@ -539,14 +599,97 @@ end_POD(char *s)
 }
 
 /*
- * FIXME: we're only implementing a single here-doc (documentation says they
- * can be stacked), and don't check for unbalanced quotes).
+ * Ruby allows more than one here-document to begin on a line.  They "stack",
+ * and are processed in succession.
  */
-static char *
-begin_HERE(char *s, int *quoted, int *strip_here)
+static void
+make_here_tag(char *value, int quote, int strip)
 {
-    char *base;
-    char *marker;
+    size_t size = 0;
+    HERE_TAGS *data = type_alloc(HERE_TAGS, (char *) 0, (size_t) 1, &size);
+
+    if (data != 0) {
+	HERE_TAGS *p = here_tags;
+	HERE_TAGS *q = 0;
+
+	while (p != 0) {
+	    q = p;
+	    p = p->next;
+	}
+	if (q != 0)
+	    q->next = data;
+	else
+	    here_tags = data;
+	data->next = p;
+	data->value = value;
+	data->quote = quote;
+	data->strip = strip;
+	DPRINTF(("make_here_tag(%s) %squoted %sstripped\n", value,
+		 quote ? "" : "un",
+		 strip ? "" : "un"));
+    }
+}
+
+static char *
+free_here_tag(void)
+{
+    HERE_TAGS *next = here_tags->next;
+    char *result = next ? next->value : 0;
+
+    free(here_tags->value);
+    free(here_tags);
+    here_tags = next;
+
+    return result;
+}
+
+/*
+ * Workaround for ruby bug: ignoring their documentation, which notes that you
+ * must separate a string from the "<<" operator to distinguish it from a here
+ * document tag, some of the source code happens to use " " or "\n" without an
+ * intervening space.
+ */
+static int
+valid_HERE(char *s)
+{
+    int ok = is_QIDENT(s);
+    int delim;
+    if (ok) {
+	switch (*s) {
+	default:
+	    delim = 0;
+	    break;
+	case DQUOTE:
+	    /* FALLTHRU */
+	case SQUOTE:
+	    delim = *s;
+	    break;
+	}
+	if (delim) {
+	    switch (ok) {
+	    case 3:
+		if (s[0] == s[2] && s[1] == ' ')
+		    ok = 0;
+		break;
+	    case 4:
+		if (s[0] == s[3] && s[1] == '\\' && s[2] == 'n')
+		    ok = 0;
+		break;
+	    }
+	}
+    }
+    return ok;
+}
+
+/*
+ * Mark the beginning of a here-document.
+ */
+static int
+begin_HERE(char *s)
+{
+    char *base = s;
+    char *first;
+    char *marker = 0;
     int ok;
     int strip = 0;
 
@@ -555,33 +698,40 @@ begin_HERE(char *s, int *quoted, int *strip_here)
 	&& s[1] == '<'
 	&& !isBlank(s[2])) {
 	s += 2;
-	base = s;
 	if (*s == '-') {
 	    strip = 1;
 	    ++s;
 	}
-	if ((ok = is_QIDENT(s)) != 0) {
-	    unsigned temp = 0;
+	first = s;
+	if ((ok = valid_HERE(s)) != 0) {
+	    size_t temp = 0;
+	    int delim = (*s == SQUOTE || *s == DQUOTE) ? *s : 0;
+	    int quote = (*s == SQUOTE);
 
 	    s += ok;
-	    *quoted = 0;
-	    *strip_here = 1;
 
-	    if ((marker = do_alloc((char *) 0, ok + 1, &temp)) != 0) {
+	    if ((marker = do_alloc((char *) 0, (size_t) (ok + 1), &temp)) != 0) {
 		char *d = marker;
-		while (base != s) {
-		    if (isIdent(*base))
-			*d++ = *base;
-		    else
-			*quoted = (*base != DQUOTE);
-		    base++;
+		if (delim) {
+		    ++first;
+		}
+		while (first != s) {
+		    if (delim) {
+			if (*first == delim)
+			    break;
+			*d++ = *first;
+		    } else {
+			if (isIdent(*first))
+			    *d++ = *first;
+		    }
+		    first++;
 		}
 		*d = 0;
+		make_here_tag(marker, quote, strip);
 	    }
-	    return marker;
 	}
     }
-    return 0;
+    return (marker ? (int) (s - base) : 0);
 }
 
 static char *
@@ -596,11 +746,60 @@ skip_BLANKS(char *s)
 	++s;
     }
     if (s != base) {
-	flt_puts(base, s - base, "");
+	flt_puts(base, (int) (s - base), "");
     }
     return s;
 }
 
+/*
+ * Check for ERB (embedded ruby).  Ruby on Rails uses templates which fit into
+ * the special case of ruby embedded in ruby.
+ */
+static int
+is_ERB(char *s)
+{
+    int found = 0;
+#define OPS(s) { s, sizeof(s) - 1 }
+    static const struct {
+	const char *ops;
+	int len;
+    } table[] = {
+	/* 3 */
+	OPS("<%="),
+	    OPS("<%#"),
+	    OPS("<%-"),
+	    OPS("-%>"),
+	/* 2 */
+	    OPS("<%"),
+	    OPS("%>"),
+    };
+
+    if (FltOptions('e') && ispunct(CharOf(*s))) {
+	unsigned n;
+	for (n = 0; n < TABLESIZE(table); ++n) {
+	    if (ATLEAST(s, table[n].len)
+		&& table[n].ops[0] == *s
+		&& !memcmp(s, table[n].ops, (size_t) table[n].len)) {
+		found = table[n].len;
+		break;
+	    }
+	}
+	/* special-case comments */
+	if (found == 3 && s[2] == '#') {
+	    s += found;
+	    while (ATLEAST(s, 2)) {
+		if (!memcmp(s, "%>", 2)) {
+		    found += 2;
+		    break;
+		}
+		s++;
+		found++;
+	    }
+	}
+    }
+    return found;
+}
+
 /*
  * Return the number of characters if an operator is found, otherwise zero.
  */
@@ -672,7 +871,7 @@ is_OPERATOR(char *s)
 	for (n = 0; n < TABLESIZE(table); ++n) {
 	    if (ATLEAST(s, table[n].len)
 		&& table[n].ops[0] == *s
-		&& !memcmp(s, table[n].ops, table[n].len)) {
+		&& !memcmp(s, table[n].ops, (size_t) table[n].len)) {
 		found = table[n].len;
 		break;
 	    }
@@ -693,6 +892,7 @@ is_REGEXP(char *s, int left_delim, int right_delim)
     int len;
     int level = 0;
     int range = 0;
+    int block = (left_delim != L_BLOCK);
 
     while (MORE(s)) {
 	if (left_delim != right_delim) {
@@ -708,10 +908,10 @@ is_REGEXP(char *s, int left_delim, int right_delim)
 	}
 	if ((len = is_ESCAPED(s)) != 0) {
 	    s += len;
-	} else if (*s == R_BLOCK && range) {
+	} else if (block && (*s == R_BLOCK) && range) {
 	    range = 0;
 	    ++s;
-	} else if (*s == L_BLOCK && !range) {
+	} else if (block && (*s == L_BLOCK) && !range) {
 	    range = 1;
 	    ++s;
 	} else if ((len = var_embedded(s)) != 0) {
@@ -723,7 +923,7 @@ is_REGEXP(char *s, int left_delim, int right_delim)
 	    while (MORE(s) && isalpha(CharOf(*s))) {
 		++s;
 	    }
-	    found = s - base;
+	    found = (int) (s - base);
 	    break;
 	} else {
 	    ++s;
@@ -732,6 +932,8 @@ is_REGEXP(char *s, int left_delim, int right_delim)
     return found;
 }
 
+#define valid_delimiter(c) (isgraph(CharOf(c)) && !isalnum(CharOf(c)))
+
 static int
 balanced_delimiter(char *s)
 {
@@ -751,12 +953,45 @@ balanced_delimiter(char *s)
 	result = R_ANGLE;
 	break;
     default:
-	result = *s;
+	if (valid_delimiter(*s))
+	    result = *s;
+	else
+	    result = 0;
 	break;
     }
     return result;
 }
 
+static int
+is_Balanced(char *s, int length, int *left, int *right)
+{
+    int delim;
+    int n;
+
+    *left = 0;
+    *right = 0;
+    if (*s == '%') {
+	for (n = 1; n < length; ++n) {
+	    if (isalnum(CharOf(s[n])))
+		continue;
+	    delim = balanced_delimiter(s + n);
+	    if (delim == 0) {
+		*left = n;
+		break;
+	    } else if (delim == s[n]) {
+		*left = n + 1;
+		*right = 1;
+		break;
+	    } else {
+		*left = n + 1;
+		*right = 1;
+		break;
+	    }
+	}
+    }
+    return *left;
+}
+
 /*
  * Check for a regular expression, returning its length
  */
@@ -771,7 +1006,7 @@ is_Regexp(char *s, int *delim)
     } else if (ATLEAST(s, 4)
 	       && s[0] == '%'
 	       && s[1] == 'r'
-	       && !isalnum(CharOf(s[2]))) {
+	       && valid_delimiter(s[2])) {
 
 	*delim = balanced_delimiter(s + 2);
 	found = 2 + is_REGEXP(s + 2, s[2], *delim);
@@ -799,8 +1034,12 @@ is_Symbol(char *s, int *delim, int *err)
 	    if ((found = is_STRINGS(s, err, *s, *s, 0)) != 0)
 		*delim = DQUOTE;
 	    break;
+	case BQUOTE:
+	    found = 1;
+	    break;
 	default:
 	    found = is_MKEYWORD(s, 0);
+	    break;
 	}
     }
 
@@ -823,31 +1062,39 @@ is_String(char *s, int *delim, int *err)
 	switch (*s) {
 	case '%':
 	    if (ATLEAST(s, 4)) {
-		switch (s[1]) {
-		case 'q':
-		case 'w':
-		    found = is_STRINGS(s + 2,
+		int modifier = 0;
+		int single = 0;
+		char *base = s;
+
+		++s;
+		if (isalpha(CharOf(*s))) {
+		    modifier = *s++;
+		    switch (modifier) {
+		    case 'q':
+		    case 'w':	/* FIXME - sword */
+			single = 1;
+			break;
+		    case 'x':
+		    case 'Q':
+		    case 'W':	/* FIXME - dword */
+			break;
+		    case 'I':	/* FIXME - dword */
+		    case 'i':	/* FIXME - sword */
+		    case 'r':	/* FIXME - regexp */
+		    case 's':	/* FIXME - symbol */
+			break;
+		    }
+		}
+		if (valid_delimiter(*s)) {
+		    found = is_STRINGS(s,
 				       err,
-				       s[2],
-				       balanced_delimiter(s + 2),
+				       *s,
+				       balanced_delimiter(s),
 				       1);
-		    if (found != 0) {
-			found += 2;
-			*delim = SQUOTE;
-		    }
-		    break;
-		case 'x':
-		case 'Q':
-		    found = is_STRINGS(s + 2,
-				       err,
-				       s[2],
-				       balanced_delimiter(s + 2),
-				       0);
-		    if (found != 0) {
-			found += 2;
-			*delim = DQUOTE;
-		    }
-		    break;
+		}
+		if (found != 0) {
+		    found += (int) (s - 1 - base);
+		    *delim = single ? SQUOTE : DQUOTE;
 		}
 	    }
 	    break;
@@ -879,7 +1126,7 @@ line_size(char *s)
 	    break;
 	s++;
     }
-    return s - base;
+    return (int) (s - base);
 }
 
 static char *
@@ -891,32 +1138,68 @@ put_newline(char *s)
 }
 
 /*
- * pattern: #(\{[^}]*\}|{IDENT})
+ * pattern: #\{{EXPR}*\}
  * returns: length of the token
  */
 static int
 var_embedded(char *s)
 {
-    int found = 0;
+    char *base = s;
     int delim;
+    int level = 0;
+    int had_op = 1;
+    int ignore;
+    int ok;
 
     if (*s == '#' && MORE(++s)) {
 	if (*s == L_CURLY) {
-	    for (found = 1; ATLEAST(s, found); ++found) {
-		/* FIXME: this check for %q{xxx} is too naive... */
-		if (s[found] == '%' && (strchr("wqQ", s[found + 1]) != 0)) {
-		    int ignore;
-		    found += is_String(s + found, &delim, &ignore);
-		} else if (s[found] == R_CURLY) {
-		    ++found;
-		    break;
+	    ++level;
+	    ++s;
+	    while (MORE(s)) {
+		if ((*s == '%' || had_op)
+		    && (ok = is_Regexp(s, &delim)) != 0) {
+		    had_op = 0;
+		    s += ok;
+		} else if ((ok = is_String(s, &delim, &ignore)) != 0) {
+		    had_op = 0;
+		    s += ok;
+		} else if ((ok = is_CHAR(s, &ignore)) != 0
+			   && (ok != 2 || (s[1] != L_CURLY && s[1] != R_CURLY))) {
+		    had_op = 0;
+		    s += ok;
+		} else if ((ok = is_NUMBER(s, &ignore)) != 0) {
+		    had_op = 0;
+		    s += ok;
+		} else if ((ok = is_KEYWORD(s)) != 0) {
+		    had_op = 0;
+		    s += ok;
+		} else if ((ok = is_VARIABLE(s)) != 0) {
+		    had_op = 0;
+		    s += ok;
+		} else if ((ok = is_OPERATOR(s)) != 0) {
+		    had_op = 1;
+		    if (*s == L_CURLY) {
+			++level;
+		    } else if (*s == R_CURLY) {
+			if (--level <= 0) {
+			    ++s;
+			    break;
+			}
+		    }
+		    s += ok;
+		} else {
+		    ++s;
 		}
 	    }
 	} else {
-	    found = is_VARIABLE(s);
+	    if ((ok = is_VARIABLE(s)) != 0) {
+		++ok;
+	    } else {
+		s = base;
+	    }
 	}
     }
-    return found ? found + 1 : 0;
+    return (int) (s - base);
 }
 
 /*
@@ -967,13 +1250,25 @@ put_remainder(char *s, char *attr, int quoted)
 static char *
 put_COMMENT(char *s, int ok)
 {
-    int skip = (skip_BLANKS(s) - s);
+    int skip = (int) (skip_BLANKS(s) - s);
     ok -= skip;
     s += skip;
     flt_puts(s, ok, Comment_attr);
     return s + ok;
 }
 
+static char *
+put_ERB(char *s, int ok, int *had_op)
+{
+    if (ok > 3) {
+	flt_puts(s, ok, Comment_attr);
+    } else {
+	flt_puts(s, ok, Action_attr);
+    }
+    *had_op = 1;
+    return s + ok;
+}
+
 static char *
 put_KEYWORD(char *s, int ok, int *had_op)
 {
@@ -981,9 +1276,9 @@ put_KEYWORD(char *s, int ok, int *had_op)
     char save = s[ok];
 
     s[ok] = '\0';
-    attr = keyword_attr(s);
+    attr = get_keyword_attr(s);
     s[ok] = save;
-    if ((attr == 0 || *attr == '\0') && isupper(CharOf(s[0])))
+    if (isEmpty(attr) && isupper(CharOf(s[0])))
 	attr = Type_attr;
     flt_puts(s, ok, attr);
     *had_op = (attr == Keyword_attr);
@@ -994,7 +1289,7 @@ static char *
 put_OPERATOR(char *s, int ok, int *had_op)
 {
     flt_puts(s, ok, "");
-    if (strchr("(|&=~!,;", *s) != 0)
+    if (strchr("[(|&=~!,;", *s) != 0)
 	*had_op = 1;
     return s + ok;
 }
@@ -1006,7 +1301,7 @@ put_VARIABLE(char *s, int ok)
     char save = s[ok];
 
     s[ok] = '\0';
-    attr = keyword_attr(s);
+    attr = get_keyword_attr(s);
     s[ok] = save;
     flt_puts(s, ok, (attr != 0 && *attr != '\0') ? attr : Ident2_attr);
     return s + ok;
@@ -1020,21 +1315,32 @@ put_REGEXP(char *s, int length, int delim)
     char *last;
     int len;
     int range = 0;
+    int on_end = 0;
+    int block = (delim != R_BLOCK);
+    int level = 0;
+    int extended = 0;
 
+    for (last = s + length - 1; last != s && isalpha(CharOf(*last)); --last) {
+	if (*last == 'x') {
+	    extended = 1;
+	    break;
+	}
+    }
     if (*s == '%') {
-	flt_puts(s, 2, Keyword_attr);
-	s += 2;
+	flt_puts(s, 3, Keyword_attr);
+	s += 3;
 	first = s;
+	on_end = 1;
     }
     flt_bfr_begin(String_attr);
     while (s < base + length) {
 	if ((len = is_ESCAPED(s)) != 0) {
 	    flt_bfr_append(s, len);
 	    s += len;
-	} else if (*s == R_BLOCK && range) {
+	} else if (block && (*s == R_BLOCK) && range) {
 	    range = 0;
 	    flt_bfr_append(s++, 1);
-	} else if (*s == L_BLOCK && !range) {
+	} else if (block && (*s == L_BLOCK) && !range) {
 	    range = 1;
 	    flt_bfr_append(s++, 1);
 	} else if ((len = var_embedded(s)) != 0) {
@@ -1042,13 +1348,32 @@ put_REGEXP(char *s, int length, int delim)
 	    s += len;
 	} else if (range) {
 	    flt_bfr_append(s++, 1);
-	} else if (s != first && *s == delim) {
-	    flt_bfr_append(s, 1);
-	    last = ++s;
+	} else if (*s == L_PAREN && delim != R_PAREN) {
+	    ++level;
+	    flt_bfr_append(s++, 1);
+	} else if (*s == R_PAREN && delim != R_PAREN) {
+	    --level;
+	    flt_bfr_append(s++, 1);
+	} else if (extended && (len = is_BLANK(s)) != 0) {
+	    last = s;
+	    flt_bfr_embed(last, len, "");
+	    s += len;
+	} else if (extended && (*s == '#')) {
+	    last = s;
+	    while (MORE(s) && !isreturn(CharOf(*s))) {
+		++s;
+	    }
+	    flt_bfr_embed(last, (int) (s - last), Comment_attr);
+	} else if (s != first && level == 0 && *s == delim) {
+	    if (!on_end)
+		flt_bfr_append(s++, 1);
+	    last = s;
+	    if (on_end)
+		++s;
 	    while (MORE(s) && isalpha(CharOf(*s))) {
 		++s;
 	    }
-	    flt_bfr_embed(last, s - last, Keyword_attr);
+	    flt_bfr_embed(last, (int) (s - last), Keyword_attr);
 	    break;
 	} else {
 	    flt_bfr_append(s, 1);
@@ -1059,6 +1384,42 @@ put_REGEXP(char *s, int length, int delim)
     return s;
 }
 
+static char *
+put_String(char *s, int ok, int delim, int err, int *had_op)
+{
+    int on_left = 0;
+    int on_end = 0;
+    int embed = (delim == DQUOTE);
+
+    *had_op = 0;
+    if (is_Balanced(s, ok, &on_left, &on_end)) {
+	flt_puts(s, on_left, Keyword_attr);
+	s += on_left;
+	ok -= on_left;
+    }
+    if (embed) {
+	if (err) {
+	    flt_error("unexpected quote");
+	    s = put_embedded(s, ok, Error_attr);
+	} else {
+	    s = put_embedded(s, ok, String_attr);
+	}
+    } else {
+	if (err) {
+	    flt_error("unterminated string");
+	    flt_puts(s, ok, Error_attr);
+	} else {
+	    flt_puts(s, ok, String_attr);
+	}
+	s += ok;
+    }
+    if (on_end) {
+	flt_puts(s, 1, Keyword_attr);
+	++s;
+    }
+    return s;
+}
+
 /******************************************************************************
  ******************************************************************************/
 
@@ -1071,10 +1432,10 @@ init_filter(int before GCC_UNUSED)
 static void
 do_filter(FILE *input GCC_UNUSED)
 {
-    static unsigned used;
+    static size_t used;
     static char *line;
 
-    unsigned actual = 0;
+    size_t actual = 0;
     size_t request = 0;
     States state = eCODE;
     TType this_tok = tNULL;
@@ -1085,12 +1446,11 @@ do_filter(FILE *input GCC_UNUSED)
     int ok;
     int err;
     int delim;
-    int strip_here = 0;		/* strip leading blanks in here-document */
-    int quote_here = 0;		/* tokens in here-document are quoted */
     int had_op = 1;		/* true to allow regex */
 
     (void) input;
 
+    Action_attr = class_attr(NAME_ACTION);
     Comment_attr = class_attr(NAME_COMMENT);
     Error_attr = class_attr(NAME_ERROR);
     Ident_attr = class_attr(NAME_IDENT);
@@ -1126,6 +1486,8 @@ do_filter(FILE *input GCC_UNUSED)
 	s = the_file;
 	while (MORE(s)) {
 	    if (*s == '\n') {
+		if (marker != 0)
+		    state = eHERE;
 		in_line = -1;
 		if (state == eCODE)
 		    had_op = 1;
@@ -1152,6 +1514,10 @@ do_filter(FILE *input GCC_UNUSED)
 		    Parsed(this_tok, tKEYWORD);
 		    s = put_KEYWORD(s, ok, &had_op);
 		    state = eCODE;
+		} else if ((ok = is_ERB(s)) != 0) {
+		    Parsed(this_tok, tERB);
+		    s = put_ERB(s, ok, &had_op);
+		    state = eCODE;
 		} else if ((ok = is_OPERATOR(s)) != 0) {
 		    Parsed(this_tok, tOPERATOR);
 		    s = put_OPERATOR(s, ok, &had_op);
@@ -1182,6 +1548,10 @@ do_filter(FILE *input GCC_UNUSED)
 		    Parsed(this_tok, tKEYWORD);
 		    s = put_KEYWORD(s, ok, &had_op);
 		    state = eCODE;
+		} else if ((ok = is_ERB(s)) != 0) {
+		    Parsed(this_tok, tERB);
+		    s = put_ERB(s, ok, &had_op);
+		    state = eCODE;
 		} else if ((ok = is_OPERATOR(s)) != 0) {
 		    Parsed(this_tok, tOPERATOR);
 		    s = put_OPERATOR(s, ok, &had_op);
@@ -1194,11 +1564,11 @@ do_filter(FILE *input GCC_UNUSED)
 		break;
 	    case eCODE:
 		if ((last_tok == tKEYWORD || last_tok == tOPERATOR)
-		    && (marker = begin_HERE(s, &quote_here, &strip_here)) != 0) {
-		    ok = 1;	/* FIXME: should allow other tokens on line */
+		    && (ok = begin_HERE(s)) != 0) {
 		    Parsed(this_tok, tHERE);
-		    state = eHERE;
-		    s = put_remainder(s, String_attr, quote_here);
+		    flt_puts(s, ok, String_attr);
+		    s += ok;
+		    marker = here_tags->value;
 		} else if ((in_line < 0 && begin_POD(s + 1))
 			   || ((s == the_file && begin_POD(s)))) {
 		    DPRINTF(("...POD\n"));
@@ -1212,14 +1582,15 @@ do_filter(FILE *input GCC_UNUSED)
 		    Parsed(this_tok, tBLANK);
 		    flt_puts(s, ok, "");
 		    s += ok;
-		} else if (had_op && (ok = is_Regexp(s, &delim)) != 0) {
+		} else if ((*s == '%' || had_op)
+			   && (ok = is_Regexp(s, &delim)) != 0) {
 		    Parsed(this_tok, tREGEXP);
 		    s = put_REGEXP(s, ok, delim);
 		} else if ((ok = is_CHAR(s, &err)) != 0) {
 		    Parsed(this_tok, tCHAR);
 		    had_op = 0;
 		    if (err) {
-			flt_error("not a number");
+			flt_error("not a number: %.*s", ok, s);
 			flt_puts(s, ok, Error_attr);
 		    } else {
 			flt_puts(s, ok, Number_attr);
@@ -1229,7 +1600,7 @@ do_filter(FILE *input GCC_UNUSED)
 		    Parsed(this_tok, tNUMBER);
 		    had_op = 0;
 		    if (err) {
-			flt_error("not a number");
+			flt_error("not a number: %.*s", ok, s);
 			flt_puts(s, ok, Error_attr);
 		    } else {
 			flt_puts(s, ok, Number_attr);
@@ -1237,45 +1608,31 @@ do_filter(FILE *input GCC_UNUSED)
 		    s += ok;
 		} else if ((ok = is_MKEYWORD(s, 1)) != 0) {
 		    Parsed(this_tok, tKEYWORD);
-		    if (ok == 5 && !strncmp(s, "alias", ok))
+		    if (ok == 5 && !strncmp(s, "alias", (size_t) ok))
 			state = eALIAS;
-		    else if (ok == 5 && !strncmp(s, "class", ok))
+		    else if (ok == 5 && !strncmp(s, "class", (size_t) ok))
 			state = eCLASS;
-		    else if (ok == 3 && !strncmp(s, "def", ok))
+		    else if (ok == 3 && !strncmp(s, "def", (size_t) ok))
 			state = eDEF;
+		    else if (ok == 7 && !strncmp(s, "__END__", (size_t) ok))
+			state = eEND;
 		    s = put_KEYWORD(s, ok, &had_op);
 		} else if ((ok = is_VARIABLE(s)) != 0) {
 		    Parsed(this_tok, tVARIABLE);
 		    s = put_VARIABLE(s, ok);
 		    had_op = 0;
-		} else if (ATLEAST(s, (ok = 2)) && !strncmp(s, "?\"", ok)) {
+		} else if (ATLEAST(s, (ok = 2))
+			   && !strncmp(s, "?\"", (size_t) ok)) {
 		    Parsed(this_tok, tVARIABLE);
 		    s = put_VARIABLE(s, ok);	/* csv.rb uses it, undocumented */
 		    had_op = 0;
+		} else if ((ok = is_ERB(s)) != 0) {
+		    Parsed(this_tok, tERB);
+		    s = put_ERB(s, ok, &had_op);
+		    state = eCODE;
 		} else if ((ok = is_String(s, &delim, &err)) != 0) {
 		    Parsed(this_tok, tSTRING);
-		    had_op = 0;
-		    if (*s == '%') {
-			flt_puts(s, 2, Keyword_attr);
-			s += 2;
-			ok -= 2;
-		    }
-		    if (delim == DQUOTE) {
-			if (err) {
-			    flt_error("unexpected quote");
-			    s = put_embedded(s, ok, Error_attr);
-			} else {
-			    s = put_embedded(s, ok, String_attr);
-			}
-		    } else {
-			if (err) {
-			    flt_error("unterminated string");
-			    flt_puts(s, ok, Error_attr);
-			} else {
-			    flt_puts(s, ok, String_attr);
-			}
-			s += ok;
-		    }
+		    s = put_String(s, ok, delim, err, &had_op);
 		} else if ((ok = is_OPERATOR(s)) != 0) {
 		    Parsed(this_tok, tOPERATOR);
 		    s = put_OPERATOR(s, ok, &had_op);
@@ -1286,18 +1643,28 @@ do_filter(FILE *input GCC_UNUSED)
 		}
 		break;
 	    case eHERE:
-		if (end_marker(s + is_BLANK(s), marker, 1)) {
+		if (here_tags == 0) {
 		    state = eCODE;
-		    free(marker);
-		    marker = 0;
+		} else if (end_marker(s + (here_tags->strip
+					   ? is_BLANK(s)
+					   : 0),
+				      marker, 1)) {
+		    if ((marker = free_here_tag()) == 0)
+			state = eCODE;
 		}
-		s = put_remainder(s, String_attr, quote_here);
+		s = put_remainder(s, String_attr,
+				  (here_tags
+				   ? here_tags->quote
+				   : 0));
 		break;
 	    case ePOD:
 		if (end_POD(s))
 		    state = eCODE;
 		s = put_remainder(s, Comment_attr, 1);
 		break;
+	    case eEND:
+		s = put_remainder(s, Comment_attr, 1);
+		break;
 	    }
 
 	    switch (this_tok) {
@@ -1306,6 +1673,7 @@ do_filter(FILE *input GCC_UNUSED)
 	    case tCOMMENT:
 		continue;
 	    case tCHAR:
+	    case tERB:
 	    case tHERE:
 	    case tNUMBER:
 	    case tREGEXP:
@@ -1325,6 +1693,10 @@ do_filter(FILE *input GCC_UNUSED)
 	}
 	free(the_file);
     }
+    while (here_tags != 0) {
+	flt_error("expected tag:%s", here_tags->value);
+	(void) free_here_tag();
+    }
 }
 
 #if NO_LEAKS
