@@ -1,29 +1,35 @@
 /*
  * find the next error in mentioned in the shell output window.
  * written for vile by paul fox.
- * rewritten to use regular expressions by T.Dickey (dickey@clark.net)
+ * rewritten to use regular expressions by T.Dickey
  *
- * Copyright (c) 1990-1999 by Paul Fox and Thomas Dickey
+ * Copyright (c) 1990-2016 by Paul Fox and Thomas Dickey
  *
- * $Header: /usr/build/vile/vile/RCS/finderr.c,v 1.85 2000/01/07 01:34:42 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/finderr.c,v 1.151 2016/07/24 17:14:14 tom Exp $
  *
  */
 
 #include "estruct.h"
 #include "edef.h"
+#include "nevars.h"
 
 #if OPT_FINDERR
 
-#define W_VERB  0
-#define W_FILE  1
-#define W_LINE  2
-#define W_COLM  3
-#define W_TEXT  4
-#define W_LAST	5
+typedef enum {
+    W_VERB = 0
+    ,W_FILE
+    ,W_LINE
+    ,W_COLM
+    ,W_TEXT
+    ,W_LAST
+} ErrTokens;
 
 typedef struct {
+    char *exp_text;
     regexp *exp_comp;
     int words[W_LAST];
+    int lBase;
+    int cBase;
 } ERR_PATTERN;
 
 static LINE *getdot(BUFFER *bp);
@@ -32,11 +38,27 @@ static void putdotback(BUFFER *bp, LINE *dotp);
 static char febuff[NBUFN];	/* name of buffer to find errors in */
 static int newfebuff = TRUE;	/* is the name new since last time? */
 
+/*
+ * These variables are set as a side-effect of decode_exp().
+ */
 static TBUFF *fe_verb;
 static TBUFF *fe_file;
 static TBUFF *fe_text;
-static int fe_colm;
-static int fe_line;
+static int cC_base;
+static int lL_base;
+static long fe_colm;
+static long fe_line;
+
+/*
+ * Beginning of line.
+ */
+#define P_BOL	"^"
+
+/*
+ * Text from running "ant", which puts the program name before each message,
+ * in square brackets.
+ */
+#define P_ANT	P_BOL "\\(\\s\\+\\[[^]]\\+\\]\\s\\+\\)\\?"
 
 /*
  * This is the list of predefined regular expressions for the error
@@ -51,6 +73,7 @@ static int fe_line;
  *      %F - range of characters to match filename.
  *      %B - range of characters to match scratch-buffer.
  *      %L - line number (this has to be an integer)
+ *      %C - column number (this has to be an integer)
  *      %T - text to display in the message line. If no field is given,
  *              the error finder will display the entire line from
  *              the error-buffer.
@@ -71,12 +94,10 @@ static int fe_line;
 static const
 char *const predefined[] =
 {
-    "^\"%[^\" \t]\", line %L:%T",	/* various C compilers */
-    "^%[^: \t]:\\s*%L:\\s*%T",	/* "grep -n" */
-#if OPT_MSDOS_PATH
-    "^%F:\\s*%L:\\s*%T",	/* "grep -n", handles */
-					/* dos drive letter   */
-#endif
+    P_ANT "\"%[^\" \t]\", line %L:%T",	/* various C compilers */
+    P_ANT "%F:\\s*%L:%C:\\s*%T",	/* antic */
+    P_ANT "%F:\\s*%L:\\s*%T",	/* "grep -n" */
+
 #if SYS_VMS
     "[ \t]*At line number %L in %[^;].*",	/* crude support for DEC C
 						 * compilers. Unfortunately,
@@ -84,59 +105,71 @@ char *const predefined[] =
 						 * text on previous _lines_.
 						 */
 #endif
-#if SYS_APOLLO
-    " Line %L of \"%[^\" \t]\"",	/* C compiler */
-#endif
-#if SYS_SUNOS && SYSTEM_HAS_LINT_PROG
+
+#if SYS_SUNOS && defined(SYSTEM_HAS_LINT_PROG)
     "%[^:( \t](%L):%T",		/* bsd lint) */
     "  ::  %[^( \t](%L)",	/* bsd lint) */
     "used[ \t]*([ \t]%[^(](%L)[ \t]*)",		/* bsd lint) */
 #endif
+
     /* ultrix, sgi, osf1 (alpha only?)  use:                            */
     /*      compiler-name: Error: filename, line line-number ...        */
     "[^ ]\\+ [^ ]\\+ \"%[^, \t\"]\", line %L",
     "[^ ]\\+ [^ ]\\+ %[^, \t], line %L",
     "[^ ]\\+ \"%[^\"]\", line %L",	/* HP/UX C compiler */
     "File = %F, Line = %L",	/* SGI MIPSpro 7.3 compilers    */
+    "^\"%F\":line %L:%T",	/* Solaris lex */
 #if defined(_AIX)
-    "^\"%[^\" \t]\", line %L\\.[0-9]\\+:%T",	/* AIX C compilers */
+    "^\"%[^\" \t]\", line %L\\.%C:%T",	/* AIX C compilers */
 #endif
 #if defined(clipper) || defined(__clipper__)
-    "^\"%[^\" \t]\", line %L (col. [0-9]\\+):%T",	/* CLIX C compiler */
+    "^\"%[^\" \t]\", line %L (col. %C\\+):%T",	/* CLIX C compiler */
 #endif
-    "^%[^(](%L)[ \t]\\+:%T",
 
-#if SYS_UNIX && SYSTEM_HAS_LINT_PROG
+#ifdef _MSC_VER
+    "^[ \t]*%F\\s*(%L)\\s*:%T",	/* MSVC++ */
+#endif
+
+    "^%[^\"(](%L)[ \t]\\?:%T",	/* weblint */
+#if SYS_UNIX && defined(SYSTEM_HAS_LINT_PROG)
     /* sys5 lint */
     "^    [^ \t]\\+[ \t]\\+%[^(](%L)$",
     "^    [^(].*( arg %L ) \t%[^( \t](%L) :: [^(]\\+(%L))",
     "^    .* :: %[^(](%L)",
 #endif
-#if CC_CSETPP
-    "^%[^(](%L:%C) : %T",
-#endif
-#if CC_TURBO
-    "^Error %[^ ] %L:",
-    "^Warning %[^ ] %L:",
+
+#if SYS_OS2 || SYS_OS2_EMX
+    "^%[^(](%L:%C) : %T",	/* CSet compiler */
 #endif
 #if CC_WATCOM
     "^%[^(](%L): %T",
 #endif
+
+    /* Borland C++ */
+#if CC_NEWDOSCC
+    "^Error\\( [^ ]\\+\\)\\? %F %L: %T",
+    "^Warning\\( [^ ]\\+\\)\\? %F %L: %T",
+#endif
+
     "^%B:%L:%T",		/* "pp" in scratch buf */
-    "^[^:]\\+: %V directory `%[^']'",	/* GNU make */
+    "^[^:]\\+: %V directory [`']%[^']'",	/* GNU make */
     "%T at %F line %L.*",	/* perl 5 */
-    "%F\\[%L\\]:%T",		/* hgrep */
+    "^%F\\[%L\\]:%T",		/* hgrep */
     "^\"%[^\"]\", line %L, col %C, %T",		/* ncurses, atac */
     "^\"%[^\"]\", line %L, %T",	/* ncurses */
+    "^%F:%L\\.%C\\(-\\d\\+\\)\\?:%T",	/* bison */
+    "^%F: \\a - line %L of \"%F\", %T",		/* byacc */
+    "^\\[%[^:]:%L\\]: %T",	/* cppcheck */
+    "^\\[%[^:]:%L\\]\\( -> \\[[^]]\\+\\]\\)\\?: %T",	/* cppcheck */
 };
 
 static ERR_PATTERN *exp_table = 0;
-static ALLOC_T exp_count = 0;
+static size_t exp_count = 0;
 
 void
 set_febuff(const char *name)
 {
-    (void) strncpy0(febuff, name, NBUFN);
+    (void) strncpy0(febuff, name, (size_t) NBUFN);
     newfebuff = TRUE;
 }
 
@@ -146,15 +179,67 @@ get_febuff(void)
     return febuff;
 }
 
+static const char *
+get_token_name(ErrTokens n)
+{
+    const char *result;
+
+    switch (n) {
+    case W_VERB:
+	result = "verb";
+	break;
+    case W_FILE:
+	result = "filename";
+	break;
+    case W_LINE:
+	result = "line";
+	break;
+    case W_COLM:
+	result = "column";
+	break;
+    case W_TEXT:
+	result = "text";
+	break;
+    default:
+	result = "unknown";
+	break;
+    }
+    return result;
+}
+
+static int
+marks_in(const char *expr)
+{
+    int result = 0;
+    int escaped = FALSE;
+
+    while (*expr != EOS) {
+	if (escaped) {
+	    if (*expr == L_PAREN)
+		result++;
+	    escaped = FALSE;
+	} else if (*expr == BACKSLASH) {
+	    escaped = TRUE;
+	}
+	expr++;
+    }
+    return result;
+}
+
 /*
  * Convert a given error-pattern to regular expression
  */
 
+#define	APP_T(S) if (pass == 1) want += strlen(S); \
+		 else dst = lsprintf(dst, "%s", S)
 #define	APP_S(S) if (pass == 1) want += sizeof(S); \
 		 else dst = lsprintf(dst, "%s", S)
 #define APP_C    if (pass != 1) *dst++ = *src
 
-static void
+/*
+ * Convert the pattern; returns false if malloc's fail.
+ */
+static int
 convert_pattern(ERR_PATTERN * errp, LINE *lp)
 {
     static const char before[] = "\\(";
@@ -169,11 +254,13 @@ convert_pattern(ERR_PATTERN * errp, LINE *lp)
     int word;
     int mark;
     int range;
-    size_t want = llength(lp);
-    char *first = lp->l_text;
+    int status = TRUE;
+    size_t want = (size_t) llength(lp);
+    char *first = lvalue(lp);
     char *last = first + want;
 
     (void) memset(errp, 0, sizeof(*errp));
+    TPRINTF(("error-pattern %.*s\n", (int) want, first));
 
     /* In the first pass, find the number of fields we'll substitute.
      * Then allocate a new string that's a genuine regular expression
@@ -184,6 +271,8 @@ convert_pattern(ERR_PATTERN * errp, LINE *lp)
 		APP_C;
 		if (++src == last)
 		    break;
+		if (*src == L_PAREN)	/* a group we don't own... */
+		    word++;
 		APP_C;
 	    } else if (*src == '%') {
 		mark = -1;
@@ -192,7 +281,13 @@ convert_pattern(ERR_PATTERN * errp, LINE *lp)
 		    mark = W_VERB;
 		    break;
 		case 'F':
-		    mark = W_FILE;
+		    if (tb_values(filename_expr) != 0) {
+			APP_S(before);
+			APP_T(tb_values(filename_expr));
+			APP_S(after);
+			errp->words[W_FILE] = ++word;
+			word += marks_in(tb_values(filename_expr));
+		    }
 		    break;
 		case 'B':
 		    APP_S("\\(\\[[^:]\\+]\\)");
@@ -202,15 +297,19 @@ convert_pattern(ERR_PATTERN * errp, LINE *lp)
 		    APP_S(remain);
 		    errp->words[W_TEXT] = ++word;
 		    break;
+		case 'c':
 		case 'C':
+		    errp->cBase = ((*src) == 'C');
 		    APP_S(number);
 		    errp->words[W_COLM] = ++word;
 		    break;
+		case 'l':
 		case 'L':
+		    errp->lBase = ((*src) == 'L');
 		    APP_S(number);
 		    errp->words[W_LINE] = ++word;
 		    break;
-		case LBRACK:
+		case L_BLOCK:
 		    range = TRUE;
 		    APP_S(before);
 		    APP_C;
@@ -218,7 +317,7 @@ convert_pattern(ERR_PATTERN * errp, LINE *lp)
 			src++;
 			APP_C;
 		    }
-		    if (src[1] == RBRACK) {
+		    if (src[1] == R_BLOCK) {
 			src++;
 			APP_C;
 		    }
@@ -231,7 +330,7 @@ convert_pattern(ERR_PATTERN * errp, LINE *lp)
 		    APP_S(normal);
 		    errp->words[mark] = ++word;
 		}
-	    } else if ((*src == RBRACK) && range) {
+	    } else if ((*src == R_BLOCK) && range) {
 		APP_C;
 		APP_S(after);
 		range = FALSE;
@@ -259,17 +358,34 @@ convert_pattern(ERR_PATTERN * errp, LINE *lp)
 	    }
 	}
 	if (pass == 1) {
-	    dst = temp = typeallocn(char, want);
-	    if (dst == 0)
+	    beginDisplay();
+	    dst = temp = typeallocn(char, want + 1);
+	    endofDisplay();
+	    if (dst == 0) {
+		status = no_memory("convert_pattern");
 		break;
-	} else
+	    }
+	} else {
 	    *dst = EOS;
+	}
     }
     if (temp != 0) {
-	exp = regcomp(temp, TRUE);
-	free(temp);
+#if OPT_TRACE
+	TRACE(("COMPILE %s\n", temp));
+	for (word = 0; word < W_LAST; word++)
+	    TRACE(("word[%d] = %d (%s)\n", word, errp->words[word],
+		   get_token_name((ErrTokens) word)));
+#endif
+	TPRINTF(("-> %s\n", temp));
+	exp = regcomp(temp, strlen(temp), TRUE);
+	/* FIXME:  this might be null if the pattern was incorrect, or if we
+	 * ran out of memory.  We only want the latter condition.
+	 */
     }
+    errp->exp_text = temp;
     errp->exp_comp = exp;
+
+    return status;
 }
 
 /*
@@ -279,16 +395,22 @@ static void
 free_patterns(void)
 {
     if (exp_table != 0) {
-	while (exp_count != 0)
-	    free((char *) (exp_table[--exp_count].exp_comp));
+	beginDisplay();
+	while (exp_count-- != 0) {
+	    free(exp_table[exp_count].exp_text);
+	    free((char *) (exp_table[exp_count].exp_comp));
+	}
 	free((char *) exp_table);
+	exp_table = 0;
+	exp_count = 0;
+	endofDisplay();
     }
 }
 
 #if OPT_UPBUFF
 /* ARGSUSED */
-static int
-update_patterns(BUFFER *bp GCC_UNUSED)
+int
+free_err_exps(BUFFER *bp GCC_UNUSED)
 {
     free_patterns();
     return TRUE;
@@ -305,7 +427,8 @@ load_patterns(void)
 {
     BUFFER *bp;
     LINE *lp;
-    SIZE_T n;
+    size_t n;
+    int status = TRUE;
 
     /* find the error-expressions buffer */
     if ((bp = find_b_name(ERRORS_BufName)) == 0) {
@@ -325,21 +448,31 @@ load_patterns(void)
 
     /* any change makes the patterns obsolete */
 #if OPT_UPBUFF
-    update_scratch(ERRORS_BufName, update_patterns);
-    bp->b_rmbuff = update_patterns;
+    update_scratch(ERRORS_BufName, free_err_exps);
+    bp->b_rmbuff = free_err_exps;
 #endif
 
     if (exp_count == 0) {
-	exp_count = bp->b_linecount;
+	beginDisplay();
+	exp_count = (size_t) bp->b_linecount;
 	exp_table = typeallocn(ERR_PATTERN, exp_count);
-	for (n = 0; n < W_LAST; n++)
-	    exp_table->words[n] = -1;
+	endofDisplay();
 
-	n = 0;
-	for_each_line(lp, bp)
-	    convert_pattern(&exp_table[n++], lp);
+	if (exp_table != 0) {
+	    for (n = 0; n < W_LAST; n++)
+		exp_table->words[n] = -1;
+	    n = 0;
+	    for_each_line(lp, bp) {
+		if (!convert_pattern(&exp_table[n++], lp)) {
+		    status = FALSE;
+		    break;
+		}
+	    }
+	} else {
+	    status = no_memory("load_patterns");
+	}
     }
-    return TRUE;
+    return status;
 }
 
 /*
@@ -348,7 +481,7 @@ load_patterns(void)
  * if count is out of range.
  */
 static ERR_PATTERN *
-next_pattern(ALLOC_T count)
+next_pattern(size_t count)
 {
     ERR_PATTERN *result = 0;
 
@@ -360,40 +493,153 @@ next_pattern(ALLOC_T count)
 /*
  * Decode the matched ERR_PATTERN
  */
-static void
+static int
 decode_exp(ERR_PATTERN * exp)
 {
+    /* *INDENT-OFF* */
+    static struct {
+	ErrTokens code;
+	TBUFF **buffer;
+	long *number;
+    } lookup[] = {
+	{ W_VERB, &fe_verb, 0 },
+	{ W_FILE, &fe_file, 0 },
+	{ W_LINE, 0, &fe_line },
+	{ W_COLM, 0, &fe_colm },
+	{ W_TEXT, &fe_text, 0 },
+    };
+    /* *INDENT-ON* */
+
     regexp *p = exp->exp_comp;
-    int n;
+    int j, n;
+    int failed = FALSE;
     TBUFF *temp;
 
-    tb_free(&fe_verb);
-    tb_free(&fe_file);
-    tb_free(&fe_text);
-    fe_colm = 1;
-    fe_line = 0;
+    TRACE(("decode_exp{%s}\n", exp->exp_text));
 
-    n = 0;
-    for (n = 1; (n < NSUBEXP) && p->startp[n] && p->endp[n]; n++) {
+    for (j = 0; j < W_LAST; j++) {
+	if (lookup[j].buffer != 0) {
+	    tb_free(lookup[j].buffer);
+	} else {
+	    *(lookup[j].number) = 0;
+	}
+    }
+    cC_base = exp->cBase;
+    lL_base = exp->lBase;
+    fe_colm = cC_base;
+
+    /*
+     * Count the atoms separately from the loop indices because when
+     * we do
+     *          setv $filename-expr='\([a-zA-Z]:\)\?[^ ^I\[:]'
+     * the resulting array will have an extra entry with null pointers
+     * for the nested atom in the %F expression:
+     *          \([a-zA-Z]:\)
+     */
+    for (n = 1; !failed && (n < NSUBEXP); n++) {
+	if (p->startp[n] == 0 || p->endp[n] == 0)
+	    continue;		/* discount nested atom */
+	if (p->startp[n] >= p->endp[n])
+	    continue;		/* discount empty atom */
 	temp = 0;
 	if (tb_bappend(&temp,
-		p->startp[n],
-		(ALLOC_T) (p->endp[n] - p->startp[n])) == 0
-	    || tb_append(&temp, EOS) == 0)
-	    return;
-
-	if (n == exp->words[W_VERB]) {
-	    fe_verb = temp;
-	} else if (n == exp->words[W_FILE]) {
-	    fe_file = temp;
-	} else if (n == exp->words[W_TEXT]) {
-	    fe_text = temp;
+		       p->startp[n],
+		       (size_t) (p->endp[n] - p->startp[n])) == 0
+	    || tb_append(&temp, EOS) == 0) {
+	    (void) no_memory("finderr");
+	    failed = TRUE;
+	} else if (tb_length(temp) == 0) {
+	    mlforce("BUG: marker %d is empty string", n);
+	    failed = TRUE;
 	} else {
-	    if (n == exp->words[W_LINE])
-		fe_line = atoi(tb_values(temp));
-	    else if (n == exp->words[W_COLM])
-		fe_colm = atoi(tb_values(temp));
-	    tb_free(&temp);
+	    for (j = 0; j < W_LAST; j++) {
+		ErrTokens code = lookup[j].code;
+		if (n == exp->words[code]) {
+		    if (lookup[j].buffer) {
+			*(lookup[j].buffer) = temp;
+			TRACE(("matched %s:%s\n", get_token_name(code),
+			       tb_values(temp)));
+		    } else {
+			*(lookup[j].number) = vl_atol(tb_values(temp), 10, &failed);
+			TRACE(("matched %s:%s(%ld)\n", get_token_name(code),
+			       tb_values(temp), *(lookup[j].number)));
+			if (failed) {
+			    mlforce("BUG: \"%s\" (marker %d) is not a number: %s",
+				    get_token_name(code),
+				    n,
+				    tb_values(temp));
+			}
+			tb_free(&temp);
+		    }
+		}
+	    }
+	}
+    }
+
+    if (!failed
+	&& (fe_colm != cC_base && fe_line == 0)) {
+	mlforce("BUG: found column %ld but no line", fe_colm);
+	failed = TRUE;
+    }
+
+    /*
+     * There's not enough room on the message line to show which expression
+     * failed - but update $error-expr anyway.
+     */
+    if (failed)
+	var_ERROR_EXPR((TBUFF **) 0, exp->exp_text);
+
+    return failed;
+}
+
+#define set_tabstop_val(bp,value) set_b_val(bp, VAL_TAB, value)
+
+static void
+goto_column(void)
+{
+    int saved_tabstop = tabstop_val(curbp);
+
+    if (error_tabstop > 0)
+	set_tabstop_val(curbp, error_tabstop);
+
+    gocol((int) (fe_colm ? fe_colm - cC_base : 0));
+
+    if (error_tabstop > 0)
+	set_tabstop_val(curbp, saved_tabstop);
+}
+
+#define DIRLEVELS 20
+static int dir_level = 0;
+static char *dir_stack[DIRLEVELS + 1];
+
+static void
+freeDirs(void)
+{
+    beginDisplay();
+    while (dir_level) {
+	free(dir_stack[dir_level]);
+	--dir_level;
+    }
+    endofDisplay();
+}
+
+static void
+updateDirs(const char *errverb, const char *errfile)
+{
+    if (!strcmp("Entering", errverb)) {
+	if (dir_level < DIRLEVELS) {
+	    beginDisplay();
+	    ++dir_level;
+	    dir_stack[dir_level] = strmalloc(errfile);
+	    endofDisplay();
+	}
+    } else if (!strcmp("Leaving", errverb)) {
+	if (dir_level > 0) {
+	    beginDisplay();
+	    if (dir_stack[dir_level] != 0)
+		free(dir_stack[dir_level]);
+	    --dir_level;
+	    endofDisplay();
 	}
     }
 }
@@ -408,28 +654,25 @@ decode_exp(ERR_PATTERN * exp)
 int
 finderr(int f GCC_UNUSED, int n GCC_UNUSED)
 {
-    register BUFFER *sbp;
-    register int s;
+    BUFFER *sbp;
+    int status;
     LINE *dotp;
     int moveddot = FALSE;
     ERR_PATTERN *exp;
+    size_t count;
 
     char *errverb;
     char *errfile;
     char *errtext;
     char ferrfile[NFILEN];
-    ALLOC_T len;
+    size_t len;
 
-    static int oerrline = -1;
+    static long oerrline = -1;
     static TBUFF *oerrfile;
     static TBUFF *oerrtext;
 
     static LINE *odotp = 0;
 
-#define DIRLEVELS 20
-    static int l = 0;
-    static char *dirs[DIRLEVELS];
-
     if (!comp_err_exps(FALSE, 1))
 	return (FALSE);
 
@@ -439,21 +682,25 @@ finderr(int f GCC_UNUSED, int n GCC_UNUSED)
 	return (FALSE);
     }
     if (newfebuff) {
+	beginDisplay();
 	oerrline = -1;
 	oerrfile = tb_init(&oerrfile, EOS);
 	oerrtext = tb_init(&oerrtext, EOS);
-	while (l)
-	    free(dirs[l--]);
+	freeDirs();
 	odotp = 0;
+	endofDisplay();
     }
     dotp = getdot(sbp);
+    if (dotp == 0) {
+	TRACE(("getdot returns null\n"));
+	return (FALSE);
+    }
 
     if (newfebuff || dotp != odotp) {
 
 	LINE *tdotp;
 
-	while (l)
-	    free(dirs[l--]);
+	freeDirs();
 
 	tdotp = lforw(buf_head(sbp));
 
@@ -463,11 +710,11 @@ finderr(int f GCC_UNUSED, int n GCC_UNUSED)
 	 * have jumped out of sequence we need to recalibrate
 	 * the directory stack against our current position.
 	 */
+	TRACE(("check for Entering/Leaving lines\n"));
 	while (tdotp != dotp) {
 
 	    if (lisreal(tdotp)) {
-		ALLOC_T count = 0;
-
+		count = 0;
 		while ((exp = next_pattern(count++)) != 0) {
 		    if (exp->words[W_VERB] > 0)
 			if (lregexec(exp->exp_comp, tdotp, 0, llength(tdotp)))
@@ -475,20 +722,14 @@ finderr(int f GCC_UNUSED, int n GCC_UNUSED)
 		}
 
 		if (exp != 0) {
-		    decode_exp(exp);
+		    if (decode_exp(exp))
+			return ABORT;
 
 		    errverb = tb_values(fe_verb);
 		    errfile = tb_values(fe_file);
 
 		    if (errverb != 0 && errfile != 0) {
-			if (!strcmp("Entering", errverb)) {
-			    if (l < DIRLEVELS) {
-				dirs[++l] = strmalloc(errfile);
-			    }
-			} else if (!strcmp("Leaving", errverb)) {
-			    if (l > 0)
-				free(dirs[l--]);
-			}
+			updateDirs(errverb, errfile);
 		    }
 		} else if (interrupted()) {
 		    kbd_alarm();
@@ -500,19 +741,23 @@ finderr(int f GCC_UNUSED, int n GCC_UNUSED)
     }
     newfebuff = FALSE;
 
+    TRACE(("look for matching line\n"));
     for_ever {
 	/* To use this line, we need both the filename and the line
 	 * number in the expected places, and a different line than
 	 * last time.
 	 */
 	if (lisreal(dotp)) {
-	    ALLOC_T count = 0;
-
+	    count = 0;
 	    while ((exp = next_pattern(count++)) != 0
-		&& !lregexec(exp->exp_comp, dotp, 0, llength(dotp))) ;
+		   && !lregexec(exp->exp_comp, dotp, 0, llength(dotp))) {
+		;
+	    }
 
 	    if (exp != 0) {
-		decode_exp(exp);
+		TRACE(("matched TEXT:%.*s\n", llength(dotp), lvalue(dotp)));
+		if (decode_exp(exp))
+		    return ABORT;
 
 		errverb = tb_values(fe_verb);
 		errfile = tb_values(fe_file);
@@ -527,16 +772,8 @@ finderr(int f GCC_UNUSED, int n GCC_UNUSED)
 			&& errtext != 0
 			&& strcmp(tb_values(oerrtext), errtext))
 			break;
-		} else if (errverb != 0
-		    && errfile != 0) {
-		    if (!strcmp("Entering", errverb)) {
-			if (l < DIRLEVELS) {
-			    dirs[++l] = strmalloc(errfile);
-			}
-		    } else if (!strcmp("Leaving", errverb)) {
-			if (l > 0)
-			    free(dirs[l--]);
-		    }
+		} else if (errverb != 0 && errfile != 0) {
+		    updateDirs(errverb, errfile);
 		}
 	    }
 	}
@@ -547,8 +784,7 @@ finderr(int f GCC_UNUSED, int n GCC_UNUSED)
 	    mlwarn("[No more errors in %s buffer]", febuff);
 	    /* start over at the top of file */
 	    putdotback(sbp, lforw(buf_head(sbp)));
-	    while (l)
-		free(dirs[l--]);
+	    freeDirs();
 	    return FALSE;
 	}
 	dotp = lforw(dotp);
@@ -561,7 +797,7 @@ finderr(int f GCC_UNUSED, int n GCC_UNUSED)
 
     odotp = dotp;
 
-    (void) lengthen_path(pathcat(ferrfile, dirs[l], errfile));
+    (void) lengthen_path(pathcat(ferrfile, dir_stack[dir_level], errfile));
 
     if (strcmp(ferrfile, curbp->b_fname)) {
 	/* if we must change windows */
@@ -575,18 +811,18 @@ finderr(int f GCC_UNUSED, int n GCC_UNUSED)
 	    make_current(curwp->w_bufp);
 	    upmode();
 	} else {
-	    s = getfile(ferrfile, TRUE);
-	    if (s != TRUE)
-		return s;
+	    status = getfile(ferrfile, TRUE);
+	    if (status != TRUE)
+		return status;
 	}
     }
     if (errtext) {
 	mlforce("%s", errtext);
 	len = strlen(errtext);
     } else {
-	mlforce("Error: %*S", dotp->l_used, dotp->l_text);
-	errtext = dotp->l_text;
-	len = dotp->l_used;
+	mlforce("Error: %.*s", llength(dotp), lvalue(dotp));
+	errtext = lvalue(dotp);
+	len = (size_t) llength(dotp);
     }
     if ((oerrtext = tb_init(&oerrtext, EOS)) != 0) {
 	tb_bappend(&oerrtext, errtext, len);
@@ -594,29 +830,49 @@ finderr(int f GCC_UNUSED, int n GCC_UNUSED)
     }
     /* it's an absolute move */
     curwp->w_lastdot = DOT;
-    s = gotoline(TRUE, -(curbp->b_lines_on_disk - fe_line + 1));
-    DOT.o = fe_colm ? fe_colm - 1 : 0;
+    if (fe_line <= lL_base) {
+	status = gotobob(f, n);
+    } else if ((fe_line - lL_base) >= curbp->b_lines_on_disk) {
+	status = gotoeob(f, n);
+    } else {
+	status = gotoline(TRUE,
+			  (int) -(curbp->b_lines_on_disk - fe_line + lL_base));
+    }
+    goto_column();
 
     oerrline = fe_line;
     (void) tb_scopy(&oerrfile, errfile);
+    if (status == TRUE) {
+	TBUFF *match = 0;
+	var_ERROR_EXPR((TBUFF **) 0, exp_table[count - 1].exp_text);
+	if (tb_bappend(&match, lvalue(dotp), (size_t) llength(dotp))
+	    && tb_append(&match, EOS) != 0) {
+	    var_ERROR_MATCH((TBUFF **) 0, tb_values(match));
+	    tb_free(&match);
+	}
+	updatelistvariables();
+    }
 
-    return s;
+    return status;
 }
 
 static LINE *
 getdot(BUFFER *bp)
 {
-    register WINDOW *wp;
+    LINE *result = bp->b_dot.l;
+    WINDOW *wp;
+
     if (bp->b_nwnd) {
 	/* scan for windows holding that buffer,
 	   pull dot from the first */
 	for_each_visible_window(wp) {
 	    if (wp->w_bufp == bp) {
-		return wp->w_dot.l;
+		result = wp->w_dot.l;
+		break;
 	    }
 	}
     }
-    return bp->b_dot.l;
+    return result;
 }
 
 static void
@@ -652,7 +908,7 @@ finderrbuf(int f GCC_UNUSED, int n GCC_UNUSED)
 
     (void) strcpy(name, febuff);
     if ((s = ask_for_bname("Buffer to scan for \"errors\": ",
-		name, sizeof(name))) == ABORT)
+			   name, sizeof(name))) == ABORT)
 	return s;
     if (s == FALSE) {
 	set_febuff(OUTPUT_BufName);
@@ -664,6 +920,78 @@ finderrbuf(int f GCC_UNUSED, int n GCC_UNUSED)
     return TRUE;
 }
 
+#define ERR_PREFIX 8
+
+static void
+make_err_regex_list(int dum1 GCC_UNUSED, void *ptr GCC_UNUSED)
+{
+    BUFFER *bp;
+    LINE *lp;
+    char temp[NSTRING];
+
+    bprintf("--- Error Meta-Expressions and Resulting Regular Expressions ");
+    bpadc('-', term.cols - DOT.o);
+
+    if (exp_table == 0)
+	load_patterns();
+
+    if (exp_table != 0
+	&& (bp = find_b_name(ERRORS_BufName)) != 0) {
+	size_t j = 0;
+	b_set_left_margin(curbp, ERR_PREFIX);
+	for_each_line(lp, bp) {
+	    int k, first = TRUE;
+	    if (j >= exp_count)
+		break;
+	    bprintf("\n%7d ", (int) j);
+	    bputsn_xcolor(lvalue(lp), llength(lp), XCOLOR_STRING);
+	    bputc('\n');
+	    bprintf("%.*s", ERR_PREFIX, " ");
+	    bputsn_xcolor(exp_table[j].exp_text, -1, XCOLOR_REGEX);
+	    for (k = 0; k < W_LAST; k++) {
+		if (exp_table[j].words[k] != 0) {
+		    if (first) {
+			bprintf("\n%.*s", ERR_PREFIX, " ");
+			first = FALSE;
+		    } else {
+			bprintf(", ");
+		    }
+		    bputsn_xcolor(get_token_name((ErrTokens) k), -1, XCOLOR_ENUM);
+		    bputc('=');
+		    sprintf(temp, "\\%d", exp_table[j].words[k]);
+		    bputsn_xcolor(temp, -1, XCOLOR_REGEX);
+		}
+	    }
+	    ++j;
+	}
+    }
+}
+
+/* ARGSUSED */
+static int
+show_ErrRegex(BUFFER *bp GCC_UNUSED)
+{
+    return liststuff(ERR_REGEX_BufName,
+		     FALSE, make_err_regex_list, 0, (void *) 0);
+}
+
+int
+show_err_regex(int f GCC_UNUSED, int n GCC_UNUSED)
+{
+    return show_ErrRegex(curbp);
+}
+
+#if OPT_UPBUFF
+void
+update_err_regex(void)
+{
+    update_scratch(ERR_REGEX_BufName, show_ErrRegex);
+}
+
+#else
+#define update_err_regex()	/*nothing */
+#endif
+
 /*
  * (Re)compile the error-expressions buffer.  This is needed as an entrypoint
  * so that macros can manipulate the set of expressions (including reading it
@@ -677,6 +1005,7 @@ comp_err_exps(int f GCC_UNUSED, int n GCC_UNUSED)
 	mlforce("[No error-expressions are defined.]");
 	return (FALSE);
     }
+    update_err_regex();
     return TRUE;
 }
 #endif
