@@ -44,7 +44,7 @@
  *	tgetc_avail()     true if a key is avail from tgetc() or below.
  *	keystroke_avail() true if a key is avail from keystroke() or below.
  *
- * $Header: /usr/build/vile/vile/RCS/input.c,v 1.323 2009/03/20 19:47:52 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/input.c,v 1.365 2015/09/04 01:05:11 tom Exp $
  *
  */
 
@@ -215,7 +215,7 @@ mlquickask(const char *prompt, const char *respchars, int *cp)
 		break;
 	    }
 
-	    if (strchr(respchars, *cp)) {
+	    if (vl_index(respchars, *cp)) {
 		result = TRUE;
 		break;
 	    }
@@ -239,9 +239,17 @@ mlreply_reg(const char *prompt,
     int status;
     int c;
 
+    regs_kbd_default = (at_dft != -1);
     if (clexec || isnamedcmd) {
-	if ((status = mlreply(prompt, cbuf, 2)) != TRUE)
+	TBUFF *result = 0;
+	tb_scopy(&result, cbuf);
+	if ((status = kbd_reply(prompt, &result, eol_history, '\n',
+				regs_kbd_options(), regs_completion)) != TRUE) {
+	    tb_free(&result);
 	    return status;
+	}
+	strcpy(cbuf, tb_values(result));
+	tb_free(&result);
 	c = cbuf[0];
     } else {
 	c = keystroke();
@@ -249,7 +257,7 @@ mlreply_reg(const char *prompt,
 	    return ABORT;
     }
 
-    if (c == '@' && at_dft != -1) {
+    if (c == UNAME_KCHR && regs_kbd_default) {
 	c = at_dft;
     } else if (reg2index(c) < 0) {
 	mlwarn("[Invalid register name]");
@@ -265,15 +273,16 @@ mlreply_reg(const char *prompt,
  * ":put" commands).  The register-name, if given, is first.
  */
 int
-mlreply_reg_count(int state,	/* negative=register, positive=count, zero=either */
+mlreply_reg_count(const char *cmd,	/* command-name, for clarity */
+		  int state,	/* negative=register, positive=count, zero=either */
 		  int *retp,	/* returns the register-index or line-count */
 		  int *next)	/* returns 0/1=register, 2=count */
 {
     int status;
     char prompt[80];
     char expect[80];
-    char buffer[10];
-    UINT length;
+    char *buffer;
+    TBUFF *result = 0;
 
     *expect = EOS;
     if (state <= 0)
@@ -282,15 +291,15 @@ mlreply_reg_count(int state,	/* negative=register, positive=count, zero=either *
 	(void) strcat(expect, " or");
     if (state >= 0) {
 	(void) strcat(expect, " line-count");
-	length = sizeof(buffer);
-    } else
-	length = 2;
+    }
 
-    (void) lsprintf(prompt, "Specify%s: ", expect);
-    *buffer = EOS;
-    status = kbd_string(prompt, buffer, length, ' ', 0, no_completion);
+    (void) lsprintf(prompt, "Specify%s (%s): ", expect, cmd);
+    tb_scopy(&result, "");
+    status = kbd_reply(prompt, &result, eol_history, '\n',
+		       regs_kbd_options(), regs_completion);
 
     if (status == TRUE) {
+	buffer = tb_values(result);
 	if (state <= 0
 	    && isAlpha(buffer[0])
 	    && buffer[1] == EOS
@@ -306,6 +315,7 @@ mlreply_reg_count(int state,	/* negative=register, positive=count, zero=either *
 	    status = ABORT;
 	}
     }
+    tb_free(&result);
     return status;
 }
 
@@ -319,7 +329,7 @@ mlreply_reg_count(int state,	/* negative=register, positive=count, zero=either *
 int
 mlreply(const char *prompt, char *buf, UINT bufn)
 {
-    return kbd_string(prompt, buf, bufn, '\n', KBD_NORMAL, no_completion);
+    return kbd_string(prompt, buf, (size_t) bufn, '\n', KBD_NORMAL, no_completion);
 }
 
 int
@@ -337,7 +347,11 @@ mlreply_no_bs(const char *prompt, char *buf, UINT bufn)
     doing_shell = TRUE;
     init_filec(FILECOMPLETION_BufName);
 #endif
-    s = kbd_string(prompt, buf, bufn, '\n', KBD_EXPAND | KBD_SHPIPE, shell_complete);
+    s = kbd_string(prompt,
+		   buf,
+		   (size_t) bufn,
+		   '\n',
+		   KBD_EXPAND | KBD_SHPIPE, shell_complete);
 #if COMPLETE_FILES
     doing_shell = FALSE;
 #endif
@@ -348,7 +362,7 @@ mlreply_no_bs(const char *prompt, char *buf, UINT bufn)
 int
 mlreply_no_opts(const char *prompt, char *buf, UINT bufn)
 {
-    return kbd_string(prompt, buf, bufn, '\n', 0, no_completion);
+    return kbd_string(prompt, buf, (size_t) bufn, '\n', 0, no_completion);
 }
 
 /* the numbered buffer names increment each time they are referenced */
@@ -409,7 +423,7 @@ get_recorded_char(int eatit)	/* consume the character? */
 	    if (!itb_more(buffer = dotcmd)) {
 		if (!eatit) {
 		    if (dotcmdrep > 1)
-			return itb_get(buffer, 0);
+			return itb_get(buffer, (size_t) 0);
 		} else {	/* at the end of last repetition?  */
 		    if (--dotcmdrep < 1) {
 			dotcmdactive = 0;
@@ -468,7 +482,7 @@ get_recorded_char(int eatit)	/* consume the character? */
 void
 unkeystroke(int c)
 {
-    mapungetc(c | NOREMAP);
+    mapungetc((int) ((UINT) c | NOREMAP));
 }
 
 int
@@ -549,7 +563,7 @@ tgetc(int quoted)
 	/* read a character from the terminal */
 	not_interrupted();
 	if (setjmp(read_jmp_buf)) {
-	    c = kcod2key(intrc);
+	    c = kcod2key((UINT) intrc);
 	    TRACE(("setjmp/getc:%c (%#x)\n", c, c));
 #if defined(BUG_LINUX_2_0_INTR) && DISP_TERMCAP
 	    /*
@@ -570,7 +584,7 @@ tgetc(int quoted)
 	    } while (c == -1);
 	}
 	(void) im_waiting(FALSE);
-	if (quoted || (c != kcod2key(intrc)))
+	if (quoted || (c != kcod2key((UINT) intrc)))
 	    record_char(c);
     }
 
@@ -578,10 +592,10 @@ tgetc(int quoted)
 #if OPT_TRACE
     {
 	char temp[NSTRING];
-	TRACE(("tgetc(%s) = %s\n", (quoted
-				    ? "quoted"
-				    : "unquoted"),
-	       kcod2prc(c, temp)));
+	TRACE(("tgetc(%s) = %s(%#x)\n", (quoted
+					 ? "quoted"
+					 : "unquoted"),
+	       kcod2prc(c, temp), c));
     }
 #endif
     return c;
@@ -636,9 +650,16 @@ adjust_chartype(CHARTYPE *mask)
 {
     int whole_line = FALSE;
 
-    /* if from gototag(), grab from the beginning of the string */
-    if (b_val(curbp, MDTAGWORD)
-	&& *mask == vl_ident) {
+#if OPT_EVAL
+    if (vl_get_it_all) {
+	whole_line = TRUE;
+	if (*mask == 0)		/* special case for $line variable */
+	    *mask = ~(*mask);
+    } else
+#endif
+	if (b_val(curbp, MDTAGWORD)
+	    && *mask == vl_ident) {
+	/* if from gototag(), grab from the beginning of the string */
 	whole_line = TRUE;
     } else if (b_is_directory(curbp)
 	       && *mask == SCREEN_STRING) {
@@ -667,7 +688,9 @@ read_by_ctype(char *buf, size_t bufn, CHARTYPE inclchartype, int whole_line)
     }
 #endif
 
-    TRACE((T_CALLED "read_by_ctype(incl=%#lx)\n", (ULONG) inclchartype));
+    TRACE((T_CALLED "read_by_ctype(incl=%#lx, whole=%d)\n",
+	   (ULONG) inclchartype, whole_line));
+
     if ((rc = vl_ctype2tbuff(&temp, inclchartype, whole_line)) == TRUE) {
 	if ((len = tb_length(temp)) >= bufn)
 	    len = bufn - 1;
@@ -776,27 +799,31 @@ screen_to_ident(char *buf, size_t bufn)
 int
 vl_ctype2tbuff(TBUFF **result, CHARTYPE inclchartype, int whole_line)
 {
+    int okay;
     int i = 0;
     MARK save_dot;
-    int first = -1;
-    int last = -1;
+    int first_ch = -1;
+    int last_ch = -1;
+    int first_off = -1;
+    int save_off;
+    int last_off = -1;
+    int at_end = 0;
 
-    TRACE((T_CALLED "vl_ctype2tbuff(incl=%#lx)\n", (ULONG) inclchartype));
+    TRACE((T_CALLED "vl_ctype2tbuff(incl=%#lx, whole=%d)\n",
+	   (ULONG) inclchartype, whole_line));
 
     save_dot = DOT;
 
-    tb_init(result, EOS);
-
     /*
      * If we're processing the whole line, find the beginning of the token
      * by searching to the left.
      */
     if (whole_line
 	&& DOT.o > 0
-	&& istype(inclchartype, char_at(DOT))) {
+	&& HasCType(inclchartype, CharAtDot())) {
 	while (DOT.o > 0) {
 	    backchar_to_bol(TRUE, 1);
-	    if (!istype(inclchartype, char_at(DOT))) {
+	    if (!HasCType(inclchartype, CharAtDot())) {
 		forwchar_to_eol(TRUE, 1);
 		break;
 	    }
@@ -806,18 +833,19 @@ vl_ctype2tbuff(TBUFF **result, CHARTYPE inclchartype, int whole_line)
     /*
      * Search for the start of the token, and then the end of the token.
      */
+    first_off = DOT.o;
     while (!is_at_end_of_line(DOT)) {
-	last = CharOf(char_at(DOT));
-	if (first < 0)
-	    first = last;
+	last_ch = CharAtDot();
+	if (first_ch < 0)
+	    first_ch = last_ch;
 #if OPT_WIDE_CTYPES
 	if (i == 0) {
 	    if (inclchartype & vl_scrtch) {
-		if (first != SCRTCH_LEFT[0])
+		if (first_ch != SCRTCH_LEFT[0])
 		    inclchartype &= ~vl_scrtch;
 	    }
 	    if (inclchartype & vl_shpipe) {
-		if (first != SHPIPE_LEFT[0])
+		if (first_ch != SHPIPE_LEFT[0])
 		    inclchartype &= ~vl_shpipe;
 	    }
 	}
@@ -825,32 +853,38 @@ vl_ctype2tbuff(TBUFF **result, CHARTYPE inclchartype, int whole_line)
 	/* allow "[!command]" */
 	if ((inclchartype & vl_scrtch)
 	    && (i == 1)
-	    && (last == SHPIPE_LEFT[0])) {
+	    && (last_ch == SHPIPE_LEFT[0])) {
 	    /*EMPTY */ ;
-	    /* guard against things like "[Buffer List]" on VMS */
 	} else if ((inclchartype & vl_pathn)
-		   && !ispath(last)
+		   && !HasCType(vl_pathn, last_ch)
 		   && (inclchartype == vl_pathn)) {
+	    /* guard against things like "[Buffer List]" on VMS */
 	    break;
 	} else
 #endif
-	if (inclchartype && !istype(inclchartype, last))
+	if (inclchartype && !HasCType(inclchartype, last_ch)) {
 	    break;
-	tb_append(result, last);
-	if (!forwchar_to_eol(TRUE, 1))
+	}
+	if (!forwchar_to_eol(TRUE, 1)) {
+	    at_end = 1;
 	    break;
+	}
 	i++;
 #if OPT_WIDE_CTYPES
 	if (inclchartype & vl_scrtch) {
 	    if ((inclchartype & vl_pathn)
-		&& ispath(char_at(DOT)))
+		&& HasCType(vl_pathn, CharAtDot())) {
 		continue;
-	    if (last == SCRTCH_RIGHT[0])
+	    }
+	    if (last_ch == SCRTCH_RIGHT[0]) {
 		break;
+	    }
 	}
 #endif
     }
-    tb_append(result, EOS);
+    last_off = DOT.o + at_end;
+    if (!inclchartype)
+	last_off += len_record_sep(curbp);
 
 #if OPT_WIDE_CTYPES
 #if OPT_VMS_PATH
@@ -859,57 +893,162 @@ vl_ctype2tbuff(TBUFF **result, CHARTYPE inclchartype, int whole_line)
     } else
 #endif
     if (inclchartype & vl_scrtch) {
-	if (last != SCRTCH_RIGHT[0])
-	    tb_init(result, EOS);
+	if (last_ch != SCRTCH_RIGHT[0]) {
+	    last_off = first_off - 1;
+	}
     }
 #endif
 
+    tb_init(result, EOS);
+    save_off = first_off;
+    while (first_off < last_off && first_off < llength(DOT.l)) {
+	tb_append(result, lgetc(DOT.l, first_off));
+	++first_off;
+    }
+    tb_append(result, EOS);
     DOT = save_dot;
 
-    returnCode(tb_length(*result) != 0);
+    okay = (tb_length(*result) != 0);
+
+    /*
+     * Set side-effect variable $get-offset and $get-length, for scripts.
+     */
+#if OPT_EVAL
+    if (okay) {
+	vl_get_offset = (C_NUM) save_off;
+	vl_get_length = ((C_NUM) tb_length(*result)) - 1;
+	if (vl_get_at_dot) {
+	    if ((vl_get_offset + vl_get_length <= DOT.o) ||
+		(vl_get_offset > DOT.o)) {
+		vl_get_offset = -1;
+		vl_get_length = -1;
+	    }
+#if OPT_MULTIBYTE
+	    else if (b_is_utfXX(curbp)) {
+		int offset = chars_to_bol(DOT.l, vl_get_offset);
+		int length = count_chars(DOT.l, vl_get_offset, vl_get_length);
+		vl_get_offset = offset;
+		vl_get_length = length;
+	    }
+#endif
+	}
+    } else {
+	vl_get_offset = -1;
+	vl_get_length = -1;
+    }
+#endif
+
+    returnCode(okay);
 }
 
-/*
- * Get a string matching the given expression, which includes the current
- * position.  If the "whole_line" parameter is true, allow the string to
- * start before the current position.
- */
 #if OPT_CURTOKENS
-int
-vl_regex2tbuff(TBUFF **result, REGEXVAL * rexp, int whole_line)
+static void
+vl_regex2tbuff_best(TBUFF **result, regexp * exp)
 {
     C_NUM given = DOT.o;
     C_NUM length;
     C_NUM offset;
-    C_NUM match_off = -1;
-    C_NUM match_len = -1;
+    char *line_text = lvalue(DOT.l);
 
-    TRACE((T_CALLED "vl_regex2tbuff\n"));
+    vl_get_offset = -1;
+    vl_get_length = -1;
+    while (given >= 0) {
+	if (lregexec(exp, DOT.l, given, llength(DOT.l))) {
+	    offset = (C_NUM) (exp->startp[0] - line_text);
+	    length = (C_NUM) (exp->endp[0] - exp->startp[0]);
+	    if ((length > vl_get_length)
+		&& (offset + length > DOT.o)
+		&& (offset <= DOT.o || !vl_get_at_dot)) {
+		vl_get_offset = offset;
+		vl_get_length = length;
+	    }
+	}
+	--given;
+    }
+    if (vl_get_length > 0) {
+	tb_bappend(result, line_text + vl_get_offset, (size_t) vl_get_length);
+	tb_append(result, EOS);
+    }
+}
+
+static void
+vl_regex2tbuff_dot(TBUFF **result, regexp * exp)
+{
+    C_NUM start = DOT.o;
+    C_NUM given = DOT.o;
+    C_NUM length;
+    C_NUM offset;
+    char *line_text = lvalue(DOT.l);
+
+    vl_get_offset = -1;
+    vl_get_length = -1;
+    while (given >= 0) {
+	if (lregexec(exp, DOT.l, given, llength(DOT.l))) {
+	    offset = (C_NUM) (exp->startp[0] - line_text);
+	    length = (C_NUM) (exp->endp[0] - exp->startp[0]);
+	    if (offset <= DOT.o) {
+		if ((length > vl_get_length || vl_get_offset > DOT.o)
+		    && (offset + length > DOT.o)) {
+		    vl_get_offset = offset;
+		    vl_get_length = length;
+		}
+	    } else if ((length > vl_get_length) && (offset + length >= DOT.o)) {
+		vl_get_offset = offset;
+		vl_get_length = length;
+	    }
+	}
+	if ((vl_get_offset >= 0
+	     && vl_get_offset <= DOT.o
+	     && vl_get_offset + vl_get_length > DOT.o)) {
+	    vl_get_length -= (start - given);
+	    vl_get_offset = start;
+	    break;
+	}
+	--given;
+    }
+    if (vl_get_length > 0) {
+	tb_bappend(result, line_text + vl_get_offset, (size_t) vl_get_length);
+	tb_append(result, EOS);
+    }
+}
+
+/*
+ * Get a string matching the given expression, which includes the current
+ * position.
+ *
+ * The "whole_line" parameter selects one of two cases for matching:
+ *
+ * 1. true, try successively smaller offsets, looking only for the best match.
+ * This is used for tags and directory-buffers.
+ *
+ * 2. false, try smaller offsets, constrained by preferring to match a string
+ * including dot.  This is used for matching $identifier
+ */
+int
+vl_regex2tbuff(TBUFF **result, REGEXVAL * rexp, int whole_line)
+{
+    regexp *exp = rexp->reg;
+
+    TRACE((T_CALLED "vl_regex2tbuff(pat=%s, whole=%d)\n",
+	   NonNull(rexp->pat), whole_line));
 
     tb_init(result, EOS);
 
-    if (rexp != 0 && rexp->pat != 0 && rexp->pat[0] && rexp->reg) {
-	regexp *exp = rexp->reg;
-	char *line_text = lvalue(DOT.l);
-
-	while (given >= 0) {
-	    if (lregexec(exp, DOT.l, given, llength(DOT.l))) {
-		offset = exp->startp[0] - line_text;
-		length = exp->endp[0] - exp->startp[0];
-		if ((length > match_len) && (offset + length >= DOT.o)) {
-		    match_off = offset;
-		    match_len = length;
-		}
-	    }
-	    if (!whole_line)
-		break;
-	    --given;
-	}
-	if (match_len > 0) {
-	    tb_bappend(result, line_text + match_off, match_len);
-	    tb_append(result, EOS);
+    if (rexp->pat != 0 && rexp->pat[0] && rexp->reg) {
+	if (whole_line) {
+	    vl_regex2tbuff_best(result, exp);
+	} else {
+	    vl_regex2tbuff_dot(result, exp);
 	}
     }
+#if OPT_MULTIBYTE
+    if (vl_get_offset >= 0 && vl_get_length > 0 && b_is_utfXX(curbp)) {
+	int offset = chars_to_bol(DOT.l, vl_get_offset);
+	int length = count_chars(DOT.l, vl_get_offset, vl_get_length);
+	vl_get_offset = offset;
+	vl_get_length = length;
+    }
+#endif
 
     returnCode(tb_length(*result) != 0);
 }
@@ -972,7 +1111,7 @@ add_backslashes2(char *text, const char *find)
 	if (strpbrk(text, find) != 0) {
 	    temp = tb_init(&temp, EOS);
 	    for (n = 0; text[n] != EOS; ++n) {
-		if (strchr(find, text[n]))
+		if (vl_index(find, text[n]))
 		    temp = tb_append(&temp, BACKSLASH);
 		temp = tb_append(&temp, text[n]);
 	    }
@@ -1092,6 +1231,11 @@ show1Char(int c)
     kbd_flush();
 }
 
+#define isRangeCmd(s) \
+	((s)[0] == SQUOTE || isDigit(s[0]))
+#define editingRangeCmd(buf) \
+	((buf) != 0 && isRangeCmd(buf))
+
 /*
  * Macro result is true if the buffer is a normal :-command with a leading "!",
  * or is invoked from one of the places that supplies an implicit "!" for shell
@@ -1115,7 +1259,7 @@ expandChar(TBUFF **buf,
     char *buffer = tb_values(*buf);
     int shell = editingShellCmd(buffer, options);
     int expand = ((options & KBD_EXPAND) || shell);
-    int exppat = (options & KBD_EXPPAT);
+    int exppat = (options & KBD_EXPPAT) != 0;
 
     if (buffer == 0)
 	return FALSE;
@@ -1125,7 +1269,7 @@ expandChar(TBUFF **buf,
 	return FALSE;
 
     /* Is this a character that we know about? */
-    if (strchr(global_g_val_ptr(GVAL_EXPAND_CHARS), c) == 0)
+    if (vl_index(global_g_val_ptr(GVAL_EXPAND_CHARS), c) == 0)
 	return FALSE;
 
     switch (c) {
@@ -1144,7 +1288,7 @@ expandChar(TBUFF **buf,
 	if (isInternalName(cp)) {
 	    cp = bp->b_bname;
 	} else if (!global_g_val(GMDEXPAND_PATH)) {
-	    cp = shorten_path(strcpy(str, cp), FALSE);
+	    cp = shorten_path(vl_strncpy(str, cp, sizeof(str)), FALSE);
 #if OPT_MSDOS_PATH
 	    /* always use backslashes if invoking external prog */
 	    if (shell)
@@ -1197,12 +1341,12 @@ expandChar(TBUFF **buf,
 
     if (cp != NULL) {
 	while ((c = *cp++) != EOS) {
-	    tb_insert(buf, cpos++, c);
+	    tb_insert(buf, (size_t) cpos++, c);
 	    showChar(c);
 	}
 	kbd_flush();
     }
-    *position = cpos;
+    *position = (size_t) cpos;
     return TRUE;
 }
 
@@ -1227,12 +1371,12 @@ kbd_kill_response(TBUFF *buffer, size_t *position, int c)
     char *buf = tb_values(buffer);
     TBUFF *tmp = 0;
     int cpos = (int) *position;
-    UINT mark = cpos;
+    size_t mark = *position;
 
     if (buf != 0) {
 	tmp = tb_copy(&tmp, buffer);
 	while (cpos > 0) {
-	    cpos--;
+	    cpos -= TbBytesBefore(tmp, cpos);
 	    kbd_erase();
 	    if (c == wkillc) {
 		if (!isSpace(buf[cpos])) {
@@ -1247,8 +1391,8 @@ kbd_kill_response(TBUFF *buffer, size_t *position, int c)
 	if (vl_echo)
 	    kbd_flush();
 
-	*position = cpos;
-	buffer->tb_used = cpos;
+	*position = (size_t) cpos;
+	buffer->tb_used = (size_t) cpos;
 	if (mark < tb_length(tmp)) {
 	    tb_bappend(&buffer, tb_values(tmp) + mark, tb_length(tmp) - mark);
 	}
@@ -1278,7 +1422,7 @@ kbd_show_response(TBUFF **dst,	/* string with escapes */
 	if ((c == BACKSLASH) || (c == eolchar && eolchar != '\n')) {
 	    if (options & KBD_QUOTES)
 		tb_append(dst, BACKSLASH);	/* add extra */
-	} else if (strchr(global_g_val_ptr(GVAL_EXPAND_CHARS), c) != 0) {
+	} else if (vl_index(global_g_val_ptr(GVAL_EXPAND_CHARS), c) != 0) {
 	    if (c == EXPC_RPAT && !(options & KBD_EXPPAT))
 		/*EMPTY */ ;
 	    else if (c == EXPC_SHELL && !(options & KBD_SHPIPE))
@@ -1348,7 +1492,7 @@ kbd_pushback(TBUFF *buf, int skip)
 	clexec = TRUE;
 	execstr = tb_values(PushBack);
 	buffer[skip] = EOS;
-	buf->tb_used = skip;
+	buf->tb_used = (size_t) skip;
     }
 }
 
@@ -1472,17 +1616,9 @@ shiftMiniBuffer(int offs)
     static const int slack = 5;
     int shift = w_val(wminip, WVAL_SIDEWAYS);
     int adjust;
-    BUFFER *savebp;
-    WINDOW *savewp;
-    MARK savemk;
+    KBD_DATA save;
 
-    beginDisplay();
-    savebp = curbp;
-    savewp = curwp;
-    savemk = MK;
-
-    curbp = bminip;
-    curwp = wminip;
+    kbd_start(&save);
 
     adjust = offs - (shift + LastMsgCol - 1);
     if (adjust >= 0)
@@ -1490,12 +1626,9 @@ shiftMiniBuffer(int offs)
     else if (shift > 0 && (slack + adjust) < 0)
 	mvleftwind(TRUE, slack - adjust);
 
-    curbp = savebp;
-    curwp = savewp;
-    MK = savemk;
+    kbd_finish(&save);
 
     kbd_flush();
-    endofDisplay();
 }
 
 static int
@@ -1526,22 +1659,15 @@ reallyEditMiniBuffer(TBUFF **buf,
     int edited = FALSE;
     const CMDFUNC *cfp;
     int save_insertmode = insertmode;
-    BUFFER *savebp;
-    WINDOW *savewp;
-    MARK savemk;
+    KBD_DATA save;
 
-    beginDisplay();
-    savebp = curbp;
-    savewp = curwp;
-    curbp = bminip;
-    curwp = wminip;
-    savemk = MK;
+    kbd_start(&save);
     ++no_minimsgs;
 
     TRACE((T_CALLED
-	   "editMiniBuffer(%d:%s) called with c=%#x, miniedit=%d, dot=%d\n",
+	   "editMiniBuffer(%d:%s) called with c=%#x, miniedit=%d, dot=%d, quoted:%d\n",
 	   (int) *cpos, tb_visible(*buf),
-	   c, miniedit, DOT.o));
+	   c, miniedit, DOT.o, quoted));
 
 #if OPT_MULTIBYTE
     /*
@@ -1557,7 +1683,7 @@ reallyEditMiniBuffer(TBUFF **buf,
 	&& ((char2int(c) >= (int) iBIT(MinCBits))
 	    || (char2int(c) >= 128 && b_is_utfXX(bminip)))) {
 	UCHAR temp[10];
-	int used = vl_conv_to_utf8(temp, c, sizeof(temp));
+	int used = vl_conv_to_utf8(temp, (UINT) c, sizeof(temp));
 	int n;
 
 	for (n = 0; n < used; ++n) {
@@ -1578,7 +1704,9 @@ reallyEditMiniBuffer(TBUFF **buf,
 	    C_NUM first = (C_NUM) * cpos + margin;
 	    int old_clexec = clexec;
 	    int old_named = isnamedcmd;
+#if OPT_B_LIMITS
 	    int old_margin = b_left_margin(bminip);
+#endif
 	    REGIONSHAPE old_shape = regionshape;
 
 	    /*
@@ -1619,14 +1747,17 @@ reallyEditMiniBuffer(TBUFF **buf,
 	    if (DOT.o > llength(DOT.l)) {
 		DOT.o = llength(DOT.l);
 	    }
-	    *cpos = DOT.o - margin;
+	    *cpos = (size_t) (DOT.o - margin);
 
 	    /*
 	     * Copy the data back from the minibuffer into our working TBUFF.
 	     */
 	    tb_init(buf, EOS);
-	    if (llength(DOT.l) > margin)
-		tb_bappend(buf, lvalue(DOT.l) + margin, llength(DOT.l) - margin);
+	    if (llength(DOT.l) > margin) {
+		tb_bappend(buf,
+			   lvalue(DOT.l) + margin,
+			   (size_t) (llength(DOT.l) - margin));
+	    }
 
 	    /*
 	     * Below are some workarounds for making it appear that we're doing the
@@ -1663,13 +1794,13 @@ reallyEditMiniBuffer(TBUFF **buf,
 	    miniedit = FALSE;
 	    if (!vl_echo || qpasswd) {
 		char tmp = (char) c;
-		tb_bappend(buf, &tmp, 1);
+		tb_bappend(buf, &tmp, (size_t) 1);
 		if (qpasswd)
 		    show1Char(c);
 	    } else if (llength(lp) >= margin) {
 		show1Char(c);
 		tb_init(buf, EOS);
-		tb_bappend(buf, lvalue(lp) + margin, llength(lp) - margin);
+		tb_bappend(buf, lvalue(lp) + margin, (size_t) (llength(lp) - margin));
 	    }
 	    *cpos += 1;
 	    edited = TRUE;
@@ -1686,11 +1817,9 @@ reallyEditMiniBuffer(TBUFF **buf,
     }
 
     --no_minimsgs;
-    curbp = savebp;
-    curwp = savewp;
-    MK = savemk;
+
+    kbd_finish(&save);
     kbd_flush();
-    endofDisplay();
 
     returnCode(edited);
 }
@@ -1732,7 +1861,13 @@ read_quoted(int count, int inscreen)
 	    c = '0';
 	    digs = 5;		/* 4 plus the fake '0' */
 	    base = 16;
-	    str = "unicode";
+	    str = "unicode (hex)";
+	    limit = 0xffff;
+	} else if (c == 'U') {
+	    c = '0';
+	    digs = 6;		/* 5 plus the fake '0' */
+	    base = 10;
+	    str = "unicode (dec)";
 	    limit = 0xffff;
 	} else
 #endif
@@ -1752,7 +1887,7 @@ read_quoted(int count, int inscreen)
 	}
 	do {
 	    if (isbackspace(c)) {
-		value /= base;
+		value /= (UINT) base;
 		if (--i < 0)
 		    break;
 	    } else {
@@ -1765,7 +1900,7 @@ read_quoted(int count, int inscreen)
 		    delta = '0';
 		else
 		    break;
-		value = value * base + c - delta;
+		value = (value * (UINT) base) + (UINT) (c - delta);
 		i++;
 	    }
 
@@ -1853,6 +1988,7 @@ kbd_reply(const char *prompt,	/* put this out first */
     size_t cpos;		/* current position */
     int status;
     int shell;
+    int range;
     int margin;
     size_t save_len;
     MARK saveMK;		/* FIXME: may be lost by name-completion swbuffer */
@@ -1869,6 +2005,11 @@ kbd_reply(const char *prompt,	/* put this out first */
     WINDOW *old_wp = curwp;
     int old_reading;
 
+    (void) old_wp;
+    (void) old_bname;
+
+    assert(extbuf != 0);
+
     TRACE((T_CALLED "kbd_reply(prompt=%s, extbuf=%s, options=%#x)\n",
 	   TRACE_NULL(prompt),
 	   tb_visible(*extbuf),
@@ -1884,7 +2025,7 @@ kbd_reply(const char *prompt,	/* put this out first */
      * Allowing the user to change the string to something like "ERROR_VALUE"
      * would be confusing and is not really necessary.
      */
-    if ((extbuf != 0 && *extbuf != 0)
+    if (*extbuf != 0
 	&& (isErrorVal(tb_values(*extbuf))
 	    || isTB_ERRS(*extbuf))) {
 	tb_init(extbuf, EOS);
@@ -1909,9 +2050,11 @@ kbd_reply(const char *prompt,	/* put this out first */
 	}
 
 	TRACE(("...getting token from: %s\n", str_visible0(execstr)));
-	execstr = ((options & KBD_REGLUE) != 0 && pushed_back)
-	    ? get_token2(execstr, extbuf, endfunc, eolchar, &actual)
-	    : get_token(execstr, extbuf, endfunc, eolchar, &actual);
+	execstr = (dontmap
+		   ? get_token1(execstr, extbuf, endfunc, eolchar, &actual)
+		   : (((options & KBD_REGLUE) != 0 && pushed_back)
+		      ? get_token2(execstr, extbuf, endfunc, eolchar, &actual)
+		      : get_token(execstr, extbuf, endfunc, eolchar, &actual)));
 	StrToBuff(*extbuf);	/* FIXME: get_token should use TBUFF */
 	TRACE(("...got token, result:  %s\n", tb_visible(*extbuf)));
 	status = (tb_length(*extbuf) != 0);
@@ -2010,11 +2153,11 @@ kbd_reply(const char *prompt,	/* put this out first */
      * field left/right within the margins.
      */
     margin = llength(wminip->w_dot.l);
-    cpos = kbd_show_response(&buf,
-			     tb_values(*extbuf),
-			     tb_length(*extbuf),
-			     eolchar,
-			     options);
+    cpos = (size_t) kbd_show_response(&buf,
+				      tb_values(*extbuf),
+				      tb_length(*extbuf),
+				      eolchar,
+				      options);
     backslashes = 0;		/* by definition, there is an even
 				   number of backslashes */
     c = -1;			/* initialize 'lastch' */
@@ -2093,6 +2236,7 @@ kbd_reply(const char *prompt,	/* put this out first */
 	if (complete != no_completion) {
 	    kbd_unquery();
 	    shell = editingShellCmd(tb_values(buf), options);
+	    range = editingRangeCmd(tb_values(buf));
 	    if (done && (options & KBD_NULLOK) && cpos == 0)
 		/*EMPTY */ ;
 	    else if ((done && (may_complete(buf, options) == ABORT))
@@ -2111,7 +2255,7 @@ kbd_reply(const char *prompt,	/* put this out first */
 			show1Char(tb_values(buf)[cpos++]);
 		    }
 		}
-		if (shell && isreturn(c)) {
+		if ((shell || range) && isreturn(c)) {
 		    /*EMPTY */ ;
 		} else if ((*complete) (options, c, tbreserve(&buf), &cpos)) {
 		    done = TRUE;
@@ -2126,7 +2270,7 @@ kbd_reply(const char *prompt,	/* put this out first */
 					     tbreserve(&buf), &cpos));
 		    }
 		    firstch = FALSE;
-		    shiftMiniBuffer((int) (cpos + margin));
+		    shiftMiniBuffer(((int) cpos + margin));
 		    continue;
 		}
 	    }
@@ -2138,7 +2282,7 @@ kbd_reply(const char *prompt,	/* put this out first */
 		remove_backslashes(buf);	/* take out quoters */
 
 	    save_len = tb_length(buf);
-	    hst_append(buf, eolchar, (options & KBD_EXPCMD));
+	    hst_append(buf, eolchar, (int) (options & KBD_EXPCMD));
 	    (void) tb_copy(extbuf, buf);
 	    status = (tb_length(*extbuf) != 0);
 
@@ -2244,8 +2388,10 @@ kbd_reply(const char *prompt,	/* put this out first */
 		    c = toUpper(c);
 	    }
 #endif
-	    if (!editMiniBuffer(&buf, &cpos, c, margin, EscOrQuo))
+	    if (!editMiniBuffer(&buf, &cpos, c, margin,
+				quotef || (c != editc && (backslashes & 1)))) {
 		continue;	/* keep firstch==TRUE */
+	    }
 	}
 	firstch = FALSE;
     }
@@ -2400,7 +2546,9 @@ get_kbd_macro(TBUFF **rp)
     *rp = tb_init(rp, EOS);
     if ((last = itb_length(KbdMacro)) != 0) {
 	for (n = 0; n < last; n++) {
-	    len = kcod2escape_seq(itb_get(KbdMacro, n), temp, sizeof(temp));
+	    len = (size_t) kcod2escape_seq(itb_get(KbdMacro, n),
+					   temp,
+					   sizeof(temp));
 	    *rp = tb_bappend(rp, temp, len);
 	}
     }
@@ -2504,7 +2652,7 @@ start_kbm(int n,		/* # of times to repeat */
 	  int macnum,		/* register to execute */
 	  ITBUFF * ptr)		/* data to interpret */
 {
-    KSTACK *sp;
+    KSTACK *sp = 0;
     ITBUFF *tp = 0;
 
     if (interrupted())
@@ -2538,6 +2686,7 @@ start_kbm(int n,		/* # of times to repeat */
 	return (itb_init(&dotcmd, esc_c) != 0
 		&& itb_init(&(sp->m_dots), esc_c) != 0);
     }
+    free(sp);
     return FALSE;
 }
 
