@@ -10,7 +10,7 @@
  * editing must be being displayed, which means that "b_nwnd" is non zero,
  * which means that the dot and mark values in the buffer headers are nonsense.
  *
- * $Header: /usr/build/vile/vile/RCS/line.c,v 1.204 2009/05/18 00:13:12 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/line.c,v 1.231 2015/09/06 21:09:26 tom Exp $
  *
  */
 
@@ -21,10 +21,12 @@
 #define poison(p,s)
 #endif
 
-#include	"estruct.h"
-#include	"edef.h"
+#include "estruct.h"
+#include "edef.h"
 
-#define roundlenup(n) ((n+NBLOCK-1) & (UINT)~(NBLOCK-1))
+#include <assert.h>
+
+#define roundlenup(n) (((size_t) (n) + NBLOCK - 1) & (size_t)~(NBLOCK-1))
 
 static int doput(int f, int n, int after, REGIONSHAPE shape);
 static int ldelnewline(void);
@@ -32,7 +34,12 @@ static int PutChar(int n, REGIONSHAPE shape);
 
 #if OPT_SHOW_REGS && OPT_UPBUFF
 static int will_relist_regs;
-static void relist_registers(void);
+static int show_Registers(BUFFER *);
+#define relist_registers() \
+    if (!will_relist_regs) { \
+	will_relist_regs = TRUE; \
+	update_scratch(REGISTERS_BufName, show_Registers); \
+    }
 #else
 #define relist_registers()
 #endif
@@ -232,22 +239,24 @@ lremove(BUFFER *bp, LINE *lp)
 #endif
 #if OPT_VIDEO_ATTRS
     {
-	AREGION *ap = bp->b_attribs;
-	while (ap != NULL) {
-	    AREGION *next = ap->ar_next;
+	AREGION **rpp = &(bp->b_attribs);
+	AREGION *ap;
+
+	while ((ap = *rpp) != NULL) {
+	    AREGION **next = &(ap->ar_next);
 	    int samestart = (ap->ar_region.r_orig.l == lp);
 	    int sameend = (ap->ar_region.r_end.l == lp);
 
 	    if (samestart && sameend) {
-		AREGION *tofree = ap;
-		free_attrib(bp, tofree);
+		free_attrib2(bp, rpp);
+		next = rpp;
 	    } else if (samestart) {
 		ap->ar_region.r_orig = mark_after;
 	    } else if (sameend) {
 		ap->ar_region.r_end.l = lback(lp);
 		ap->ar_region.r_end.o = llength(ap->ar_region.r_end.l);
 	    }
-	    ap = next;
+	    rpp = next;
 	}
     }
 #endif /* OPT_VIDEO_ATTRS */
@@ -263,6 +272,18 @@ lremove(BUFFER *bp, LINE *lp)
     }
 }
 
+/*
+ * Remove a line, and free its memory whether or not the buffer is undoable.
+ */
+void
+lremove2(BUFFER *bp, LINE *lp)
+{
+    lremove(bp, lp);
+    if (b_val(bp, MDUNDOABLE)) {
+	lfree(lp, bp);
+    }
+}
+
 /* insert spaces forward into text */
 int
 insspace(int f, int n)
@@ -374,7 +395,6 @@ lins_bytes(int n, int c)
 {
     char *cp1;
     char *cp2;
-    LINE *tmp;
     LINE *lp1;
     LINE *lp2;
     LINE *lp3;
@@ -388,6 +408,8 @@ lins_bytes(int n, int c)
 
     beginDisplay();
 
+    assert(n >= 0);
+
     lp1 = DOT.l;		/* Current line         */
     if (n == 0) {
 	;
@@ -415,27 +437,29 @@ lins_bytes(int n, int c)
 	}
     } else {
 	doto = DOT.o;		/* Save for later.      */
-	tmp = lp1;
-	nsize = (size_t) (llength(tmp) + n);
-	if (nsize > tmp->l_size) {	/* Hard: reallocate     */
+	nsize = (size_t) (llength(lp1) + 1 + doto + n);
+	if (nsize > lp1->l_size) {	/* Hard: reallocate     */
 	    /* first, create the new image */
-	    nsize = roundlenup((int) nsize);
+	    nsize = roundlenup(nsize);
 	    CopyForUndo(lp1);
 	    if ((ntext = castalloc(char, nsize)) == NULL) {
 		rc = FALSE;
 	    } else {
-		if (lvalue(lp1) && doto)	/* possibly NULL if l_size == 0 */
+		if (lvalue(lp1) && doto) {	/* possibly NULL if l_size == 0 */
+		    assert((size_t) doto < nsize);
 		    (void) memcpy(&ntext[0], &lvalue(lp1)[0], (size_t) doto);
+		}
 		(void) memset(&ntext[doto], c, (size_t) n);
 		if (lvalue(lp1)) {
 #if OPT_LINE_ATTRS
 		    UCHAR *l_attrs = lp1->l_attrs;
 		    lp1->l_attrs = 0;	/* momentarily detach */
 #endif
-		    if (llength(lp1) != doto)
+		    if (llength(lp1) > doto) {
 			(void) memcpy(&ntext[doto + n],
 				      &lvalue(lp1)[doto],
 				      (size_t) (llength(lp1) - doto));
+		    }
 		    ltextfree(lp1, curbp);
 #if OPT_LINE_ATTRS
 		    lp1->l_attrs = l_attrs;	/* reattach */
@@ -444,23 +468,28 @@ lins_bytes(int n, int c)
 		lvalue(lp1) = ntext;
 		lp1->l_size = nsize;
 		llength(lp1) += n;
+		assert((size_t) llength(lp1) <= lp1->l_size);
 	    }
 	} else {		/* Easy: in place       */
 	    CopyForUndo(lp1);
 	    changed = (WFEDIT);
-	    tmp = lp1;
 	    /* don't use memcpy:  overlapping regions.... */
-	    llength(tmp) += n;
-	    if (llength(tmp) - n > doto) {
-		cp2 = &lvalue(tmp)[llength(tmp)];
+	    llength(lp1) += n;
+	    assert((size_t) llength(lp1) <= lp1->l_size);
+	    if (llength(lp1) > (n + doto)) {
+		cp2 = &lvalue(lp1)[llength(lp1)];
 		cp1 = cp2 - n;
-		while (cp1 != &lvalue(tmp)[doto])
+		while (cp1 != &lvalue(lp1)[doto])
 		    *--cp2 = *--cp1;
 	    }
-	    for (i = 0; i < n; ++i)	/* Add the characters       */
-		lvalue(tmp)[doto + i] = (char) c;
+	    assert(llength(lp1) >= (n + doto));
+	    for (i = 0; i < n; ++i) {	/* Add the characters       */
+		lvalue(lp1)[doto + i] = (char) c;
+	    }
 	}
 	if (rc != FALSE) {
+	    int did_wminip = FALSE;
+
 	    changed = (WFEDIT);
 #if ! WINMARK
 	    if (MK.l == lp1) {
@@ -470,8 +499,12 @@ lins_bytes(int n, int c)
 #endif
 	    for_each_window(wp) {	/* Update windows       */
 		after_linsert(wp, lp1, n, doto);
+		if (wp == wminip)
+		    did_wminip = TRUE;
+	    }
+	    if (!did_wminip) {
+		after_linsert(wminip, lp1, n, doto);
 	    }
-	    after_linsert(wminip, lp1, n, doto);
 	    do_mark_iterate(mp, {
 		if (mp->l == lp1) {
 		    if (mp->o > doto)
@@ -508,11 +541,17 @@ lins_chars(int n, int c)
     UCHAR target[10];
     int nbytes;
     int nn;
+    int mapped;
 
     if ((c > 127) && b_is_utfXX(curbp)) {
 	nbytes = vl_conv_to_utf8(target, (UINT) c, sizeof(target));
     } else if (okCTYPE2(vl_wide_enc) && !vl_mb_is_8bit(c)) {
-	nbytes = vl_conv_to_utf8(target, (UINT) c, sizeof(target));
+	nbytes = 1;
+	if (b_is_utfXX(curbp)) {
+	    nbytes = vl_conv_to_utf8(target, (UINT) c, sizeof(target));
+	} else if (vl_ucs_to_8bit(&mapped, c)) {
+	    c = mapped;
+	}
     } else {
 	nbytes = 1;
     }
@@ -538,7 +577,7 @@ lins_chars(int n, int c)
  * current window. The funny ass-backwards way it does things is not a botch;
  * it just makes the last line in the file not a special case. Return TRUE if
  * everything works out and FALSE on error (memory allocation failure). The
- * update of dot and mark is a bit easier then in the above case, because the
+ * update of dot and mark is a bit easier than in the above case, because the
  * split forces more updating.
  */
 int
@@ -696,11 +735,11 @@ ldel_bytes(B_COUNT nbytes, int kflag)
     char *cp2;
     LINE *dotp;
     LINE *nlp;
-    int doto;
+    C_NUM doto;
     B_COUNT uchunk;
     long schunk;
     WINDOW *wp;
-    int i;
+    C_NUM i;
     int status = TRUE;
     B_COUNT len_rs = (B_COUNT) len_record_sep(curbp);
 
@@ -712,9 +751,14 @@ ldel_bytes(B_COUNT nbytes, int kflag)
 	    status = FALSE;
 	    break;
 	}
-	uchunk = (B_COUNT) (llength(dotp) - doto);	/* Size of chunk.  */
-	if (uchunk > nbytes)
+	if (doto > llength(dotp)) {
+	    uchunk = 0;
+	} else {
+	    uchunk = (B_COUNT) (llength(dotp) - doto);	/* Size of chunk.  */
+	}
+	if (uchunk > nbytes) {
 	    uchunk = nbytes;
+	}
 
 	schunk = (long) uchunk;
 	if (schunk < 0) {
@@ -738,7 +782,11 @@ ldel_bytes(B_COUNT nbytes, int kflag)
 		lremove(curbp, nlp);
 		lines_deleted++;
 		TossToUndo(nlp);
-		nbytes -= line_length(nlp);
+		assert((B_COUNT) line_length(nlp) <= nbytes);
+		if (line_length(nlp) > nbytes)
+		    nbytes = 0;
+		else
+		    nbytes -= line_length(nlp);
 		nlp = lforw(dotp);
 	    }
 	    if (status != TRUE)
@@ -770,10 +818,10 @@ ldel_bytes(B_COUNT nbytes, int kflag)
 	}
 	while (cp2 != lvalue(dotp) + llength(dotp))
 	    *cp1++ = *cp2++;
-	llength(dotp) -= schunk;
+	llength(dotp) -= (int) schunk;
 #if ! WINMARK
 	if (MK.l == dotp && MK.o > doto) {
-	    MK.o -= schunk;
+	    MK.o -= (C_NUM) schunk;
 	    if (MK.o < doto)
 		MK.o = doto;
 	}
@@ -781,7 +829,7 @@ ldel_bytes(B_COUNT nbytes, int kflag)
 	for_each_window(wp) {	/* Fix windows          */
 	    if (wp->w_dot.l == dotp
 		&& wp->w_dot.o > doto) {
-		wp->w_dot.o -= schunk;
+		wp->w_dot.o -= (C_NUM) schunk;
 		if (wp->w_dot.o < doto)
 		    wp->w_dot.o = doto;
 	    }
@@ -795,7 +843,7 @@ ldel_bytes(B_COUNT nbytes, int kflag)
 #endif
 	    if (wp->w_lastdot.l == dotp
 		&& wp->w_lastdot.o > doto) {
-		wp->w_lastdot.o -= schunk;
+		wp->w_lastdot.o -= (C_NUM) schunk;
 		if (wp->w_lastdot.o < doto)
 		    wp->w_lastdot.o = doto;
 	    }
@@ -803,17 +851,19 @@ ldel_bytes(B_COUNT nbytes, int kflag)
 	do_mark_iterate(mp, {
 	    if (mp->l == dotp
 		&& mp->o > doto) {
-		mp->o -= schunk;
+		mp->o -= (C_NUM) schunk;
 		if (mp->o < doto)
 		    mp->o = doto;
 	    }
 	});
 #if OPT_LINE_ATTRS
-	if (!lattr_shift(curbp, dotp, doto, -schunk)) {
+	if (!lattr_shift(curbp, dotp, doto, (int) -schunk)) {
 	    status = FALSE;
 	    break;
 	}
 #endif
+	assert(uchunk != 0);
+	assert(uchunk <= nbytes);
 	nbytes -= uchunk;
     }
     return (status);
@@ -827,7 +877,25 @@ ldel_bytes(B_COUNT nbytes, int kflag)
 int
 ldel_chars(B_COUNT nchars, int kflag)
 {
-    return ldel_bytes((B_COUNT) count_bytes(DOT.l, DOT.o, (int) nchars), kflag);
+    LINE *lp = DOT.l;
+    int off = DOT.o;
+    B_COUNT total = 0;
+
+    while (nchars != 0) {
+	int test_bytes = count_bytes(lp, off, (int) nchars);
+	int test_chars = count_chars(lp, off, test_bytes);
+
+	total += (B_COUNT) test_bytes;
+	if ((B_COUNT) test_chars >= nchars)
+	    break;
+
+	lp = lforw(DOT.l);
+	off = 0;
+
+	total += (B_COUNT) len_record_sep(curbp);;
+	nchars -= (B_COUNT) (1 + test_chars);
+    }
+    return ldel_bytes(total, kflag);
 }
 #endif
 
@@ -847,7 +915,7 @@ lrepl_ctype(CHARTYPE type, const char *np, int length)
 
     TRACE((T_CALLED "lrepl_ctype:%s%lx:%.*s\n",
 	   type ? "word" : "line",
-	   (ULONG) type, length, NONNULL(np)));
+	   (ULONG) type, length, np));
 
     if ((status = check_editable(curbp)) == TRUE) {
 
@@ -856,7 +924,7 @@ lrepl_ctype(CHARTYPE type, const char *np, int length)
 	if (type != 0) {	/* it's an exact region */
 	    regionshape = rgn_EXACT;
 	    while (DOT.o < llength(DOT.l)
-		   && istype(type, char_at(DOT))) {
+		   && HasCType(type, CharAtDot())) {
 		if ((status = forwdelchar(FALSE, 1)) != TRUE)
 		    break;
 	    }
@@ -910,7 +978,7 @@ lrepl_regex(REGEXVAL * rexp, const char *np, int length)
 	mayneedundo();
 
 	if (lregexec(exp, DOT.l, DOT.o, llength(DOT.l))) {
-	    int old = exp->endp[0] - exp->startp[0];
+	    int old = (int) (exp->endp[0] - exp->startp[0]);
 	    if (old > 0) {
 		regionshape = rgn_EXACT;
 		status = forwdelchar(TRUE, old);
@@ -1025,6 +1093,7 @@ ldelnewline(void)
     FreeAndNull(lp2->l_attrs);
 #endif
     llength(lp1) += add;
+    assert((size_t) llength(lp1) <= lp1->l_size);
     set_lforw(lp1, lforw(lp2));
     set_lback(lforw(lp2), lp1);
     dumpuline(lp1);
@@ -1090,7 +1159,7 @@ kdone(void)
 	kregwidth = 0;
 	kcharpending = -1;
     }
-    kregflag &= ~KNEEDCLEAN;
+    kregflag = (USHORT) (kregflag & ~KNEEDCLEAN);
     kbs[ukb].kbflag = kregflag;
     relist_registers();
 }
@@ -1207,7 +1276,7 @@ reg2index(int c)
     else if (isLower(c))
 	n = c - 'a' + 10;	/* named buffs are in 10 through 36 */
     else if (isUpper(c))
-	n = c - 'A' + 10;
+	n = c - 'A' + 10;	/* ignore case of register-name */
 #if OPT_SELECTIONS
     else if (c == SEL_KCHR)
 	n = SEL_KREG;
@@ -1459,7 +1528,7 @@ doput(int f, int n, int after, REGIONSHAPE shape)
     if (shape == rgn_FULLLINE) {
 	if (after && !is_header_line(DOT, curbp))
 	    DOT.l = lforw(DOT.l);
-	DOT.o = 0;
+	DOT.o = b_left_margin(curbp);
     } else {
 	if (after && !is_at_end_of_line(DOT))
 	    forwchar(TRUE, 1);
@@ -1739,7 +1808,7 @@ execkreg(int f, int n)
 	    i = (int) KbSize(jj, tkp);
 	    sp = (char *) tkp->d_chunk + i - 1;
 	    while (i--) {
-		mapungetc((int) ((*sp--) | YESREMAP));
+		mapungetc((int) ((UINT) (*sp--) | YESREMAP));
 	    }
 	    kbcount--;
 	}
@@ -1757,7 +1826,7 @@ loadkreg(int f, int n GCC_UNUSED)
     ksetup();
     *respbuf = EOS;
     s = mlreply_no_opts("Load register with: ",
-			respbuf, sizeof(respbuf));
+			respbuf, (UINT) sizeof(respbuf));
     if (s != TRUE)
 	return FALSE;
     if (f)
@@ -1855,29 +1924,84 @@ showkreg(int f, int n GCC_UNUSED)
 }
 
 #if OPT_UPBUFF
-
 static int
 show_Registers(BUFFER *bp)
 {
     b_clr_obsolete(bp);
     return showkreg(show_all_chars, 1);
 }
-
-static void
-relist_registers(void)
-{
-
-    if (will_relist_regs)	/* have we already done this? */
-	return;
-
-    will_relist_regs = TRUE;
-
-    update_scratch(REGISTERS_BufName, show_Registers);
-}
 #endif /* OPT_UPBUFF */
 
 #endif /* OPT_SHOW_REGS */
 
+#if OPT_REGS_CMPL
+KBD_OPTIONS
+regs_kbd_options(void)
+{
+    return KBD_MAYBEC;
+}
+
+static int
+reg_has_data(int src)
+{
+    int rc = FALSE;
+    int j = reg2index(src);
+    if (j >= 0 ||
+	(regs_kbd_default && (src == UNAME_KCHR))) {
+	int jj = index2ukb(j);
+	if (jj >= 0 && jj < NKREGS) {
+	    if (kbs[jj].kbufh) {
+		rc = TRUE;
+	    }
+	}
+    }
+    return rc;
+}
+
+/*
+ * Register names are a single ASCII character.
+ */
+static const char **
+init_regs_cmpl(char *buf, size_t cpos)
+{
+    int dst, src;
+    char **result = typeallocn(char *, 96);
+    for (dst = 0, src = 32; src < 127; ++src) {
+	if (isUpper(src))	/* register names are caseless */
+	    continue;
+	if (reg_has_data(src)) {
+	    if (cpos == 0 || *buf == src) {
+		char value[2];
+		value[0] = (char) src;
+		value[1] = 0;
+		result[dst++] = strmalloc(value);
+	    }
+	}
+    }
+    result[dst] = 0;
+    return (const char **) result;
+}
+
+int
+regs_completion(DONE_ARGS)
+{
+    size_t cpos = *pos;
+    int status = FALSE;
+    const char **nptr;
+
+    kbd_init();			/* nothing to erase */
+    buf[cpos] = EOS;		/* terminate it for us */
+
+    beginDisplay();
+    if ((nptr = init_regs_cmpl(buf, cpos)) != 0) {
+	status = kbd_complete(PASS_DONE_ARGS, (const char *) nptr, sizeof(*nptr));
+	free(TYPECAST(char *, nptr));
+    }
+    endofDisplay();
+    return status;
+}
+#endif
+
 /* For memory-leak testing (only!), releases all kill-buffer storage. */
 #if NO_LEAKS
 void
