@@ -17,12 +17,12 @@
  * distributable status.  This version of vile is distributed under the
  * terms of the GNU Public License (see COPYING).
  *
- * Copyright (c) 1992-2008 by Paul Fox and Thomas Dickey
+ * Copyright (c) 1992-2015,2016 by Paul Fox and Thomas Dickey
  *
  */
 
 /*
- * $Header: /usr/build/vile/vile/RCS/main.c,v 1.664 2009/05/24 12:48:26 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/main.c,v 1.734 2016/11/07 00:30:15 bod Exp $
  */
 
 #define realdef			/* Make global definitions not external */
@@ -51,6 +51,10 @@
 #include <processes.h>
 #endif
 
+#if OPT_FILTER
+#include <filters.h>
+#endif
+
 #if OPT_PERL
 #if defined(DECL_ENVIRON)
 extern char **environ;
@@ -72,10 +76,15 @@ static int cmd_mouse_motion(const CMDFUNC * cfp);
 static void get_executable_dir(void);
 static void global_val_init(void);
 static void main_loop(void);
-static void make_startup_file(char *name);
+static void make_startup_file(const char *name);
 static void siginit(int enabled);
 
-extern const int nametblsize;
+#if OPT_MULTIBYTE
+static void pre_init_ctype(void);
+#endif
+
+extern const int nametbl_size;
+extern const int glbstbl_size;
 
 /*--------------------------------------------------------------------------*/
 
@@ -119,9 +128,19 @@ restore_buffer_state(BUFFER *save_bp, UINT save_flags)
     }
 }
 
-static int
-run_startup_commands(BUFFER *cmds)
+static void
+hide_and_discard(BUFFER **bp)
 {
+    if (zotwp(*bp)) {
+	zotbuf(*bp);
+	*bp = 0;
+    }
+}
+
+static int
+run_startup_commands(BUFFER **bp)
+{
+    BUFFER *cmds = *bp;
     int result = TRUE;
     BUFFER *save_bp = NULL;	/* saves curbp when doing startup commands */
     LINE *lp;
@@ -136,10 +155,7 @@ run_startup_commands(BUFFER *cmds)
 	LINE *nlp = lforw(lp);
 
 	if (lisreal(lp) && !llength(lp)) {
-	    lremove(cmds, lp);
-	    if (b_val(cmds, MDUNDOABLE)) {
-		lfree(lp, cmds);
-	    }
+	    lremove2(cmds, lp);
 	}
 	lp = nlp;
     }
@@ -165,7 +181,7 @@ run_startup_commands(BUFFER *cmds)
     }
     if (result) {		/* remove the now unneeded buffer */
 	b_set_scratch(cmds);	/* make sure it will go */
-	(void) zotbuf(cmds);
+	hide_and_discard(bp);
     }
 
     returnCode(result);
@@ -174,17 +190,30 @@ run_startup_commands(BUFFER *cmds)
 static int
 run_and_discard(BUFFER **bp)
 {
-    int rc = run_startup_commands(*bp);
-    zotbuf(*bp);
-    *bp = 0;
+    int rc = run_startup_commands(bp);
+
+    hide_and_discard(bp);
     return rc;
 }
 
+static void
+add_cmdarg(BUFFER *bp, const char *cmd, const char *arg)
+{
+    TBUFF *tb = 0;
+    if (tb_scopy(&tb, arg) != 0
+	&& tb_enquote(&tb) != 0) {
+	b2printf(bp, cmd, tb_values(tb));
+    }
+    tb_free(&tb);
+}
+
 static void
 setup_command(BUFFER *opts_bp, char *param)
 {
     char *p1;
     char *p2;
+    const CMDFUNC *cfp;
+
     /*
      * Check for special cases where a leading number should be treated
      * as a repeat-count (see docmd(), which does something similar).
@@ -198,7 +227,24 @@ setup_command(BUFFER *opts_bp, char *param)
 	    param = skip_blanks(p2 + 1);
 	}
     }
-    b2printf(opts_bp, "execute-named-command %s\n", param);
+
+    /*
+     * As a special case, quote simple commands such as forward and reverse
+     * search.
+     */
+    cfp = DefaultKeyBinding(CharOf(*param));
+    if (isPunct(CharOf(*param))
+	&& ((cfp == &f_forwsearch)
+	    || (cfp == &f_backsearch)
+#if OPT_SHELL
+	    || (cfp == &f_capturecmd)
+	    || (cfp == &f_operfilter)
+#endif
+	)) {
+	add_cmdarg(opts_bp, "execute-named-command %s\n", param);
+    } else {
+	b2printf(opts_bp, "execute-named-command %s\n", param);
+    }
 }
 
 static int
@@ -216,7 +262,7 @@ default_fill(void)
 static char *
 get_set_locale(const char *value)
 {
-    char *result = setlocale(LC_ALL, value);
+    char *result = setlocale(LC_CTYPE, value);
     if (result != 0)
 	result = strmalloc(result);
     return result;
@@ -235,14 +281,74 @@ set_posix_locale(void)
 #endif
 #endif /* OPT_LOCALE */
 
+/*
+ * For command-line use, run the appropriate syntax filter on the current
+ * buffer, writing the attributed text to the standard output.  This assumes
+ * that if filters are built-in, that the appropriate one is available.
+ */
+static void
+filter_to_stdio(FILE *fp)
+{
+    static char *my_macro;
+    int s;
+
+    TRACE(("SyntaxFilter-only:\n"));
+    TRACE(("\tcurfname:%s\n", curbp->b_fname));
+
+#if OPT_FILTER && OPT_MAJORMODE
+    if (curbp->majr != 0 && flt_lookup(curbp->majr->shortname)) {
+	const CMDFUNC *cmd = &f_operattrdirect;
+	clexec = TRUE;
+	filter_only = TRUE;
+	gotobob(FALSE, 0);
+	havemotion = &f_gotoeob;
+	(void) CMD_U_FUNC(cmd) (FALSE, 0);
+    } else
+#endif
+    {
+	/*
+	 * Rely upon the macro to construct the proper filter path.
+	 */
+	if (my_macro == 0)
+	    my_macro = strmalloc("HighlightFilter");
+
+	/*
+	 * If we used a built-in filter, its output is written as a side effect
+	 * of calling flt_putc, etc.  Otherwise we have to write it ourselves
+	 * now.
+	 */
+	s = docmd(my_macro, TRUE, FALSE, 0);
+	if (s == TRUE) {
+	    int nlines;
+	    B_COUNT nchars;
+	    REGION region;
+
+	    gotoeob(FALSE, 1);
+	    setmark();
+	    gotobob(FALSE, 1);
+
+	    regionshape = rgn_FULLLINE;
+	    if (getregion(curbp, &region)) {
+
+		ffp = fp;
+		ffstatus = file_is_pipe;
+
+		write_region(curbp, &region, TRUE, &nlines, &nchars);
+	    }
+	}
+    }
+}
+
 /*--------------------------------------------------------------------------*/
 
 int
 MainProgram(int argc, char *argv[])
 {
+    static char dft_vileinit[] = "vileinit.rc";
     int tt_opened;
     BUFFER *bp;
     int carg;			/* current arg to scan */
+    int literal = FALSE;	/* force args to be interpreted as filenames */
     char *vileinit = NULL;	/* the startup file or VILEINIT var */
     int startstat = TRUE;	/* result of running startup */
     BUFFER *havebp = NULL;	/* initial buffer to read */
@@ -264,9 +370,16 @@ MainProgram(int argc, char *argv[])
     *startkey = EOS;
 #endif
 
+    /*
+     * This is needed for a fallback case in locales.
+     */
+#if OPT_MULTIBYTE
+    pre_init_ctype();
+#endif
+
 #if OPT_LOCALE
     {
-	char *env = "";
+	const char *env = "";
 	char *old_locale = get_set_locale("");
 	char *old_encoding = 0;
 
@@ -286,16 +399,10 @@ MainProgram(int argc, char *argv[])
 	    && (((env = getenv("LC_ALL")) != 0 && *env != 0) ||
 		((env = getenv("LC_CTYPE")) != 0 && *env != 0) ||
 		((env = getenv("LANG")) != 0 && *env != 0))) {
-	    char *utf;
 	    char *tmp;
 
 	    TRACE(("Checking for UTF-8 suffix of '%s'\n", env));
-	    if (((utf = strstr(env, ".UTF-8")) != 0
-		 || (utf = strstr(env, ".utf-8")) != 0
-		 || (utf = strstr(env, ".UTF8")) != 0
-		 || (utf = strstr(env, ".utf8")) != 0)
-		&& (tmp = strmalloc(env)) != 0) {
-		tmp[utf - env] = EOS;
+	    if ((tmp = vl_narrowed(env)) != 0) {
 		vl_init_8bit(env, tmp);
 		env = tmp;
 	    } else {
@@ -331,7 +438,7 @@ MainProgram(int argc, char *argv[])
 	 * corresponding encoding.
 	 */
 	if (!okCTYPE2(vl_real_enc)) {
-	    vl_real_enc.encoding = 0;
+	    vl_get_encoding(&vl_real_enc.encoding, env);
 	} else {
 	    vl_get_encoding(&vl_real_enc.encoding, vl_real_enc.locale);
 	}
@@ -339,7 +446,9 @@ MainProgram(int argc, char *argv[])
 	if (!okCTYPE2(vl_real_enc)
 	    || isEmpty(vl_real_enc.encoding)) {
 	    TRACE(("...checking original locale '%s'\n", NonNull(old_locale)));
-	    if (vl_is_utf8_encoding(old_encoding)) {
+	    if (vl_8bit_builtin()) {
+		TRACE(("using built-in locale data\n"));
+	    } else if (vl_is_utf8_encoding(old_encoding)) {
 		TRACE(("original encoding is UTF-8\n"));
 		vl_narrow_enc.locale = 0;
 	    } else {
@@ -367,7 +476,8 @@ MainProgram(int argc, char *argv[])
 #endif
 
 #if OPT_NAMEBST
-    build_namebst(nametbl, 0, nametblsize - 1);
+    build_namebst(nametbl, 0, nametbl_size - 1, 0);
+    build_namebst(glbstbl, 0, glbstbl_size - 1, GLOBOK);
 #endif
     vl_ctype_init(global_g_val(GVAL_PRINT_LOW),
 		  global_g_val(GVAL_PRINT_HIGH));
@@ -393,8 +503,8 @@ MainProgram(int argc, char *argv[])
      * or setgid'd.
      */
 #if defined(HAVE_SETUID) && defined(HAVE_SETGID) && defined(HAVE_GETEGID) && defined(HAVE_GETEUID)
-    setgid(getegid());
-    setuid(geteuid());
+    IGNORE_RC(setgid(getegid()));
+    IGNORE_RC(setuid(geteuid()));
 #endif
 
     get_executable_dir();
@@ -418,7 +528,15 @@ MainProgram(int argc, char *argv[])
      */
     siginit(TRUE);
 #if OPT_DUMBTERM
-    if (isatty(fileno(stdin))
+    /* try to avoid staircasing if -F option goes directly to terminal */
+    for (carg = 1; carg < argc; ++carg) {
+	if (!strcmp(argv[carg], "-F")) {
+	    filter_only = -TRUE;
+	    break;
+	}
+    }
+    if (!filter_only
+	&& isatty(fileno(stdin))
 	&& isatty(fileno(stdout))) {
 	tt_opened = open_terminal(&dumb_term);
     } else
@@ -439,8 +557,13 @@ MainProgram(int argc, char *argv[])
 	char *param = argv[carg];
 
 	/* evaluate switches */
-	if (*param == '-') {
+	if (*param == '-' && !literal) {
 	    ++param;
+	    /* all arguments following -- are interpreted as filenames */
+	    if (!strcmp(param, "-")) {
+		literal = TRUE;
+		continue;
+	    }
 #if DISP_BORLAND || SYS_VMS
 	    /* if it's a digit, it's probably a screen
 	       resolution */
@@ -491,9 +614,12 @@ MainProgram(int argc, char *argv[])
 		case 'E':
 		    b2printf(opts_bp, "set noview\n");
 		    break;
+		case 'F':
+		    filter_only = -TRUE;
+		    break;
 		case 'g':	/* -g for initial goto */
 		case 'G':
-		    b2printf(opts_bp, "%s goto-line\n", GetArgVal(param));
+		    add_cmdarg(opts_bp, "%s goto-line\n", GetArgVal(param));
 		    break;
 		case 'h':	/* -h for initial help */
 		case 'H':
@@ -502,7 +628,7 @@ MainProgram(int argc, char *argv[])
 
 		case 'i':
 		case 'I':
-		    vileinit = "vileinit.rc";
+		    vileinit = dft_vileinit;
 		    /*
 		     * If the user has no startup file, make a simple
 		     * one that points to this one.
@@ -510,7 +636,7 @@ MainProgram(int argc, char *argv[])
 		    if (cfg_locate(vileinit, LOCATE_SOURCE) != 0
 			&& cfg_locate(startup_file, LOCATE_SOURCE) == 0)
 			make_startup_file(vileinit);
-		    b2printf(init_bp, "source %s\n", vileinit);
+		    add_cmdarg(init_bp, "source %s\n", vileinit);
 		    break;
 
 #if OPT_ENCRYPT
@@ -530,14 +656,20 @@ MainProgram(int argc, char *argv[])
 #endif
 		case 's':	/* -s <pattern> */
 		case 'S':
-		    b2printf(opts_bp, "search-forward %s\n", GetArgVal(param));
+		    add_cmdarg(opts_bp, "search-forward %s\n", GetArgVal(param));
 		    break;
 #if OPT_TAGS
 		case 't':	/* -t for initial tag lookup */
 		case 'T':
-		    b2printf(opts_bp, "tag %s\n", GetArgVal(param));
+		    add_cmdarg(opts_bp, "tag %s\n", GetArgVal(param));
 		    break;
 #endif
+		case 'U':
+		    set_global_b_val(MDDOS, system_crlf = TRUE);
+		    break;
+		case 'u':
+		    set_global_b_val(MDDOS, system_crlf = FALSE);
+		    break;
 		case 'v':	/* -v is view mode */
 		    b2printf(opts_bp, "set view\n");
 		    break;
@@ -560,12 +692,11 @@ MainProgram(int argc, char *argv[])
 		default:	/* unknown switch */
 		    print_usage(GOODEXIT);
 		}
-
-	} else if (*param == '+') {	/* alternate form of -g */
+	} else if (*param == '+' && !literal) {		/* alternate form of -g */
 	    setup_command(opts_bp, GetArgVal(param));
-	} else if (*param == '@') {
+	} else if (*param == '@' && !literal) {
 	    vileinit = ++param;
-	    b2printf(init_bp, "source %s\n", param);
+	    add_cmdarg(init_bp, "source %s\n", param);
 	} else if (*param != EOS) {
 
 	    /* must be a filename */
@@ -593,7 +724,7 @@ MainProgram(int argc, char *argv[])
      * in this case).
      */
 #if DISP_TERMCAP || DISP_CURSES || DISP_ANSI
-    if (!isatty(fileno(stdout))) {
+    if (!filter_only && !isatty(fileno(stdout))) {
 	fprintf(stderr, "vile: ex mode is not implemented\n");
 	ExitProgram(BADEXIT);
     }
@@ -625,21 +756,22 @@ MainProgram(int argc, char *argv[])
 #if !SYS_WINNT
 #if !DISP_X11
 #if SYS_UNIX
-	char *tty = 0;
+	const char *tty = 0;
 #else
 	FILE *in;
-	int fd;
 #endif /* SYS_UNIX */
 #endif /* DISP_X11 */
 #endif /* !SYS_WINNT */
 	BUFFER *lastbp = havebp;
 	int nline = 0;
+	int fd;
 
 	bp = bfind(STDIN_BufName, BFARGS);
 	make_current(bp);	/* pull it to the front */
 	if (!havebp)
 	    havebp = bp;
-	ffp = fdopen(dup(fileno(stdin)), "r");
+	fd = dup(fileno(stdin));
+	ffp = (fd >= 0) ? fdopen(fd, "r") : 0;
 #if !DISP_X11
 # if SYS_UNIX
 # if defined(HAVE_TTYNAME)
@@ -695,7 +827,7 @@ MainProgram(int argc, char *argv[])
 # else
 #  if SYS_VMS
 	fd = open("tt:", O_RDONLY, S_IREAD);	/* or sys$command */
-#  else				/* e.g., DOS-based systems */
+#  else	/* e.g., DOS-based systems */
 	fd = fileno(stderr);	/* this normally cannot be redirected */
 #  endif
 	if ((fd >= 0)
@@ -704,15 +836,14 @@ MainProgram(int argc, char *argv[])
 	    && (in = fdopen(fd, "r")) != 0) {
 	    *stdin = *in;	/* FIXME: won't work on opaque FILE's */
 	}
-#  endif			/* SYS_WINNT */
-# endif				/* SYS_UNIX */
+#  endif /* SYS_WINNT */
+# endif	/* SYS_UNIX */
 #endif /* DISP_X11 */
 
 #if OPT_ENCRYPT
 	if (*startkey != EOS) {
-	    strcpy(bp->b_cryptkey, startkey);
-	    make_local_b_val(bp, MDCRYPT);
-	    set_b_val(bp, MDCRYPT, TRUE);
+	    vl_strncpy(bp->b_cryptkey, startkey, sizeof(bp->b_cryptkey));
+	    set_local_b_val(bp, MDCRYPT, TRUE);
 	}
 #endif
 	if (ffp != 0) {
@@ -733,30 +864,36 @@ MainProgram(int argc, char *argv[])
     }
 #endif
 
-    if (!tt_opened)
-	siginit(TRUE);
-    (void) open_terminal((TERM *) 0);
-    term.kopen();		/* open the keyboard */
-    term.rev(FALSE);
-
-    /*
-     * The terminal driver sets default colors and $palette during the
-     * open_terminal function.  Save that information as the 'default'
-     * color scheme.  We have to do this before reading .vilerc
-     */
+    if (filter_only) {
 #if OPT_COLOR_SCHEMES
-    init_scheme();
+	init_scheme();
+#endif
+    } else {
+	if (!tt_opened)
+	    siginit(TRUE);
+	(void) open_terminal((TERM *) 0);
+	term.kopen();		/* open the keyboard */
+	term.rev(FALSE);
+
+	/*
+	 * The terminal driver sets default colors and $palette during the
+	 * open_terminal function.  Save that information as the 'default'
+	 * color scheme.  We have to do this before reading .vilerc
+	 */
+#if OPT_COLOR_SCHEMES
+	init_scheme();
 #endif
 
-    if (vtinit() != TRUE)	/* allocate display memory */
-	tidy_exit(BADEXIT);
+	if (vtinit() != TRUE)	/* allocate display memory */
+	    tidy_exit(BADEXIT);
 
-    winit(TRUE);		/* windows */
+	winit(TRUE);		/* windows */
 
 #if OPT_MENUS
-    if (delay_menus)
-	vlmenu_load(FALSE, 1);
+	if (delay_menus)
+	    vlmenu_load(FALSE, 1);
 #endif
+    }
 
     set_global_b_val(VAL_FILL, default_fill());
 
@@ -774,8 +911,7 @@ MainProgram(int argc, char *argv[])
 #if OPT_DOSFILES
     /* an empty non-existent buffer defaults to line-style
        favored by the OS */
-    make_local_b_val(bp, MDDOS);
-    set_b_val(bp, MDDOS, CRLF_LINES);
+    set_local_b_val(bp, MDDOS, system_crlf);
 #endif
     fix_cmode(bp, FALSE);
     swbuffer(bp);
@@ -810,10 +946,10 @@ MainProgram(int argc, char *argv[])
 		if (!startstat)
 		    goto begin;
 	    } else {
-		zotbuf(init_bp);
+		init_bp = zotbuf2(init_bp);
 	    }
 	} else {		/* find and run .vilerc */
-	    zotbuf(init_bp);
+	    init_bp = zotbuf2(init_bp);
 	    if (do_source(startup_file, 1, TRUE) != TRUE) {
 		startstat = FALSE;
 		goto begin;
@@ -866,7 +1002,7 @@ MainProgram(int argc, char *argv[])
     purge_msgs();
 #endif
     if (startstat == TRUE)	/* else there's probably an error message */
-	mlforce(msg);
+	mlforce("%s", msg);
 
   begin:
 #if SYS_VMS
@@ -878,48 +1014,81 @@ MainProgram(int argc, char *argv[])
 	(void) term.setdescrip(init_descrip);
     }
 #endif
-    (void) update(FALSE);
+    if (filter_only) {
+#ifndef filter_to_stdio
+
+	/*
+	 * Process files in the order they appear on the command-line.
+	 */
+	set_global_g_val(GMDABUFF, FALSE);
+	sortlistbuffers();
+	firstbuffer(FALSE, 0);
+
+	do {
+	    filter_only = TRUE;
+	    filter_to_stdio(stdout);
+	    filter_only = FALSE;	/* ...so nextbuffer doesn't write stdout */
+	    fflush(stdout);
+	} while (nextbuffer(FALSE, 0) == TRUE);
+	ExitProgram(GOODEXIT);
+#else
+	fprintf(stderr,
+		"The -F option is not supported in this configuration\n");
+#endif
+
+    } else {
+	(void) update(FALSE);
 
 #if DISP_NTWIN
-    /* Draw winvile's main window, for the very first time. */
-    winvile_start();
+	/* Draw winvile's main window, for the very first time. */
+	winvile_start();
 #endif
 
 #if OPT_POPUP_MSGS
-    if (global_g_val(GMDPOPUP_MSGS) && (startstat != TRUE)) {
-	bp = bfind(MESSAGES_BufName, BFINVS);
-	bsizes(bp);
-	TRACE(("Checking size of popup messages: %d\n", bp->b_linecount));
-	if (bp->b_linecount > 1) {
-	    int save = global_g_val(GMDPOPUP_MSGS);
-	    set_global_g_val(GMDPOPUP_MSGS, TRUE);
-	    popup_msgs();
-	    tb_init(&mlsave, EOS);
-	    set_global_g_val(GMDPOPUP_MSGS, save);
+	if (global_g_val(GMDPOPUP_MSGS) && (startstat != TRUE)) {
+	    bp = bfind(MESSAGES_BufName, BFINVS);
+	    bsizes(bp);
+	    TRACE(("Checking size of popup messages: %d\n", bp->b_linecount));
+	    if (bp->b_linecount > 1) {
+		int save = global_g_val(GMDPOPUP_MSGS);
+		set_global_g_val(GMDPOPUP_MSGS, TRUE);
+		popup_msgs();
+		tb_init(&mlsave, EOS);
+		set_global_g_val(GMDPOPUP_MSGS, save);
+	    }
 	}
-    }
-    if (global_g_val(GMDPOPUP_MSGS) == -TRUE)
-	set_global_g_val(GMDPOPUP_MSGS, FALSE);
+	if (global_g_val(GMDPOPUP_MSGS) == -TRUE)
+	    set_global_g_val(GMDPOPUP_MSGS, FALSE);
 #endif
 
 #ifdef GMDCD_ON_OPEN
-    /* reset a one-shot, e.g., for winvile's drag/drop code */
-    if (global_g_val(GMDCD_ON_OPEN) < 0) {
-	set_directory_from_file(curbp);
-	set_global_g_val(GMDCD_ON_OPEN, FALSE);
-    }
+	/* reset a one-shot, e.g., for winvile's drag/drop code */
+	if (global_g_val(GMDCD_ON_OPEN) < 0) {
+	    set_directory_from_file(curbp);
+	    set_global_g_val(GMDCD_ON_OPEN, FALSE);
+	}
 #endif
 
-    /* We won't always be able to show messages before the screen is
-     * initialized.  Give it one last chance.
-     */
-    if ((startstat != TRUE) && tb_length(mlsave))
-	mlforce("%.*s", (int) tb_length(mlsave), tb_values(mlsave));
+	/*
+	 * These should be gone, but may have persisted to this point if
+	 * there was some problem with the scripting.  Try again.
+	 */
+	hide_and_discard(&init_bp);
+	hide_and_discard(&opts_bp);
 
-    /* process commands */
-    main_loop();
+	/*
+	 * Force the last message, if any, onto the status line.
+	 */
+	if (tb_length(mlsave)) {
+	    tb_append(&mlsave, EOS);
+	    mlforce("%.*s", (int) tb_length(mlsave), tb_values(mlsave));
+	}
 
-    /* NOTREACHED */
+	/* process commands */
+	main_loop();
+
+	/* NOTREACHED */
+    }
     return BADEXIT;
 }
 
@@ -964,6 +1133,12 @@ main_loop(void)
 	kbd_mac_check();
 	c = kbd_seq();
 
+	/*
+	 * Now that we have started a command, reset "$_".  If we did this at a
+	 * lower level, we could not test for it in macros, etc.
+	 */
+	tb_scopy(&last_macro_result, status2s(TRUE));
+
 	/* reset the contents of the command/status line */
 	if (kbd_length() > 0) {
 	    mlerase();
@@ -1066,19 +1241,20 @@ get_executable_dir(void)
     if (last_slash(prog_arg) == NULL) {
 	/* If there are no slashes, we can guess where we came from,
 	 */
-	if ((s = cfg_locate(prog_arg, FL_PATH | FL_EXECABLE)) != 0)
+	if ((s = cfg_locate(prog_arg, FL_PATH | FL_EXECABLE | FL_INSECURE))
+	    != 0)
 	    s = strmalloc(s);
     } else {
 	/* if there _are_ slashes, then argv[0] was either
 	 * absolute or relative. lengthen_path figures it out.
 	 */
-	s = strmalloc(lengthen_path(strcpy(temp, prog_arg)));
+	s = strmalloc(lengthen_path(vl_strncpy(temp, prog_arg, sizeof(temp))));
     }
     if (s == 0)
 	return;
 
     t = pathleaf(s);
-    if (t != s) {
+    if (t != s && t != 0) {
 	/*
 	 * On a unix host, 't' points past slash.  On a VMS host,
 	 * 't' points to first char after the last ':' or ']' in
@@ -1087,10 +1263,10 @@ get_executable_dir(void)
 	free(prog_arg);
 	prog_arg = strmalloc(t);
 	*t = EOS;
-	exec_pathname = strmalloc(lengthen_path(strcpy(temp, s)));
+	exec_pathname = strmalloc(lengthen_path(vl_strncpy(temp, s, sizeof(temp))));
 #ifndef VILE_STARTUP_PATH
 	/* workaround for DJGPP, to add executable's directory to startup path */
-	if ((s = vile_getenv("VILE_STARTUP_PATH")) == 0) {
+	if (vile_getenv("VILE_STARTUP_PATH") == 0) {
 	    append_to_path_list(&startup_path, exec_pathname);
 	}
 #endif
@@ -1104,7 +1280,7 @@ get_executable_dir(void)
     HKEY hkey;
     char buffer[256];
 
-    for (j = 0; j < TABLESIZE(rootkeys); ++j) {
+    for (j = 0; j < (int) TABLESIZE(rootkeys); ++j) {
 	if (RegOpenKeyEx(rootkeys[j],
 			 VILE_SUBKEY,
 			 0,
@@ -1137,6 +1313,9 @@ tidy_exit(int code)
 #if DISP_NTWIN
     winvile_cleanup();
 #endif
+#if DISP_NTCONS
+    term.close();
+#endif
 #if DISP_X11
     term.close();		/* need this if $xshell left subprocesses */
 #endif
@@ -1208,6 +1387,36 @@ no_memory(const char *s)
 
 #define DFT_IDENTIFIER_EXPR    "\\w\\+"
 
+#if OPT_MULTIBYTE
+/* default value assumes POSIX locale, since it is used before vl_ctyle init */
+#define DFT_LATIN1_EXPR	"^\\(\
+aa\\|\
+af\\|\
+br\\|\
+de\\|\
+en\\|\
+es\\|\
+fr\\|\
+fi\\|\
+ga\\|\
+gb\\|\
+gl\\|\
+gv\\|\
+id\\|\
+it\\|\
+kl\\|\
+ms\\|\
+nl\\|\
+nn\\|\
+om\\|\
+pt\\|\
+so\\|\
+sv\\|\
+tl\\)\
+_[A-Za-z]\\+$\
+"
+#endif
+
 			/* where do paragraphs start? */
 #define DFT_PARAGRAPHS  "^\\.[ILPQ]P\\>\\|^\\.P\\>\\|\
 ^\\.LI\\>\\|^\\.[plinb]p\\>\\|^\\.\\?\\s*$"
@@ -1285,6 +1494,7 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 	    setINT(GMDALTTABPOS, FALSE);	/* emacs-style tab positioning */
 	    setINT(GMDERRORBELLS, TRUE);	/* alarms are noticeable */
 	    setINT(GMDEXPAND_PATH, FALSE);
+	    setINT(GMDFILENAME_IC, DFT_FILE_IC);
 	    setINT(GMDIMPLYBUFF, FALSE);	/* imply-buffer */
 	    setINT(GMDINSEXEC, FALSE);	/* allow ^F/^B, etc., to be interpreted during insert mode */
 	    setINT(GMDMAPLONGER, FALSE);	/* favor longer maps */
@@ -1296,7 +1506,6 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 	    setINT(GMDRONLYVIEW, FALSE);	/* Set view-on-readonly */
 	    setINT(GMDSAMEBANGS, FALSE);	/* use same "!!" data for ^X-! */
 	    setINT(GMDSMOOTH_SCROLL, FALSE);
-	    setINT(GMDSPACESENT, TRUE);		/* add two spaces after each sentence */
 	    setINT(GMDUNIQ_BUFS, FALSE);	/* only one buffer per file */
 	    setINT(GMDWARNRENAME, TRUE);	/* warn before renaming a buffer */
 	    setINT(GMDWARNREREAD, TRUE);	/* warn before rereading a buffer */
@@ -1377,9 +1586,6 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 #ifdef GMDXTERM_TITLE
 	    setINT(GMDXTERM_TITLE, FALSE);	/* xterm window-title */
 #endif
-#ifdef GVAL_BACKUPSTYLE
-	    setTXT(GVAL_BACKUPSTYLE, DFT_BACKUPSTYLE);
-#endif
 #ifdef GVAL_BCOLOR
 	    setINT(GVAL_BCOLOR, C_BLACK);	/* background color */
 #endif
@@ -1410,6 +1616,9 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 #ifdef GVAL_ICURSOR
 	    setTXT(GVAL_ICURSOR, "0");	/* no insertion cursor */
 #endif
+#ifdef GVAL_KEEP_POS
+	    setINT(GVAL_KEEP_POS, KPOS_VILE);
+#endif
 #ifdef GVAL_MCOLOR
 	    setINT(GVAL_MCOLOR, VAREV);		/* show in reverse */
 #endif
@@ -1450,7 +1659,7 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 	    setINT(MDAUTOWRITE, FALSE);		/* auto-write */
 	    setINT(MDBACKLIMIT, TRUE);	/* limit backspacing to insert point */
 	    setINT(MDCINDENT, FALSE);	/* C-style indent */
-	    setINT(MDDOS, CRLF_LINES);	/* on by default on DOS, off others */
+	    setINT(MDDOS, system_crlf);
 	    setINT(MDIGNCASE, FALSE);	/* exact matches */
 	    setINT(MDINS_RECTS, FALSE);		/* insert-mode for rectangles */
 	    setINT(MDLOADING, FALSE);	/* asynchronously loading a file */
@@ -1460,7 +1669,7 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 	    setINT(MDREADONLY, FALSE);	/* readonly */
 	    setINT(MDSHOWMAT, FALSE);	/* show-match */
 	    setINT(MDSHOWMODE, TRUE);	/* show-mode */
-	    setINT(MDSWRAP, TRUE);	/* scan wrap */
+	    setINT(MDSPACESENT, TRUE);	/* add two spaces after each sentence */
 	    setINT(MDTABINSERT, TRUE);	/* allow tab insertion */
 	    setINT(MDTAGSRELTIV, FALSE);	/* path relative tag lookups */
 	    setINT(MDTAGWORD, FALSE);	/* tag entire word */
@@ -1468,7 +1677,8 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 	    setINT(MDUNDOABLE, TRUE);	/* undo stack active */
 	    setINT(MDUNDO_DOS_TRIM, FALSE);	/* undo dos trimming */
 	    setINT(MDVIEW, FALSE);	/* view-only */
-	    setINT(MDWRAP, FALSE);	/* wrap */
+	    setINT(MDWRAPSCAN, TRUE);	/* scan wrap */
+	    setINT(MDWRAPWORDS, FALSE);		/* wrap */
 	    setINT(MDYANKMOTION, TRUE);		/* yank-motion */
 	    setINT(VAL_ASAVECNT, 256);	/* autosave count */
 	    setINT(VAL_FILL, default_fill());	/* column for paragraph reformat */
@@ -1524,6 +1734,9 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 #ifdef MDMODELINE
 	    setINT(MDMODELINE, FALSE);
 #endif
+#ifdef MDREUSE_POS
+	    setINT(MDREUSE_POS, FALSE);		/* reuse position */
+#endif
 #ifdef MDTAGIGNORECASE
 	    setINT(MDTAGIGNORECASE, FALSE);	/* ignore case in tags */
 #endif
@@ -1533,6 +1746,9 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 #ifdef VAL_AUTOCOLOR
 	    setINT(VAL_AUTOCOLOR, 0);	/* auto syntax coloring timeout */
 #endif
+#ifdef VAL_BACKUPSTYLE
+	    setTXT(VAL_BACKUPSTYLE, DFT_BACKUPSTYLE);
+#endif
 #ifdef VAL_BYTEORDER_MARK
 	    setINT(VAL_BYTEORDER_MARK, ENUM_UNKNOWN);	/* byteorder-mark NONE */
 #endif
@@ -1601,7 +1817,8 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
 	    setINT(WMDNONPRINTOCTAL, FALSE);	/* unprintable-as-octal */
 	    setINT(WVAL_SIDEWAYS, 0);	/* list-mode */
 #ifdef WMDLINEWRAP
-	    setINT(WMDLINEWRAP, FALSE);		/* line-wrap */
+	    setINT(WMDLINEBREAK, FALSE);	/* linebreak */
+	    setINT(WMDLINEWRAP, FALSE);		/* linewrap */
 #endif
 #ifdef WMDRULER
 	    setINT(WMDRULER, FALSE);	/* ruler */
@@ -1648,19 +1865,19 @@ init_mode_value(struct VAL *d, MODECLASS v_class, int v_which)
  Char %{$curchar}d of %{$bchars}d\
  (%P%%) char is 0x%{$char}x or 0%{$char}o"
 
-static char *
+static const char *
 default_help_file(void)
 {
-    char *result;
+    const char *result;
     if ((result = vile_getenv("VILE_HELP_FILE")) == 0)
 	result = DFT_HELP_FILE;
     return result;
 }
 
-static char *
+static const char *
 default_libdir_path(void)
 {
-    char *result;
+    const char *result;
     if ((result = vile_getenv("VILE_LIBDIR_PATH")) == 0)
 	result = DFT_LIBDIR_PATH;
     return result;
@@ -1687,10 +1904,10 @@ default_menu_file(void)
 }
 #endif
 
-static char *
+static const char *
 default_startup_file(void)
 {
-    char *result;
+    const char *result;
     if ((result = vile_getenv("VILE_STARTUP_FILE")) == 0)
 	result = DFT_STARTUP_FILE;
     return result;
@@ -1728,6 +1945,14 @@ default_startup_path(void)
     return result;
 }
 
+#if OPT_MULTIBYTE
+static void
+pre_init_ctype(void)
+{
+    tb_scopy(&latin1_expr, DFT_LATIN1_EXPR);
+}
+#endif
+
 #if OPT_EVAL
 /*
  * Returns a string representing the default/initial value of the given
@@ -1745,9 +1970,17 @@ init_state_value(int which)
 	value = DFT_FILENAME_EXPR;
 	break;
 #endif
+    case VAR_FONT:
+	value = default_font;
+	break;
     case VAR_HELPFILE:
 	value = default_help_file();
 	break;
+#if OPT_MULTIBYTE
+    case VAR_LATIN1_EXPR:
+	value = DFT_LATIN1_EXPR;
+	break;
+#endif
     case VAR_LIBDIR_PATH:
 	value = default_libdir_path();
 	break;
@@ -1778,6 +2011,11 @@ init_state_value(int which)
     case VAR_STARTUP_PATH:
 	result = default_startup_path();
 	break;
+#if !SMALLER
+    case VAR_EMPTY_LINES:
+	value = "1";
+	break;
+#endif
     }
     return (value != 0) ? strmalloc(value) : result;
 }
@@ -1826,6 +2064,49 @@ copy_kbindtbl(BINDINGS * dst)
 extern void flt_array(void);
 #endif
 
+/*
+ * Initialize compiled-in state variable values.
+ */
+static void
+state_val_init(void)
+{
+#if OPT_EVAL
+    int i;
+
+    /*
+     * Note that not all state variables can be initialized explicitly, many
+     * are either read-only, or have too many side-effects to set blindly.
+     */
+    for (i = 0; i < Num_StateVars; i++) {
+	char *s;
+	if ((s = init_state_value(i)) != 0) {
+	    set_state_variable(statevars[i], s);
+	    free(s);
+	}
+    }
+#else
+    helpfile = default_help_file();
+    startup_file = strmalloc(default_startup_file());
+    startup_path = default_startup_path();
+    libdir_path = default_libdir_path();
+#if OPT_MENUS
+    menu_file = default_menu_file();
+#endif
+#if OPT_MLFORMAT
+    modeline_format = strmalloc(DFT_MLFORMAT);
+#endif
+#if OPT_MULTIBYTE
+    tb_scopy(&latin1_expr, DFT_LATIN1_EXPR);
+#endif
+#if OPT_POSFORMAT
+    position_format = strmalloc(DFT_POSFORMAT);
+#endif
+#if OPT_FINDERR
+    tb_scopy(&filename_expr, DFT_FILENAME_EXPR);
+#endif
+#endif
+}
+
 static void
 global_val_init(void)
 {
@@ -1854,6 +2135,7 @@ global_val_init(void)
 	make_local_val(global_w_values.wv, i);
 	init_mode_value(&global_w_values.wv[i], WIN_MODE, i);
     }
+    state_val_init();
 
 #if OPT_MAJORMODE
     /*
@@ -1873,37 +2155,6 @@ global_val_init(void)
 #endif
 #endif /* OPT_MAJORMODE */
 
-#if OPT_EVAL
-    /*
-     * Note that not all state variables can be initialized explicitly, many
-     * are either read-only, or have too many side-effects to set blindly.
-     */
-    for (i = 0; i < Num_StateVars; i++) {
-	char *s;
-	if ((s = init_state_value(i)) != 0) {
-	    set_state_variable(statevars[i], s);
-	    free(s);
-	}
-    }
-#else
-    helpfile = default_help_file();
-    startup_file = strmalloc(default_startup_file());
-    startup_path = default_startup_path();
-    libdir_path = default_libdir_path();
-#if OPT_MENUS
-    menu_file = default_menu_file();
-#endif
-#if OPT_MLFORMAT
-    modeline_format = strmalloc(DFT_MLFORMAT);
-#endif
-#if OPT_POSFORMAT
-    position_format = strmalloc(DFT_POSFORMAT);
-#endif
-#if OPT_FINDERR
-    filename_expr = strmalloc(DFT_FILENAME_EXPR);
-#endif
-#endif
-
     /*
      * Initialize the "normal" bindings for insert and command mode to
      * the motion-commands that are safe if insert-exec mode is enabled.
@@ -2079,7 +2330,7 @@ siginit(int enabled)
 	void *ptrfunc = wat_crit_handler;
 	_harderr(ptrfunc);
     }
-#  else				/* CC_TURBO */
+#  else	/* CC_TURBO */
     _harderr(dos_crit_handler);
 #  endif
 # endif
@@ -2109,7 +2360,7 @@ do_num_proc(int *cp, int *fp, int *np)
 
     c = *cp;
 
-    if (isCntrl(c) || isSpecial(c))
+    if (isCntrl(c) || !is8Bits(c))
 	return;
 
     f = *fp;
@@ -2121,7 +2372,7 @@ do_num_proc(int *cp, int *fp, int *np)
 	n = 0;			/* start with a zero default */
 	f = TRUE;		/* there is a # arg */
 	mflag = 1;		/* current minus flag */
-	while ((isDigit(c) && !isSpecial(c)) || (c == '-')) {
+	while ((isDigit(c) && is8Bits(c)) || (c == '-')) {
 	    if (c == '-') {
 		/* already hit a minus or digit? */
 		if ((mflag == -1) || (n != 0))
@@ -2165,8 +2416,10 @@ do_rept_arg_proc(int *cp, int *fp, int *np)
     f = TRUE;			/* there is a # arg */
     mflag = 0;			/* that can be discarded. */
     mlwrite("arg: %d", n);
-    while ((isDigit(c = kbd_seq()) && !isSpecial(c))
-	   || c == reptc || c == '-') {
+    for (;;) {
+	c = kbd_seq();
+	if (!(isDigit(c) || (c == reptc) || (c == '-')))
+	    break;
 	if (c == reptc)
 	    n = n * 4;
 
@@ -2593,13 +2846,13 @@ showmemory(int f, int n)
 #endif /* SYS_MSDOS */
 
 char *
-strncpy0(char *t, const char *f, size_t l)
+strncpy0(char *dest, const char *src, size_t destlen)
 {
-    if (t != 0 && f != 0 && l != 0) {
-	(void) strncpy(t, f, l);
-	t[l - 1] = EOS;
+    if (dest != 0 && src != 0 && destlen != 0) {
+	(void) strncpy(dest, src, destlen);
+	dest[destlen - 1] = EOS;
     }
-    return t;
+    return dest;
 }
 
 /*
@@ -2616,6 +2869,19 @@ vl_strncpy(char *dest, const char *src, size_t destlen)
     return strncpy0(dest, src, srclen);
 }
 
+char *
+vl_strncat(char *dest, const char *src, size_t destlen)
+{
+    size_t srclen = (src != 0) ? (strlen(src) + 1) : 0;
+    size_t oldlen = strlen(dest);
+
+    if (srclen > (destlen - oldlen))
+	srclen = (destlen - oldlen);
+
+    (void) strncpy0(dest + oldlen, src, srclen);
+    return dest;
+}
+
 /*
  * Wrapper for environment variables that tell vile where to find its
  * components.
@@ -2633,7 +2899,7 @@ vile_getenv(const char *name)
 	HKEY hkey;
 	char buffer[256];
 
-	for (j = 0; j < TABLESIZE(rootkeys); ++j) {
+	for (j = 0; j < (int) TABLESIZE(rootkeys); ++j) {
 	    if (RegOpenKeyEx(rootkeys[j],
 			     VILE_SUBKEY W32_STRING("\\Environment"),
 			     0,
@@ -2727,9 +2993,9 @@ newprocessgroup(int f GCC_UNUSED, int n GCC_UNUSED)
     (void) setpgrp(0, 0);
 #   else
     (void) setpgrp();
-#   endif			/* HAVE_BSD_SETPGRP */
-#  endif			/* HAVE_SETSID */
-# endif				/* SYS_VMS */
+#   endif /* HAVE_BSD_SETPGRP */
+#  endif /* HAVE_SETSID */
+# endif	/* SYS_VMS */
 #endif /* DISP_X11 */
     return TRUE;
 }
@@ -2749,7 +3015,7 @@ cmd_mouse_motion(const CMDFUNC * cfp)
 }
 
 static void
-make_startup_file(char *name)
+make_startup_file(const char *name)
 {
     FILE *fp;
     char temp[NFILEN];
@@ -2796,7 +3062,7 @@ int
 valid_window(WINDOW *test)
 {
     beginDisplay();
-    if (test != NULL && test != wminip) {
+    if (test != NULL && test != wminip && test != wnullp) {
 	WINDOW *wp;
 	int valid = FALSE;
 
@@ -2954,9 +3220,11 @@ free_all_leaks(void)
     vt_leaks();
 
     /* whatever is left over must be a leak */
+#if OPT_TRACE
     show_alloc();
     show_elapsed();
     trace_leaks();
+#endif
 #ifdef HAVE__NC_FREEALL
     _nc_freeall();
 #endif
