@@ -2,7 +2,7 @@
  *		The routines in this file handle the conversion of pathname
  *		strings.
  *
- * $Header: /usr/build/vile/vile/RCS/path.c,v 1.166 2009/05/26 21:25:37 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/path.c,v 1.179 2016/07/16 14:33:07 tom Exp $
  *
  *
  */
@@ -301,13 +301,13 @@ pathcat(char *dst, const char *path, const char *cleaf)
     if (cleaf == 0)
 	cleaf = "";
 
-    leaf = vl_strncpy(save_leaf, cleaf, NFILEN);	/* leaf may be in dst */
+    leaf = vl_strncpy(save_leaf, cleaf, (size_t) NFILEN);	/* leaf may be in dst */
 
     if (isEmpty(path)) {
 	(void) strcpy(dst, leaf);
     } else {
 
-	path = vl_strncpy(save_path, path, NFILEN);	/* path may be in dst */
+	path = vl_strncpy(save_path, path, (size_t) NFILEN);	/* path may be in dst */
 
 #if OPT_MSDOS_PATH
 	if (is_msdos_drive(save_path) != 0
@@ -323,7 +323,7 @@ pathcat(char *dst, const char *path, const char *cleaf)
 #endif
 	if (is_abs_pathname(leaf)) {
 	    (void) strcpy(dst, leaf);
-	} else {
+	} else if (leaf != 0) {
 	    have = strlen(strcpy(dst, path));
 	    if (have + strlen(leaf) + 2 < NFILEN) {
 		s = dst + have - 1;
@@ -389,7 +389,7 @@ save_user(const char *name, const char *path)
 
     if (name != NULL
 	&& path != NULL
-	&& (q = typealloc(UPATH)) != NULL) {
+	&& (q = typecalloc(UPATH)) != NULL) {
 	TRACE(("save_user(name=%s, path=%s)\n", name, path));
 	if ((q->name = strmalloc(name)) != NULL
 	    && (q->path = strmalloc(path)) != NULL) {
@@ -431,7 +431,7 @@ find_user(const char *name)
 	     * so lookup the current uid, and then lookup
 	     * the name based on that.
 	     */
-	    p = getpwuid((int) getuid());
+	    p = getpwuid((uid_t) getuid());
 	}
 
 	/* if either of the above lookups worked
@@ -883,12 +883,10 @@ resolve_directory(char *path_name, char **file_namep)
 #endif /* defined(GMDRESOLVE_LINKS) */
 #endif /* defined(HAVE_REALPATH) */
 
-#if OPT_CASELESS
-
 /*
  * The function case_correct_path is intended to determine the true
  * case of all pathname components of a syntactically canonicalized
- * pathname for operating systems on which OPT_CASELESS applies.
+ * pathname for operating systems which use caseless filenames.
  */
 
 #if SYS_WINNT
@@ -983,9 +981,7 @@ case_correct_path(char *old_file, char *new_file)
     return;
 }
 
-#else /* !SYS_WINNT */
-
-#if SYS_OS2
+#elif SYS_OS2
 
 int
 is_case_preserving(const char *name)
@@ -1083,7 +1079,7 @@ case_correct_path(char *old_file, char *new_file)
     return;
 }
 
-#else /* !SYS_OS2 */
+#else
 
 static void
 case_correct_path(char *old_file, char *new_file)
@@ -1093,12 +1089,8 @@ case_correct_path(char *old_file, char *new_file)
     return;
 }
 
-#endif /* !SYS_OS2 */
-
 #endif /* !SYS_WINNT */
 
-#endif /* OPT_CASELESS */
-
 /* canonicalize a pathname, to eliminate extraneous /./, /../, and ////
 	sequences.  only guaranteed to work for absolute pathnames */
 static char *
@@ -1118,9 +1110,8 @@ canonpath(char *ss)
 	    returnString(s);
 
 #if OPT_MSDOS_PATH
-#if !OPT_CASELESS
-	(void) mklower(ss);	/* MS-DOS is case-independent */
-#endif
+	if (!global_g_val(GMDFILENAME_IC))
+	    (void) mklower(ss);	/* MS-DOS is case-independent */
 	if (is_slashc(*ss))
 	    *ss = SLASHC;
 	/* pretend the drive designator isn't there */
@@ -1168,8 +1159,9 @@ canonpath(char *ss)
 
 	    if (real != 0) {
 		(void) strcpy(s, real);
-	    } else if ((leaf = pathleaf(strcpy(temp, s))) != 0) {
-		strcpy(temp2, leaf);
+	    } else if ((leaf = pathleaf(vl_strncpy(temp, s,
+						   sizeof(temp)))) != 0) {
+		vl_strncpy(temp2, leaf, sizeof(temp2));
 		if (leaf == temp + 1)
 		    leaf[0] = EOS;
 		else
@@ -1194,6 +1186,12 @@ canonpath(char *ss)
 
 	    p++;
 	    pp++;		/* leave the leading slash */
+#if OPT_UNC_PATH
+	    if (is_slashc(*pp)) {
+		p++;
+		pp++;
+	    }
+#endif
 	    while (*pp) {
 		if (is_slashc(*pp)) {
 		    pp++;
@@ -1288,9 +1286,8 @@ canonpath(char *ss)
 	}
 #endif
 
-#if OPT_CASELESS
-	case_correct_path(ss, ss);
-#endif
+	if (global_g_val(GMDFILENAME_IC))
+	    case_correct_path(ss, ss);
     }
     returnString(ss);
 }
@@ -1408,10 +1405,10 @@ shorten_path(char *path, int keep_cwd)
 	    *temp = EOS;
 	if (is_slashc(*ff)) {
 	    found = TRUE;
-	    (void) strcpy(path, strcat(temp, ff + 1));
+	    (void) strcpy(path, vl_strncat(temp, ff + 1, sizeof(temp)));
 	} else if (slp == ff - 1) {
 	    found = TRUE;
-	    (void) strcpy(path, strcat(temp, ff));
+	    (void) strcpy(path, vl_strncat(temp, ff, sizeof(temp)));
 	}
     }
 
@@ -1420,13 +1417,17 @@ shorten_path(char *path, int keep_cwd)
 	if (slp != path) {
 	    /* if we mismatched in the last component of cwd, then the file
 	       is under '..' */
-	    if (last_slash(cwd) == 0)
-		(void) strcpy(path, strcat(strcpy(temp, ".."), slp));
+	    if (last_slash(cwd) == 0) {
+		(void) strcpy(path,
+			      vl_strncat(vl_strncpy(temp, "..", sizeof(temp)),
+					 slp,
+					 sizeof(temp)));
+	    }
 	}
     }
 
     /* we're off by more than just '..', so use absolute path */
-# endif				/* SYS_UNIX || SYS_MSDOS */
+# endif	/* SYS_UNIX || SYS_MSDOS */
 #endif /* OPT_VMS_PATH */
 
     TRACE(("     -> '%s' shorten\n", path));
@@ -1579,7 +1580,7 @@ lengthen_path(char *path)
 #if OPT_UNC_PATH
 	if (drive == EOS) {
 	    W32_CHAR uncdir[NFILEN];
-	    GetCurrentDirectory(sizeof(uncdir), uncdir);
+	    GetCurrentDirectory(sizeof(uncdir) / sizeof(uncdir[0]), uncdir);
 	    cwd = bsl_to_sl_inplace(asc_charstring(uncdir));
 	    free_cwd = TRUE;
 	} else
@@ -1603,12 +1604,12 @@ lengthen_path(char *path)
 #if OPT_VMS_PATH
 	vms2unix_path(temp, cwd);
 #else
-	(void) strcpy(temp, cwd);
+	(void) vl_strncpy(temp, cwd, sizeof(temp));
 #endif
 	len = (int) strlen(temp);
 	temp[len++] = SLASHC;
-	(void) vl_strncpy(temp + len, f, sizeof(temp) - len);
-	(void) strcpy(path, temp);
+	(void) vl_strncpy(temp + len, f, sizeof(temp) - (size_t) len);
+	(void) vl_strncpy(path, temp, NFILEN);
     }
 #if OPT_MSDOS_PATH
     if (is_msdos_drive(path) == 0) {	/* ensure that we have drive too */
@@ -1617,7 +1618,7 @@ lengthen_path(char *path)
 	    temp[0] = (char) curdrive();
 	    temp[1] = ':';
 	    (void) vl_strncpy(temp + 2, path, sizeof(temp) - 2);
-	    (void) strcpy(path, temp);
+	    (void) vl_strncpy(path, temp, NFILEN);
 	}
     }
     if (free_cwd)
@@ -1635,32 +1636,36 @@ lengthen_path(char *path)
 int
 is_abs_pathname(char *path)
 {
+    int result = FALSE;
     char *f;
-    if ((f = is_appendname(path)) != 0)
-	return is_pathname(f);
+
+    if (path == 0)
+	result = FALSE;
+    else if ((f = is_appendname(path)) != 0)
+	result = is_pathname(f);
 
 #if OPT_VMS_PATH
-    if (is_vms_pathname(path, -TRUE)
-	&& (strchr(path, L_BLOCK) != 0
-	    || strchr(path, ':') != 0))
-	return TRUE;
+    else if (is_vms_pathname(path, -TRUE)
+	     && (strchr(path, L_BLOCK) != 0
+		 || strchr(path, ':') != 0))
+	result = TRUE;
 #endif
 
 #if OPT_MSDOS_PATH
-    if ((f = is_msdos_drive(path)) != 0)
-	return is_abs_pathname(f);
+    else if ((f = is_msdos_drive(path)) != 0)
+	result = is_abs_pathname(f);
 #endif
 
 #if SYS_UNIX || OPT_MSDOS_PATH || SYS_VMS
 #if SYS_UNIX
-    if (path[0] == CH_TILDE)
-	return TRUE;
+    else if (path[0] == CH_TILDE)
+	result = TRUE;
 #endif
-    if (is_slashc(path[0]))
-	return TRUE;
+    else if (is_slashc(path[0]))
+	result = TRUE;
 #endif /* SYS_UNIX || OPT_MSDOS_PATH || SYS_VMS */
 
-    return FALSE;
+    return result;
 }
 
 /*
@@ -1865,24 +1870,32 @@ is_nonfile(char *path)
  * more entries can be parsed.
  */
 const char *
-parse_pathlist(const char *list, char *result)
+parse_pathlist(const char *list, char *result, int *first)
 {
+    TRACE(("parse_pathlist(%s)\n", TRACE_NULL(list)));
     if (list != NULL && *list != EOS) {
 	int len = 0;
 
-	while (*list && (*list != vl_pathchr)) {
-	    if (len < NFILEN - 1)
-		result[len++] = *list;
-	    list++;
-	}
-	if (len == 0)		/* avoid returning an empty-string */
+	if (first && *first && *list == vl_pathchr) {
 	    result[len++] = '.';
+	} else {
+	    if (*list == vl_pathchr)
+		++list;
+	    while (*list && (*list != vl_pathchr)) {
+		if (len < NFILEN - 1)
+		    result[len++] = *list;
+		list++;
+	    }
+	    if (len == 0)	/* avoid returning an empty-string */
+		result[len++] = '.';
+	}
 	result[len] = EOS;
-
-	if (*list == vl_pathchr)
-	    ++list;
-    } else
+    } else {
 	list = NULL;
+    }
+    TRACE(("...parse_pathlist(%s) ->'%s'\n", TRACE_NULL(list), result));
+    if (first)
+	*first = 0;
     return list;
 }
 
@@ -2103,16 +2116,12 @@ find_in_path_list(const char *path_list, char *path)
     char temp[NFILEN];
     int found = FALSE;
     char find[NFILEN];
+    int first = TRUE;
 
     vl_strncpy(find, SL_TO_BSL(path), sizeof(find));
     TRACE(("find_in_path_list\n\t%s\n\t%s\n", TRACE_NULL(path_list), find));
-    while ((path_list = parse_pathlist(path_list, temp)) != 0) {
-#if OPT_CASELESS
-	if (!stricmp(temp, find))
-#else
-	if (!strcmp(temp, find))
-#endif
-	{
+    while ((path_list = parse_pathlist(path_list, temp, &first)) != 0) {
+	if (!cs_strcmp(global_g_val(GMDFILENAME_IC), temp, find)) {
 	    found = TRUE;
 	    break;
 	}
@@ -2159,41 +2168,46 @@ prepend_to_path_list(char **path_list, char *path)
 #endif
 
 /*
- * Append the given path to a path-list
+ * Append the given path(s) to a path-list
  */
 void
 append_to_path_list(char **path_list, const char *path)
 {
     char *s, *t;
     size_t need;
+    char working[NFILEN];
     char find[NFILEN];
+    const char *in_work = working;
 #if SYS_UNIX
     struct stat sb;
 #endif
+    int first = TRUE;
 
-    vl_strncpy(find, SL_TO_BSL(path), sizeof(find));
-    if (!find_in_path_list(*path_list, find)
+    vl_strncpy(working, SL_TO_BSL(path), sizeof(working));
+    while ((in_work = parse_pathlist(in_work, find, &first)) != 0) {
+	if (!find_in_path_list(*path_list, find)
 #if SYS_UNIX
-	&& file_stat(find, &sb) == 0
+	    && file_stat(find, &sb) == 0
 #endif
-	) {
-	need = strlen(find) + 2;
-	if ((t = *path_list) != 0) {
-	    need += strlen(t);
-	} else {
-	    t = empty_string;
-	}
-	if ((s = typeallocn(char, need)) != 0) {
-	    if (*t != EOS)
-		lsprintf(s, "%s%c%s", t, vl_pathchr, find);
-	    else
-		strcpy(s, find);
-	    if (*path_list != 0)
-		free(*path_list);
-	    *path_list = s;
+	    ) {
+	    need = strlen(find) + 2;
+	    if ((t = *path_list) != 0) {
+		need += strlen(t);
+	    } else {
+		t = empty_string;
+	    }
+	    if ((s = typeallocn(char, need)) != 0) {
+		if (*t != EOS)
+		    lsprintf(s, "%s%c%s", t, vl_pathchr, find);
+		else
+		    strcpy(s, find);
+		if (*path_list != 0)
+		    free(*path_list);
+		*path_list = s;
+	    }
 	}
     }
-    TRACE(("append_to_path_list\n\t%s\n\t%s\n", *path_list, find));
+    TRACE(("...append_to_path_list\n\t%s\n", TRACE_NULL(*path_list)));
 }
 
 /*
@@ -2240,7 +2254,7 @@ path_trunc(char *path, int max_path_len, char *trunc_buf, int trunc_buf_len)
 	max_len -= sizeof(">>>") - 1;
 #else
 	strcpy(trunc_buf, "...");
-	max_len -= sizeof("...") - 1;
+	max_len -= (int) sizeof("...") - 1;
 #endif
 	strcat(trunc_buf, path + path_len - max_len);
     }
@@ -2263,8 +2277,10 @@ append_libdir_to_path(void)
     if (libdir_path != 0
 	&& (tmp = getenv("PATH")) != 0
 	&& (env = strmalloc(tmp)) != 0) {
+	int first = TRUE;
+
 	cp = libdir_path;
-	while ((cp = parse_pathlist(cp, buf)) != 0) {
+	while ((cp = parse_pathlist(cp, buf, &first)) != 0) {
 	    append_to_path_list(&env, buf);
 	}
 	if (strcmp(tmp, env)) {
@@ -2273,11 +2289,11 @@ append_libdir_to_path(void)
 		putenv(tmp);
 		TRACE(("putenv %s\n", tmp));
 	    }
-	} else {
-	    free(env);
 	}
+	free(env);
     }
 }
+
 #else
 void
 append_libdir_to_path(void)
