@@ -2,7 +2,7 @@
  * This file contains the command processing functions for a number of random
  * commands. There is no functional grouping here, for sure.
  *
- * $Header: /usr/build/vile/vile/RCS/random.c,v 1.330 2009/05/26 21:25:37 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/random.c,v 1.357 2016/07/16 14:28:05 tom Exp $
  *
  */
 
@@ -39,11 +39,6 @@
 #include <lib$routines.h>
 #endif
 
-#ifdef WIN32
-#include "w32vile.h"		/* for VL_ELAPSED */
-#undef WIN32_LEAN_AND_MEAN
-#endif
-
 #define DIRS_FORCE    0		/* force DirStack buffer on screen if dirs_idx > 0.
 				 * If DirStack on screen and dirs_idx == 0, kill
 				 * buffer (if possible).
@@ -92,21 +87,18 @@ set_rdonly(BUFFER *bp, const char *name, int mode)
     b_clr_changed(bp);		/* assumes text is loaded... */
     bp->b_active = TRUE;
 
-    make_local_b_val(bp, mode);
-    set_b_val(bp, mode, TRUE);
-
-    make_local_b_val(bp, VAL_TAB);
-    set_b_val(bp, VAL_TAB, 8);
-
-    make_local_b_val(bp, MDDOS);
-    set_b_val(bp, MDDOS, CRLF_LINES);
+    set_local_b_val(bp, mode, TRUE);
+    set_local_b_val(bp, VAL_TAB, 8);
+    set_local_b_val(bp, MDDOS, system_crlf);
 
     fix_cmode(bp, FALSE);
 }
 
-/* generic "lister", which takes care of popping a window/buffer pair under
-	the given name, and calling "func" with a couple of args to fill in
-	the buffer */
+/*
+ * This is a generic "lister", which takes care of popping a window/buffer pair
+ * under the given name, and calling "func" with a couple of args to fill in
+ * the buffer.
+ */
 int
 liststuff(const char *name,
 	  int appendit,
@@ -114,11 +106,14 @@ liststuff(const char *name,
 	  int iarg,
 	  void *vargp)
 {
+    int rc = FALSE;
     BUFFER *bp;
-    int s;
     WINDOW *wp;
     int alreadypopped;
     BUFFER *ocurbp = curbp;
+    L_NUM save_line = -1;
+    C_NUM save_off = -1;
+    int save_top = -1;
 
     TRACE((T_CALLED
 	   "liststuff(name=%s, appendit=%d, func=%p, iarg=%d, vargp=%p)\n",
@@ -126,44 +121,70 @@ liststuff(const char *name,
 
     /* create the buffer list buffer   */
     bp = bfind(name, BFSCRTCH);
-    if (bp == NULL)
-	returnCode(FALSE);
+    if (bp != NULL) {
 
-    if (!appendit && (s = bclear(bp)) != TRUE)	/* clear old text (?) */
-	returnCode(s);
-    b_set_scratch(bp);
-    alreadypopped = (bp->b_nwnd != 0);
-    if (popupbuff(bp) == FALSE) {
-	(void) zotbuf(bp);
-	returnCode(FALSE);
-    }
+#ifdef MDREUSE_POS
+	if ((appendit <= 0)
+	    && b_val(bp, MDREUSE_POS)
+	    && (wp = bp2any_wp(bp)) != NULL) {
+	    save_line = line_no(bp, wp->w_dot.l);
+	    save_off = wp->w_dot.o;
+	    save_top = line_no(bp, wp->w_line.l);
+	}
+#endif
 
-    if ((wp = bp2any_wp(bp)) != NULL) {
-	make_local_w_val(wp, WMDNUMBER);
-	set_w_val(wp, WMDNUMBER, FALSE);
-    }
-    if (appendit) {
-	(void) gotoeob(FALSE, 1);
-	MK = DOT;
-    }
-    /* call the passed in function, giving it both the integer and
-       character pointer arguments */
-    (*func) (iarg, vargp);
-    if (alreadypopped && appendit) {
-	(void) gomark(FALSE, 1);
-	(void) forwbline(FALSE, 1);
-	(void) reposition(FALSE, 1);
-    } else {			/* if we're not appending, go to the top */
-	(void) gotobob(FALSE, 1);
-    }
-    set_rdonly(bp, non_filename(), MDVIEW);
+	/* clear old text (?) */
+	if ((appendit > 0) || (rc = bclear(bp)) == TRUE) {
 
-    if (alreadypopped)		/* don't switch to the popup if it wasn't there */
-	swbuffer(ocurbp);
-    else
-	shrinkwrap();		/* only resize if it's fresh */
+	    b_set_scratch(bp);
+	    alreadypopped = (bp->b_nwnd != 0);
+	    if (popupbuff(bp) == FALSE) {
+		(void) zotbuf(bp);
+	    } else {
 
-    returnCode(TRUE);
+		if ((wp = bp2any_wp(bp)) != NULL) {
+		    make_local_w_val(wp, WMDNUMBER);
+		    set_w_val(wp, WMDNUMBER, FALSE);
+		}
+		if (appendit > 0) {
+		    (void) gotoeob(FALSE, 1);
+		    MK = DOT;
+		}
+		/*
+		 * call the passed in function, giving it both the integer and
+		 * character pointer arguments
+		 */
+		(*func) (iarg, vargp);
+		if (alreadypopped && (appendit > 0)) {
+		    (void) gomark(FALSE, 1);
+		    (void) forwbline(FALSE, 1);
+		    (void) reposition(FALSE, 1);
+		} else if (wp != 0 && (appendit < 0 && save_line >= 0)) {
+		    (void) vl_gotoline(save_top);
+		    wp->w_line.l = wp->w_dot.l;
+		    wp->w_line.o = 0;
+		    if (save_line != save_top)
+			(void) vl_gotoline(save_line);
+		    gocol(save_off);
+		} else {
+		    /* if we're not appending, go to the top */
+		    (void) gotobob(FALSE, 1);
+		}
+		set_rdonly(bp, non_filename(), MDVIEW);
+
+		if (alreadypopped) {
+		    /* don't switch to the popup if it wasn't there */
+		    swbuffer(ocurbp);
+		} else {
+		    /* only resize if it's fresh */
+		    shrinkwrap();
+		}
+
+		rc = TRUE;
+	    }
+	}
+    }
+    returnCode(rc);
 }
 
 /*
@@ -214,7 +235,7 @@ showcpos(int f GCC_UNUSED, int n GCC_UNUSED)
 	if (lp == DOT.l) {
 	    predlines = numlines;
 	    predchars = numchars + DOT.o;
-	    curchar = char_at_mark(DOT);
+	    curchar = CharAtDot();
 	}
 	/* on to the next line */
 	++numlines;
@@ -351,7 +372,7 @@ char_no(BUFFER *the_buffer, MARK the_mark)
 {
     LINE *lp;
     B_COUNT curchar = 0;
-    int len_rs = len_record_sep(curbp);
+    B_COUNT len_rs = (B_COUNT) len_record_sep(curbp);
 
     for_each_line(lp, the_buffer) {
 	if (lp == the_mark.l) {
@@ -387,7 +408,7 @@ gotochr(int f, int n)
 {
     LINE *lp;
     B_COUNT len;
-    int len_rs = len_record_sep(curbp);
+    B_COUNT len_rs = (B_COUNT) len_record_sep(curbp);
 
     if (!f) {
 	DOT.l = lback(buf_head(curbp));
@@ -671,7 +692,7 @@ forceblank(int f, int n)
     B_COUNT n_arg;
     C_NUM nchar;
     int s = TRUE;
-    int len_rs = len_record_sep(curbp);
+    B_COUNT len_rs = (B_COUNT) len_record_sep(curbp);
 
     n_arg = (B_COUNT) need_at_least(f, n, 0);
 
@@ -694,13 +715,13 @@ forceblank(int f, int n)
     }
 
     DOT.l = lforw(lp1);
-    DOT.o = 0;
+    DOT.o = b_left_margin(curbp);
 
     if (n_arg == nld) {		/* things are just right */
 	/*EMPTY */ ;
     } else if (n_arg < nld) {	/* delete (nld - n_arg) lines */
 	DOT.l = lp2;
-	DOT.o = 0;
+	DOT.o = b_left_margin(curbp);
 	if (nchar) {
 	    backchar(TRUE, nchar);
 	    s = ldel_chars((B_COUNT) nchar, FALSE);
@@ -713,8 +734,9 @@ forceblank(int f, int n)
 
     /* scan forward */
     while ((firstchar(DOT.l) < 0) &&
-	   (DOT.l != buf_head(curbp)))
-	DOT.l = lforw(DOT.l);
+	   (DOT.l != buf_head(curbp))) {
+	dot_next_bol();
+    }
 
     return s;
 }
@@ -1047,7 +1069,7 @@ current_directory(int force)
     cwd = realpath(s ? s : ".", current_dirname);
     if (cwd == 0)
 #endif
-	cwd = getcwd(current_dirname, NFILEN);
+	cwd = getcwd(current_dirname, (size_t) NFILEN);
 #else
 # ifdef HAVE_GETWD
     cwd = getwd(current_dirname);
@@ -1080,9 +1102,8 @@ current_directory(int force)
     bsl_to_sl_inplace(cwd);
     lengthen_path(cwd);
     cwd[NFILEN - 1] = EOS;
-#if !OPT_CASELESS
-    mklower(cwd);
-#endif
+    if (!global_g_val(GMDFILENAME_IC))
+	mklower(cwd);
 #endif
 
 #if SYS_MSDOS || SYS_OS2
@@ -1224,7 +1245,7 @@ vl_chdir(int f GCC_UNUSED, int n GCC_UNUSED)
     status = mlreply_dir("Change to directory: ", &last, cdirname);
 #if SYS_UNIX || SYS_VMS || SYS_WINNT
     if (status == FALSE) {	/* empty reply, go HOME */
-	(void) strcpy(cdirname, "~");
+	(void) vl_strncpy(cdirname, "~", sizeof(cdirname));
 	status = TRUE;
     }
 #endif
@@ -1337,7 +1358,7 @@ set_directory(const char *dir)
     upmode();
 
     TRACE(("set_directory(%s)\n", dir));
-    exdp = strcpy(exdir, dir);
+    exdp = vl_strncpy(exdir, dir, sizeof(exdir));
 
     if (doglob(exdp)) {
 #if SYS_MSDOS || SYS_OS2
@@ -1368,7 +1389,7 @@ set_directory(const char *dir)
 	if (!dirs_add_active) {
 	    /* Save current directory for subsequent "cd -". */
 
-	    (void) strcpy(prevdir, current_directory(FALSE));
+	    (void) vl_strncpy(prevdir, current_directory(FALSE), sizeof(prevdir));
 	}
 #if OPT_VMS_PATH
 	if (!*exdp)
@@ -1398,9 +1419,10 @@ set_directory(const char *dir)
 	 * want here.
 	 */
 	if (is_directory(exdp)) {
+	    int first = TRUE;
 	    /* For each comma-delimited component in CDPATH */
 	    cdpath = get_cdpath();
-	    while ((cdpath = parse_pathlist(cdpath, cdpathdir)) != 0) {
+	    while ((cdpath = parse_pathlist(cdpath, cdpathdir, &first)) != 0) {
 		int len;
 		char *tmp, newdir[NFILEN];
 
@@ -1410,7 +1432,7 @@ set_directory(const char *dir)
 		    char **globvec = glob_string(cdpathdir);
 
 		    if (glob_length(globvec) == 1) {
-			strcpy(cdpathdir, globvec[0]);
+			vl_strncpy(cdpathdir, globvec[0], sizeof(cdpathdir));
 			glob_free(globvec);
 			tmp = &cdpathdir[len - 1];
 			if (*tmp == R_BLOCK &&
@@ -1450,13 +1472,14 @@ set_directory(const char *dir)
 	 * directory in CDPATH.
 	 */
 	if (!is_pathname(exdp)) {
+	    int first = TRUE;
 	    /* For each appropriately delimited component in CDPATH */
 	    cdpath = get_cdpath();
-	    while ((cdpath = parse_pathlist(cdpath, cdpathdir)) != 0) {
+	    while ((cdpath = parse_pathlist(cdpath, cdpathdir, &first)) != 0) {
 		char **globvec = glob_string(cdpathdir);
 
 		if (glob_length(globvec) == 1) {
-		    strcpy(cdpathdir, globvec[0]);
+		    vl_strncpy(cdpathdir, globvec[0], sizeof(cdpathdir));
 		    glob_free(globvec);
 		    if (cd_and_pwd(pathcat(cdpathdir, cdpathdir, exdp))) {
 			return TRUE;
@@ -1475,7 +1498,7 @@ set_directory(const char *dir)
 #endif
 
     outlen = (int) ((unsigned) (term.cols - 1) - (sizeof(CHANGE_FAILED) - 3));
-    mlforce(CHANGE_FAILED, path_trunc(exdir, outlen, tmp, sizeof(tmp)));
+    mlforce(CHANGE_FAILED, path_trunc(exdir, outlen, tmp, (int) sizeof(tmp)));
     return FALSE;
 
 #undef CHANGE_FAILED
@@ -1487,7 +1510,7 @@ set_directory_from_file(BUFFER *bp)
     TRACE((T_CALLED "set_directory_from_file(%s)\n", bp->b_fname));
     if (!isInternalName(bp->b_fname)) {
 	char name[NFILEN];
-	char *leaf = pathleaf(vl_strncpy(name, bp->b_fname, NFILEN));
+	char *leaf = pathleaf(vl_strncpy(name, bp->b_fname, (size_t) NFILEN));
 	if (leaf != 0
 	    && leaf != name) {
 	    *leaf = EOS;
@@ -1594,26 +1617,24 @@ run_a_hook(HOOK * hook)
 {
     int status = FALSE;
 
-    if (!hook->latch && hook->proc[0]) {
-	MARK save_pre_op_dot;	/* ugly hack */
-	int save_dotcmdactive;	/* another ugly hack */
+    if (hook->proc[0] && !doingopcmd) {
+	if (!DisableHook(hook)) {
+	    MARK save_pre_op_dot;	/* ugly hack */
+	    int save_dotcmdactive;	/* another ugly hack */
 
-	TRACE((T_CALLED "run_a_hook(%s)\n", name_of_hook(hook)));
+	    save_dotcmdactive = dotcmdactive;
+	    save_pre_op_dot = pre_op_dot;
 
-	save_dotcmdactive = dotcmdactive;
-	save_pre_op_dot = pre_op_dot;
+	    TPRINTF(("running %s HOOK with %s, current %s\n",
+		     name_of_hook(hook), hook->proc, curbp->b_bname));
+	    status = docmd(hook->proc, TRUE, FALSE, 1);
 
-	DisableHook(hook);
-	TPRINTF(("running %s HOOK with %s, current %s\n",
-		 name_of_hook(hook), hook->proc, curbp->b_bname));
-	status = docmd(hook->proc, TRUE, FALSE, 1);
+	    pre_op_dot = save_pre_op_dot;
+	    dotcmdactive = save_dotcmdactive;
+	}
 	EnableHook(hook);
-
-	pre_op_dot = save_pre_op_dot;
-	dotcmdactive = save_dotcmdactive;
-	returnCode(status);
     }
-    return FALSE;
+    return (status);
 }
 
 int
@@ -1639,8 +1660,8 @@ vl_elapsed(VL_ELAPSED * first, int begin)
 {
     double result;
 
-#if SYS_UNIX && !SYS_MINGW
-#define	SECS(tv)	(tv.tv_sec + (tv.tv_usec / 1.0e6))
+#if defined(HAVE_GETTIMEOFDAY) && (SYS_UNIX && !SYS_MINGW)
+#define	SECS(tv)	((double) tv.tv_sec + ((double) tv.tv_usec / 1.0e6))
     VL_ELAPSED tv1;
     gettimeofday(&tv1, 0);
     if (begin)
@@ -1777,8 +1798,7 @@ autocolor(void)
 			     * Make a local autocolor mode to show the user
 			     * what the actual interval is.
 			     */
-			    make_local_b_val(bp, VAL_AUTOCOLOR);
-			    set_b_val(bp, VAL_AUTOCOLOR, next_time);
+			    set_local_b_val(bp, VAL_AUTOCOLOR, next_time);
 			    /*
 			     * Set counter to skip autocolor intervals.
 			     */
@@ -1900,7 +1920,7 @@ do_pushd(int uindx,		/* user-specified dirstack index */
     int rc;
     char oldcwd[NFILEN], *path;
 
-    strcpy(oldcwd, current_directory(TRUE));
+    vl_strncpy(oldcwd, current_directory(TRUE), sizeof(oldcwd));
 
     beginDisplay();
     if ((uindx == 0 && sign > 0) || (uindx == dirs_idx && sign < 0)) {
@@ -1947,6 +1967,7 @@ pushd(int f GCC_UNUSED, int n GCC_UNUSED)
 {
     long dirlocn;
     int failed, rc, sign;
+    int ch;
     char oldcwd[NFILEN], newcwd[NFILEN], *tmp, *cp;
     static TBUFF *last;
 
@@ -1971,7 +1992,7 @@ pushd(int f GCC_UNUSED, int n GCC_UNUSED)
 
 	/* handle simple case first:  pushd without +|-n arg */
 	if (*cp != '+' && *cp != '-') {
-	    strcpy(oldcwd, current_directory(TRUE));
+	    vl_strncpy(oldcwd, current_directory(TRUE), sizeof(oldcwd));
 	    if ((rc = pushd_popd_set_dir(newcwd)) != TRUE)
 		return (rc);
 
@@ -1986,7 +2007,8 @@ pushd(int f GCC_UNUSED, int n GCC_UNUSED)
 		sign = (*cp == '-') ? -1 : 1;
 		cp++;
 		rc = FALSE;
-		if (isDigit(*cp++)) {
+		ch = *cp++;
+		if (isDigit(ch)) {
 		    cp = skip_number(cp);
 		    if (*cp == '\0')
 			rc = TRUE;
@@ -2009,7 +2031,7 @@ pushd(int f GCC_UNUSED, int n GCC_UNUSED)
 		    mlforce("[Pushd index out-of-range]");
 		    return (FALSE);
 		}
-		rc = do_pushd(dirlocn, sign);
+		rc = do_pushd((int) dirlocn, sign);
 	    }
 	}
     }
@@ -2028,6 +2050,7 @@ popd(int f GCC_UNUSED, int n GCC_UNUSED)
     long dirlocn;
     static TBUFF *last;
     int sign = 0, rc, failed;
+    int ch;
 
     rc = mlreply2("Directory stack entry: ", &last);
     if (rc == ABORT)
@@ -2061,7 +2084,8 @@ popd(int f GCC_UNUSED, int n GCC_UNUSED)
     if (*cp == '-' || *cp == '+') {
 	sign = (*cp == '-') ? -1 : 1;
 	cp++;
-	if (isDigit(*cp++)) {
+	ch = *cp++;
+	if (isDigit(ch)) {
 	    cp = skip_number(cp);
 	    if (*cp == '\0')
 		rc = TRUE;
@@ -2085,7 +2109,7 @@ popd(int f GCC_UNUSED, int n GCC_UNUSED)
 	mlforce("[Popd index out-of-range]");
 	return (FALSE);
     }
-    if ((rc = do_popd(dirlocn, sign)) == TRUE)
+    if ((rc = do_popd((int) dirlocn, sign)) == TRUE)
 	(void) vl_dirs(0, 0);
     return (rc);
 }
@@ -2224,7 +2248,7 @@ vl_dirs_add(int f GCC_UNUSED, int n GCC_UNUSED)
 	 * might be accessible via CDPATH
 	 */
 
-	strcpy(savedcwd, current_directory(TRUE));
+	vl_strncpy(savedcwd, current_directory(TRUE), sizeof(savedcwd));
 	dirs_add_active = TRUE;
 	if (set_directory(newdir)) {
 	    rc = dirstack_extend(current_directory(TRUE), "vl_dirs_add");
@@ -2272,7 +2296,7 @@ vl_atol(const char *str, int base, int *failed)
 	    if (errno == EDOM || errno == ERANGE) {
 		*failed = TRUE;
 	    } else {
-		if (*(prem = skip_blanks(prem)) != EOS) {
+		if (*skip_blanks(prem) != EOS) {
 		    *failed = TRUE;	/* trailing garbage */
 		}
 	    }
@@ -2312,7 +2336,7 @@ vl_atoul(const char *str, int base, int *failed)
 	    if (errno == EDOM || errno == ERANGE) {
 		*failed = TRUE;
 	    } else {
-		if (*(prem = skip_blanks(prem)) != EOS)
+		if (*skip_blanks(prem) != EOS)
 		    *failed = TRUE;	/* trailing garbage */
 	    }
 	}
@@ -2346,8 +2370,33 @@ vl_stricmp(const char *a, const char *b)
 	    ch2 = toUpper(ch2);
 	cmp = ch1 - ch2;
 	if (cmp != 0 || ch1 == EOS || ch2 == EOS)
-	    return cmp;
+	    break;
     }
+    return cmp;
+}
+#endif
+
+#ifndef vl_strnicmp
+/*
+ * Compare two strings ignoring case
+ */
+int
+vl_strnicmp(const char *a, const char *b, size_t n)
+{
+    int ch1, ch2, cmp = 0;
+
+    while (n-- != 0) {
+	ch1 = CharOf(*a++);
+	ch2 = CharOf(*b++);
+	if (isLower(ch1))
+	    ch1 = toUpper(ch1);
+	if (isLower(ch2))
+	    ch2 = toUpper(ch2);
+	cmp = ch1 - ch2;
+	if (cmp != 0 || ch1 == EOS || ch2 == EOS)
+	    break;
+    }
+    return cmp;
 }
 #endif
 
