@@ -1,7 +1,7 @@
 /*	Spawn:	various DOS access commands
  *		for MicroEMACS
  *
- * $Header: /usr/build/vile/vile/RCS/spawn.c,v 1.202 2009/05/26 23:46:03 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/spawn.c,v 1.216 2016/07/13 09:03:29 tom Exp $
  *
  */
 
@@ -53,7 +53,7 @@ static int spawn1(int rerun, int pressret);
 #include	<iodef.h>
 
 #if DISP_X11
-#define VMSVT_DATA /* nothing */
+#define VMSVT_DATA		/* nothing */
 #else
 #define VMSVT_DATA extern
 #endif
@@ -88,7 +88,7 @@ x_window_SHELL(const char *cmd)
     if ((pid = fork()) > 0) {
 	waitpid(pid, 0, 0);
     } else if (pid == 0) {
-	if ((pid = fork()) == 0) {
+	if (fork() == 0) {
 #endif
 	    /*
 	     * We would use the -display option of xterm, but that
@@ -118,7 +118,7 @@ x_window_SHELL(const char *cmd)
 	    if (tmp != 0) {
 		char *result = tb_values(tmp);
 		TRACE(("executing '%s'\n", result));
-		(void) system(result);
+		IGNORE_RC(system(result));
 		tb_free(&tmp);
 	    }
 #ifdef HAVE_WAITPID
@@ -145,7 +145,7 @@ spawncli(int f GCC_UNUSED, int n GCC_UNUSED)
 #if	SYS_UNIX
 #define	OK_SPAWN
     term.clean(TRUE);
-    file_stat(0, 0);
+    (void) file_stat(0, 0);
     term.openup();
 #if	DISP_X11 && !SMALLER
     (void) x_window_SHELL((char *) 0);
@@ -231,7 +231,7 @@ bktoshell(int f, int n)		/* suspend and wait to wake up */
 
     beginDisplay();
     term.clean(TRUE);
-    file_stat(0, 0);
+    (void) file_stat(0, 0);
 
 /* #define simulate_job_control_for_debug */
 # ifdef simulate_job_control_for_debug
@@ -397,6 +397,7 @@ spawn1(int rerun, int pressret)
 
 #if SYS_UNIX
 #if DISP_X11
+    (void) pressret;
 #if defined(HAVE_WAITPID) && !SMALLER
     (void) x_window_SHELL(line);
 #else
@@ -404,7 +405,7 @@ spawn1(int rerun, int pressret)
 #endif
 #else
     term.clean(TRUE);
-    file_stat(0, 0);
+    (void) file_stat(0, 0);
 
     (void) system_SHELL(line);
 
@@ -561,8 +562,7 @@ capturecmd(int f, int n)
     set_bname(curbp, OUTPUT_BufName);
 
     /* make this window in VIEW mode, update buffer's mode lines */
-    make_local_b_val(curwp->w_bufp, MDVIEW);
-    set_b_val(curwp->w_bufp, MDVIEW, TRUE);
+    set_local_b_val(curwp->w_bufp, MDVIEW, TRUE);
     curwp->w_flag |= WFMODE;
 
 #if OPT_FINDERR
@@ -591,7 +591,10 @@ write_kreg_to_pipe(void *writefp)
     kregcirculate(FALSE);
     kp = kbs[ukb].kbufh;
     while (kp != NULL) {
-	fwrite((char *) kp->d_chunk, 1, (size_t) KbSize(ukb, kp), fw);
+	IGNORE_RC(fwrite((char *) kp->d_chunk,
+			 (size_t) 1,
+			 (size_t) KbSize(ukb, kp),
+			 fw));
 	kp = kp->d_next;
     }
 #if SYS_UNIX && ! TEST_DOS_PIPES
@@ -630,7 +633,10 @@ write_region_to_pipe(void *writefp)
     LINE *lp = DOT.l;
 
     while (lp != last) {
-	fwrite((char *) lvalue(lp), sizeof(char), (size_t) llength(lp), fw);
+	IGNORE_RC(fwrite((char *) lvalue(lp),
+			 sizeof(char),
+			   (size_t) llength(lp),
+			 fw));
 	vl_putc('\n', fw);
 	lp = lforw(lp);
     }
@@ -778,6 +784,9 @@ filterregion(void)
 	    (void) firstnonwhite(FALSE, 1);
 	    (void) setmark();
 	    end_kill();
+	} else {
+	    fclose(fw);
+	    npclose(fr);
 	}
     }
 #else
@@ -800,52 +809,63 @@ open_region_filter(void)
     static char oline[NLINE];	/* command line send to shell */
     char line[NLINE];		/* command line send to shell */
     FILE *fr, *fw;
+#endif
     int s;
 
-    /* get the filter name and its args */
-    if ((s = mlreply_no_bs("!", oline, NLINE)) != TRUE)
-	return (s);
-    (void) strcpy(line, oline);
-    if ((s = inout_popen(&fr, &fw, line)) != TRUE) {
-	mlforce("[Couldn't open pipe or command]");
-	return s;
-    }
+    TRACE((T_CALLED "open_region_filter\n"));
 
-    if (!softfork()) {
+#if SYS_UNIX || SYS_MSDOS || (SYS_OS2 && CC_CSETPP) || SYS_WINNT
+    /* get the filter name and its args */
+    if ((s = mlreply_no_bs("!", oline, NLINE)) == TRUE) {
+	(void) strcpy(line, oline);
+	if ((s = inout_popen(&fr, &fw, line)) == TRUE) {
+	    if (!softfork()) {
 #if !(SYS_WINNT && defined(GMDW32PIPES))
-	write_region_to_pipe(fw);
+		write_region_to_pipe(fw);
 #else
-	/* This is a Win32 environment with compiled Win32 pipe support. */
-	if (global_g_val(GMDW32PIPES)) {
-	    /*
-	     * w32pipes mode enabled -- create child thread to blast
-	     * region to write pipe.
-	     */
-	    if (_beginthread(write_region_to_pipe, 0, fw) == (unsigned long) -1) {
-		mlforce("[Can't create Win32 write pipe]");
+		/* This is a Win32 environment with compiled Win32 pipe
+		 * support.
+		 */
+		if (global_g_val(GMDW32PIPES)) {
+		    ULONG code = (ULONG) _beginthread(write_region_to_pipe,
+						      0,
+						      fw);
+
+		    /*
+		     * w32pipes mode enabled -- create child thread to blast
+		     * region to write pipe.
+		     */
+		    if (code == (ULONG) (-1)) {
+			mlforce("[Can't create Win32 write pipe]");
+			(void) fclose(fw);
+			(void) npclose(fr);
+			s = FALSE;
+		    }
+		} else {
+		    /*
+		     * Single-threaded parent process writes region to pseudo
+		     * write pipe (temp file).
+		     */
+		    write_region_to_pipe(fw);
+		}
+#endif
+	    }
+	    if (s == TRUE) {
+#if ! ((SYS_OS2 && CC_CSETPP) || SYS_WINNT)
 		(void) fclose(fw);
-		(void) npclose(fr);
-		return (FALSE);
+#endif
+		ffp = fr;
+		count_fline = 0;
 	    }
 	} else {
-	    /*
-	     * Single-threaded parent process writes region to pseudo
-	     * write pipe (temp file).
-	     */
-	    write_region_to_pipe(fw);
+	    mlforce("[Couldn't open pipe or command]");
 	}
-#endif
     }
-#if ! ((SYS_OS2 && CC_CSETPP) || SYS_WINNT)
-    (void) fclose(fw);
-#endif
-    ffp = fr;
-    count_fline = 0;
-    return s;
 #else
     mlforce("[Region filtering not available -- try buffer filtering]");
-    return FALSE;
+    s = FALSE;
 #endif
+    returnCode(s);
 }
 #endif /* OPT_SELECTIONS */
 
@@ -903,7 +923,7 @@ vile_filter(int f GCC_UNUSED, int n GCC_UNUSED)
 #if	SYS_UNIX
     bottomleft();
     term.clean(TRUE);
-    file_stat(0, 0);
+    (void) file_stat(0, 0);
     if ((t = strchr(line, '|')) != 0) {
 	char temp[NLINE];
 	(void) strcpy(temp, t);
@@ -993,18 +1013,20 @@ set_envvar(int f GCC_UNUSED, int n GCC_UNUSED)
     char *both;
     int rc;
 
-    if ((rc = mlreply2("Environment variable: ", &var)) == ABORT) {
-	/* EMPTY */ ;
-    } else if ((rc = mlreply2("Value: ", &val)) == ABORT) {
-	/* EMPTY */ ;
-    } else if (tb_length(var) == 0
-	       || tb_length(val) == 0
-	       || (both = typeallocn(char,
-				     tb_length(var) + tb_length(val))) == 0) {
-	rc = no_memory("set_envvar");
-    } else {
-	lsprintf(both, "%s=%s", tb_values(var), tb_values(val));
-	putenv(both);	/* this will leak.  i think it has to. */
+    if ((rc = mlreply2("Environment variable: ", &var)) != ABORT) {
+	if ((rc = mlreply2("Value: ", &val)) != ABORT) {
+	    size_t len_var = tb_length(var);
+	    size_t len_val = tb_length(val);
+
+	    if (len_var != 0
+		&& len_val != 0
+		&& (both = typeallocn(char, len_var + len_val + 1)) != 0) {
+		lsprintf(both, "%s=%s", tb_values(var), tb_values(val));
+		putenv(both);	/* this will leak.  i think it has to. */
+	    } else {
+		rc = no_memory("set_envvar");
+	    }
+	}
     }
 
     return rc;
@@ -1115,6 +1137,18 @@ parse_findcfg_mode(FINDCFG * pcfg, char *inputstr)
     return (rc);
 }
 
+static void
+free_vector(char ***vec, size_t vec_elements)
+{
+    char **base;
+    ULONG i;
+
+    base = *vec;
+    for (i = 0; i < vec_elements; i++, base++)
+	(void) free(*base);
+    (void) free(*vec);
+}
+
 /*
  * Cruise through the user's shell command, stripping out unquoted tokens
  * that include wildcard characters.   Save each token in a vector.  Return
@@ -1123,7 +1157,11 @@ parse_findcfg_mode(FINDCFG * pcfg, char *inputstr)
 static char *
 extract_wildcards(char *cmd, char ***vec, size_t *vecidx, const char *fnname)
 {
-    char **base, *cp, *anchor, buf[NFILEN * 2];
+    char **temp;
+    char **base;
+    char *cp;
+    char *anchor;
+    char buf[NFILEN * 2];
     int delim;
     size_t idx, len;
 
@@ -1140,33 +1178,38 @@ extract_wildcards(char *cmd, char ***vec, size_t *vecidx, const char *fnname)
 	if (*cp == '\'' || *cp == '"') {
 	    delim = *cp++;
 	    while (*cp) {
-		if (*cp == '\\' && cp[1] == delim)
+		if (*cp == '\\' && cp[1] == delim) {
 		    cp += 2;
-		else if (*cp == delim) {
+		} else if (*cp == delim) {
 		    cp++;
 		    break;
-		} else
+		} else {
 		    cp++;
+		}
 	    }
 	} else {
 	    cp++;
 	    while (*cp && (!isSpace(*cp)))
 		cp++;
-	    len = cp - anchor;
+	    len = (size_t) (cp - anchor);
 	    strncpy(buf, anchor, len);
 	    buf[len] = EOS;
 	    if (string_has_wildcards(buf)) {
 		memset(anchor, ' ', len);	/* blank out wildcard in cmd */
 		if (idx >= len) {
 		    len *= 2;
-		    base = castrealloc(char *, base, sizeof(*base));
-		    if (base == NULL) {
+		    temp = castrealloc(char *, base, sizeof(*base));
+		    if (temp == NULL) {
+			free_vector(&base, idx);
 			(void) no_memory(fnname);
 			return (NULL);
+		    } else {
+			base = temp;
 		    }
 		}
 		base[idx] = castalloc(char, len + 1);
 		if (base[idx] == NULL) {
+		    free_vector(&base, idx);
 		    (void) no_memory(fnname);
 		    return (NULL);
 		}
@@ -1181,18 +1224,6 @@ extract_wildcards(char *cmd, char ***vec, size_t *vecidx, const char *fnname)
     return (cmd);
 }
 
-static void
-free_vector(char ***vec, size_t vec_elements)
-{
-    char **base;
-    ULONG i;
-
-    base = *vec;
-    for (i = 0; i < vec_elements; i++, base++)
-	(void) free(*base);
-    (void) free(*vec);
-}
-
 static const char *
 determine_quoted_delimiter(void)
 {
@@ -1246,24 +1277,30 @@ add_token_to_cmd(char **cmd,
 		 const char *funcname)
 {
     int rc = TRUE;
+    char *tmp;
     size_t toklen = strlen(token);
 
-    if (*cmdidx + toklen + 2 > *cmdlen) {
-	char *tmp = *cmd;
-
-	*cmdlen *= 2;
-	tmp = castrealloc(char, tmp, *cmdlen);
-	if (tmp == NULL) {
-	    (void) free(*cmd);
-	    rc = no_memory(funcname);
-	    return (rc);
+    if ((tmp = *cmd) != 0) {
+	if (*cmdidx + toklen + 2 > *cmdlen) {
+	    *cmdlen *= 2;
+	    tmp = castrealloc(char, tmp, *cmdlen);
+	    if (tmp == NULL) {
+		(void) free(*cmd);
+		*cmd = 0;
+	    } else {
+		*cmd = tmp;
+	    }
 	}
-	*cmd = tmp;
     }
-    strcpy(*cmd + *cmdidx, token);
-    *cmdidx += toklen;
-    (*cmd)[*cmdidx] = ' ';
-    (*cmdidx)++;
+    if (*cmd != 0) {
+	strcpy(*cmd + *cmdidx, token);
+	*cmdidx += toklen;
+	(*cmd)[*cmdidx] = ' ';
+	(*cmdidx)++;
+	rc = TRUE;
+    } else {
+	rc = no_memory(funcname);
+    }
     return (rc);
 }
 
@@ -1360,6 +1397,14 @@ ck_find_cmd(char *cmd, int *allocd_storage, int prepend_bang)
     return (cmd);
 }
 
+#if SYS_UNIX
+#define EGREP_OPT_CASELESS ""
+#define EGREP_TAG_CASELESS "[Tt][Aa][Gg][Ss]"
+#else
+#define EGREP_OPT_CASELESS "i"
+#define EGREP_TAG_CASELESS "tags"
+#endif
+
 /*
  * FUNCTION
  *   find_dirs_only(char *cmd, FINDINFO *pinfo, int prepend_bang)
@@ -1399,6 +1444,7 @@ find_dirs_only(char *cmd, FINDINFO * pinfo, int prepend_bang)
     const char *path, *fnname;
     char *rslt, buf[NFILEN * 2];
     const char *qdelim;
+    int first = TRUE;
 
     fnname = "find_dirs_only";
     outlen = sizeof(buf);
@@ -1419,7 +1465,7 @@ find_dirs_only(char *cmd, FINDINFO * pinfo, int prepend_bang)
     path = pinfo->dir_list;
 
     /* add directory list to find command */
-    while ((path = parse_pathlist(path, buf)) != NULL) {
+    while ((path = parse_pathlist(path, buf, &first)) != NULL) {
 	if (!add_token_to_cmd(&rslt, &outidx, &outlen, buf, fnname))
 	    return (NULL);
     }
@@ -1453,23 +1499,16 @@ find_dirs_only(char *cmd, FINDINFO * pinfo, int prepend_bang)
      * ========================== FIXME ==================================
      */
     sprintf(buf,
-	    "| egrep -v%s %s(RCS|CVS)/%s",
-#if SYS_UNIX
-	    "",
-#else
-	    "i",
-#endif
+	    "| egrep -v" EGREP_OPT_CASELESS " %s(RCS|CVS)/%s",
 	    qdelim,
 	    qdelim);
-    if (!add_token_to_cmd(&rslt, &outidx, &outlen, buf, fnname))
-	return (NULL);
-
-    /* finish off with xargs */
-    if (!add_token_to_cmd(&rslt, &outidx, &outlen, "| xargs", fnname))
-	return (NULL);
-    if (!add_token_to_cmd(&rslt, &outidx, &outlen, cmd, fnname))
+    if (!add_token_to_cmd(&rslt, &outidx, &outlen, buf, fnname)) {
 	rslt = NULL;
-    else {
+    } else if (!add_token_to_cmd(&rslt, &outidx, &outlen, "| xargs", fnname)) {
+	rslt = NULL;
+    } else if (!add_token_to_cmd(&rslt, &outidx, &outlen, cmd, fnname)) {
+	rslt = NULL;
+    } else if (rslt != 0) {
 	rslt[outidx] = EOS;	/* terminate cmd string */
 	if (outidx != 0) {
 	    char *cp;
@@ -1538,6 +1577,7 @@ find_all_files(char *cmd, FINDINFO * pinfo, int prepend_bang)
     const char *path, *fnname;
     char *xargstr, **vec, *rslt, buf[NFILEN * 2];
     const char *qdelim;
+    int first = TRUE;
 
     fnname = "find_all_files";
     if ((xargstr = extract_wildcards(cmd, &vec, &vecidx, fnname)) == NULL)
@@ -1580,7 +1620,7 @@ find_all_files(char *cmd, FINDINFO * pinfo, int prepend_bang)
     path = pinfo->dir_list;
 
     /* add directory list to find command */
-    while ((path = parse_pathlist(path, buf)) != NULL) {
+    while ((path = parse_pathlist(path, buf, &first)) != NULL) {
 	if (!add_token_to_cmd(&rslt, &outidx, &outlen, buf, fnname)) {
 	    free_vector(&vec, vecidx);
 	    return (NULL);
@@ -1615,13 +1655,8 @@ find_all_files(char *cmd, FINDINFO * pinfo, int prepend_bang)
     /* add wildcards (if any) to find command */
     for (i = 0; i < vecidx; i++) {
 	sprintf(buf,
-		"%s-%sname %s%s%s",
+		"%s-" EGREP_OPT_CASELESS "name %s%s%s",
 		(i != 0) ? "-o " : "",
-#if SYS_UNIX
-		"",
-#else
-		"i",		/* find uses case insensitive filename search */
-#endif
 		qdelim,
 		vec[i],
 		qdelim);
@@ -1652,28 +1687,17 @@ find_all_files(char *cmd, FINDINFO * pinfo, int prepend_bang)
      * ========================== FIXME ==================================
      */
     sprintf(buf,
-	    "| egrep -v%s %s((RCS|CVS)/|/%s$)%s",
-#if SYS_UNIX
-	    "",
-#else
-	    "i",
-#endif
+	    "| egrep -v" EGREP_OPT_CASELESS
+	    " %s((RCS|CVS)/|/" EGREP_TAG_CASELESS "$)%s",
 	    qdelim,
-#if SYS_UNIX
-	    "[Tt][Aa][Gg][Ss]",
-#else
-	    "tags",
-#endif
 	    qdelim);
-    if (!add_token_to_cmd(&rslt, &outidx, &outlen, buf, fnname))
-	return (NULL);
-
-    /* finish off with xargs */
-    if (!add_token_to_cmd(&rslt, &outidx, &outlen, "| xargs", fnname))
-	return (NULL);
-    if (!add_token_to_cmd(&rslt, &outidx, &outlen, xargstr, fnname))
+    if (!add_token_to_cmd(&rslt, &outidx, &outlen, buf, fnname)) {
 	rslt = NULL;
-    else {
+    } else if (!add_token_to_cmd(&rslt, &outidx, &outlen, "| xargs", fnname)) {
+	rslt = NULL;
+    } else if (!add_token_to_cmd(&rslt, &outidx, &outlen, xargstr, fnname)) {
+	rslt = NULL;
+    } else if (rslt != 0) {
 	rslt[outidx] = EOS;	/* terminate cmd string */
 	if (outidx != 0) {
 	    char *cp;
