@@ -1,7 +1,7 @@
 /*
  * debugging support -- tom dickey.
  *
- * $Header: /usr/build/vile/vile/RCS/trace.c,v 1.92 2009/06/02 20:47:07 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/trace.c,v 1.116 2014/10/01 08:40:10 tom Exp $
  *
  */
 
@@ -80,7 +80,10 @@ static unsigned used_indent;
 
 static int trace_depth;
 
-#if OPT_ELAPSED
+static char null_string[] = "<null>";
+static char empty_string[] = "";
+
+#if OPT_ELAPSED && !defined(show_elapsed)
 
 #if 0
 static void ETrace(const char *fmt,...) GCC_PRINTFLIKE(1,2);
@@ -151,9 +154,19 @@ dpy_elapsed(BI_NODE * a, int level GCC_UNUSED)
 	  q->total_time / q->total_calls);
 }
 
-#define BI_TREE0 {{0}, 0, {0, 0}}
+static void
+xcg_elapsed(BI_NODE * a, BI_NODE * b)
+{
+    BI_DATA temp = a->value;
+    a->value = b->value;
+    b->value = temp;
+}
+
+#define BI_DATA0 {0,0}
+#define BI_TREE0 0, 0, BI_DATA0
+
 static BI_TREE elapsed_tree =
-{new_elapsed, old_elapsed, dpy_elapsed, 0, 0, BI_TREE0};
+{new_elapsed, old_elapsed, dpy_elapsed, xcg_elapsed, BI_TREE0};
 
 static BI_DATA *elapsed_stack[100];
 static int elapsed_sp = 0;
@@ -271,6 +284,7 @@ Trace(const char *fmt,...)
 	    (void) fflush(stdout);
 	    (void) fflush(stderr);
 	    trace_depth = 0;
+	    fp = 0;
 	}
 
 	va_end(ap);
@@ -307,6 +321,14 @@ retrace_string(char *code)
     return code;
 }
 
+const char *
+retrace_cstring(const char *code)
+{
+    Trace(T_RETURN "%s%s\n", TRACE_NULL(code),
+	  isErrorVal(code) ? " (ERROR)" : "");
+    return code;
+}
+
 void
 retrace_void(void)
 {
@@ -333,16 +355,44 @@ alloc_visible(size_t need)
 static char *
 my_vischr(char *buffer, int ch)
 {
-    ch = CharOf(ch);
-    if (ch == 127) {
-	strcpy(buffer, "^?");
-    } else if (ch < 32) {
-	sprintf(buffer, "^%c", ch | '@');
-    } else if (ch >= 128) {
-	sprintf(buffer, "\\%03o", ch);
-    } else {
-	buffer[0] = (char) ch;
-	buffer[1] = '\0';
+    if (buffer != 0) {
+	if (ch == 127) {
+	    strcpy(buffer, "^?");
+	} else if (ch < 32) {
+	    sprintf(buffer, "^%c", ch | '@');
+#if OPT_MULTIBYTE
+	} else if (ch >= 256) {
+	    sprintf(buffer, "\\u%04x", ch);
+#endif
+	} else if (ch >= 128) {
+	    sprintf(buffer, "\\x%02x", ch);
+	} else if (ch == '\\') {
+	    sprintf(buffer, "\\\\");
+	} else {
+	    buffer[0] = (char) ch;
+	    buffer[1] = '\0';
+	}
+    }
+    return buffer;
+}
+
+static char *
+my_visattr(char *buffer, int ch)
+{
+    if (buffer != 0) {
+	if (ch & (VACOLOR | VASPCOL)) {
+	    *buffer = 'C';
+	} else if (ch & (int) VASEL) {
+	    *buffer = 'S';
+	} else if (ch & (int) VAREV) {
+	    *buffer = 'R';
+	} else if (ch & (int) VAUL) {
+	    *buffer = '_';
+	} else if (ch) {
+	    *buffer = '#';
+	} else {
+	    *buffer = '.';
+	}
     }
     return buffer;
 }
@@ -361,8 +411,9 @@ trace_indent(int level, int marker)
 	else
 	    visible_indent = typereallocn(char, visible_indent, need);
 
-	if (visible_indent != 0)
-	    memset(visible_indent, 0, need);
+	assert(visible_indent != 0);
+
+	memset(visible_indent, 0, need);
     }
 
     *visible_indent = EOS;
@@ -385,19 +436,19 @@ visible_buff(const char *buffer, int length, int eos)
 	length = 0;
     }
 
-    result = alloc_visible(need);
-
-    for (j = 0; j < length; j++) {
-	int c = buffer[j] & 0xff;
-	if (eos && !c) {
-	    break;
-	} else {
-	    char *dst = result + k;
-	    my_vischr(dst, c);
-	    k += (unsigned) strlen(dst);
+    if ((result = alloc_visible(need)) != 0) {
+	for (j = 0; j < length; j++) {
+	    int c = CharOf(buffer[j]);
+	    if (eos && !c) {
+		break;
+	    } else {
+		char *dst = result + k;
+		my_vischr(dst, c);
+		k += (unsigned) strlen(dst);
+	    }
 	}
+	result[k] = 0;
     }
-    result[k] = 0;
     endofDisplay();
     return result;
 }
@@ -418,15 +469,44 @@ visible_video_text(const VIDEO_TEXT * buffer, int length)
 	length = 0;
     }
 
-    result = alloc_visible(need);
+    if ((result = alloc_visible(need)) != 0) {
+	for (j = 0; j < length; j++) {
+	    int c = buffer[j] & ((1 << (8 * sizeof(VIDEO_TEXT))) - 1);
+	    char *dst = result + k;
+	    my_vischr(dst, c);
+	    k += (unsigned) strlen(dst);
+	}
+	result[k] = 0;
+    }
+    endofDisplay();
+    return result;
+}
 
-    for (j = 0; j < length; j++) {
-	int c = buffer[j] & ((1 << (8 * sizeof(VIDEO_TEXT))) - 1);
-	char *dst = result + k;
-	my_vischr(dst, c);
-	k += (unsigned) strlen(dst);
+char *
+visible_video_attr(const VIDEO_ATTR * buffer, int length)
+{
+    int j;
+    unsigned k = 0;
+    unsigned need = (unsigned) (((length > 0) ? length : 0) + 1);
+    char *result;
+
+    beginDisplay();
+    if (buffer == 0) {
+	static const VIDEO_ATTR dummy[] =
+	{0};
+	buffer = dummy;
+	length = 0;
+    }
+
+    if ((result = alloc_visible(need)) != 0) {
+	for (j = 0; j < length; j++) {
+	    int c = buffer[j] & ((1 << (8 * sizeof(VIDEO_ATTR))) - 1);
+	    char *dst = result + k;
+	    my_visattr(dst, c);
+	    ++k;
+	}
+	result[k] = 0;
     }
-    result[k] = 0;
     endofDisplay();
     return result;
 }
@@ -438,7 +518,7 @@ char *
 str_visible(const char *s)
 {
     if (s == 0)
-	return "<null>";
+	return null_string;
     return visible_buff(s, (int) strlen(s), FALSE);
 }
 
@@ -449,7 +529,7 @@ char *
 str_visible0(const char *s)
 {
     if (s == 0)
-	return "<null>";
+	return null_string;
     return visible_buff(s, (int) strlen(s), TRUE);
 }
 
@@ -486,11 +566,16 @@ itb_visible(ITBUFF * p)
 		    strcat(result, temp);
 		used += strlen(temp);
 	    }
-	    if (!pass)
-		*(result = alloc_visible(1 + used)) = EOS;
+	    if (!pass) {
+		if ((result = alloc_visible(1 + used)) == 0) {
+		    break;
+		}
+		*result = EOS;
+	    }
 	}
     } else {
-	*(result = alloc_visible(1)) = EOS;
+	if ((result = alloc_visible(1)) != 0)
+	    *result = EOS;
     }
     endofDisplay();
     return result;
@@ -500,7 +585,7 @@ char *
 lp_visible(LINE *p)
 {
     if (p == 0)
-	return "";
+	return empty_string;
     return visible_buff(lvalue(p), llength(p), FALSE);
 }
 
@@ -515,7 +600,7 @@ fail_alloc(char *msg, char *ptr)
 #if NO_LEAKS
     show_alloc();
 #endif
-    Trace(bad_form);
+    Trace("%s", bad_form);
     abort();
 }
 
@@ -540,9 +625,9 @@ InitArea(void)
 	area = typecallocn(AREA, DOALLOC);
 	if (area == 0)
 	    abort();
-	Trace("Initialized doalloc (%d * %d) = %ld\n",
-	      DOALLOC,
-	      sizeof(AREA),
+	Trace("Initialized doalloc (%ld * %ld) = %ld\n",
+	      (long) DOALLOC,
+	      (long) sizeof(AREA),
 	      (long) sizeof(AREA) * (long) DOALLOC);
     }
 }
@@ -693,15 +778,17 @@ void
 dofree(void *oldp)
 {
     check_opt_working();
-    count_freed++;
-    LOG_PTR("dealloc ", oldp);
+    if (oldp != 0) {
+	count_freed++;
+	LOG_PTR("dealloc ", oldp);
 
-    if (OK_FREE(oldp)) {
-	free(oldp);
-	return;
+	if (OK_FREE(oldp)) {
+	    free(oldp);
+	    return;
+	}
+
+	fail_alloc("free (not found)", oldp);
     }
-
-    fail_alloc("free (not found)", oldp);
 }
 
 /*
@@ -822,7 +909,7 @@ trace_line(LINE *lp, BUFFER *bp)
 }
 
 void
-trace_mark(char *name, MARK *mk, BUFFER *bp)
+trace_mark(const char *name, MARK *mk, BUFFER *bp)
 {
     Trace("%s %d.%d\n", name, line_no(bp, mk->l), mk->o);
 }
@@ -930,13 +1017,19 @@ void
 trace_buffer(BUFFER *bp)
 {
     LINE *lp;
-    Trace("trace_buffer(%s) dot=%p%s fn=%s\n",
-	  bp->b_bname,
-	  (void *) bp->b_dot.l,
-	  bp == curbp ? " (curbp)" : "",
-	  NonNull(bp->b_fname));
-    for_each_line(lp, bp) {
-	trace_line(lp, bp);
+    if (bp != 0) {
+	Trace("trace_buffer(%s)%s %s dot=%p%s fn=%s\n",
+	      bp->b_bname,
+	      (bp == bminip ? " MINI" : ""),
+	      encoding2s(b_val(bp, VAL_FILE_ENCODING)),
+	      (void *) bp->b_dot.l,
+	      bp == curbp ? " (curbp)" : "",
+	      NonNull(bp->b_fname));
+	for_each_line(lp, bp) {
+	    trace_line(lp, bp);
+	}
+    } else {
+	Trace("trace_buffer()\n");
     }
 }
 
@@ -944,7 +1037,7 @@ void
 trace_all_buffers(const char *fn, int ln)
 {
     BUFFER *bp;
-    TRACE((T_CALLED "trace_all_buffers %s@%d\n", fn, ln));
+    Trace(T_CALLED "trace_all_buffers %s@%d\n", fn, ln);
     for_each_buffer(bp)
 	trace_buffer(bp);
     returnVoid();
@@ -958,12 +1051,16 @@ trace_window(WINDOW *wp)
     int had_line = FALSE;
 
     if (wp == 0
-	|| wp->w_bufp == 0
-	|| wp->w_bufp->b_bname == 0)
+	|| wp->w_bufp == 0)
 	return;
 
-    Trace("trace_window(%s) top=%d, rows=%d, head=%p, line=%p dot=%p%s\n",
+    Trace("trace_window(%s)%s top=%d, rows=%d, head=%p, line=%p dot=%p%s\n",
 	  wp->w_bufp->b_bname,
+	  ((wp == wminip)
+	   ? " MINI"
+	   : ((wp == wnullp)
+	      ? " NULL"
+	      : "")),
 	  wp->w_toprow,
 	  wp->w_ntrows,
 	  (void *) win_head(wp),
@@ -985,6 +1082,16 @@ trace_window(WINDOW *wp)
     }
 }
 
+void
+trace_all_windows(const char *fn, int ln)
+{
+    WINDOW *wp;
+    Trace(T_CALLED "trace_all_windows %s@%d\n", fn, ln);
+    for_each_window(wp)
+	trace_window(wp);
+    returnVoid();
+}
+
 #define TRACE_CTYPE(mask) if (table[n] & chrBIT(mask)) strcat(buffer, "|" #mask)
 void
 trace_ctype(CHARTYPE *table, int first, int last)
@@ -1015,12 +1122,13 @@ trace_ctype(CHARTYPE *table, int first, int last)
     }
 }
 
-#if OPT_WORKING && (OPT_TRACE > 2)
+#if OPT_WORKING && OPT_TRACE
 #undef beginDisplay
 void
 beginDisplay(void)
 {
-    Trace(T_CALLED "beginDisplay(%d)\n", ++im_displaying);
+    ++im_displaying;
+    TRACE2((T_CALLED "beginDisplay(%d)\n", im_displaying));
 }
 
 #undef endofDisplay
@@ -1029,7 +1137,7 @@ endofDisplay(void)
 {
     assert(im_displaying > 0);
     --im_displaying;
-    returnVoid();		/* matches beginDisplay */
+    return2Void();		/* matches beginDisplay */
 }
 #endif
 
@@ -1043,7 +1151,7 @@ close_me(void)
 #if OPT_ELAPSED
     btree_freeup(&elapsed_tree);
 #endif
-    Trace(bad_form);
+    Trace("%s", bad_form);
 
     FreeAndNull(visible_result);
     FreeAndNull(visible_indent);
