@@ -17,7 +17,7 @@
  *   "FAILED" may not be used to test an OLE return code.  Use SUCCEEDED
  *   instead.
  *
- * $Header: /usr/build/vile/vile/RCS/w32ole.cpp,v 1.30 2008/06/01 20:07:04 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/w32ole.cpp,v 1.34 2015/05/22 01:06:03 tom Exp $
  */
 
 #include "w32vile.h"
@@ -146,9 +146,9 @@ oleauto_init(OLEAUTO_OPTIONS *opts)
      * Register Vile application object in the Running Object Table (ROT).
      * This allows controllers to connect to a running application object
      * instead of creating a new instance.  Use weak registration so that
-     * the ROT releases it's reference when all external references are
+     * the ROT releases its reference when all external references are
      * released.  If strong registration is used, the ROT will not release
-     * it's reference until RevokeActiveObject is called and so will keep
+     * its reference until RevokeActiveObject is called and so will keep
      * the object alive even after all external references have been
      * released.
      */
@@ -212,7 +212,7 @@ oleauto_exit(int code)
 
 /* ----------------------- C++ Helper Functions --------------------- */
 
-#if OPT_TRACE
+#if OPT_TRACE && !(defined(_UNICODE) || defined(UNICODE))
 static char *visible_wcs(OLECHAR *source)
 {
     static char *result = 0;
@@ -278,7 +278,7 @@ static char *visible_wcs(OLECHAR *source)
  * used/copied before the conversion routine is called again.
  *
  */
-#if !(defined(_UNICODE) || defined(UNICODE))
+#if 0 // !(defined(_UNICODE) || defined(UNICODE))
 
 static char *
 ConvertToAnsi(OLECHAR *szW)
@@ -305,9 +305,11 @@ ConvertToAnsi(OLECHAR *szW)
     TRACE(("ConvertToAnsi %d->%d:%s\n", len, strlen(ansibuf), ansibuf));
     return (ansibuf);
 }
+#endif
 
+#if !(defined(_UNICODE) || defined(UNICODE))
 static OLECHAR *
-ConvertToUnicode(char *szA)
+ConvertToUnicode(const char *szA)
 {
     size_t len;
 
@@ -350,6 +352,7 @@ LoadTypeInfo(ITypeInfo **pptinfo, REFCLSID clsid)
     return NOERROR;
 }
 
+#if VC6_ADDIN
 static int
 ole_err_to_msgline(HRESULT hr, char *msg)
 {
@@ -385,6 +388,7 @@ ole_err_to_msgline(HRESULT hr, char *msg)
     LocalFree(lclbuf);
     return (FALSE);
 }
+#endif
 
 /* ------------------------ Class Factory -------------------------- */
 
@@ -482,7 +486,7 @@ vile_oa::Create(vile_oa **ppvile, BOOL visible)
     pvile->m_bVisible = (visible) ? VARIANT_TRUE : VARIANT_FALSE;
 
     // Name
-    tmp = TO_OLE_STRING(prognam);
+    tmp = reinterpret_cast<OLECHAR *>(TO_OLE_STRING(prognam));
     if (! (tmp && (pvile->m_bstrName = SysAllocString(tmp)) != 0))
         return E_OUTOFMEMORY;
 
@@ -623,7 +627,7 @@ vile_oa::get_FullName(BSTR *pbstr)
         /* Extract server path from registry. */
         sprintf(key, "CLSID\\%s\\LocalServer32", CLSID_VILEAUTO_KEY);
         if (RegOpenKeyEx(HKEY_CLASSES_ROOT,
-                         w32_charstring(key),
+                         reinterpret_cast<LPTSTR>(w32_charstring(key)),
                          0,
                          KEY_QUERY_VALUE,
                          &hk) != ERROR_SUCCESS)
@@ -640,7 +644,7 @@ vile_oa::get_FullName(BSTR *pbstr)
         }
         if ((cp = strchr(value, ' ')) != NULL)
             *cp = '\0';
-        tmp = TO_OLE_STRING(value);
+        tmp = reinterpret_cast<OLECHAR *>(TO_OLE_STRING(value));
         if (! (tmp && (m_bstrFullName = SysAllocString(tmp)) != 0))
             return (E_OUTOFMEMORY);
     }
@@ -677,7 +681,7 @@ STDMETHODIMP
 vile_oa::VileKeys(BSTR keys)
 {
     HRESULT hr   = NOERROR;
-    char    *msg = FROM_OLE_STRING(keys);
+    char    *msg = _com_util::ConvertBSTRToString(keys);
 
     TRACE(("VileKeys %d bytes:\n%s\n", strlen(msg), msg));
     while (*msg)
