@@ -2,7 +2,7 @@
  * Window management. Some of the functions are internal, and some are
  * attached to keys that the user actually types.
  *
- * $Header: /usr/build/vile/vile/RCS/window.c,v 1.118 2008/12/05 01:33:09 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/window.c,v 1.127 2013/04/13 12:37:00 tom Exp $
  *
  */
 
@@ -11,7 +11,7 @@
 
 #if OPT_PERL || OPT_TCL
 /* Window id to assign to w_id field for next visible window allocated */
-static ULONG w_id_next = 1;
+static WIN_ID w_id_next = 1;
 
 /* Fake windows are given a window id of 0 */
 #define FAKE_WINDOW_ID 0
@@ -129,11 +129,11 @@ int
 reposition(int f, int n)
 {
     if (f) {
-	int an;
+	long an;
 	/* clamp the value at the size of the window */
-	an = absol(n);
+	an = absol((long) n);
 	if (an > curwp->w_ntrows)
-	    curwp->w_force = curwp->w_ntrows * (n / an);
+	    curwp->w_force = (int) (curwp->w_ntrows * (n / an));
 	else
 	    curwp->w_force = n;
 	curwp->w_flag |= WFFORCE;
@@ -492,9 +492,10 @@ delwp(WINDOW *thewp)
     }
 
     /* find receiving window and give up our space */
-    if (thewp == wheadp
-	|| ((thewp->w_split_hist & 1) == 0 && thewp->w_wndp)
-	|| !visible) {
+    if (thewp->w_wndp != 0
+	&& (thewp == wheadp
+	    || ((thewp->w_split_hist & 1) == 0)
+	    || !visible)) {
 	/* merge with next window down */
 	wp = thewp->w_wndp;
 	if (visible) {
@@ -506,9 +507,9 @@ delwp(WINDOW *thewp)
 	    if (wp->w_split_hist & 1)
 		wp->w_split_hist >>= 1;
 	}
-	if (thewp == wheadp)
+	if (thewp == wheadp) {
 	    wheadp = wp;
-	else {
+	} else {
 	    WINDOW *pwp = wheadp;
 	    while (pwp->w_wndp != thewp)
 		pwp = pwp->w_wndp;
@@ -562,7 +563,7 @@ save_window_modes(BUFFER *bp)
 
     if (bp->b_nwnd != 0) {
 	beginDisplay();
-	result = typecallocn(W_VALUES, bp->b_nwnd);
+	result = typecallocn(W_VALUES, (size_t) bp->b_nwnd);
 	if (result != 0) {
 	    n = 0;
 	    for_each_window(wp) {
@@ -886,14 +887,16 @@ shrinkwrap(void)
 	for (wp = wheadp;
 	     wp->w_wndp != curwp && wp->w_wndp != NULL;
 	     wp = wp->w_wndp) ;
-	curwp = wp;
-	nrows = curwp->w_ntrows + curwp->w_wndp->w_ntrows - 1;
-	/* don't take more than 3/4 of its rows */
-	snrows = (nrows * 3) / 4;
-	if (nlines > snrows)
-	    nlines = snrows;
-	resize(TRUE, nrows - nlines + 1);
-	curwp = savewp;
+	if (wp->w_wndp != 0) {
+	    curwp = wp;
+	    nrows = curwp->w_ntrows + curwp->w_wndp->w_ntrows - 1;
+	    /* don't take more than 3/4 of its rows */
+	    snrows = (nrows * 3) / 4;
+	    if (nlines > snrows)
+		nlines = snrows;
+	    resize(TRUE, nrows - nlines + 1);
+	    curwp = savewp;
+	}
     }
 }
 
@@ -1079,8 +1082,8 @@ init_window(WINDOW *wp, BUFFER *bp)
     if (valid_buffer(bp)) {
 	wp->w_line.l = lforw(buf_head(bp));
 	wp->w_line.o = 0;
-	wp->w_dot.l = lforw(buf_head(bp));
-	wp->w_dot.o = 0;
+	wp->w_dot.l = wp->w_line.l;
+	wp->w_dot.o = wp->w_line.o;
     } else {
 	wp->w_line = nullmark;
 	wp->w_dot = nullmark;
@@ -1113,6 +1116,23 @@ clone_window(WINDOW *dst, WINDOW *src)
     dst->w_flag |= WFMODE | WFHARD;	/* Quite nasty.         */
 }
 
+static WINDOW *
+new_WINDOW(int screen)
+{
+    WINDOW *wp = typecalloc(WINDOW);
+
+    if (wp == NULL) {
+	fprintf(stderr, "Cannot allocate windows\n");
+	ExitProgram(BADEXIT);
+    } else {
+	init_window(wp, (BUFFER *) 0);
+	wp->w_ntrows = (screen
+			? (term.rows - 2)	/* "-1" for mode line.  */
+			: 1);	/* command-line         */
+    }
+    return wp;
+}
+
 /*
  * Initialize all of the windows.
  */
@@ -1123,23 +1143,10 @@ winit(int screen)
 
     TRACE((T_CALLED "winit(%d)\n", screen));
 
-    wp = typecalloc(WINDOW);	/* First window         */
-    if (wp == NULL) {
-	fprintf(stderr, "Cannot allocate windows\n");
-	ExitProgram(BADEXIT);
-    }
+    wp = new_WINDOW(screen);	/* First window         */
     wheadp = wp;
     curwp = wp;
 
-    init_window(wp, (BUFFER *) 0);
-    wp->w_wndp = NULL;		/* Initialize window    */
-    wp->w_toprow = 0;
-    wp->w_ntrows = (screen
-		    ? (term.rows - 2)	/* "-1" for mode line.  */
-		    : 1);	/* command-line         */
-    wp->w_force = 0;
-    wp->w_bufp = NULL;
-
 #if OPT_PERL || OPT_TCL
     wp->w_id = w_id_next;
     ++w_id_next;
@@ -1155,6 +1162,9 @@ winit(int screen)
 	(void) delink_bp(btempp);
 #endif
     } else {
+	/* create a dummy window, to hold buffers before display is active */
+	wnullp = new_WINDOW(1);
+
 	/* create the command-buffer */
 	TRACE(("winit creating bminip & wminip\n"));
 	wminip = wp;
@@ -1169,6 +1179,8 @@ winit(int screen)
 	make_local_w_val(wminip, WMDLIST);
 	set_w_val(wminip, WMDLIST, TRUE);
 #ifdef WMDLINEWRAP
+	make_local_w_val(wminip, WMDLINEBREAK);
+	set_w_val(wminip, WMDLINEBREAK, FALSE);
 	make_local_w_val(wminip, WMDLINEWRAP);
 	set_w_val(wminip, WMDLINEWRAP, FALSE);
 #endif
@@ -1258,7 +1270,7 @@ pop_fake_win(WINDOW *oldwp, BUFFER *oldbp)
 /* Find and return the window with the given window id.  Return NULL
    if not found */
 WINDOW *
-id2win(int id)
+id2win(WIN_ID id)
 {
     WINDOW *wp;
     for_each_visible_window(wp) {
@@ -1269,7 +1281,7 @@ id2win(int id)
 }
 
 /* Return the window id associated with the given window */
-int
+WIN_ID
 win2id(WINDOW *wp)
 {
     return wp->w_id;
