@@ -3,7 +3,7 @@
  * paragraph at a time.  There are all sorts of word mode commands.  If I
  * do any sentence mode commands, they are likely to be put in this file.
  *
- * $Header: /usr/build/vile/vile/RCS/word.c,v 1.93 2009/05/17 23:33:39 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/word.c,v 1.107 2015/03/13 08:42:56 tom Exp $
  *
  */
 
@@ -50,7 +50,7 @@ wrapword(int f, int n)
 	if (f) {
 	    LINE *lp = DOT.l;
 	    to_delete = 0L;
-	    if (DOT.o >= 0 && !n && !isSpace(lgetc(lp, DOT.o))) {
+	    if (DOT.o >= 0 && !isSpace(lgetc(lp, DOT.o))) {
 		int tmp = 0;
 		for (each_character(c)) {
 		    if (isSpace(lgetc(lp, c))) {
@@ -309,14 +309,15 @@ backword(int f, int n)
 static int
 join_region(int exact)
 {
+    BUFFER *bp = curbp;
     int status;
     int doto, c;
     LINE *end;
     REGION region;
     int done = FALSE;
 
-    if ((status = getregion(&region)) == TRUE
-	&& (status = !is_last_line(region.r_orig, curbp)) == TRUE) {
+    if ((status = getregion(bp, &region)) == TRUE
+	&& (status = !is_last_line(region.r_orig, bp)) == TRUE) {
 
 	DOT = region.r_orig;
 	end = region.r_end.l;
@@ -410,12 +411,13 @@ comment_prefix(void)
 
 #if OPT_MULTIBYTE
 static int
-byte_to_columns(char text, int size)
+byte_to_columns(int text, int size)
 {
+    char ch = (char) text;
     int result = 0;
 
     if (b_is_utfXX(curbp)) {
-	result = (mb_cellwidth(curwp, &text, 1) * size);
+	result = (mb_cellwidth(curwp, &ch, 1) * size);
     } else {
 	result = size;
     }
@@ -423,13 +425,11 @@ byte_to_columns(char text, int size)
 }
 
 static int
-string_to_columns(TBUFF *text)
+any_wcs_width(const char *value, long bytes)
 {
-    long bytes = (long) tb_length(text);
     int result = 0;
 
     if (b_is_utfXX(curbp)) {
-	char *value = tb_values(text);
 
 	while (bytes > 0) {
 	    int step = vl_conv_to_utf32((UINT *) 0, value, (B_COUNT) bytes);
@@ -440,16 +440,27 @@ string_to_columns(TBUFF *text)
 	    bytes -= step;
 	}
     } else {
-	result = bytes;
+	result = (int) bytes;
     }
 
     return result;
 }
+
+int
+str_wcs_width(const char *value)
+{
+    return any_wcs_width(value, (long) strlen(value));
+}
+
+int
+tb_wcs_width(TBUFF *text)
+{
+    return any_wcs_width(tb_values(text), (long) tb_length(text));
+}
+
 #define Byte2Column(s,n) byte_to_columns(s, n)
-#define String2Columns(s) string_to_columns(s)
 #else
 #define Byte2Column(s,n) n
-#define String2Columns(s) tb_length(s)
 #endif
 
 #define cplus_comment_start(c) \
@@ -473,7 +484,7 @@ string_to_columns(TBUFF *text)
 		    if ((s = lins_bytes(1, tb_values(str)[i])) != TRUE) \
 			return s; \
 		} \
-		clength += String2Columns(str)
+		clength += tb_columns(str)
 
 #define fmt_c_preprocessor(cp) \
 		(tb_length(*cp) == 1 \
@@ -482,11 +493,12 @@ string_to_columns(TBUFF *text)
 
 #define word_finishes_c_comment(wp) \
 		(tb_length(*wp) >= 2 \
-		&& memcmp("/*", tb_values(*wp), 2) == 0)
+		&& memcmp("/*", tb_values(*wp), (size_t) 2) == 0)
 
 static int
 do_formatting(TBUFF **wp, TBUFF **cp)
 {
+    BUFFER *bp = curbp;
     int c;			/* current char during scan     */
     int clength;		/* position on line during fill */
     int plength;		/* prefix to ignore during fill */
@@ -500,18 +512,18 @@ do_formatting(TBUFF **wp, TBUFF **cp)
     int sentence;		/* was the last char a period?  */
     int secondindent;
     int s;
-    int fillcolumn = getfillcol(curbp);
+    int fillcolumn = getfillcol(bp);
 
     if (!sameline(MK, DOT)) {
 	REGION region;
 
-	if (getregion(&region) != TRUE)
+	if (getregion(bp, &region) != TRUE)
 	    return FALSE;
 	if (sameline(region.r_orig, MK))
 	    swapmark();
     }
     pastline = MK.l;
-    if (pastline != buf_head(curbp))
+    if (pastline != buf_head(bp))
 	pastline = lforw(pastline);
 
     finished = FALSE;
@@ -519,7 +531,7 @@ do_formatting(TBUFF **wp, TBUFF **cp)
 	if (DOT.l == pastline)	/* FIXME */
 	    return setmark();
 	while (dot_at_section_break()) {
-	    DOT.l = lforw(DOT.l);
+	    dot_next_bol();
 	    if (DOT.l == pastline)
 		return setmark();
 	}
@@ -528,7 +540,7 @@ do_formatting(TBUFF **wp, TBUFF **cp)
 
 	/* go forward to get the indent for the second
 	   and following lines */
-	DOT.l = lforw(DOT.l);
+	dot_next_bol();
 
 	if (DOT.l != pastline) {
 	    secondindent = indentlen(DOT.l);
@@ -552,10 +564,10 @@ do_formatting(TBUFF **wp, TBUFF **cp)
 		       (size_t) (plength - DOT.o));
 	} else if (cplus_comment_start(c)) {
 	    is_comment = TRUE;
-	    tb_bappend(cp, "//", 2);
+	    tb_bappend(cp, "//", (size_t) 2);
 	} else if (c_comment_start(c)) {
 	    is_comment = TRUE;
-	    tb_bappend(cp, "*", 1);
+	    tb_bappend(cp, "*", (size_t) 1);
 	}
 
 	/* scan through lines, filling words */
@@ -569,8 +581,10 @@ do_formatting(TBUFF **wp, TBUFF **cp)
 
 	    /* get the next character */
 	    if (is_at_end_of_line(DOT)) {
+		MARK saved = DOT;
+
 		c = ' ';
-		DOT.l = lforw(DOT.l);
+		dot_next_bol();
 		if (DOT.l == pastline) {
 		    finished = TRUE;
 		} else if (dot_at_section_break()) {
@@ -578,8 +592,6 @@ do_formatting(TBUFF **wp, TBUFF **cp)
 		}
 
 		if ((s = comment_prefix()) >= 0) {
-		    int save = DOT.o;
-
 		    (void) firstnonwhite(FALSE, 1);
 		    s -= DOT.o;
 
@@ -602,15 +614,14 @@ do_formatting(TBUFF **wp, TBUFF **cp)
 			if (DOT.l != pastline
 			    && !dot_at_section_break()) {
 			    B_COUNT spcs = (B_COUNT) DOT.o;
-			    DOT.o = 0;
+			    DOT.o = b_left_margin(curbp);
 			    s = ldel_bytes(spcs, FALSE);
 			    if (s != TRUE)
 				return s;
 			}
 		    }
-		    DOT.o = save;
 		}
-		DOT.l = lback(DOT.l);
+		DOT = saved;
 		at_nl = TRUE;
 	    } else {
 		c = char_at(DOT);
@@ -631,7 +642,7 @@ do_formatting(TBUFF **wp, TBUFF **cp)
 	    if (!isBlank(c)) {
 		tb_append(wp, c);
 		sentence = FALSE;
-		if (global_g_val(GMDSPACESENT)) {
+		if (b_val(bp, MDSPACESENT)) {
 		    /* was it the end of a "sentence"? */
 		    if (c == '?' || c == ':' || c == '!') {
 			sentence = TRUE;
@@ -646,8 +657,8 @@ do_formatting(TBUFF **wp, TBUFF **cp)
 			     * add "etc.", but that is a valid sentence ending,
 			     * and would be unclear without a comma otherwise.
 			     */
-			    sentence = memcmp(ptr, "e.g.", 4)
-				&& memcmp(ptr, "i.e.", 4);
+			    sentence = memcmp(ptr, "e.g.", (size_t) 4)
+				&& memcmp(ptr, "i.e.", (size_t) 4);
 			} else {
 			    sentence = TRUE;
 			}
@@ -657,7 +668,7 @@ do_formatting(TBUFF **wp, TBUFF **cp)
 		/* at a word break with a word waiting */
 		/* calculate tentative new length
 		   with word added */
-		newlen = (int) (clength + 1 + String2Columns(*wp));
+		newlen = (int) (clength + 1 + tb_columns(*wp));
 		if (newlen <= fillcolumn) {
 		    /* add word to current line */
 		    if (!firstflag) {
@@ -666,7 +677,7 @@ do_formatting(TBUFF **wp, TBUFF **cp)
 		} else {
 		    /* fix the leading indent now, if
 		       some spaces should be tabs */
-		    if (b_val(curbp, MDTABINSERT))
+		    if (b_val(bp, MDTABINSERT))
 			entabline((void *) TRUE, 0, 0);
 		    if (lnewline() == FALSE)
 			return FALSE;
@@ -696,9 +707,9 @@ do_formatting(TBUFF **wp, TBUFF **cp)
 	/* catch the case where we're done with a line not because
 	   there's no more room, but because we're done processing a
 	   section or the region */
-	if (b_val(curbp, MDTABINSERT))
+	if (b_val(bp, MDTABINSERT))
 	    entabline((void *) TRUE, 0, 0);
-	DOT.l = lforw(DOT.l);
+	dot_next_bol();
     }
     return setmark();
 }
@@ -728,24 +739,25 @@ formatregion(void)
 int
 wordcount(int f GCC_UNUSED, int n GCC_UNUSED)
 {
+    BUFFER *bp = curbp;
     LINE *lp;			/* current line to scan */
     int offset;			/* current char to scan */
     ULONG size;			/* size of region left to count */
     int ch;			/* current character to scan */
     int wordflag;		/* are we in a word now? */
     int lastflag;		/* were we just in a word? */
-    long nwhite;		/* total # of blanks */
+    B_COUNT nwhite;		/* total # of blanks */
     long nwords;		/* total # of words */
-    long nchars;		/* total number of chars */
+    B_COUNT nchars;		/* total number of chars */
     int nlines;			/* total number of lines in region */
     double avgch;		/* average number of chars/word */
     int status;			/* status return code */
     REGION region;		/* region to look at */
-    const char *ending = get_record_sep(curbp);
-    int len_rs = len_record_sep(curbp);
+    const char *ending = get_record_sep(bp);
+    B_COUNT len_rs = (B_COUNT) len_record_sep(bp);
 
     /* make sure we have a region to count */
-    if ((status = getregion(&region)) != TRUE) {
+    if ((status = getregion(bp, &region)) != TRUE) {
 	return (status);
     }
     lp = region.r_orig.l;
@@ -787,11 +799,11 @@ wordcount(int f GCC_UNUSED, int n GCC_UNUSED)
 
     /* and report on the info */
     if (nwords > 0L)
-	avgch = ((1.0 * nchars) / nwords);
+	avgch = ((1.0 * (double) nchars) / (double) nwords);
     else
 	avgch = 0.0;
 
-    mlforce("lines %d, words %ld, chars %ld, avg chars/word %f",
+    mlforce("lines %d, words %ld, chars %lu, avg chars/word %f",
 	    nlines, nwords, nchars + nwhite, avgch);
 
     return (TRUE);
